package SAP_functions;

#"@(#) SGeSAP/LX - 12.20.00 - $Rev: 1819 $ - Last modified 02/02/17"
#
# **********************************************************************
# *                                                                    *
# * SGeSAP common perl module                                          *
# *                                                                    *
# * (c) Copyright 2015-2016 Hewlett Packard Enterprise Co.             *
# * (c) Copyright 2011-2015 Hewlett-Packard Corp.                      *
# *                                                                    *
# *                                                                    *
# * Destination Filename:    /opt/cmcluster/sap/lib/SAP_functions.pm   *
# *                                                                    *
# * Description: common perl module used used by sapgetconf, sapverify,*
# *              deploysappkgs                                         *
# *                                                                    *
# * Parameters:  see usage                                             *
# *                                                                    *
# **********************************************************************

#============================================================================

=pod

=head1 NAME

SAP_functions.pm - Common SGeSAP perl functions

=head1 USAGE

    use SAP_functions;

=head1 EXPORTED FUNCTIONS

    usage_perldoc()
    setup_script_runtime()
    cleanup_n_close()
    getparam()
    setparam()
    printSGMgr()
    LogEnterSub ()
    LogReturnSub ()
    LogVerbose ()
    LogDebug()
    LogInfo ()
    LogError()
    LogExtra()
    LogRuntimeError()
    timestamp_and_exit()
    call_func_with_timeout_sig()
    call_func_with_timeout_thr()

=cut

#============================================================================

#---------------------------------------------------
# load perl modules
#---------------------------------------------------
 
use Carp;
use Carp qw/longmess cluck confess/;
use File::Basename;
use Sys::Hostname;
use Cwd;
use Data::Dumper;
use Time::HiRes qw(gettimeofday tv_interval);
use threads;
use Thread::Queue;
use B;
use POSIX 'strftime';


#---------------------------------------------------
# export the subroutines
#---------------------------------------------------
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
    usage_perldoc
    setup_script_runtime 
    cleanup_n_close
    getparam
    setparam
    printSGMgr
    LogEnterSub 
    LogReturnSub 
    LogVerbose 
    LogDebug
    LogInfo 
    LogError
    LogExtra
    LogRuntimeError
    timestamp_and_exit
    call_func_with_timeout_sig
    call_func_with_timeout_thr
    search_enqorpkg
);


#---------------------------------------------------
# $params{verbose}=2 enables early debugging/extra
# output e.g. bootstrap output of script
#---------------------------------------------------
$params{verbose}=0;
LogExtra("In SAP_functions.pm");

#---------------------------------------------------
# set system specifics
#---------------------------------------------------

my $host   = hostname();
my $node   = $host unless defined $node;
my $user   = getpwuid($<);
my $mypid  = $$;
my $cwd    = getcwd();
my $osname = $^O;

# Serviceguard cluster config file
my $CMCONF="/etc/cmcluster.conf";   # config file
my $CMCONF_FH;                      # file handle
my (
    $SGROOT,
    $SGLIB,
    $SGSBIN,
    $SGLBIN,
    $SGCONF,
    $SGRUN, 
);


# split $0 into basename, directory and suffix
my ($basename, $dirname, $suffix) = fileparse($0,qr/\.[^.]*/);

# an optional script configuration file for $0
my $conf_file;

# logs if the script config file got executed
my $conf_file_executed="none";

# config or environment params are preceded with uppercase scriptname
my $SCRIPT= uc($basename);

# save number of errors encountered during sapverify run
$params{errorcount} = 0; 

# PREFIX used for logging to package log
my $PREFIX="$user\@$host $basename\[$mypid\]: ";

#---------------------------------------------------
#  bootstrap: if exists read the script specific 
#  configuration file of $0 and and save in params{}
#---------------------------------------------------
read_script_conf_file();

#---------------------------------------------------
# bootstrap: read the /etc/cmcluster.conf params an 
# save in params{}
#---------------------------------------------------
getcmclusterconf();
#############################################################################
#
# BEGIN SECTION FOR GENERAL FUNCTIONS
#
#############################################################################

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         usage_perldoc()
 # PURPOSE:      Print usage using pod/perldoc syntax 
 # PARAMETERS:   Global $0 is the name of this script
 # RETURNS:      None - script exits
 # COMMENTS:     This function gets executed when "script -help" is run. It
 #               runs perldoc on itself
 #============================================================================

=cut

sub usage_perldoc {
    system("perldoc", $0);    
    exit;
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         printOn()
 # PURPOSE:      Prints to stdout without a timestamp (n=no timestamp)
 # PARAMETERS:   @_  arguments
 # RETURNS:      None
 # COMMENTS:     None
 #============================================================================

=cut

sub printOn {

    my $tmp = join( "", @_, "\n" );
    print STDOUT "$tmp";
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         printO()
 # PURPOSE:      Prints to stdout with a timestamp
 # PARAMETERS:   @_  arguments
 # RETURNS:      None
 # COMMENTS:     None
 #============================================================================

=cut 

sub printO {

#    my $tmp = join( "", scalar localtime(), " ", @_, "\n" );

    my $datestring = strftime "%b %e %H:%M:%S ", localtime;
    my $tmp = join( "", $datestring, "$PREFIX", @_, "\n" );
    print STDOUT "$tmp";
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         printL()
 # PURPOSE:      Prints  to logfile $LOG_FH with a timestamp. If logfile
 #               unavailable prints to stdout.
 # PARAMETERS:   @_  arguments
 # RETURNS:      None
 # COMMENTS:     None
 #============================================================================

=cut

sub printL {

#    my $tmp = join( "", scalar localtime(), " ", @_, "\n" );
    my $datestring = strftime "%b %e %H:%M:%S ", localtime;
    my $tmp = join( "", $datestring, "$PREFIX", @_, "\n" );
    if ( defined $LOG_FH ) {

        #  $LOG_FH descriptor is available
        print $LOG_FH "$tmp";
    }
    else {

        #  $LOG_FH descriptor is closed
        print STDOUT "$tmp";
    }

}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         printEn()
 # PURPOSE:      prints to stderr without a timestamp (n=no timestamp)
 # PARAMETERS:   @_  arguments
 # RETURNS:      None
 # COMMENTS:     None
 #============================================================================

=cut 

sub printEn {

    my $tmp = join( "", @_, "\n" );
    print STDERR "$tmp";
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         printE()
 # PURPOSE:      prints to stderr with a timestamp
 # PARAMETERS:   @_ = arguments
 # RETURNS:      None
 # COMMENTS:     None
 #============================================================================

=cut

sub printE {

    my $tmp = join( "", scalar localtime(), " ", @_, "\n" );
    print STDERR "$tmp";

}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         LogEnterSub()
 # PURPOSE:      Log the name and line of a subroutine line that was entered
 #               as well as who is calling this subroutine
 # PARAMETERS:   @_ = arguments to function being called
 # RETURNS:      None
 # COMMENTS:     None
 #============================================================================

=cut

sub LogEnterSub {

    if ($params{tracesubs}) {

        # Line number and subroutine of who is calling this function
        my ( $called_frome_line, $sub_name ) = ( caller(1) )[ 2, 3 ];

        # The current line number of the subroutine being executed,
        # and make the $current_line 4 digits wide
        my $current_line = sprintf "%4s", ( caller(0) )[2];

        printL "ENTER  at LINE: $current_line $sub_name(@_) " . 
               "Called from line: $called_frome_line";
    }
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         LogReturnSub()
 # PURPOSE:      Log the the subroutine line, subroutine and the return value
 # PARAMETERS:   @_  = arguments to function being called
 # RETURNS:      None
 # COMMENTS:     None
 #============================================================================

=cut

sub LogReturnSub {
    my ($rc) = @_;
    $rc = "" unless defined $rc;
#    my @rc = @_;
#    @rc = "" unless @rc;

    if ($params{tracesubs}) {

        # subroutine of returning function
        my ($sub_name) = ( caller(1) )[3];

        # The current line number the "return" is being called from
        # and make $current_line 4 digits wide
        my ($current_line) = sprintf "%4s", ( caller(0) )[2];

        printL "RETURN at LINE: $current_line $sub_name() rc=[$rc]";
    }
    return $rc;
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         LogVerbose() 
 # PURPOSE:      Log to log file only if $params{verbose} flag is set
 # PARAMETERS:   @_  = arguments to function being called
 # RETURNS:      None
 # COMMENTS:     - Indent one space so output aligns with LogEnterSub and
 #                 LogReturnSub subroutines 
 #               - default it to log verbose output to STDERR
 #               - if "logging to logfile" is enabled, disable logging to STDERR
 #============================================================================

=cut

sub LogVerbose {

    if ($params{verbose} > 0) {
#        printEn " ", @_ ;
        printEn " ", @_ if (  $params{logging} != 1)  ;
        printL " ", @_ if ( defined $LOG_FH );
    }
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         LogDebug()
 # PURPOSE:      Log only if $params{debug} flag is set
 # PARAMETERS:   @_ = arguments to function being called
 # RETURNS:      None
 # COMMENTS:     Prepend DEBUG string  to output line
 #============================================================================

=cut

sub LogDebug {

    if ( $params{debug} ) {
        printE "  DEBUG: ", @_ if ( !defined $LOG_FH );
        printL "  DEBUG: ", @_ if ( defined $LOG_FH );
    }
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         LogExtra()
 # PURPOSE:      Log extra/real early on - just for initial debugging
 # PARAMETERS:   
 # RETURNS:      None
 # COMMENTS:     increase verbose to 2 for early/extra logging
 #============================================================================

=cut

sub LogExtra {

    if ((defined $params{verbose}) &&  ($params{verbose} > 1)) {
        printEn " ",@_;
    }
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         LogError()
 # PURPOSE:      Log to stderr and $LOG_FH
 # PARAMETERS:   @_  = arguments to function being called
 # RETURNS:      None
 # COMMENTS:     Logs a error to standard error and log file.
 #               Also counts number of errors encountered for oveall report
 #============================================================================

=cut

sub LogError {
    printEn @_;
    printL @_ if ( defined $LOG_FH );

}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         LogInfo()
 # PURPOSE:      Log to stdout and $LOG_FH
 # PARAMETERS:   @_  = arguments to function being called
 # RETURNS:      None
 # COMMENTS:     Logs a error to standard error and log file.
 #               Also counts number of errors encountered for oveall report
 #============================================================================

=cut

sub LogInfo {

    printOn     @_;
    printL " ", @_ if ( defined $LOG_FH );
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         printSGMgr()
 # PURPOSE:      Creates output for ServiceGuard Manager input
 # PARAMETERS:   @_     = arguments to function being called
 # RETURNS:      None
 # COMMENTS:
 #============================================================================

=cut

sub printSGMgr {

    printOn     @_;
    printL " ", @_ if ( defined $LOG_FH );

}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         LogRuntimeError()
 # PURPOSE:      Log a runtime error (opening files, pipes...)
 # PARAMETERS:   @_     = arguments to function being called
 # RETURNS:      None
 # COMMENTS:     This script could be executing in various contexts. If being
 #               run out of an SGeSAP monitors it should log the verification
 #               errors found but shouldn't force the package to stop as the
 #               verification error could be from another cluster node. 
 #============================================================================

=cut

sub LogRuntimeError {
    my ( @string ) = @_;

    my $message  = longmess();

    printO "ERROR: @string";
    printO "\n\t$message";

    printL "ERROR: @string" if ( defined $LOG_FH );
    printL "\n\t$message" if ( defined $LOG_FH ) ;

    $params{errorcount}++; 

    timestamp_and_exit(2);
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         timestamp_and_exit()
 # PURPOSE:      Log to stdout and $LOG_FH
 # PARAMETERS:   @_     = arguments to function being called
 # RETURNS:      None
 # COMMENTS:     Before script exits, it should always log 
 #============================================================================

=cut

sub timestamp_and_exit {
    my $EXIT = shift;

    if ($params{verbose} > 0) {
        printO "Exit from script $0 with EXIT=[$EXIT]";
        printL "Exit from script $0 with EXIT=[$EXIT]" if ( defined $LOG_FH );
    }
    exit $EXIT;
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         OpenLogFile()
 # PURPOSE:      Open the log file $LOG_FH
 # PARAMETERS:   $LOGF   = name of the logfile
 # RETURNS:      None
 # COMMENTS:     Note: since we are just about to open the log file itself - 
 #               can't log with LogEnterSub yet....
 #============================================================================

=cut

sub OpenLogFile {
#    LogEnterSub(@_);

    my $LOGF = shift;

    # open and log a initial timestamp to the logfile
    my $timestamp = localtime;

    open( $LOG_FH, ">>", "$LOGF" ) or confess "Can't open [$LOGF]: $!";
    $LOG_FH->autoflush(1);

#    LogVerbose("\n");
#    LogVerbose("********** START [$basename] on [$host] **********");

#    LogReturnSub();
    return;
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         CloseLogFile()
 # PURPOSE:      Close the log file $LOG_FH
 # PARAMETERS:   $LOGF   = name of the logfile
 # RETURNS:      None
 # COMMENTS:     
 #============================================================================

=cut

sub CloseLogFile {
#    LogEnterSub(@_);

    my $LOGF = shift;

    # open and log a initial timestamp to the logfile
    my $timestamp = localtime;

    LogVerbose("********** END   [$basename] on [$host] **********");
    LogVerbose("\n");

    close( $LOG_FH) or confess "Can't close [$LOGF]: $!";

#    LogReturnSub();
    return;
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         cleanup_n_close()
 # PURPOSE:      .....close logfile with timestamp
 # PARAMETERS:   
 # RETURNS:      
 # COMMENTS:     This function gets executed when script has completed
 #============================================================================

=cut

sub cleanup_n_close {
 
        CloseLogFile($params{LOGF});
}


=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         getparam()
 # PURPOSE:      get parameters defined / used in the SAP_functions.pm module
 #               and make them visible to any script using the module
 # PARAMETERS:   $param_name = name of the parameter to get
 # RETURNS:      the value of the parameter
 # COMMENTS:     
 #============================================================================

=cut

sub getparam {
    LogEnterSub(@_);
    my $param_name = shift;
    my $value;

    if ( exists $params{$param_name} ) {
        $value = $params{$param_name};
    }
    else {
        $value = "";
    }

    LogDebug("getparam(): param_name=[$param_name] value=[$value]");

    LogReturnSub($value);
    return $value;
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         setparam()
 # PURPOSE:      set parameters defined / used in the SAP_functions.pm module 
 #               from an "outside" script 
 # PARAMETERS:   $param_name  = name of the parameter
 #               $param_value = value of the parameter
 # RETURNS:      
 # COMMENTS:     
 #============================================================================

=cut

sub setparam {
    LogEnterSub(@_);
    my $param_name  = shift;
    my $param_value = shift;

    $params{$param_name} = $param_value;
    LogDebug "setparam() \$param_name=[$param_name] \$param_value=[$param_value]";

    LogReturnSub();
    return;
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         warning_handler()
 # PURPOSE:      Signal handler for catching warnings. Log warnings to STDERR 
 #               and the log file
 # PARAMETERS:
 # RETURNS:
 # COMMENTS:
 #============================================================================

=cut

sub warning_handler{

    local $_ = "Warning: @_";
    printL("$_") if ( defined $LOG_FH ); 
    printE("$_");

    my $message  = longmess();
    printL("$_") if ( defined $LOG_FH ); 
    printE " $message";
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         call_func_with_timeout_sig()
 # PURPOSE:      Execute a "user function call" under watchdog supervision. 
 #               Implementation based on alarm signal handler.
 # PARAMETERS:   $timeout      = time after which the function call times out
 #               $function_ref = reference to a function to call
 #               @params       =  parameters of function
 # RETURNS:      
 # COMMENTS:     The goal is to prevent any NFS related hangs of this programm.
 #               Note: this only works if the programm itself is forked.
 #               It is then possible to kill the child hanging on NFS I/O. 
 #               This is a low level / local timeout alarm
 #
 #               Takes into consideration that a global / high level alarm
 #               might have been established. Before setting up the local alarm
 #               save global alarm elapsed time in $prev_state. Now execute
 #               function under a local alarm supervsion. After the function
 #               call has completed, establich the global timer again.
 #
 #               See:
 #                   perldoc Time::HiRes
 #                   perldoc -f alarm
 #
 #              Also Checkout function call_func_with_timeout_thr()
 #============================================================================

=cut 

sub call_func_with_timeout_sig {
    LogEnterSub(@_);

    my ( $timeout, $function_ref, @params ) = @_;
    my @result;                     # save result of function call
    my $timeout_occurred = 0;       # if function call timed out
    my $prev_status;                # save elapsed time of any higher-level alarm
    my $start_time = [gettimeofday];# Time::HiRes

#    LogDebug "seconds microseconds =[@{$start_time}];
    LogDebug "elapsed=[". int ( tv_interval($start_time) ) . "]";

    eval {

        # setup signal-handler in local scope
        local ( $SIG{ALRM} ) = sub { croak "alarm for call_func_with_timeout_sig\n"; };

        # set up low-level alarm
        LogDebug "Set up low-level alarm timeout=[$timeout]";
        $prev_status = alarm($timeout);
        LogDebug "High-level alarm elapsed time saved in prev_status=[$prev_status]";

        # call the function
        @result = &$function_ref(@params);

        # cancel the low-level alarm
        alarm(0);
    };

    if ($@) {

        # only interested in alarms  - propagate all other exceptions
        croak "Other exception occured \$@=[$@]" unless  $@  =~ m#^alarm for call_func_with_timeout_sig\n# ;    

        # exception was an alarm
        LogDebug "Low-level alarm=[$@]";
        $timeout_occurred = 1;

    }

    # reset high-level alarm if necessary
    if ($prev_status) {
        my $new_alarm = $prev_status - int( tv_interval($start_time) );
        if ($new_alarm) {
            alarm($new_alarm);
            LogDebug "reset alarm to $new_alarm secs";
        }
        else {

            # what do I do now? maybe set it for one sec in future
            alarm(1);
        }
    }

    # return 
    if ($timeout_occurred) {

        # get the function name of the function reference for logging purposes
        my $obj  = B::svref_2object $function_ref;
        my $name = $obj->GV->NAME;
        my $pkg  = $obj->GV->STASH->NAME;

        # don't exit on error - just report warnings
        if ($params{warningmode}) {
            LogInfo("Timeout of function: $pkg\::$name() after timeout=[$timeout]");
            return undef; 
        } else {
            LogRuntimeError("Timeout of function: $pkg\::$name()");
            timestamp_and_exit(2);
            croak "timeout occurred";
        }

    }
    else {
        #return result of function call

        LogDebug "\@result=[@result]";
        return ( LogReturnSub( wantarray ? @result : $result[0]));
    }

}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         call_func_with_timeout_thr()
 # PURPOSE:      Execute a "user function call" under watchdog supervision. 
 #               Implementation based on threads and queues
 # PARAMETERS:  
 # RETURNS:      
 # COMMENTS:     The goal is to prevent any NFS related hanging of this script.
 #               Note: No need to fork a process as in call_func_with_timeout_sig. 
 #============================================================================

=cut

sub call_func_with_timeout_thr {
    LogEnterSub(@_);
    my ( $timeout, $function_ref, @params ) = @_;


    #---------------------------------------------------
    # create a queue for msg-passing
    # timeout thread: will queue a 'timeout' if timer expired
    # work thread:    will queue a 'success' if didn't hang
    #---------------------------------------------------
    my $q = Thread::Queue->new();

    # create threads for timeout and function-call
    my $timeout_thr = threads->create( \&do_timeout, $q, $timeout );
    my ($work_thr) = threads->create( \&exec_function, $q, $function_ref, @params, );

    # block until one of the threads finishes
    my $thr_status = $q->dequeue();

    if ( $thr_status eq 'timeout' ) {

        #---------------------------------------------------
        # function timed out
        #---------------------------------------------------
        LogDebug "function timed out";

        # When the thread finishes, ->join() will return the
        # return value(s) of the entry point function.
        $timeout_thr->join();
        LogDebug "timeout-func join finished";

        # Signal the thread to terminate, and then detach
        # it so that it will get cleaned up automatically
        $work_thr->kill('KILL')->detach();

        # get the name of the function for logging purposes
        my $obj  = B::svref_2object $function_ref;
        my $name = $obj->GV->NAME;
        my $pkg  = $obj->GV->STASH->NAME;

        LogRuntimeError("Timeout of function: $pkg\::$name()");
        timestamp_and_exit(2);

        # timeout occured: return empty list in array context,
        # undef in scalar context
        LogReturnSub();
        return wantarray ? () : undef;
    }
    else {

        #---------------------------------------------------
        # function finished before timeout
        #---------------------------------------------------
        LogDebug "function finished successfully";

        # Signal the timeout thread to terminate, and then detach
        # it so that it will get cleaned up automatically
        $timeout_thr->kill('KILL')->detach();

        # When the thread finishes, ->join() will return the
        # return value(s) of the entry point function.

        # see http://rt.perl.org/rt3/Public/Bug/Display.html?id=70602
        LogDebug "Scalars leaked - call_func_with_timeout_thr - before join" ;
        my (@func_results) = $work_thr->join();
        LogDebug "Scalars leaked - call_func_with_timeout_thr - after join" ;

        LogDebug "func_results: ", join( '/', @func_results );

        # return results in list or scalar context. In list return
        #  undef if no result, in scalar return 1
        LogReturnSub();
        if (wantarray) {
            return @func_results ? @func_results : (undef);
        }
        else {
            if (@func_results) {
                return defined( $func_results[0] ) ? $func_results[0] : 1;
            }
            else {
                return 1;
            }
        }
    }
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         exec_function()
 # PURPOSE:      Execute a function in a thread
 # PARAMETERS:   $q            = queue - logs return status of executed function
 #               $function_ref = reference to a function 
 #               @params       = parameters of function
 # RETURNS:      
 # COMMENTS:     
 #============================================================================

=cut

sub exec_function
{
	my ($q, $func_ref, @params) = @_;

    # Setup thread 'cancellation' signal handler
	$SIG{'KILL'} = sub { threads->exit(); };

	# execute function
	my (@output) = &$func_ref(@params);

    # execute function finished successfullyi - put status in queue
	$q->enqueue('success');

	return (@output);
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         do_timeout()
 # PURPOSE:      Setup a timeout function in a thread
 # PARAMETERS:   $q            = queue - logs return status of timeout thread
 #               $timeout      = time after which tiemout is triggered
 # RETURNS:      
 # COMMENTS:     
 #============================================================================

=cut

sub do_timeout
{
	my ($q, $timeout) = @_;

    # Setup thread 'cancellation' signal handler
	$SIG{'KILL'} = sub { threads->exit(); };

	# sleep for $timeout time
	select(undef, undef, undef, $timeout);

    # timeout got triggered - put timeout status in queue
	$q->enqueue('timeout');
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         search_enqorpkg
 # PURPOSE:      Read the cluster config and check if a enqor MNP package
 #               is configured. 
 # PARAMETERS:   1. Reference to array containing cmviewcl output
 #               2. Basename of enqor monitor (usually sapenqor.mon)
 # RETURNS:      Name of enqor pkg if configured, undef otherwise
 # COMMENTS:     Works for sapgetconf and deploysappkgs
 #               (not used by sapgetconf yet)
 #               Expects output to be in 
 #                 cmviewcl -v -f line -l package
 #               format
 #               Currently the search criteria are
 #               type=multinode AND service_cmd contains sapenqor.mon
 #               With the usage of the enqor module this can be changed
 #               to check for the presence of the sgesap/enqor module in pkg.
 #
 #============================================================================


=cut

sub search_enqorpkg {
  my $linesArr = shift;  # Ref to Array
  my $enqormon = shift;
  my $multinode_pkg;
  foreach (@{$linesArr}) {
  	# package:enqor|type=multi_node
    if ( $_ =~ m#^package:(\w+)\|type=multi_node$# ) {
    	$multinode_pkg = $1;
        next;
    }
    # package:enqor|service:enqorS7S_mon|command=$SGCONF/monitors/sgesap/sapenqor.mon
    if ( $_ =~ m#^package:(\w+)\|service:([0-9A-Za-z_.\-]+)\|command=.+/$enqormon$# ) {
    	return ($1) if ($multinode_pkg = $1);
    } 	
  }
  return (undef);	
}


=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         read_script_conf_file()
 # PURPOSE:      if it exists read in a script config file. Can contain 
 #               additional debugging and verbose params. 
 # PARAMETERS:   
 # RETURNS:      
 # COMMENTS:     Works for sapgetconf and sapverify.
 #               Script config file not required for normal use.
 #               Any user environment ENV{} params set will override config 
 #               file params.
 #               Note: No SG package params used yet.
 #============================================================================


=cut

sub read_script_conf_file {

    my $key;    

    #---------------------------------------------------
    # check if a config file has been specified in the
    # user environment. if not check if parameter
    # $params{"${SCRIPT}_CONF"} was set by script. if not
    # use default config file e.g. in the same directory
    # the binary is located:
    #   sapverify  -> sapverify.conf
    #   sapgetconf -> sapgetconf.conf
    #---------------------------------------------------

    $key = "${SCRIPT}_CONF";
    $conf_file = $ENV{ "$key" }                 if defined $ENV{ "$key" };
    $conf_file = $params{"${SCRIPT}_CONF"}      if defined $params{"${SCRIPT}_CONF"};
    $conf_file = $dirname . $basename . ".conf" unless defined $conf_file;

    #---------------------------------------------------
    # use perl's require to init runtime environment from 
    # the optional $conf_file. 
    #---------------------------------------------------

    if (-s $conf_file) { 
        if ( eval qq{ require "$conf_file" } ) {
            $conf_file_executed="success";
            LogExtra("SAP_functions::read_script_conf_file() Command succeeded: [require \"$conf_file\"]");
        }
        else {
            $conf_file_executed="failure";
            LogExtra("SAP_functions::read_script_conf_file() Command skipped: [require \"$conf_file\"]");
        }
    }
    LogExtra("read_script_conf_file() - Done scanning conf_file=[$conf_file]");

    # 1/0  = on/off more verbose output
    $key = "${SCRIPT}_VERBOSE";
    $params{verbose} = $ENV{ "$key" }               if defined $ENV{ "$key" };
#    $params{verbose} = $ENV{ "$key" }               if ((defined $ENV{ "$key" }) and (!defined $params{verbose})) ;
    $params{verbose} = 0                            unless defined $params{verbose};

    # 1/0  = trace ENTER/RETURN of subroutine calls
    $key = "${SCRIPT}_TRACESUBS";
    $params{tracesubs} = $ENV{ "$key" }             if defined $ENV{ "$key" };
    $params{tracesubs} = 0                          unless defined $params{tracesubs};

    # 1/0  = dump data structures
    $key = "${SCRIPT}_DUMPDATA";
    $params{dumpdata} = $ENV{ "$key" }             if defined $ENV{ "$key" };
    $params{dumpdata} = 0                          unless defined $params{dumpdata};

    # 1/0  = on/off debugging output
    $key = "${SCRIPT}_DEBUG";

    # if not early debugging enabled
    if (!defined $params{debug}) {
    $params{debug} = $ENV{ "$key" }                 if defined $ENV{ "$key" };
    }
    $params{debug} = 0                              unless defined $params{debug};

    # timeout this script after $params{timeout} seconds
    $key = "${SCRIPT}_TIMEOUT";
#    $params{timeout} = $ENV{ "$key" }               if defined $ENV{ "$key" };
    $params{timeout} = $ENV{ "$key" }               if ((defined $ENV{ "$key" }) and (!defined $params{timeout})) ;
    $params{timeout} = 60                           unless defined $params{timeout};

    # fork a child (that can be killed) to prevent hung NFS accesses
    $key = "${SCRIPT}_FORK";
    $params{fork} = $ENV{ "$key" }                  if defined $ENV{ "$key" };
    $params{fork} = 1                               unless defined $params{fork};

    # run in testmode
    $key = "${SCRIPT}_TESTMODE";
    $params{testmode} = $ENV{ "$key" }              if defined $ENV{ "$key" };
    $params{testmode} = 0                           unless defined $params{testmode};

    # allow extended/advanced features
    $key = "${SCRIPT}_EXTENDED";
    $params{extended} = $ENV{ "$key" }            if defined $ENV{ "$key" };
    $params{extended} = 1                         unless defined $params{extended};

    # map warnings to errors -e.g all errors are flagged as warnings 
    # internal use only
    $key = "${SCRIPT}_WARNINGMODE";
    $params{warningmode} = $ENV{ "$key" }            if defined $ENV{ "$key" };
    $params{warningmode} = 1                         unless defined $params{warningmode};

    #  API/Features (allow external overwrite)
    $key = "${SCRIPT}_APIVERSION";
    $params{apiversion} = $ENV{ "$key" }            if defined $ENV{ "$key" };
    $params{apiversion} = 0                         unless defined $params{apiversion};

    #  enable SGMgr output for services (not in 5.10 release and PHSS_40911)
    $key = "${SCRIPT}_SERVOUT";
    $params{servoutsgmgr} = $ENV{ "$key" }          if defined $ENV{ "$key" };
    $params{servoutsgmgr} = 1                       unless defined $params{servoutsgmgr};

    # check if output should be logged to a file?
    $key = "${SCRIPT}_LOGGING";
    $params{logging} = $ENV{"$key"}                if defined $ENV{"$key"};
    $params{logging} = 0                           unless defined $params{logging};

}


=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:         getcmclusterconf()
 # PURPOSE:      get system wide Serviceguard config from /etc/cmcluster.conf
 # PARAMETERS:   
 # RETURNS:      cmcluster parameters in params{} hash 
 # COMMENTS:     
 #============================================================================


=cut

sub getcmclusterconf {
    LogEnterSub(@_);

    LogExtra "getcmclusterconf() reading /etc/cmcluster.conf";
    open( $CMCONF_FH, "<", "$CMCONF" ) or confess "Error: Can't open [$CMCONF]: $!";

    while (<$CMCONF_FH>) {

        ## skip uninteresting lines
        chomp;       # no newline
        s/#.*//;     # no comments
        s/^\s+//;    # no leading white
        s/\s+$//;    # no trailing white
        next unless length;    # anything left?

        if (/^\s*(\w+)\s*=\s*(.+)$/) {
#            LogDebug "$CMCONF one=[$1] two=[$2]";
            $params{$1} = $2;
        }
    }

    # override /etc/cmcluster.conf params if user environment is set
    $params{SGROOT}=$ENV{SGROOT}                        if defined $ENV{SGROOT};
    LogRuntimeError("Couldn't find SGROOT in $CMCONF")  if !defined $params{SGROOT};

    $params{SGLIB}=$ENV{SGLIB}                          if defined $ENV{SGLIB};
    LogRuntimeError("Couldn't find SGLIB in $CMCONF")   if !defined $params{SGLIB};

    $params{SGSBIN}=$ENV{SGSBIN}                        if defined $ENV{SGSBIN};
    LogRuntimeError("Couldn't find SGSBIN in $CMCONF")  if !defined $params{SGSBIN};

    $params{SGLBIN}=$ENV{SGLBIN}                        if defined $ENV{SGLBIN};
    LogRuntimeError("Couldn't find SGLBIN in $CMCONF")  if !defined $params{SGLBIN};

    $params{SGCONF}=$ENV{SGCONF}                        if defined $ENV{SGCONF};
    LogRuntimeError("Couldn't find SGCONF in $CMCONF")  if !defined $params{SGCONF};

    $params{SGRUN}=$ENV{SGRUN}                          if defined $ENV{SGRUN};
    LogRuntimeError("Couldn't find SGRUN in $CMCONF")   if !defined $params{SGRUN};

    # export the cmcluster config into user environment for forked processes...
    $ENV{SGROOT}=$params{SGROOT};
    $ENV{SGLIB}=$params{SGLIB};
    $ENV{SGSBIN}=$params{SGSBIN};
    $ENV{SGLBIN}=$params{SGLBIN};
    $ENV{SGCONF}=$params{SGCONF};
    $ENV{SGRUN}=$params{SGRUN};
    
}

=head1 PURPOSE

 #===  FUNCTION  =============================================================
 # NAME:        setup_script_runtime()
 # PURPOSE:     
 # PARAMETERS:  
 # RETURNS:      
 # COMMENTS:    will get and set the runtime environment for sapgetconf, 
 #              deploysappkgs, sapverify 
 #============================================================================

=cut

sub setup_script_runtime {
    LogEnterSub(@_);

    LogExtra("setup_script_runtime() \@_[@_]");

    my $tmp_f = "$params{SGCONF}/sapverify.rules";
    $key = "${SCRIPT}_RULES";
    $params{RULESF} = $ENV{ "$key" }                if defined $ENV{ "$key" };
    $params{RULESF} = $tmp_f                        unless defined $params{RULESF};

    if ($params{logging}) {

        # get/set the log file for STDOUT and STDERR
        $key = "${SCRIPT}_LOG";
        if (defined $ENV{ "$key" } ) {
            $params{LOGF} = $ENV{ "$key" } 
        }
        else {
            if (-d "$params{SGRUN}/log/") {
                if (($basename eq "sapverify") && (defined $params{sgpkg})) {
                    $params{LOGF} = "$params{SGRUN}/log/$params{sgpkg}.log";
                } else {
                    $params{LOGF} = "$params{SGRUN}/log/$basename.log";
                }
            } else {
                 $params{LOGF} = "$basename.log";
            }
        }

        LogDebug("\$params{LOGF} [$params{LOGF}]");
        OpenLogFile($params{LOGF});
    }

    #---------------------------------------------------
    # logfile $LOG_FH is finally open. to be consistent and
    # and since this is the "first" subroutine getting 
    # executed, log a "delayed" LogEnterSub() anyway
    #---------------------------------------------------

    LogEnterSub(@_);

    #establish signal handler for warnings after logfile was opened
    $SIG{__WARN__} = \&warning_handler;

    LogVerbose("Selected configuration variables:");
    LogVerbose("node=[$node]");
    LogVerbose("cwd=[$cwd]");

    LogVerbose("conf_file_executed=[$conf_file_executed]");
    LogVerbose("conf_file=[$conf_file]")               if ($conf_file_executed ne "none");

    LogVerbose("debug=[$params{debug}]");
    LogVerbose("verbose=[$params{verbose}]");
    LogVerbose("timeout=[$params{timeout}]");
    LogVerbose("logging=[$params{logging}]");
    LogVerbose("warningmode=[$params{warningmode}]");
#    LogVerbose("fork=[$params{fork}]");
#    LogVerbose("testmode=[$params{testmode}]");

    LogVerbose("apiversion=[$params{apiversion}]")     if ($basename eq "sapgetconf");
    LogVerbose("servoutsgmgr=[$params{servoutsgmgr}]") if ($basename eq "sapgetconf");
 
    LogVerbose("SAPVERIFY_RULES=[$params{RULESF}]")    if ($basename eq "sapverify");
    LogVerbose("LOGF=[$params{LOGF}]")                 if $params{logging};

    LogVerbose("SGSBIN=$params{SGSBIN}");
    LogVerbose("SGRUN=$params{SGRUN}");
    LogVerbose("SGCONF=$params{SGCONF}");

    LogExtra ( Data::Dumper->Dump( [ \%ENV ], ["ENV"] ));
    LogExtra ( Data::Dumper->Dump( [ \%params ], ["params"] ));

    LogReturnSub();
    return();
}

#============================================================================
# Don't delete the following line
#============================================================================
1;
