#! /usr/bin/perl -w
#
my $WHATSTRING=
q("@(#) SGeSAP/LX - 12.20.00 - $Rev: 1819 $ - Last modified 05/11/17");
#
# **********************************************************************
# *                                                                    *
# * SGeSAP Easy Deployment Support                                     *
# *                                                                    *
# * (c) Copyright 2011-2017 Hewlett Packard Enterprise Co.             *
# *                                                                    *
# * Description: Script to be called by SGMgr (or commandline)         *
# *              to provide easy deployment function for SGeSAP        *
# *                                                                    *
# * Parameters:  see Usage                                             *
# *                                                                    *
# **********************************************************************

# NOTE:
# Messages on stdout of the from
# <some text> = <list of packages>
# are consumed by the program calling this script
# (usually SGMgr, deploysappkgs itself).
# They should be modified with care (if at all)
#
use strict;
use Getopt::Long qw (:config no_ignore_case); # qw(GetOptions HelpMessage :config bundling no_ignore_case);
use Data::Dumper;
use Pod::Usage;
use List::Util qw(first);
use Sys::Hostname;
use File::Basename;

# use generic SG easy deploy tools package (only for screen_and_log)
use lib '/etc/cmcluster/lib';
my $EDToolsmod="EDTools";

# SAP_functions module.
# Three different locations. HP-UX,SLES RHEL
use lib qw(
   /opt/cmcluster/sap/lib
   /opt/cmcluster/lib
   /usr/local/cmcluster/lib
);
use SAP_functions;

my $host = hostname(); # not exported by SAP_functions.pm

my $USAGE = <<"END_USAGE";
Usage:  \${SGSBIN}/deploysappkgs [-a] [-i rw_ip[,ro_ip]] [-o option(s)] combi|multi <SAP-SID>
  combi: combine entities into minimal number of packages
  multi: put entities into maximal number of packages possible
  -a: immediately apply generated packages (w/o confirmation)
  -o: possible options are (commma separated)
        smart_quorum (also to change assignment)
        safesync
  -i: client ip (or host) (range start) for primary HANA (and secondary HANA)
      Examples: -i=hostA
                -i=10.3.1.61,10.3.1.66
END_USAGE

my $theapp="deploysappkgs";  # could use basename or similar
my $sgeasyTool = undef;
sub printL; # forward declaration

# EDTools.pm might no be there (Linux). Just eval it and check output.
eval "use $EDToolsmod";
if ($@ =~ /^$/) {
  printL "$EDToolsmod found\n";
  $sgeasyTool = new $EDToolsmod($theapp);
}
else {
  printL "$EDToolsmod not found\n";
  # leaves $sgeasyTool undefined
}

=head1 NAME

deploysappkgs -  deploy one or multiple SGeSAP package(s)

=cut


=head1 SYNOPSIS
deploysappkgs multi|combi SAPSID

old style (obsolete)
deploysappkgs -n package name/prefix -s SAPSID  [-o "options string"]
=cut

# "Options"

my $testmode = 0;
# So far testmode only allows "fake" input of
# - cmviewcl -f line -v -l package  (cmviewcl.dspt)
# - sapgetconf -f scs sap_instance <SID> (sapgetconf.fline.dspt)
# This allows to influence package selection mechanism
# Output of "cmgetconf -p <pkg>" and "sapgetconf easy ..." aren't
# faked so far.

my $localhdb = 0;
# localhdb is used to make sure invoking this script on other nodes
# won't in turn reinvoke again (only used in HANA context)

my $applypkgs = 0;
my $checkpkgs = 0;
# immediately apply/check packages using cmappyly/checkconf.
# Triggered by -a (or --apply) or -c (--check)
my $special_opts = "";
# special option(s), string. Triggered by -o (or --option)
my $ip_opts = "";
# ip option(s), string. Triggered by -i (or --ip)

# end of "Options"

# values returned from cmgetconf
my @pkgcmconf = ( );  # array of name/value tuples. sequence as read in from file
my %pkgcmhash = ( );  # hash for each name containing value(s) as array
# values returned from sapgetconf, same format as array and hash above
my @pkgsapconf = ( );
my %pkgsaphash = ( );

my @deployxtractions = ();  # filled in creatpkgconfig

my @modifiedPkgs=();
my @modifiedPkgsStatusUp=();

my @clusnodes=();

my $pkgname;
my $pkgprefix;
my $sapsid;
my $dbsid; # SIDs may be different (esp. for HDB)
my $multipkg=0;   # 1 if multiple packages should be created, 0 if combined

my $pkgpath=($ENV{"SGCONF"}||"/etc/cmcluster")."/sdpkg-"; # used for generated configs (this is the prefix)

my @addmods=();  # additional modules needed by discovered sapgetconf attributes
my $nfsmodadded=0; # indicator if nfs/hanfs module was added to the package
# indicator file if hanfs is installed residing in /etc/cmcluster/modules
my $nfsmodname="nfs/hanfs";
$nfsmodname="tkit/nfs/nfs" if ($^O eq "linux");

# Flag to see if repdb module is installed. Only in this
# case replicated db (repdbinstance for sybase replication)
# will be queried with sapgetconf
my $repdbmodname=$ENV{"SGCONF"}."sgesap/repdbinstance";
my $repdb_enabled=0;
$repdb_enabled=1 if (-f $repdbmodname);
# Turn off unconditionally. Support was never announced
$repdb_enabled=0;

my $arridx = 0;
my $aref;

# commands to get the config of an existing package and to collect the config
# information for a give SAP SID
my $cmd_cmgetconf=qq[$ENV{"SGSBIN"}/cmgetconf -v 0 -p];  # packagename to be appended

# Get location of sapgetconf. Will fall through until found. Uses the link
# and relies on environment variable settings (need to be exported!).
# But first look into same directory where deploysappkgs is run from.
#
my $cmd_sapgetconf=dirname($0)."/sapgetconf"; # as link
$cmd_sapgetconf = $ENV{PWD}."/".$cmd_sapgetconf unless (substr($cmd_sapgetconf,0,1) eq "/") ; # make absolute
$cmd_sapgetconf .= ".pl" unless -x $cmd_sapgetconf; # as script
printL "--- Same dir sapgetconf: $cmd_sapgetconf\n" if -x $cmd_sapgetconf;
# .. else: look into OS specific paths
$cmd_sapgetconf=qq[/opt/cmcluster/sap/bin/sapgetconf] unless -x $cmd_sapgetconf;  # HP-UX link

$cmd_sapgetconf=qq[$ENV{"SGROOT"}/sap/bin/sapgetconf] unless -x $cmd_sapgetconf; # Linux link 1st
$cmd_sapgetconf=qq[$ENV{"SGSBIN"}/sapgetconf] unless -x $cmd_sapgetconf; # Linux 2nd
$cmd_sapgetconf=qq[sapgetconf] unless -x $cmd_sapgetconf; # rely on PATH
my $cmd_sapgceasy=qq[$cmd_sapgetconf easy ]; #append sapsid=SID[sapinstance=INST]*,dbinstance=SID

# Command to get sap config information in "-f line" format. Used to get list of instances and their associated IP addresses.
# Used (a) to make sure SCS and ERS have different IPs and (b) to find "matching" packages based on the configured IP
# "-f SCS" hardcoded so far (TBD: provide as program option)
 # replace "THESID" before calling
my $cmd_sapgcinst_filter="-f scs";
#my $cmd_sapgcinst_filter="";
my $cmd_sapgcinst=qq[$cmd_sapgetconf $cmd_sapgcinst_filter sapinstance \$THESID ; $cmd_sapgetconf dbinstance \$THESID];
# append repdb related parameters
$cmd_sapgcinst .= qq[ repdbinstance \$THESID] if ($repdb_enabled);
# for getting nfs config ("pattern" can be added as 2nd param)
my $cmd_sapgcnfs=qq[$cmd_sapgetconf nfs ];

# determine path to nwrksupp binary
my $nwrksupp=dirname($0)."/nwrksupp"; # as link
$nwrksupp=qq[/opt/cmcluster/sap/bin/nwrksupp"] unless -x $nwrksupp;  # HP-UX link
$nwrksupp=qq[$ENV{"SGROOT"}/sap/bin/nwrksupp] unless -x $nwrksupp; # Linux link 1st
$nwrksupp=qq[$ENV{"SGSBIN"}/nwrksupp] unless -x $nwrksupp; # Linux 2nd

#############################################################################
#
# SAP instance related datastructures
#
#############################################################################

# defined above:
#my @pkgsapconf = ( );
#my %pkgsaphash = ( );

# The assocated hash filled by cmd above.
# Instance names are the hashkeys, each hashvalue is an array with
# 0 : replicated instances for ERS, undef otherwise
# 1:  virtual hostname (or undef, also undef if configured to a physical host of the cluster)
# 2..n: list of IPs
# ToDo:
# - List of IPs should go into a separate arry
# - List of used VG/DGs should be added
# - undef for physical host (probably) not implemented yet
my %sapgcinsthash=();

# Separate hash to hold information returned by "sapgetconf nfs".
# 0: export spec (string, tbd)
# 1: virtual hostname (or undef)
# 2: list of IPs (array)
# 3: list of VGs (array, tbd)
# 4: list of MPs (array, tbd)
my %sapgcnfshash=();

# Array containing arrays (or hashes?) of SAP instances
# sharing one or more resource(s) (so far only IPs are considered, VGs later)
# This array will be used to figure out which instances should/can
# go into a common package
# Can also be used (later) to check for illegal resources sharing
# between SCS and ERS type of SAP instances.
my @sapinstshrdres = ();

#############################################################################
#
# Package related datastructures
#
#############################################################################

# defined above
#my @pkgcmconf = ( );
#my %pkgcmhash = ( );

# Hash for each package referencing an array.
# Package occurs in hash if one of the following
# elements are in the package
# The array contains at index (have to be unique, use these constants ony
# to index array)
# 0: package style (modular/legacy)
# 1: package status (up/down)
# 2: sapsid if in package, undef otherwise
# 3: dbsid if in package, undef otherwise
# 4: reference to array of configured SAP instances, if any, undef (or empty?) otherwise
# 5: reference to array of configured SAP virt. hosts, if any
# 8: hdbsid if in package, undef otherwise
# 9: reference to array of configured HDB instances in package (if any)
# 6: reference to an ip subnet hash (contains configured IP subnet/addresses)
# 7: active IPs array
# 10: nodes package is configured to (node_name) (usually only releveant if all_cluster_nodes
#       is not true in cmviewcl outout) (ref to arry)

my %cluspkgconfighash=();
# Symbolic names for index into array
# (DB virtual hostname - attribute doesn't exists in module)
my $pkgstyle_idx = 0; # modular (or legacy)
my $pkgstatus_idx = 1;
my $sapsid_idx   = 2;
my $dbsid_idx    = 3;
my $sapinst_idx  = 4;
my $sapvhost_idx = 5;
my $cip_idx      = 6;
my $aip_idx      = 7;
my $hdbsid_idx   = 8;
my $hdbinst_idx  = 9;
my $pkgnodelist_idx = 10;
my $pkgtype_idx = 11; # failover or multi_node
#
# The ip subnet hash contains the subnet(s) as key(s)
# and an array of corresponing ip addresses as values
#
# Not sure if we really need mark packages with style "legacy"
# as they won't have all the other attributes anyway.


# cluster resource hash cross reference: "resource used by package"
# contains (symbolic names for) resource type (VG, FS name, FS mountpoint,...)
# as keys and another hash as value. That hash has the resource name as key and
# the package(s) where used as value.
my %clusrsrchash=();

# Pkg IPs to pkg(s) map (created from  %cluspkgconfighash above)
# key: IP, value: array of pkg(s) (usually 1 entry)
# (obselete with introduction of %clusrsrchash ??)
my %clusip2pkgmap = ();

# package candidates for instances. Keys: instancename, values: array of packages and their
# selection criteria. Each package in such an array allocate two indices:
# 1st: package name, 2nd: (array of)selection criteria
#(Instance are mapped to packages via IP addresses)
my %instPkgCandidates=();


# selection criteria (indexes) (so far the selection goes upwards from 0 to
my $select_on_IP_idx       = 0; # package selected based on IP
my $select_on_multIP_idx   = 1; # package selected because it contains more multiple necessary IPs (future use)
my $select_on_noSID_idx    = 2; # package selected because it doesn't contain a SAP SID (but a matching IP)
my $select_on_sameSID_idx  = 3; # package selected based on (same) SAP SID  (unrelated to IP)
my $select_on_sameInst_idx = 4; # package selected based on (same) SAP instance (implies "sameSID", also unrelated to IP)
my $select_on_SIDInst_idx  = 5; # package selected based on instance of the same SAP SID (contains also 4 if not empty)
my $select_on_sameDBSID_idx = 6; # DB SID is the same as given SID (may also be in sameSID). TODO: use for HDB also - or creata a new??
my $select_on_VHost_idx    = 7; # selection done based on SAP virtual hostname
my $select_on_combined_idx = 8; # package selected on predicates applied to the previous sets
my $select_shared_res      = 9; # contains instance which need to be grouped together because of resource sharing (e.g. IP)
my $select_preferred       = 10; # preferred other inst to group with (priority list)
my $select_final           = 11; # final selection
my $select_on_hdbrepl      = 12; # new entry: select on HDB replication setup
my $select_on_nodelist     = 13; # pkg contains the required nodelist (multinode pkgs only so far)
my $select_on_pkgtype      = 14; # pkg is of required package type (failover, multi_node,..)

# Remarks:
# - a package won't/should probably end up in $select_on_sameDBSID_idx if DBSID=!SAPSID
#   (although such a setup can be found by sapgetconf)
# - $select_shared_res also contains the instance itself (may be the only member)
# - $select_preferred contains the other instances the current instance prefers to be grouped with (in case of "combi").
#   This is esp. important to avoid arbitrary assignments in case more than one package can be choosen (e.g. SCS and DB
#   packages already exist and a new ASCS is to be assigned to a package. In this case the ASCS might end up at the DB,
#   but SCS should be preferred).

# Map of candidate packages to instance(s)
# Reverses the instPkgCandidates after final package cand. has been choosen
my %CandPkg2Inst=();

# -----
# cmd to fill the hash (cmd complete, removed -l package scope when metrocluster sites need to be supported)
my $cmd_cmviewclpkg=qq[$ENV{SGSBIN}/cmviewcl -f line -v];
#$cmd_cmviewclpkg=qq[cat /tmp/cmviewcl.tmp] if $testmode;
# this output is filled into an array (read by more than one sub)
my @linescmviewcl=();

# ----

my $updconffile;  # file containing an updated or newly created config

# The attributes configured in the next two hashes are considered to be
# the "key attributes" (many of them are parent attribute) acting as a
# kind of trigger. All these are also the attributes returned by sapgetconf,
# so if sapgetconf is extended, these list might need to be extended.
# All of these key attributes will trigger (require) a special handling routine

#
# The code assumes that in output of cmgetconf and sapgetconf the parent
# attribute is placed before the (group of) children/related attributes
#

# attribute => module file map on selected/parent attributes
my %SGKeyAttr = (
#  "package_name"  => "sg/basic",
  "package_description"  => "sg/basic",
  "node_name"            => "sg/basic",
  "script_log_file"      => "sg/basic",
  "fs_name"         => "sg/filesystem",
  "ip_subnet"       => "sg/package_ip",
  "resource_name"   => "sg/resource",
  "generic_resource_name"   => "sg/generic_resource",
  "service_name"    => "sg/service",
  "dependency_name" => "sg/dependency",
  "priority"        => "sg/priority",
  # the following 3 are not really parents of anything
  # but "toplevel" attributes reported by sapgetconf
  "vg"            => "sg/volume_group",
  "cvm_dg"        => "sg/volume_group",
  "vxvm_dg"       => "sg/volume_group",
  # SGeSAP attributes
  # (not all of them parent attributes, but they have "related" attributes)
  # WVE: usually the prefix is the module, so the values are redundant - sometimes)
  "sgesap/sap_global/sap_system" => "sgesap/sapinstance",  # sap_global
  "sgesap/stack/sap_instance"    => "sgesap/sapinstance",
  "sgesap/db_global/db_system"   => "sgesap/dbinstance",  # db_global
  "sgesap/db_global/db_vendor"   => "sgesap/dbinstance",
  "sgesap/repdbinstance/repdb_system" => "sgesap/repdbinstance",
  "sgesap/hdb_global/hdb_system"      => "sgesap/hdbinstance",  # hdb_global, hdbdistinstance
  "sgesap/hdb_global/hdb_userkey"     => "sgesap/hdbinstance",  # hdb_global, hdbdistinstance
  "sgesap/hdb_global/hdb_retry_count" => "sgesap/hdbinstance",  # hdb_global, hdbdistinstance
  "sgesap/hdbinstance/hdb_instance"   => "sgesap/hdbinstance",
  "sgesap/hdbinstance_global/hdb_instance"   => "sgesap/hdbinstance", # 06.00.40
  "sgesap/hdbprimary/hdb_sync_time_tolerance"   => "sgesap/hdbprimary",
  "sgesap/hdbprimary/hdb_sync_logpos_tolerance" => "sgesap/hdbprimary",
  "sgesap/hdbprimary/hdb_init_into_fullsync" => "sgesap/hdbprimary", # 06.00.80
  "sgesap/hdbdistinstance/hdb_ip_subnet"   => "sgesap/hdbdistinstance", # 06.00.40
  # "sgesap/hdbdistinstance/hdb_ip_address"  => "sgesap/hdbdistinstance", # 06.00.40 - not a key attrib
  "lc_system"                    => "sgesap/livecache",
  # NFS attributes
  # HPUX
  "nfs/hanfs_export/FILE_LOCK_MIGRATION" => "nfs/hanfs",
  "nfs/hanfs_export/XFS"       => "nfs/hanfs",
  # Linux
  "tkit/nfs/nfs/TKIT_DIR"  => "tkit/nfs/nfs",
  "tkit/nfs/nfs/XFS"       => "tkit/nfs/nfs",
  # Linux vmfs
  "vmdk_file_name"			=> "sg/vmfs"
  );


# map: key/parent attribute => related attributes
# these attributes need special treatment when they are found while working
# on the list returned by cmgetconf as such attr. are also returned by sapgetconf.
# Not all key attributes have related attributes. Some of them have an empty list
# (esp. if they aren't mandatory), some are not listed (e.g. mandatory from basic)
my %SGRelAttr = (
  "fs_name"       =>
     ["fs_directory","fs_type","fs_mount_opt","fs_umount_opt","fs_fsck_opt","fs_server"],
  "ip_subnet"     => ["ip_address","ip_subnet_node" ],
  "resource_name" => ["resource_start","resource_up_value","resource_polling_interval"],
  "generic_resource_name"
                  => ["generic_resource_evaluation_type","generic_resource_up_criteria"],
  "service_name"  =>
     ["service_cmd","service_restart","service_fail_fast_enabled","service_halt_timeout"],
  "dependency_name" => ["dependency_location","dependency_condition"],
  # The following 3 volume_group attributes are forming a (kind of related) group
  # which should be treated in "one chunk"
  "vg"            => ["cvm_dg","vxvm_dg"],
  "cvm_dg"        => ["vg","vxvm_dg"],
  "vxvm_dg"       => ["vg","cxvm_dg"],
  # SGeSAP attributes relation
  "sgesap/sap_global/sap_system" =>
     ["sgesap/sap_global/rem_comm","sgesap/sap_global/priv_comm",
      "sgesap/sap_global/cleanup_policy","sgesap/sap_global/retry_count",
      "sgesap/sap_global/parallel_startup","sgesap/sap_global/sapcontrol_usage"],

  "sgesap/stack/sap_instance"    =>
     ["sgesap/stack/sap_virtual_hostname","sgesap/stack/sap_replicated_instance"],
  "sgesap/db_global/db_system"   => ["sgesap/db_global/db_vendor"],
  "sgesap/db_global/db_vendor"   => ["sgesap/db_global/db_system"],
  # "sgesap/db_global/dbvendor"   => ["sgesap/db_global/db_system"], -- vendor comes before _system
  "sgesap/repdbinstance/repdb_system" => ["sgesap/repdbinstance/repdb_role"],
  "lc_system"                    =>
     ["lc_virtual_hostname","lc_startmode","lc_userkey"],
  # HDB entries are assumed to start with hdb_system attribute
  # (not with hdb_instance or even hdb_retry_count after _system and _instance became mandatory)
  # hdbinstance_global (.2) and hdbinstance (.1) are considered related, although they can
  # never show up together in a package config. Necessary for handling version 1 and 2 pkg configs
  # (hdbuserkey doesn't exist - anymore )
  # 06.00.50: sto attributes new to this group
  "sgesap/hdb_global/hdb_system" => ["sgesap/hdb_global/hdb_retry_count",
     "sgesap/hdbinstance/hdb_instance", "sgesap/hdbinstance_global/hdb_instance",
     "sgesap/hdb_global/hdb_sto_sample_size", "sgesap/hdb_global/hdb_sto_safety_factor",
     "sgesap/hdb_global/hdb_sto_minimum_threshold",
     "sgesap/hdb_global/hdb_userkey" ],
#  "sgesap/hdb_global/hdb_retry_count" => ["sgesap/hdb_global/hdb_system",
#     "sgesap/hdb_global/hdb_userkey", "sgesap/hdbinstance/hdb_instance",
#     "sgesap/hdbinstance_global/hdb_instance"],
#  "sgesap/hdbinstance/hdb_instance" => ["sgesap/hdb_global/hdb_system",
#     "sgesap/hdb_global/hdb_userkey", "sgesap/hdb_global/hdb_retry_count",
#     "sgesap/hdbinstance_global/hdb_instance"],
#  "sgesap/hdbinstance_global/hdb_instance" => ["sgesap/hdb_global/hdb_system",
#     "sgesap/hdb_global/hdb_userkey", "sgesap/hdb_global/hdb_retry_count",
#     "sgesap/hdbinstance/hdb_instance"],
  "sgesap/hdbprimary/hdb_sync_time_tolerance"   => ["sgesap/hdbprimary/hdb_sync_logpos_tolerance",
     "sgesap/hdbprimary/hdb_init_into_fullsync" ],
  "sgesap/hdbprimary/hdb_sync_logpos_tolerance" => ["sgesap/hdbprimary/hdb_sync_time_tolerance",
     "sgesap/hdbprimary/hdb_init_into_fullsync" ],
  "sgesap/hdbprimary/hdb_init_into_fullsync" => [],  # does not include 2 attrs above
  "sgesap/hdbdistinstance/hdb_ip_subnet"        => ["sgesap/hdbdistinstance/hdb_ip_address"],
  # package description (not really a merge), no related attributes
  "package_description" => [],
  # nfs (may or may not contain XFS, it has to contain FILE_LOCK_MIGRATION)
  # no grouping
  # HPUX
  "nfs/hanfs_export/FILE_LOCK_MIGRATION" => [],
  "nfs/hanfs_export/XFS"                 => [],
  # Linux
  "tkit/nfs/nfs/TKIT_DIR"  => [],
  "tkit/nfs/nfs/XFS"       => [],
  # Linux vmfs
  "vmdk_file_name" => [ "datastore_name", "scsi_controller", "disk_type" ],
  );


#===  FUNCTION  =============================================================
# PURPOSE:      prints stdout to logfile $logf_fh with a timestamp
# PARAMETERS:   @_  arguments
# RETURNS:      None
# COMMENTS:     None
#============================================================================
#
# temporarily taken from sapverify
# $logf_fh  not used yet
#
sub printL {

  my $logf_fh=undef;

  return unless defined $ENV{"deploysappkgs_log"};

#    if ( defined fileno LOGF ) {
    if ( defined $logf_fh ) {
        print $logf_fh scalar localtime(), " @_";
    }
    else {
        print STDERR scalar localtime(), " @_";
    }
}

#############################################################################
#
# sub sgeasy_print80
#
# print out 80 characters per line in SG easy deploypkg style output
# may use "[SUCCESS}", "[FAILURE]" at the end
# may print out dots after left part or between left and right part
#
# use SG screen_and_log from EDTools
#
#############################################################################
sub sgeasy_print80 {
# assumes the rightstgring
  my $leftstring=shift;
  my $dotted=shift;
  my $rightstring=shift;

  my $colwid = 80;
  my $lotsofdots = "." x $colwid;

  $dotted = 0 unless defined $dotted;

  my $lslen = length ($leftstring);

  my $rslen = length("[SUCCESS]") unless defined $rightstring || !$dotted; # fake length
  # this kind of assumes that only  "[SUCCESS]" and "[FAILURE]" might appear as 2nd string
  $rslen = length ($rightstring) if defined $rightstring;

  while ($lslen > $colwid) {
    # $sgeasyTool->screen_and_log (substr($leftstring,0,$colwid),$theapp.":");
	print_and_log (substr($leftstring,0,$colwid),$theapp.":");
    $lslen -= $colwid;
    $leftstring = substr($leftstring,$colwid,$lslen);
  }

  # now leftstring is smaller than 80 characters
  # if undotted, leftstring is ignored and now dots are printed
  $rightstring = "" unless defined $rightstring;
  if (!$dotted) {
    # $sgeasyTool->screen_and_log  ($leftstring.$rightstring);
	print_and_log  ($leftstring.$rightstring);
    return;
  }
  # check if leftstring and rightstring fit in line, otherwise cut it again
  if ($lslen + $rslen > $colwid)
  {
     #$sgeasyTool->screen_and_log ($leftstring.substr($lotsofdots,0,$colwid-$lslen),$theapp.":");
     #$sgeasyTool->screen_and_log (substr($lotsofdots,0,$colwid-$rslen).$rightstring,$theapp.":");
	 print_and_log ($leftstring.substr($lotsofdots,0,$colwid-$lslen),$theapp.":");
	 print_and_log (substr($lotsofdots,0,$colwid-$rslen).$rightstring,$theapp.":");
  }
  else {
     # $sgeasyTool->screen_and_log ($leftstring.substr($lotsofdots,0,$colwid-$lslen-$rslen).$rightstring,$theapp.":");
	 print_and_log ($leftstring.substr($lotsofdots,0,$colwid-$lslen-$rslen).$rightstring,$theapp.":");
  }
}

#############################################################################
#
# sub print_and_log
#
# Should be: print_and_perhaps_log
#
# Helper routine for sgeasy_print80 either using screen_and_log if $sgeasyTool
# is defined (which implies that EDtools was found) or a simply print in
# case no EDTools is available.
#
#############################################################################
sub print_and_log {
 if (defined($sgeasyTool)) {
   $sgeasyTool->screen_and_log($_[0],$_[1]);
 }
 else {
   print "$_[0]\n";
 }
}

#############################################################################
#
# sub parse_args
#
# checks args and sets variables $pkgname to value after "-n"
# and $sapsid to value after "-s"
#
#############################################################################
sub parse_args
{

    if (scalar @ARGV < 2) {
        print ($USAGE);
        return 0;
    }

    # otherwise style: (multi|combi)
    if ($ARGV[0] !~ /^(multi|combi)$/) {
        print ($USAGE);
        return 0;
    }
    $multipkg = 0;
    $multipkg =1 if ($ARGV[0] =~ /^multi$/);

    $sapsid = $ARGV[1];

    $testmode = 1 if (defined $ARGV[2] &&$ARGV[2] =~ /testmode/);
    $localhdb = 1 if (defined $ARGV[2] &&$ARGV[2] =~ /localhdb/);

    return 1;

}

#############################################################################
#
# sub chkrootprivs
#
# Check if current user belongs to root group
#
#############################################################################

sub chkrootprivs {

# getlogin returns undefined if run through cmexec - commented
#  my $curruser = getlogin(); # not the su user!
#  my ($currname, $pass, $curruid, $currgid, $quota, $comment, $gcos, $dir, $shell, $expire) = getpwnam($curruser);

#  printL "current user: $curruser - $curruid, $currgid - $<,$( \n";

  my ($rgrpname, $rgrppasswd, $rgrpgid, $rgrpmembers) = getgrnam('root'); # usually gid 0, but may have members

#  if ($rgrpgid == $currgid) {
#    printL "Current user has group root (/etc/passwd)\n";
#    return 1;
#  }
#  elsif (grep (/^$curruser$/, split (/\s+/,$rgrpmembers))) {
#    printL "$curruser is member of root group\n";
#    return 1;
#  }
  if (grep (/^$rgrpgid$/, split (/\s+/,$( ))) {
     printL "Current user is member of root group\n";
     return 1;
  }
  return 0;
}

#############################################################################
#
# Different input (and check) routines for cmviewcl, cmgetconf, sapgetconf,
#
##############################################################################

#############################################################################
#
# sub populateArray
#
# Read in config (attribute name/value pairs) from filehandle
# and fill array and hash.
#
# Arguments:
#  *FH        File handle to read attributes from
#  \@attrarr   Reference to array to add attritbe name/value pairs to
#  \%attrhash  Reference to hash to add attributes to. Attribute name is
#               hash key, hash value is array containing the attribute values
#
# So far comments aren't preserved. Should go into array.
#
# Returns: nothing
#
#############################################################################

sub populateArray {
  local *FH = shift;

  while (<FH>) {
    chomp;
     if ( $_ =~ m#^(\S+)\s+(.*)$# ) {
        #printL "found nonwhite 1=[$1] 2=[$2]\n";
        # push onto array
        push @{$_[0]}, [ $1 , $2 ];
        # push onto (array in) hash
        if ($1 ne "sgesap/stack/sap_replicated_instance")  {
          push @{${$_[1]}{"$1"}}, $2;
        }
        else {
          # push it on the same position as the last entry in
          # @{${$_[1]}{"sgesap/stack/sap_instance"}} to keep undefined attributes
          # "undef" in array
          my $lastelem=$#{$_[1]->{"sgesap/stack/sap_instance"}};
          @{${$_[1]}{"$1"}}[$lastelem]=$2;
          #
          # So far this mechanism of "sparse arrays" is only used for sap_instance
          # and related attributes as this group is the only one so far having
          # a varying number of attributes per parent.
          # TBD: "sap_virtual_hostname" might also not be defined when input
          #  is coming from cmgetconf
        }
    }
  }
}

#############################################################################
#
# sub fillsapgcHash
#
# Fills in the sapgetconf hash
#
# Arguments:
#  *FH        File handle to read attributes from
#  \%sapgchash  Reference to hash to be filled
#
#  Read input lines and checks them for
#    sgesap/stack/sap_instance:<INST>|sgesap/stack/sap_instance=<INST>)  --> hash key
#    (^^^^^ - this line is expected to appear before the other lines below)
#    sgesap/stack/sap_instance:<INST>|sgesap/stack/sap_replicated_instance=<REPLINST> --> array value index 0 (might stay undef)
#    sgesap/stack/sap_instance:<INST>|sgesap/stack/sap_virtual_hostname=<vhost>  --> array value index 1
#    ip_subnet:<subnet>|ip_address=<addr> --> remaining array values after index 2
#    Extracts <INST> [<REPLINST>] <vhost> <ip_addr>
#    (an instance can have only one vhost, but vhost can/may have multiple IPs)
#
#    For the DB the input is triggered by
#    sgesap/db_global/db_system=<DBSID>
#    All relevant information to appear afterwards
#
#    For replicated DB input is triggered by
#    sgesap/repdbinstance/repdb_system=<DBSID>
#
# This code assumes that only one sap_instance + related attributes appears in sapgetconf output (which is
# the same that only one profile is defined per instance). Otherwise last will win.
# It also assumes that first information of a sapinstance to appear in the output is
# the "...sapinstance:...|.../sapinstance..." line
#
#
# This input is instances per SID.
#
# TBD: should instances configured to (cluster) physical hosts be filtered out already here?
#
# Returns: nothing
#
#############################################################################

sub fillsapgcHash {
  local *FH = shift;
  my $currInst;
  my @ignlist = (); # instances to be ignored
  while (<FH>) {
    # print $_,"\n";
    # Doesn't filter out unsupported instances (that is != ASCS,SCS,ERS,[DVEBMGS])
    # as the filter used with sapgetconf is supposed to do this.
    # There are case when an instance appears twice (i.e. when it has profiles
    # configured with different (v)hosts). This can't be supported here and such
    # instance will be ignored.
    if ( $_ =~ m#^sgesap/stack/sap_instance:(\w+)\|sgesap/stack/sap_instance=(\w+)$# ) {
      if (defined $_[0]->{"$1"}) {
        sgeasy_print80 ("Instance $1 already encountered with $_[0]->{$1}->[1]");
        sgeasy_print80 ("Instance will be ignored");
        push @ignlist,$1;
        next;
      }
      $_[0]->{"$1"} = [undef,undef]; # ReplINST and vhost undefined yet
      $currInst=$1;
      next;
    }
    if ( $_ =~ m#^sgesap/stack/sap_instance:(\w+)\|sgesap/stack/sap_replicated_instance=(\w+)$# ) {
      $_[0]->{$currInst}->[0] = $2;
      $currInst=$1;
      next;
    }
    if ( $_ =~ m#^sgesap/stack/sap_instance:(\w+)\|sgesap/stack/sap_virtual_hostname=((\w|-)+)$# ) {
      if (defined $_[0]->{"$1"}->[1]) {
        sgeasy_print80 ("Instance $1 has virtual host $2, but $_[0]->{$1}->[1] defined");
        sgeasy_print80 ("Instance will be ignored");
        push @ignlist,$1;
        next;
      }
      $_[0]->{$currInst}->[1] = $2;
      next;
    }
    if ( $_ =~ m#^sgesap/db_global/db_system=(\w+)$# ) {
      # Check db vendor (expected to be on next line) and create a different key for HANA
      my $dbkey="DB-$1";
      #$_ = <FH>; # get next line
      #printL "$_";
      $dbkey = "H".$dbkey if ( <FH> =~ m#^sgesap/db_global/db_vendor=hana$# );
      $_[0]->{$dbkey} = [undef,undef]; # ReplINST and vhost undefined (could use ReplINST to hold DBSID)
      $currInst=$dbkey ; # start "DB instance"
      # usually $1 (the DB SID) is the same as the SAP SID, but to accomodate cases where they are different
      # it is embedded into the "instance name".
      next;
    }
    # get "replication attribute" and add as first entry in list (usually undef for DBs) (primary/sync)
    if ( $_ =~ m#^sgesap/hdb_global/_hdb_replmode=(\S+)$# ) {
      #printL "$_"; # values primary,sync or none.
      $_[0]->{$currInst}->[0] = $1;
      # do an addition of 'p' or 's'. requires creation of a new instance name and deletion of the old one in the hash
      # Starting with HANA scaleout support (sapgetconf rev >1264), this might be comma sep list of replmode,siteid,sitename
      if ((split(',',$1))[0] =~ /(^primary$|sync)/) {
        my $oldInst = $currInst;
        $currInst = "HDB".substr($1,0,1).substr($oldInst,-4);
        $_[0]->{$currInst} = $_[0]->{$oldInst};
        delete  $_[0]->{$oldInst};
      }
      next;
    }
    # get HANA nodes (sapgetconf rev > 1264)
    if ( $_ =~ m#^sgesap/hdb_global/_hdb_nodes=(\S+)$# ) {
    	# either a single node (scale-up)or a comma separated list of nodes (scale-out)
    	# not split up at this time. just put into vhost field
    	$_[0]->{$currInst}->[1] = $1;
    	next;
    }
	if ( $_ =~ m#^sgesap/repdbinstance/repdb_system=(\w+)$# ) {
      $_[0]->{"REP-$1"} = [undef,undef]; # ReplINST and vhost undefined (could use ReplINST to hold DBSID)
      $currInst="REP-$1"; # start "DB instance"
      # usually $1 (the DB SID) is the same as the SAP SID, but to accomodate cases where they are different
      # it is embedded into the "instance name".
      next;
      # There's no "db_virtual_hostname" (vhost) , just one or more IP addresses. This routine could make up the vhost
      # because sapgetconf can't report one as there's no official attribute for it. Would involve reverse lookup.
      # As long as there's no "db_virtual_hostname" it won't be possible to select a package based on that vhost.
    }
    if ( $_ =~ m#^ip_subnet:(\S+)\|ip_address=(\S+)$# ) {
      push @{$_[0]->{$currInst}},$2;
      next;
    }
  }

  # remove DVEBMGS instance if there is also a ASCS
  my $ascsinst;
  my $dvebinst;
  foreach my $inst (keys %{$_[0]}) {
    if ($inst =~ "^ASCS") { $ascsinst = $inst ; next; }
    if ($inst =~ "^DVEBMGS") { $dvebinst = $inst ; next ; }
  }
  if (defined $ascsinst && defined $dvebinst) {
    sgeasy_print80 ("Both $ascsinst and $dvebinst defined.");
    sgeasy_print80 ("Instance $dvebinst will be ignored");
    delete $_[0]->{$dvebinst}
  }

  # remove instances to be ignored (may have duplicates in ignlist)...
  # (done after ASCS/DVEBMGS check otherwise DVEBMGS would "survive"
  # if ASCS is on the "to be ignored" list)
  foreach my $inst (@ignlist) {
    delete $_[0]->{$inst} if defined $_[0]->{$inst};
  }
}

#############################################################################
#
# sub fillsapgcNFSHash
#
# Fills in the sapgetconf NFS hash
#
# Arguments:
#  *FH        File handle to read attributes from
#  \%sapgchash  Reference to hash to be filled
#
# Expects hanfs_XFS to be the first entry of the group of resources (IP,VG,FS)
#
# Indexes used
# 0: export spec (string, tbd)
# 1: virtual hostname (or undef)
# 2: list of IPs (array)
# 3: list of VGs (array, tbd)
# 4: list of MPs (array, tbd)
#
#############################################################################

sub fillsapgcNFSHash {
    local *FH = shift;
    my $currNFS;

    while (<FH>) {
        #printL $_,"\n";
		# HP-UX
        if ( $_ =~ m#^nfs/hanfs_export/XFS:\"(.+)\"\|nfs/hanfs_export/XFS=\"(.+)\"$# ) {
            # This string is also assumed to be enclosed in doublequotes
            # $1 and $2 should contain the same string now, but this isn't checked here.
            my @parts = split(/\s+/, $1);
            #printL "$1\n";
            $currNFS=$parts[$#parts];
            #printL "$currNFS \n";

            $_[0]->{"$currNFS"} = [undef,undef,[],[],[]]; # ReplINST and vhost undefined yet
            $_[0]->{"$currNFS"}->[0] = $1;
            next;
        }
		# Linux
        if ( $_ =~ m#^tkit/nfs/nfs/XFS:\"(.+)\"\|tkit/nfs/nfs/XFS=\"(.+)\"$# ) {
            # This string is also assumed to be enclosed in doublequotes
            # $1 and $2 should contain the same string now, but this isn't checked here.
            my @parts = split(/\s+/, $1);
            #printL "$1\n";
            $currNFS=$parts[$#parts];
            #printL "$currNFS \n";

            $_[0]->{"$currNFS"} = [undef,undef,[],[],[]]; # ReplINST and vhost undefined yet
            $_[0]->{"$currNFS"}->[0] = $1;
            next;
        }
        if ( $_ =~ m#^ip_subnet:(\S+)\|ip_address=(\S+)$# ) {
            #printL "$2\n";
            push @{$_[0]->{"$currNFS"}->[2]},$2;
        }
        if ( $_ =~ m#^(vg|vxvm_dg|cvm_dg):(\S+)\|(vg|vxvm_dg|cvm_dg)=(\S+)$# ) {
            #printL "$2\n";
            push @{$_[0]->{"$currNFS"}->[3]},$2;
        }
#        if ( $_ =~ m#^filesystem:(\S+)\|fs_directory=(\S+)$# ) {
        # newer SG version seem to use "filesystem" as parent attribute
        # sapgetconf still uses fs_name
        if ( $_ =~ m#^fs_name:(\S+)\|fs_directory=(\S+)$# ) {
            #printL "$2\n";
            push @{$_[0]->{"$currNFS"}->[4]},$2;
        }
    }
}

#############################################################################
#
# sub getgcdbSID
#
# get DB SID out of "instances" prefixed with DB or HDB
#
# returns the SID of such an instance if found, otherwise undef
#
#############################################################################

sub getgcdbSID {
  for my $sapinst (keys %{$_[0]}) {
      if ($sapinst =~ /^HDB(s|p)-(\S+)/) {
        printL "HDB: $sapinst -> $2\n";
        return $2;
      }
      if ($sapinst =~ /^DB-(\S+)/) {
        printL "DB: $sapinst -> $2\n";
        return $2;
      }
  }
  return undef;
}

#############################################################################
#
# sub buildshrdres
#
# Build shared resource array
#
# Arguments:
#  \%sapgchash       (ref to) hash to used to build of shared res array (read)
#  \@sapinstshrdres  (ref to) array containing instances sharing resources
#
# Returns:
#  0 - invalid resource sharing (e.g. ERS shares resource with SCS)
#  1 - valid resource sharing
#
# RESTRICTION: Doesn't check yet if ERS just shares resource with other ERS
#    (should it e.g. be allowed that it shares resources with DB?)
#
#############################################################################

sub buildshrdres {

    # go through sapgchash, for each instance:
    #    go through sapinstshrdres array, for each index:
    #       go through each element of array at that index, for each element
    #         Check if there is an overlap of IP addresses of the instance
    #         from sapgchash with that of the element.
    #         If so, add instance to that array and jump to next instance
    #         (that is skip to inner loops)
    #    if no instance could be found: push array with instance onto sapinstshrdres
    #    array (as new element)
    #
    # CAUTION: An instance 3 could end up in two different list, if it shares resource A
    #    with instance 1 and resource B with instance 2. In that case instance 1 and 2
    #    indirectly share resources via instance 3. This is not handled yset.
    #

    for my $sapinst (keys %{$_[0]}) {
        my $instshared = 0;
        for my $arrref (@{$_[1]}) {
            # arrref contains now (ref to) array of instances already sharing resource
OTHI:       for my $otherinst (@{$arrref}) {
                #print "$sapinst - $otherinst \n";
                for my $sapinstip (@{$_[0]->{$sapinst}}[2..scalar @{$_[0]->{$sapinst}}-1]) {
                #print $sapinstip,"\n";
                # check if replip is in ip list of the replicated instance
                    if (grep { $_ eq $sapinstip} @{$_[0]->{$otherinst}}[2..scalar @{$_[0]->{$otherinst}}-1]) {
                        printL  "$sapinst and $otherinst share $sapinstip\n";
                        if (($sapinst =~ /(ASCS|SCS)\d\d/) && ($otherinst =~ /ERS\d\d/)) {
                            sgeasy_print80 "Illegal resource sharing ($sapinst, $otherinst, $sapinstip)";
                            return 0;
                        }
                        if (($otherinst !~ /ERS\d\d/) && ($sapinst =~ /ERS\d\d/)) {
                            sgeasy_print80 "Illegal resource sharing ($sapinst, $otherinst, $sapinstip)";
                            return 0;
                        }
                        # add sapinst to current array (as this instance is sharing resource with at least one
                        # of the instance in the array).
                        push @{$arrref},$sapinst;
                        $instshared = 1;
                        last OTHI; # no need to check with other instances in array
                    }
                }
            }
        }
        push @{$_[1]},[$sapinst] unless $instshared; # create a new array starting with this index
    }
    return 1;
}

#############################################################################
#
# sub checksapgcHash
#
# Check the config contained in the hash.
# Requires the hash to be filled by fillsgapgcHash
#
# Arguments:
#  \%sapgchash  Reference to hash to be check (read)
#
# Checks config rules (esp. for IP):
# - ERS must not have the same IP as the SCS it replicates
#    (actually with a SAP System(=same SID) all ERS shouldn't share a IP address
#     with any of the [A]SCS. It is ok to share IPs among different ERS or among different [A]SCS
# - for now it isn't allowed to share the same SID between a HDB and SAP instances (TODO)
# - a Hana database a replication mode has to be configured, so only "instance names" like
#     HDBp|s-<SID> (p or s after HDB) are allowed, not HDB-<SID>
# - more to come:
#   -- check for shared VGs?
#
# RESTRICTION: This routine doesn't see if a ERS and "the other" SCS
#   (if there a pairs) share the same IP
#
# Returns: 1 if config is OK, 0 otherwise
#
#############################################################################

sub checksapgcHash {
  for my $sapinst (keys %{$_[0]}) {
    #print $sapinst,"\n";
    if ($sapinst =~ /^HDB-/) {
      sgeasy_print80 "No replication mode configured for $sapinst";
      return 0;
    }
    # skip also for HDB(s|p) as they might have field [1] filled now
    next if ($sapinst =~ /^HDB(s|p)-/);
    #
    next unless (defined $_[0]->{$sapinst}->[0] && defined $_[0]->{$sapinst}->[1]); # doesn't have a replicated instance
    my $replinst = $_[0]->{$sapinst}->[0];
    # loop through $replinst IP addresses and grep the replicated instance IPs
    for my $replip (@{$_[0]->{$replinst}}[2..scalar @{$_[0]->{$replinst}}-1]) {
      #print $replip,"\n";
      # check if replip is in ip list of the replicated instance
      if (grep { $_ eq $replip} @{$_[0]->{$sapinst}}[2..scalar @{$_[0]->{$sapinst}}-1]) {
        sgeasy_print80 "$replinst and replication $sapinst have  same IP address $replip configured";
        return 0;
      }
    }
  }
  return 1;
}

#############################################################################
#
# sub removePhysInst
#
# Args:
#   \%sapgcinsthash - (ref to) instance hash which also contains the vhosts
#   \@clusnode - (ref to) array of clusternodes
#
# Removes instance from sapgcinsthash if its virtual hostname matches
# one of the member of clusnodes.
#
# This should remove instances configured to physical host, but
# is far from being perfect. So far only the hostnames of the clustermember
# are considered to be physical host. It could, however, be also the case
# that the instance is configured to another interface of the members
# (though this might violate some SAP config rules).  This would need
# some digging into the output of
# cmquerycl -w local -k -n <member1> -n ... -f line -l net
# Still, it wouldn't catch cluster external hostnames, but those are highly
# unlikely for the components handled in this program.
#
# HANA instances are configured to the physical host, but aren't treated as
# SAP application instances and have no vhost reported (left undef). Therefore
# they aren't filtered here.
# Starting with sapgetconf rev >1264, HANA hosts are put into the vhost (idx 1)
# field. So that instances have to be filtered based an instance prefix.
#
#############################################################################
sub removePhysInst {
  for my $sapinst (keys %{$_[0]}) {
    next unless (defined $_[0]->{$sapinst}->[1]);
    if ($sapinst =~ /^HDB/) {
#    	# tmp cleanout of HANA scaleout instances (assign to variable to avoid deprecated @_ message)
#    	if (scalar (my @tmparr = split(",",$_[0]->{$sapinst}->[1])) > 1) {
#    		sgeasy_print80 "HANA installation for $sapinst appears to be a scaleout configuration on nodes";
#    		sgeasy_print80 "$_[0]->{$sapinst}->[1]";
#    		sgeasy_print80 "(tmp) removing it until such an installation will be supported by this script";
#    		delete $_[0]->{$sapinst};
#    	}
    	next;
    }
    my $vhost = $_[0]->{$sapinst}->[1];
    if (grep {$_ eq $vhost} @{$_[1]}) {
      sgeasy_print80 "Instance $sapinst configured to physical node $vhost";
      delete $_[0]->{$sapinst};
    }
  }
}

#############################################################################
#
# sub checkNFSexports
#
# Check if exports (XFS) returned by "sapgetconf nfs" are already setup
# somewhere. If so, remove them from hash.
# Doesn't look for exact match of XFS spec including options, as these
# may be different (exact options may not be discoverable by sapgetconf)
# As exports by legacy packages aren't shown by cmviewcl, they may not
# be considered as already packaged.
#
# Args:
#  \%sapgcnfshash  - hash to check
#  \%clusrsrchash  - resources already setup in a package
#                    ("XFS" key is used)
#
#############################################################################

sub checkNFSexports {
# (\%sapgcnfshash,  \%clusrsrchash);
  my @clusxfs = keys %{$_[1]->{"XFS"}};
  printL "clusxfs [ @clusxfs ] \n";
  for my $nfsxfs (keys %{$_[0]}) {
    printL "nfsxfs: $nfsxfs \n";
    # mask a '*' in nfsxfs to avoid interpretation as part of a regex
    my $searchpatt=$nfsxfs;
    $searchpatt =~ s/[*]/\[*\]/g;
    my @foundxfs = grep /$searchpatt/,@clusxfs;
    if (scalar @foundxfs) {
      printL "found xfs: [ @foundxfs ], removing it from sggcnfshash\n";
      sgeasy_print80 "$nfsxfs already exported by package(s) ";
      sgeasy_print80 "[ @{$_[1]->{XFS}->{$foundxfs[0]}} ]";
      delete $_[0]->{$nfsxfs};
    }
  }
}

#############################################################################
#
# sub mergeNFS2Instance
#
# Merge remaining required exports into sap instance hash.
# Each export is treated as an instance "NFS-<exportdir>".
# replicated instance, vhost are left undefined
# IPs are copied over (index
#
# These type of "instances" are then later subject to package selection.
#
# Args:
#  \%sapgcnfshash  - NFS hash to merge into instance hash
#  \%sapgcinsthash  - resources already setup in a package
#                    ("XFS" key is used)
#
#############################################################################

sub mergeNFS2Instance {
    for my $nfsxfs (keys %{$_[0]}) {
        printL "merge: $nfsxfs\n";
        my $keyname="NFS-|".$nfsxfs."|";
        $_[1]->{$keyname} = [undef,undef];
        my $startidx = 2;
        for my $nfsip (@{$_[0]->{$nfsxfs}->[2]}) {
          printL "nfsip: $nfsip\n";
          $_[1]->{$keyname}->[$startidx++] = $nfsip;
        }
    }
}

#############################################################################
#
# sub populatepkgHash
#
# Fill in the package information into the pkg hash
# (cluspkgconfighash)
#
# Arguments:
# \@linescmviewcl - Array containing output of cmviewcl cmd (input, reference)
# \%cluspkgconfighash - hash to be built up (output, reference)
#
# Restriction:
# - may not get all package names (esp. CFS legacy) using (\w+) as pattern
#
#############################################################################
sub populatepkgHash {
  my $linesArr = shift;  # Ref to hash is now $_[0];
  my %pkgs_all_cluster_nodes = ();
  foreach (@{$linesArr}) {
    # SAP SID and/or DB SID
    # It may not be sufficient to look at "db_system". It is legal to
    # specify the DB module, but leave db_system empty. It'll then default
    # to sap_system if "db_vendor" is set.
    # Output for the db_global module might occur before output of sap_global module
    # and db_vendor might occur before db_system.
    # This code is not considering a config where no sap_system is defined (because
    # module isn't included) and also db_system is left empty, but not db_vendor.
    # This can be legally applied (but shouldn't run?). It'll leave db_sid at "-TBS-"

    if ( $_ =~ m#^package:(\w+)\|sgesap/sap_global/sap_system=(\S+)$# ) {
        @{${$_[0]}{"$1"}}[$sapsid_idx]=$2;
        @{${$_[0]}{"$1"}}[$dbsid_idx] =$2 if ((defined @{${$_[0]}{"$1"}}[$dbsid_idx])
                                            && (@{${$_[0]}{"$1"}}[$dbsid_idx] eq "-TBS-"));
        next;
    }
    if ( $_ =~ m#^package:(\w+)\|sgesap/db_global/db_system=(\S+)$# ) {
        @{${$_[0]}{"$1"}}[$dbsid_idx]=$2;
        # Fill also SAPSID if not defined yet
        # -- commented as this doesn't reflect the real setup
        # ---- although a package with only db_system set might be a candidate to cluster
        #      other SAP instances into
        #@{${$_[0]}{"$1"}}[$sapsid_idx]=$2 unless defined @{${$_[0]}{"$1"}}[$sapsid_idx];
        next
    }
    if ( $_ =~ m#^package:(\w+)\|sgesap/db_global/db_vendor=(\S+)$# ) {
        @{${$_[0]}{"$1"}}[$dbsid_idx]="-TBS-" unless defined @{${$_[0]}{"$1"}}[$dbsid_idx];
        # Uses a marker to indicate that a value has to be filled in later (if not set yet)
        next
    }

    # check for livecache. If not already configured to a spefific SAPSID, make it unusable
    # for general SAP instance and other DB assignment by setting SID to "---";
    # package:MsapLC5|sgesap/livecache/lc_system:LC5|sgesap/livecache/lc_system=LC5
    if ( $_ =~ m#^package:(\w+)\|sgesap/livecache/lc_system:(\S+)|sgesap/livecache/lc_system=(\2)$# ) {
        @{${$_[0]}{"$1"}}[$dbsid_idx]="---";
        @{${$_[0]}{"$1"}}[$sapsid_idx]="---" unless defined @{${$_[0]}{"$1"}}[$sapsid_idx];
    }

    # SAP instances configured in this package. Doesn't care about virtual host (and replicated instance), just the instance.
    if ( $_ =~ m#^package:(\w+)\|sgesap/stack/sap_instance:(\w+)\|sgesap/stack/sap_instance=(\w+)$# ) {
        # $2 and $3 identical
        if (defined @{${$_[0]}{"$1"}}[$sapinst_idx]) {
          push @{@{${$_[0]}{"$1"}}[$sapinst_idx]},$3;
        }
        else {
          @{${$_[0]}{"$1"}}[$sapinst_idx] = [ $3 ];
        }
        next;
    }

    # SAP virt. hostnames configured in this package
    if ( $_ =~ m#^package:(\w+)\|sgesap/stack/sap_instance:(\w+)\|sgesap/stack/sap_virtual_hostname=(\S+)$# ) {
        if (defined @{${$_[0]}{"$1"}}[$sapvhost_idx]) {
          push @{@{${$_[0]}{"$1"}}[$sapvhost_idx]},$3;
        }
        else {
          @{${$_[0]}{"$1"}}[$sapvhost_idx] = [ $3 ];
        }
        next;
    }

    # HDB SID configured ....
    # package:<pkg>|sgesap/hdb_global/hdb_system=<HDB-SID>
    if ( ( $_ =~ m#^package:(\w+)\|sgesap/hdb_global/hdb_system=(\S+)$# )
         || ( $_ =~ m#^package:(\w+)\|sgesap/hdb_global/hdb_system:(\S+)\|sgesap/hdb_global/hdb_system=(\S+)$# )) {
        @{${$_[0]}{"$1"}}[$hdbsid_idx]=$2;
        next;
    }

    # HDB instance configured ...
    # package:<pkg>|sgesap/hdbsystem/hdb_instance:<HDB-Inst>|sgesap/hdbsystem/hdb_instance=<HDB-Inst>
    if ( ($_ =~ m#^package:(\w+)\|sgesap/hdbsystem/hdb_instance:(\w+)\|sgesap/hdbsystem/hdb_instance=(\w+)$# )
         || ($_ =~ m#^package:(\w+)\|sgesap/hdbinstance/hdb_instance:(\w+)\|sgesap/hdbinstance/hdb_instance=(\w+)$# )
         || ($_ =~ m#^package:(\w+)\|sgesap/hdb_global/hdb_system:(\w+)\|sgesap/hdbinstance/hdb_instance=(\w+)$# )
         || ($_ =~ m#^package:(\w+)\|sgesap/hdb_global/hdb_system:(\w+)\|sgesap/hdbinstance_global/hdb_instance=(\w+)$# ) ) {
        if (defined @{${$_[0]}{"$1"}}[$hdbinst_idx]) {
          push @{@{${$_[0]}{"$1"}}[$hdbinst_idx]},$3;
        }
        else {
          @{${$_[0]}{"$1"}}[$hdbinst_idx] = [ $3 ];
        }
        next;
    }

    # get clusternodes to which the pacakage if configured. ignore them if all_cluster_nodes=true is set
    if ( $_ =~ m#^package:(\w+)\|all_cluster_nodes=true# ) {
    	$pkgs_all_cluster_nodes{$1}=1;
    	next;
    }
    if ( $_ =~ m#^package:(\w+)\|node:(\S+)\|name=\2# ) {
        next if ($pkgs_all_cluster_nodes{$1});
        if (defined @{${$_[0]}{"$1"}}[$pkgnodelist_idx]) {
          push @{@{${$_[0]}{"$1"}}[$pkgnodelist_idx]},$2;
        }
        else {
          @{${$_[0]}{"$1"}}[$pkgnodelist_idx] = [ $2 ];
        }
        next;
    }




    # configured IPs (IPv6 addresses need to be tested) - modular only
    if ( $_ =~ m#^package:(\w+)\|ip_subnet:(\S+)\|ip_address=(\S+)$# ) {
        if (defined @{${$_[0]}{"$1"}}[$cip_idx]) {
          push @{@{${$_[0]}{"$1"}}[$cip_idx]->{"$2"}},$3;
        }
        else {
          @{${$_[0]}{"$1"}}[$cip_idx] = { $2 => [$3] };
        }
        next;
    }

    # active IP addresses (also in legacy) (ip_address and name are the same value)
    if ( $_ =~ m#^package:(\w+)\|ip_address:(\S+)\|name=(\S+)$# ) {
        if (defined @{${$_[0]}{"$1"}}[$aip_idx]) {
          push @{@{${$_[0]}{"$1"}}[$aip_idx]},$3;
        }
        else {
          @{${$_[0]}{"$1"}}[$aip_idx] = [ $3 ];
        }
        next;
    }

    # package type (legacy,modular), state(up,down,...)
    # and style (failover, multi_node,..)
    if ( $_ =~ m#^package:(\w+)\|style=(\S+)$# ) {
        @{${$_[0]}{"$1"}}[$pkgstyle_idx]=$2;
        next;
    }
    if ( $_ =~ m#^package:(\w+)\|status=(\S+)$# ) {
        @{${$_[0]}{"$1"}}[$pkgstatus_idx]=$2;
        next;
    }
     if ( $_ =~ m#^package:(\w+)\|type=(\S+)$# ) {
        @{${$_[0]}{"$1"}}[$pkgtype_idx]=$2;
        next;
    }
  }
}

#############################################################################
#
# sub collectUsedClusResources
#
# Fill in resource cross reference (resource used by package)
#
# Arguments:
# \@linescmviewcl - Array containing output of cmviewcl cmd (input, reference)
# \%clusrsrchash  - Hash to be built up (output, reference)
#
# PKG_NAME ("name"), MOD_NAME ("module_name")
# VG/DG ("vg","vxvm_dg","cvm_dg"), FS_NAME, FS_DIRECTORY ("fs_name"[,"fs_directory"]),
# IP_ADDRESS ("ip_address"), PKG NAME ("package_name"), SAPVHOST ("sap_virtual_hostname")
# XFS ("hanfs_exports/XFS")
# CVM_DG (CVM_DISK_GROUP, cvm_disk_group)
# CFS_MP (CFS_MOUNT_POINT, cfs_mount_point)
# CFS_VOL (CFS_VOLUME, cfs_volume)
#
# Uses "S+" as packagename pattern instead of "w+"
# as legacy also allows '-' in pkg names
#
# Only check modular packages for the other used resources
#
# Todo: makes "mapPkgIP2Pkg" obsolete
#
# According to sg/basic module package names must follow the rule
#  [0-9A-Za-z][0-9A-Za-z_.\-]*[0-9A-Za-z]
# This sub uses \S+ for parsing.
#
#############################################################################
sub collectUsedClusResources {
  my $linesArr = shift;
  foreach (@{$linesArr}) {
  	# printL "line -  $_ \n";
    # package name to have the list of used names.
    # (of course a package is using its own name ).
    # allow legacy pkg names (additional '-')
    # Just using (\S+) would be too greedy. (\S+?\|) doesn't work
    # (\1) references back to first match
    if ( $_ =~ m#^package:([\S-]+)\|name=(\1)$# ) {
        if (defined $_[0]->{"PKG_NAME"}->{"$1"}) {
            push @{$_[0]->{"PKG_NAME"}->{"$1"}},$1;
        }
        else {
            $_[0]->{"PKG_NAME"}->{"$1"} = [$1] ;
        }
        next;
    }

    # list of modules used by package
    # package:extinst-AP4|module_name:sg/basic|module_name=sg/basic
    if ( $_ =~ m#^package:([\S-]+)\|module_name:(\S+)\|module_name=(\2)$# ) {
        if (defined $_[0]->{"MOD_NAME"}->{"$2"}) {
            push @{$_[0]->{"MOD_NAME"}->{"$2"}},$1;
        }
        else {
            $_[0]->{"MOD_NAME"}->{"$2"} = [$1] ;
        }
        next;
    }

    # configured IPs (IPv6 addresses need to be tested) - modular only
    # (once again - to have all resources covered)
    if ( $_ =~ m#^package:(\S+)\|ip_subnet:(\S+)\|ip_address=(\S+)$# ) {
        if (defined $_[0]->{"IP_ADDRESS"}->{"$3"}) {
            push @{$_[0]->{"IP_ADDRESS"}->{"$3"}},$1;
        }
        else {
            $_[0]->{"IP_ADDRESS"}->{"$3"} = [$1] ;
        }
        next;
    }
    if ( $_ =~ m#^package:(\S+)\|(vg|vxvm_dg|cvm_dg):(\S+)\|(vg|vxvm_dg|cvm_dg)=(\S+)$# ) {
        # May be full path /dev/<vgname> or just <vgname> (for vg)
        # vxvm,cmv: /dev/vxvm/dsk prefix?
        # vxvm_dg,cvm_dg need to be verified
        if (defined $_[0]->{"VG/DG"}->{"$3"}) {
            #print "defined ($1, $2. $3) \n";
            push @{$_[0]->{"VG/DG"}->{"$3"}},$1;
        }
        else {
            #print "not defined ($1, $2. $3) \n";
            $_[0]->{"VG/DG"}->{"$3"} = [$1] ;
        }
        next;
    }

    if ( $_ =~ m#^package:(\S+)\|filesystem:(\S+)\|fs_name=(\S+)$# ) {
        # instead of "fs_name=" also "filesystem=" could be used
        if (defined $_[0]->{"FS_NAME"}->{"$3"}) {
            #print "defined ($1, $2. $3) \n";
            push @{$_[0]->{"FS_NAME"}->{"$3"}},$1;
        }
        else {
            #print "not defined ($1, $2. $3) \n";
            $_[0]->{"FS_NAME"}->{"$3"} = [$1] ;
        }
        next;
    }

    if ( $_ =~ m#^package:(\S+)\|filesystem:(\S+)\|fs_directory=(\S+)$# ) {
        if (defined $_[0]->{"FS_DIRECTORY"}->{"$3"}) {
            #print "defined ($1, $2. $3) \n";
            push @{$_[0]->{"FS_DIRECTORY"}->{"$3"}},$1;
        }
        else {
            #print "not defined ($1, $2. $3) \n";
            $_[0]->{"FS_DIRECTORY"}->{"$3"} = [$1] ;
        }
        next;
    }

    # SAP virt. hostnames configured
    if ( $_ =~ m#^package:(\S+)\|sgesap/stack/sap_instance:(\w+)\|sgesap/stack/sap_virtual_hostname=(\S+)$# ) {
        if (defined $_[0]->{"SAPVHOST"}->{"$3"}) {
            push @{$_[0]->{"SAPVHOST"}->{"$3"}},$1;
        }
        else {
            $_[0]->{"SAPVHOST"}->{"$3"} = [$1] ;
        }
        next;
    }

    # site aware configs (belongs to cluster config)
    # example: node:lipodb01|siteid=1
    # also scans
    #   site:<sitename>|name=<sitename>
    #   site:<sitename>|id=<siteid>
    # BE AWARE that this siteid is not related to the HANA site id at all
    # but assigned by the SG cluster during cluster creation/update
    #
    # Site nodes per site id = id to nodelist map (or do it the other way round?):
    # Remark: \w+ is not sufficient(nodename may contain '-') \S+ is a bit too
    #  relaxed, but ok as names have been check by cluster already. Same for site
    if ( $_ =~ m#^node:(\S+)\|siteid=(\d+)$# ) {
        printL "node: $1 - siteid: $2 \n";
        if (defined $_[0]->{"SITENODES"}->{"$2"}) {
            #print "defined ($1, $2. $3) \n";
            push @{$_[0]->{"SITENODES"}->{"$2"}},$1;
        }
        else {
            #print "not defined ($1, $2. $3) \n";
            $_[0]->{"SITENODES"}->{"$2"} = [$1] ;
        }
        next;
    }
    # sitename to siteid
    if ( $_ =~ m#^site:(\S+)\|id=(\d+)$# ) {
        printL "sitename: $1 - siteid: $2 \n";
        $_[0]->{"SITENAMES"}->{"$1"}="$2"; # name to id map
        next;
    }

    # NFS - modular only
	# HPUX
    # (replace package pattern by \S+)
    if ( $_ =~ m#^package:([0-9A-Za-z][0-9A-Za-z_.\-]*[0-9A-Za-z])\|nfs/hanfs_export/XFS:\"(.+)\"\|nfs/hanfs_export/XFS=\"(.+)\"$# ) {
        if (defined $_[0]->{"XFS"}->{"$3"}) {
            push @{$_[0]->{"XFS"}->{"$3"}},$1;
        }
        else {
            $_[0]->{"XFS"}->{"$3"} = [$1] ;
        }
        next;
    }
	# Linux
    if ( $_ =~ m#^package:([0-9A-Za-z][0-9A-Za-z_.\-]*[0-9A-Za-z])\|tkit/nfs/nfs/XFS:\"(.+)\"\|tkit/nfs/nfs/XFS=\"(.+)\"$# ) {
        if (defined $_[0]->{"XFS"}->{"$3"}) {
            push @{$_[0]->{"XFS"}->{"$3"}},$1;
        }
        else {
            $_[0]->{"XFS"}->{"$3"} = [$1] ;
        }
        next;
    }

    # CFS - legacy only up to 11.20 patch A
    #
    # Modular output will have
    # cmv_disk_group, cfs_mount_point, cfs_volume
    #
    # So far only legacy is scanned.
    # Should go into "VG/DG" and "FS_NAME" entries later? or maybe a better idea
    # to keep them separate to indicated that they belong to a dedicated CFS packages
    #
    # Legacy:
    # package:SG-CFS-DG-09|environment:CVM_DISK_GROUP|CVM_DISK_GROUP=dg09
    # package:SG-CFS-MP-altadis|environment:CFS_MOUNT_POINT|CFS_MOUNT_POINT=/document/altadis
    # package:SG-CFS-MP-altadis|environment:CFS_VOLUME|CFS_VOLUME=dg09/voldocumentaltadis
    #
    # CFS volume usually resides in /dev/vx/dsk
    # group w/o any prefix, volume just dg and volume, also no prefix
    #
    # Modular attributes (used in ascii config)
    # - cvm_disk_group
    # -- cvm_activation_mode (parent cvm_disk_group)
    # - cfs_mount_point
    # -- cfs_volume (parent cfs_mount_point)
    # -- cfs_mount_options (parent cfs_mount_point)
    # - ...
    # Sample modular output
    # package:mpdgonly|cvm_disk_group:cvm_dg0_indigo2|cvm_disk_group=cvm_dg0_indigo2
    # package:mpdgonly|cvm_disk_group:cvm_dg0_indigo2|cvm_activation_mode=all=sw
    # package:mpdgonly|cfs_concurrent_mount_unmount_operations=5
    # package:mpdgonly|cfs_mount_point:/mnt1|cfs_mount_point=/mnt1
    # package:mpdgonly|cfs_mount_point:/mnt1|cfs_volume=cvm_dg0_indigo2/lvol2
    # package:mpdgonly|cfs_mount_point:/mnt1|cfs_mount_options=indigo2=cluster indigo3=cluster
    # package:mpdgonly|cfs_mount_point:/mnt1|cfs_primary_policy=""
    #
    # ** Using \S+ for matching package name, as they can also contain a hyphen '-' in legacy.
    #
    # (why keep CVM DG, CFS MP and FS separately?)
    #

    # CVM diskgroup (legacy/modular)
    if ( $_ =~ m#^package:(\S+)\|environment:CVM_DISK_GROUP\|CVM_DISK_GROUP=(\S+)$# ) {
        if (defined $_[0]->{"CVM_DG"}->{"$2"}) {
            push @{$_[0]->{"CVM_DG"}->{"$2"}},$1;
        }
        else {
            $_[0]->{"CVM_DG"}->{"$2"} = [$1] ;
        }
        next;
    }
    if ( $_ =~ m#^package:(\S+)\|cvm_disk_group:(\S+)\|cvm_disk_group=(\S+)$# ) {
        if (defined $_[0]->{"CVM_DG"}->{"$2"}) {
            push @{$_[0]->{"CVM_DG"}->{"$2"}},$1;
        }
        else {
            $_[0]->{"CVM_DG"}->{"$2"} = [$1] ;
        }
        next;
    }

    # CFS mountpoint (legacy/modular)
    if ( $_ =~ m#^package:(\S+)\|environment:CFS_MOUNT_POINT\|CFS_MOUNT_POINT=(\S+)$# ) {
        if (defined $_[0]->{"CFS_MP"}->{"$2"}) {
            push @{$_[0]->{"CFS_MP"}->{"$2"}},$1;
        }
        else {
            $_[0]->{"CFS_MP"}->{"$2"} = [$1] ;
        }
        next;
    }
    if ( $_ =~ m#^package:(\S+)\|cfs_mount_point:(\S+)\|cfs_mount_point=(\S+)$# ) {
        if (defined $_[0]->{"CFS_MP"}->{"$2"}) {
            #print "defined ($1, $2) \n";
            push @{$_[0]->{"CFS_MP"}->{"$2"}},$1;
        }
        else {
            #print "not defined ($1, $2) \n";
            $_[0]->{"CFS_MP"}->{"$2"} = [$1] ;
        }
        next;
    }

    # CFS volume (legacy/modular)
    if ( $_ =~ m#^package:(\S+)\|environment:CFS_VOLUME\|CFS_VOLUME=(\S+)$# ) {
        if (defined $_[0]->{"CFS_VOL"}->{"$2"}) {
            push @{$_[0]->{"CFS_VOL"}->{"$2"}},$1;
        }
        else {
            $_[0]->{"CFS_VOL"}->{"$2"} = [$1] ;
        }
        next;
    }
    if ( $_ =~ m#^package:(\S+)\|cfs_mount_point:(\S+)\|cfs_volume=(\S+)$# ) {
        if (defined $_[0]->{"CFS_VOL"}->{"$3"}) {
            #print "defined ($1, $3) \n";
            push @{$_[0]->{"CFS_VOL"}->{"$3"}},$1;
        }
        else {
            #print "not defined ($1, $3) \n";
            $_[0]->{"CFS_VOL"}->{"$3"} = [$1] ;
        }
        next;
    }

    # Sybase toolkit ase and rs module (Linux only)
    #
    # A Sybase ASE DB (INSTANCE_NAME = <DBSID>)
    # tkit/sybase/ase/INSTANCE_NAME
    # key: TKIT_ASE
    #
    # A Sybase Replication Server
    # tkit/sybase/rs/INSTANCE_NAME contains <DBSID>_REP_*
    # key: TKIT_RS
    #
    # values can be optionally enclosed in doublequotes, but only consist of word characters

    if ( $_ =~ m#^package:([0-9A-Za-z][0-9A-Za-z_.\-]*[0-9A-Za-z])\|tkit/sybase/ase/INSTANCE_NAME=[\"]?(\w+)[\"]?$# ) {
        printL "TKIT_ASE: $1, $2 \n";
        if (defined $_[0]->{"TKIT_ASE"}->{"$2"}) {
            push @{$_[0]->{"TKIT_ASE"}->{"$2"}},$1;
        }
        else {
            $_[0]->{"TKIT_ASE"}->{"$2"} = [$1] ;
        }
        next;
    }

    if ( $_ =~ m#^package:([0-9A-Za-z][0-9A-Za-z_.\-]*[0-9A-Za-z])\|tkit/sybase/rs/INSTANCE_NAME=[\"]?(\w+)[\"]?$# ) {
        printL "TKIT_RS: $1, $2 \n";
        if (defined $_[0]->{"TKIT_RS"}->{"$2"}) {
            push @{$_[0]->{"TKIT_RS"}->{"$2"}},$1;
        }
        else {
            $_[0]->{"TKIT_RS"}->{"$2"} = [$1] ;
        }
        next;
    }
  }
}

#############################################################################
#
# sub checkUsedClusResources
#
# Check for resource used by package (other than the provided
# cluspkgconfighash). These resource usually require exclusive
# usage.
#
# Arguments:
# \%clusrsrchash  - cluster pkg resource x-ref (input, reference)
# $attrname       - attribute name
# $attrval        - attribute value
# $pkgname        - package name (maybe undef)
#
# Returns:
# List of packages using this resources (less $pkgname if defined).
#
# VG/DG ("vg","vxvm_dg","cvm_dg"),
# FS_NAME, FS_DIRECTORY ("fs_name"[,"fs_directory"]),
# IP_ADDRESS ("ip_address"),
# PKG NAME ("package_name"),?
# SAPVHOST ("sap_virtual_hostname")?
# CVM_DG,...
#
#
# Only check modular packages for the other used resources (except CFS/CVM)
#
#############################################################################
sub checkUsedClusResources {

    my $crh = shift;
    my $aname = shift,
    my $aval  = shift;
    my $pname = shift;

    printL "checkUsedClusResources: $aname,$aval,$pname\n";

    my @pkglist = ();

    # Check VGs/DGs
    # (might DG might also be CVM DGs)

    if ($aname =~ /^(vg|vxvm_dg|cvm_dg)$/ ) {
      @pkglist = @{$crh->{"VG/DG"}->{$aval}} if defined $crh->{"VG/DG"}->{$aval};
      # TODO INVESTIGATE
      # TODO there might not be an exact match to "$aval" (might have a "/dev" prefix)
      printL "Already configured (VG/DG): $aname $aval [@pkglist]\n";
    }
    # nothing found. look into CVM_DG
    if ((! scalar @pkglist) && ($aname eq "cvm_dg")) {
      @pkglist = @{$crh->{"CVM_DG"}->{$aval}} if defined $crh->{"CVM_DG"}->{$aval};
      # TODO INVESTIGATE
      # TODO $aval might not exactly match
      printL "Already configured (CVM_DG): $aname $aval [@pkglist]\n";
    }

    if ($aname eq "ip_address" ) {
      @pkglist = @{$crh->{"IP_ADDRESS"}->{$aval}} if defined $crh->{"IP_ADDRESS"}->{$aval};
      printL "Already configured: $aname $aval [@pkglist]\n";
    }

    if ($aname eq "fs_name" ) {
      @pkglist = @{$crh->{"FS_NAME"}->{$aval}} if defined $crh->{"FS_NAME"}->{$aval};
      # TODO INVESTIGATE
      # TODO there might not be an exact match to "$aval" (might have a "/dev" prefix)
      printL "Already configured (FS_NAME): $aname $aval [@pkglist]\n";
    }
    # nothing found. look into CFS_VOL
    if ((! scalar @pkglist) && ($aname eq "fs_name")) {
      #@pkglist = @{$crh->{"CFS_VOL"}->{$aval}} if defined $crh->{"CFS_VOL"}->{$aval};
      # $aval might not exactly match - as assumed in code above. Seems that with CFS modules
      # volumes are just defined using <dg>/<vol> w/o prefix "/dev/vx/dsk"
      my @cfsvol=();
      @cfsvol = grep {$aval =~ /($_)$/} keys %{$crh->{"CFS_VOL"}};
      if (scalar @cfsvol) {
        printL "Already configured (CFS_VOL): $aname $aval [@cfsvol] [@pkglist]\n";
        # assuming that there is just one matching CFS_VOL
        @pkglist = @{$crh->{"CFS_VOL"}->{$cfsvol[0]}} if defined $crh->{"CFS_VOL"}->{$cfsvol[0]};
      }
    }

    if ($aname eq "fs_directory" ) {
      @pkglist = @{$crh->{"FS_DIRECTORY"}->{$aval}} if defined $crh->{"FS_DIRECTORY"}->{$aval};
      printL "Already configured: $aname $aval [@pkglist]\n";
    }
    # nothing found. look into CFS_MP
    if ((! scalar @pkglist) && ($aname eq "fs_directory")) {
      @pkglist = @{$crh->{"CFS_MP"}->{$aval}} if defined $crh->{"CFS_MP"}->{$aval};
      printL "Already configured: $aname $aval [@pkglist]\n";
    }

    #if ($aname == "
    #
    # TODO ENHANCEMENT
    # TODO other resources here...
    #

    # return list (less pkgname if defined)
    return (@pkglist) unless defined ($pname);
    return (grep {$_ ne $pname} @pkglist);
}

#############################################################################
#
# sub mapPkgIP2Pkg
#
# Creates a hash with package IPs to the package(s).
# (contains essentiallly the same information cluspkgconfighash but with IPs as keys and (array of) package name(s) as values
#
# Argument:
# \%cluspkgconfighash - hash to read from
# \%clusip2pkgmap      - hash to be created/filled
#
# Restrictions:
# - won't consider legacy package (will have no configured IP anyway displayed by cmviewcl)
#
# Be aware that package IPs are part of the IPsubnet/array of IPs hash, which is in turn behind
# index $cip_idx of array refererence by package hash.
#
# Alternative: create hash already when collecting packge informatin
#
#############################################################################

# OBSOLETE with introduction of collectUsedResources
#   ... but still in use

sub mapPkgIP2Pkg {

  for my $pkg (keys %{$_[0]}) {
    # skip legacy
    next unless @{${$_[0]}{$pkg}}[$pkgstyle_idx] eq "modular";
    my $subniphashref = ${${$_[0]}{$pkg}}[$cip_idx] ; #ref
    next unless defined $subniphashref;
    for my $ipsubn (keys %{$subniphashref}) {
      # each hash contains a reference to an array
	  printL "ipsubn: $ipsubn\n";
#      for my $ipaddr (@{${%{$subniphashref}}{$ipsubn}}) {
      for my $ipaddr (@{$subniphashref->{$ipsubn}}) {
        # fill IPaddr->Pkg map
        if (defined $_[1]->{$ipaddr}) {
          push @{$_[1]->{$ipaddr}}, $pkg;
        }
        else {
           $_[1]->{$ipaddr} = [ $pkg ];
        }
      }
    }
  }
}

#############################################################################
#
# sub lookupPreferredInst
# Arguments:
#  \%instPkgCandidates  Reference to candiate hash (write)
#
# Fills: %instPkgCandidates, index $select_preferred
#
# For each instance lookup preferred other instances (priority list)
# to be combined with.
# Assumes that %instPkgCandidates already contains the instances as keys
# Depending on the SAP setup there is not necessary a preferred other
# instance for each instance, so the list may be empty for some of them.
#
# Preferred groupings are:
# SCS  ==> ASCS,DB (other SCS(/DVEBMGS?), then DB)
# ASCS ==> SCS,DB  (other SCS, then DB)
# ERS  ==> ERS     (other ERS)
# DVEBMGS ==> SCS,DB (java SCS, then DB)
# DB   ==> SCS,ASCS,DVEBMGS (SCS,ASCS should have same prio)
# REP ==> DB
# NFS  ==> DB (not implemented yet)
#
# "Instance names" for [H]DB, REP and NFS are
#   "[H]DB-<SID>", "REP-<SID> and "NFS-<SID>"
#
# (REP is replicated DB)
#
# Returns: 1
#
# Remark:
# This sub sets up yet another set of lists of instances grouped according
# instance types. It might be a good idea to setup a global hash
# with keys "SCS","ERS","DB","DVEBMGS","NFS" and array of instance names
# as array.
#############################################################################
sub lookupPreferredInst {

    # setup instance lists
    my @sapinstlist = keys %{$_[0]};
    my @scslist = grep (/^(ASCS|SCS)\d\d$/,@sapinstlist);
    my @erslist = grep (/^ERS\d\d$/,@sapinstlist);
    my @dvelist = grep (/^DVEBMGS\d\d$/,@sapinstlist);
    my @dblist  = grep (/^DB-\w+$/,@sapinstlist);  # could also be /DB-{\w+}3/
    my @hdblist = grep (/^HDB(p|s|)-\w+$/,@sapinstlist); # Not used so far for "cross-" preferred assignments
	my @repdblist = grep (/^REP-\w+$/,@sapinstlist);
    my @nfslist = grep (/^NFS-\|\S+\|$/,@sapinstlist);


    printL "Lists found (all,scs,ers,dvebmgs,db,hdb,repdb,nfs:\n";
    printL "[ @sapinstlist ] [ @scslist ] [ @erslist ] [ @dvelist ] [ @dblist ] [@hdblist] [ @repdblist ]  [ @nfslist ] \n";

    for my $sapinst (@sapinstlist) {
        if ($sapinst =~ /^ERS\d\d$/) {
            # assign other inst (if there) to preferred list
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, grep {$_ ne $sapinst} @erslist);
            next;
        }
        if ($sapinst =~ /^(ASCS|SCS)\d\d$/) {
            # assign other SCS inst and DB(if there) to preferred list
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, grep {$_ ne $sapinst} @scslist);
            # what about DVEBMGS for SCS if there's no ASCS?
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, @dblist);
            next;
        }
        if ($sapinst =~ /^DVEBMGS\d\d$/) {
            # assign SCS and DB(if there) to preferred list
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, grep (/^SCS\d\d$/, @scslist));
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, @dblist);
            next;
        }
        if ($sapinst =~ /^DB-\w+$/) {
            # assign scs list and dvebmgs
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, @scslist);
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, @dvelist);
            next;
        }
        if ($sapinst =~ /^HDB(p|s|)-\w+$/) {
            # no preferred assignment with others
            @{$_[0]->{$sapinst}->[$select_preferred]} = ();
        }
		if ($sapinst =~ /^REP-\w+$/) {
            # assign db
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, @dblist);
            next;
        }
        if ($sapinst =~ /^NFS-\|\S+\|$/) {
            # assign db
            push (@{$_[0]->{$sapinst}->[$select_preferred]}, @dblist);
            next;
        }
    }
    return 1;
}

#############################################################################
#
# sub lookupPkgCandidatesIP
#
# Match discovered instance IP address(es) to a maybe already configured package(s).
# Used as first steop to "assign" an instance to a matching package.
# IP address(es) are taken from discovered SAP config (sapgchash). Addresses to match to
# are looked up in clusip2pkgmap.
#
# Arguments:
#  \%sapgchash          Ref to hash of discovered instances  (read)
#  \%clusip2pkgmap      Ref to hash of IP2Pkg map (read)
#  \%instPkgCandidates  Ref to candiate hash (write)
#  \%clusrsrchash       Ref to cluster resource crossref (read)
#
# Fills: %instPkgCandidates
#
# Doesn't check for overlapping IPs of SCS and ERS anymore. Assumes that the
# appropriate check function has been run already (checksapgcHash)
#
# Creates a hash with instances as keys and arrays of package candiate names as
# values. Keep in mind that there may be a (already configured) package for all
# IP addresses returned by sapgetconf.
#
# Doesn't filter out instances configured to physical nodename of clusternode
#
# All packages are modular (non-modular already filtered out when setting
# up IP2Pkg map)
#
# Won't consider packages having the IP, but ECMT/toolkit modules configured as candidates.
# (clusrsrchash and hasTKmod is used for this)
#
# Returns: 1
#
#############################################################################

sub lookupPkgCandidatesIP {

  my @hastk=();
  my $whichtk="ecmt"; # lowercase toolkit name, module name required to start with

  # loop over all discovered instances (again).
  for my $sapinst (keys %{$_[0]}) {
    $_[2]->{$sapinst}->[$select_on_IP_idx] = [];  # initialize empty
    # get all IP addresses for this instance
     for my $instip (@{$_[0]->{$sapinst}}[2..scalar @{$_[0]->{$sapinst}}-1]) {
       next unless defined ($_[1]->{$instip}); # no packages configured with this IP?
       if (defined $_[2]->{$sapinst}->[$select_on_IP_idx]) {  # could be obsolete with above initialization
         # Extend candidate list by pushing new list after old one. May cause duplicate entries (which may even
         # be an indication that such a package is better suited because it has more matching IPs)
         push (@{$_[2]->{$sapinst}->[$select_on_IP_idx]},@{$_[1]->{$instip}});
         # printL "push\n";
       }
       else {
         # Copy over package list for this IP as starting candiate list.
         @{$_[2]->{$sapinst}->[$select_on_IP_idx]} = @{$_[1]->{$instip}};
         # printL "Copy\n"
       }
     }
     #print "inst_ip [ @{$_[2]->{$sapinst}->[$select_on_IP_idx]} ] \n";
  }

  # Check if package occurs more than once. This is a candidate for multIP.
  foreach my $sapinst (keys %{$_[0]}) {
    $_[2]->{$sapinst}->[$select_on_multIP_idx] = [];
    # Ignore all for select on IP which contains ECMT modules
    # First get all which have the toolkit for reporting
    @hastk = grep { pkgHasTKmod($_, $whichtk, $_[3]) == 1 } @{$_[2]->{$sapinst}->[$select_on_IP_idx]};
    if (scalar @hastk) {
      sgeasy_print80 ("Warning: Packages with $whichtk modules ignored for selection on IP address:");
      sgeasy_print80 ("[ @hastk ]");
      @{$_[2]->{$sapinst}->[$select_on_IP_idx]} = grep { pkgHasTKmod($_, $whichtk, $_[3]) != 1 } @{$_[2]->{$sapinst}->[$select_on_IP_idx]};
    }
    foreach my $pkgname (@{$_[2]->{$sapinst}->[$select_on_IP_idx]}) {
      #print $pkgname,"\n";
      #pkgHasTKmod ($pkgname, "sgesap",$_[3]);
      my $pkgcount = grep {$_ eq $pkgname} @{$_[2]->{$sapinst}->[$select_on_IP_idx]};
      #print $pkgcount,"\n";
      if ($pkgcount >= 2) {
          my $pkgcount2 = grep {$_ eq $pkgname} @{$_[2]->{$sapinst}->[$select_on_multIP_idx]};
          push @{$_[2]->{$sapinst}->[$select_on_multIP_idx]},$pkgname if $pkgcount2 == 0;
      }
    }
  }

  # These IP packages should also be checked if they have no SAP SID in them ("empty SID"). Otherwise they are only useful
  #          if they are the same SID.

  return 1;
}

#############################################################################
#
# sub lookupEmptySIDPkgCandidatesIP
#
# further restrict package selection  of "IP packages" (-> lookupPkgCandidatesIP)
# by selecting those w/o a SAP SID and w/o a DB SID defined
#
# Will write to the $select_on_noSID_idx array of %instPkgCandidates (hash)
# Will lookup the SID in $sapsid_idx in %cluspkgconfighash
# Assumes that array at index $select_on_IP_idx is already filled
#
# Arguments:
# \%cluspkgconfighash - hash to check for empty/undef SID
#  \%instPkgCandidates  Reference to candiate hash (read,write)
#
#############################################################################

sub lookupEmptySIDPkgCandidatesIP {

  foreach my $sapinst (keys %{$_[1]}) {
    $_[1]->{$sapinst}->[$select_on_noSID_idx] = [];
    foreach my $pkgname (@{$_[1]->{$sapinst}->[$select_on_IP_idx]}) {
      # print $pkgname," - ",$_[0]->{$pkgname}->[$sapsid_idx],"\n";
      next if defined $_[0]->{$pkgname}->[$sapsid_idx]; # has a SAP SID -> not empty
      next if defined $_[0]->{$pkgname}->[$dbsid_idx];  # has a DB SID  -> not empty
      next if defined $_[0]->{$pkgname}->[$hdbsid_idx]; # has a HDB SID -> not empty
      # push onto array if not already there
      push @{$_[1]->{$sapinst}->[$select_on_noSID_idx]},$pkgname
         unless grep {$_ eq $pkgname} @{$_[1]->{$sapinst}->[$select_on_noSID_idx]}
    }
  }
}

#############################################################################
#
# sub lookupPkgCandidatesSID
#
# Lookup package candidates based on SID and on possibly configured instances.
# This goes over all packages again, not just them selected by matching IPs.
# Assumes that instPkgCandidates is already setup by lookupPkgCandidatesIP
#
# Arguments:
# SAP SID               SAP SID to look at
# \%cluspkgconfighash - hash to check the SID and the instances
# \%instPkgCandidates -  Reference to candiate hash (read,write)
# DB SID                SID of DB might be different to SAP SID
#
#############################################################################

sub lookupPkgCandidatesSID {

  # List to be filled with packages containing the required SID
  # and another list for those packages also containing instances
  # The first will (e.g) contain a DB only pkg which also has a SAPSID defined,
  # the second won't. The third is the list of packages where the DB SID is
  # equal to the SID
  my @SIDpkgs=();
  my @SIDinstpkgs=();
  my @DBpkgs=();
  my @HDBpkgs=();

  # loop through all packages, select them based on SID
  for my $pkg (keys %{$_[1]}) {
    # skip legacy (to avoid furhter unnecessary checks)
    next unless $_[1]->{$pkg}->[$pkgstyle_idx] eq "modular";
    # check if it has a SAPSID defined and is the same as the parameter (could be combined)
    if ((defined $_[1]->{$pkg}->[$sapsid_idx]) && ($_[1]->{$pkg}->[$sapsid_idx] eq $_[0])) {
      push @SIDpkgs,$pkg;
      if ( exists($_[1]->{$pkg}->[$sapinst_idx]) &&  @{$_[1]->{$pkg}->[$sapinst_idx]} ) {
        push @SIDinstpkgs,$pkg
      }
      # if defined it may also contains an instance (to be verified)
    }
    # use DB SID instead of SAP SID when checking for DB packages
    if ((defined $_[1]->{$pkg}->[$dbsid_idx]) && ($_[1]->{$pkg}->[$dbsid_idx] eq $_[3])) {
      push @DBpkgs,$pkg;
    }
    # HDB
    if ((defined $_[1]->{$pkg}->[$hdbsid_idx]) && ($_[1]->{$pkg}->[$hdbsid_idx] eq $_[3])) {
      push @HDBpkgs,$pkg;
    }
  }
  printL "Packages with SID $_[0]: [ @SIDpkgs ]\n";
  printL "Packages with SID $_[0] and instances: [ @SIDinstpkgs ]\n";
  printL "Packages with DBSID $_[0]: [ @DBpkgs ]\n";
  printL "Packages with HDBSID $_[0]: [ @HDBpkgs ]\n";

  # @SIDpkgs now contains all packages for the given SID,
  # Assign it to all instances and loop through
  # all of them to see which packages match to the instance
  foreach my $sapinst (keys %{$_[2]}) {
    # assign found packages to array in candiate list. this is done even if @SIDpkgs is empty
    @{$_[2]->{$sapinst}->[$select_on_sameSID_idx]} = @SIDpkgs;
    @{$_[2]->{$sapinst}->[$select_on_SIDInst_idx]} = @SIDinstpkgs;
    @{$_[2]->{$sapinst}->[$select_on_sameDBSID_idx]} = @DBpkgs;
    # also "add" HDBpkgs to sameDBSID (usually they should be disjunct),
    # TODO HDB: but only if it is a HDB instance itself...
    # TODO HDB: ... however, the instance type isn't available here (yet)
    # TODO HDB: use a separate index (criteria) here??
    # >>> disable temporarily to trigger creation of new packages
    push (@{$_[2]->{$sapinst}->[$select_on_sameDBSID_idx]},@HDBpkgs);

    $_[2]->{$sapinst}->[$select_on_sameInst_idx] = [];
    # check which of those SIDpkgs match even closer by checking if they have the same instance in them
    foreach my $pkg (@SIDpkgs) {
      next unless ( exists($_[1]->{$pkg}->[$sapinst_idx]) &&  @{$_[1]->{$pkg}->[$sapinst_idx]} );

      #my @greplist= grep {$_ eq $sapinst} @{$_[1]->{$pkg}->[$sapinst_idx]};
      #print "$sapinst - $pkg - [ @{$_[1]->{$pkg}->[$sapinst_idx]} ] - [ @greplist ] \n";
      push @{$_[2]->{$sapinst}->[$select_on_sameInst_idx]},$pkg
        if grep {$_ eq $sapinst} @{$_[1]->{$pkg}->[$sapinst_idx]};
    }
  }
}

#############################################################################
#
# sub lookupPkgCandidatePkgType
#
# Lookup package candidates based package type.
# Usually this is a failover package. So far, only HANA scaleout package
# may select a multinode package.
# Selection is not based on package type attribute, but on configuration of
# either failover or multi_node module into package. If no such packages
# are defined in cluster the index is left undef.
#
# Arguments:
# SAP SID               SAP SID to look at
# \%sapgchash           (ref to) hash of discovered instances  (read)
# \%clusrsrchash        (ref to) of resources used by packages (read)
# \%cluspkgconfighash - hash to check the SID and pkg style
# \%instPkgCandidates - (ref to) candiate hash (write)
#
#############################################################################

sub lookupPkgCandidatesPkgType {

	# loop over all discovered instances (again).
    foreach my $sapinst (keys %{$_[1]}) {
    	#printL "lookupPkgCandidatesPkgType: $sapinst\n";
    	$_[4]->{$sapinst}->[$select_on_pkgtype] = []; # initialize empty
    	my $pkgmod = "sg/failover"; # majority will be failover packages
    	#printL "$_[1]->{$sapinst}->[1]\n";
    	#print @{$_[2]->{"MOD_NAME"}->{"sg/failover"}};
    	if ($sapinst =~ /^HDB(p|s)-/) {
    		$pkgmod = "sg/multi_node" if (scalar(my @tmparr = split (",",$_[1]->{$sapinst}->[1])) > 1)
    	}
		# assign if something is defined there
    	@{$_[4]->{$sapinst}->[$select_on_pkgtype]} = @{$_[2]->{"MOD_NAME"}->{$pkgmod}} if (defined $_[2]->{"MOD_NAME"}->{$pkgmod} );
    }
}

#############################################################################
#
# sub lookupPkgCandidatesNodelist
#
# Lookup package candidates based explicit nodelist of package.
#
# If instance requests to be placed on specific cluster nodes (usually a subset of
# the cluster node ) the package is selected if the nodelist of the packages
# is exactly the nodelist of the instance.
# WVE: would it be ok if package just has a subset?
# If no such package is defined in cluster the index is left undef (or emtpy?).
#
# So far this sub is only used for HDB scaleout (>1 member in nodelist)
#
# Arguments:
# SAP SID               SAP SID to look at
# \%sapgchash           (ref to) hash of discovered instances  (read)
# \%clusrsrchash        (ref to) of resources used by packages (read)
# \%cluspkgconfighash - hash to check the SID and pkg style
# \%instPkgCandidates - (ref to) candiate hash (write)
#
#############################################################################

sub lookupPkgCandidatesNodelist {

    # loop over all discovered instances
    foreach my $sapinst (keys %{$_[1]}) {
	    #printL "lookupPkgCandidatesNodelist: $sapinst\n";
	    my @nodelistpkgs = ();
    	next unless ($sapinst =~ /^HDB(s|p)/);
    	my @instnodes = split (",",$_[1]->{$sapinst}->[1]);
    	next if (scalar @instnodes <= 1);
    	# two or more nodes, loop through all packages and check
NSP:    foreach my $pkg (keys %{$_[3]}) {
     	  printL "lookupPkgCandidatesNodelist:   pkg $pkg\n";
          next unless (defined $_[3]->{$pkg}->[$pkgnodelist_idx]);  # WVE: undef or empty?
    	  foreach my $instnode (@instnodes) {
    	  	printL "lookupPkgCandidatesNodelist:     instnode $instnode\n";
    		next NSP unless grep (/^${instnode}$/,@{$_[3]->{$pkg}->[$pkgnodelist_idx]});
    	  }
    	  # pkg match
    	  printL "nodelist match\n";
    	  push (@nodelistpkgs, $pkg);
     	}
     	# collected all pkgs
     	@{$_[4]->{$sapinst}->[$select_on_nodelist]} = @nodelistpkgs;
    }
}

#############################################################################
#
# sub lookupPkgCandidatesVHost
#
# Lookup package candidates based on matching virtual hostname.
# This goes over all packages again, not just them selected by matching IPs.
# Assumes that instPkgCandidates is already setup by lookupPkgCandidatesIP
#
# Arguments:
# SAP SID               SAP SID to look at
# \%sapgchash           (ref to) hash of discovered instances  (read)
# \%clusrsrchash        (ref to) of resources used by packages (read)
# \%cluspkgconfighash - hash to check the SID and pkg style
# \%instPkgCandidates - (ref to) candiate hash (write)
#
#############################################################################

sub lookupPkgCandidatesVHost {

    # list to be filled with packages containing the required vhost (and the required SID)

    # loop over all discovered instances (again).
    for my $sapinst (keys %{$_[1]}) {
        my @vhostpkgs=();
        @{$_[4]->{$sapinst}->[$select_on_VHost_idx]} = @vhostpkgs;
        # get vhost for this instance (if any)
        #print $sapinst,"\n";
        next unless defined $_[1]->{$sapinst}->[1];
        my $vhost = $_[1]->{$sapinst}->[1];
        # print "$vhost \n";
        # check if vhost is already used by a package
        next unless defined $_[2]->{"SAPVHOST"}->{$vhost};
        # vhost already host by some packages, loop through them and select those with the same SID
        for my $pkg (@{$_[2]->{"SAPVHOST"}->{$vhost}}) {
            # print "$pkg \n";
            # check if package is modular
            # and contains the required SID (pkg has a SAPSID defined because it already has sap vhost)
            next unless $_[3]->{$pkg}->[$pkgstyle_idx] eq "modular";
            next unless $_[3]->{$pkg}->[$sapsid_idx] eq $_[0];
            push @vhostpkgs,$pkg
        }
        # collected all vhost based package candiates for that SID
        @{$_[4]->{$sapinst}->[$select_on_VHost_idx]} = @vhostpkgs;
    }
}

#############################################################################
#
# sub lookupPkgCandidatesHDBrepl
#
# Match discovered HDB instances to package based on replication information.
# i.e. match a instance reporting replication to a package with replication
#   and do match a package not reporting replication to others (not replication).
#
# Arguments:
#  \%sapgchash          Ref to hash of discovered instances  (read)
#  \%clusip2pkgmap      Ref to hash of IP2Pkg map (read)
#  \%instPkgCandidates  Ref to candiate hash (write)
#  \%clusrsrchash       Ref to cluster resource crossref (read)
#
# Fills: %instPkgCandidates - index select_on_hdbrepl
#
#############################################################################
sub lookupPkgCandidatesHDBrepl {

    # loop over all discovered instances (again).
    for my $sapinst (keys %{$_[0]}) {
      $_[2]->{$sapinst}->[$select_on_hdbrepl] = [];  # initialize empty
      next if ( $sapinst !~ "^HDB(s|p|)-"); # nothing to do
      if ($_[0]->{$sapinst}->[0] =~ /primary/) {
        # loop through all packages. Select those having hdbsystem( or hdb_global) && hdbprimary module
        # hdbprimary implies hdbsystem/hdbglobal(?), therefore only hdbprimary is considered

          if ( exists($_[3]->{"MOD_NAME"}->{"sgesap/hdbprimary"}) && @{$_[3]->{"MOD_NAME"}->{"sgesap/hdbprimary"}}) {
            @{$_[2]->{$sapinst}->[$select_on_hdbrepl]} = @{$_[3]->{"MOD_NAME"}->{"sgesap/hdbprimary"}}
          }
      }
      if ($_[0]->{$sapinst}->[0] =~ /sync/) {
        # loop through all packages. Select those having hdbsystem/hdb_global && !hdbprimary module
        @{$_[2]->{$sapinst}->[$select_on_hdbrepl]} = [];
        foreach my $hdbglobpkg (@{$_[3]->{"MOD_NAME"}->{"sgesap/hdb_global"}}) {
          push (@{$_[2]->{$sapinst}->[$select_on_hdbrepl]}, $hdbglobpkg)
            unless (grep {$_ eq $hdbglobpkg} @{$_[3]->{"MOD_NAME"}->{"sgesap/hdbprimary"}});
        }
        # shortcut for above using map function (?):
        # @{$_[2]->{$sapinst}->[$select_on_hdbrepl]}
        #   = map { grep {$_ ne $hdbglobpkg} @{$_[3]->{"MOD_NAME"}->{"sgesap/hdb_global"}} ? () : $_ }
        #     @{$_[3]->{"MOD_NAME"}->{"sgesap/hdb_global"}}
        # or 2 nested grep (?)
        # grep { my $outer=$_; grep { $_ ne $outer} @{$_[3]->{"MOD_NAME"}->{"sgesap/hdbprimary"}} }
        #                                              @{$_[3]->{"MOD_NAME"}->{"sgesap/hdb_global"}}
        # any better than the loop?
      }
      # else - do nothing (don't know what do with such a setup yet)
    }
}

#############################################################################
#
# sub mapInst2Pkgcand
#
# map instance to package candiates
#
# Creates a hash with pkg candidates as keys and (an array of) the instances
# they cover as values. This is (kind of) a reverse hash of %instPkgCandidates
# (seems this routine only makes sense if there is an overlap in package candidates
# between instance. otherwise this will end up with a hash having single element
# arrays as values. For this, however, we may need to look beyond the filtered
# (i.e. SCS) instance. Consedering a case where a pure dialog is started with
# the SCS. We wouldn't get the D with sapgetconf, but the configured SGeSAP
# instance via sapgetconf... For this we should consider all instances of a SID.
# Not just those returned by sapgetconf...)
#
# Argument:
# \%instPkgCandidates  - hash to read from
# \%pkgInstCandidates  - hash to be created/filled
#
# Will consider only the pkg candidates in $select_on_combined_idx index of candiate arrays.
#
# This map will subsequently be used to rule out pkg candidates covering only a subset
# of instances of another package candiate
#
# Assumes that  %instPkgCandidates is already filled (i.e. that the associated
# code was executed already).
# Otherwise the hash to be filled will remain empty.
#############################################################################

sub mapInst2Pkgcand {

  foreach my $sapinst (keys %{$_[0]}) {
     foreach my $candpkg (@{$_[0]->{$sapinst}->[$select_on_combined_idx]}) {
       push @{$_[1]{$candpkg}}, $sapinst
     }
  }
}

#############################################################################
#
# sub selectPkgCandiate
#
# Select a package candidate for each instance.
# ..based on strategy (TBD: multi/single) and other criteria (TBD)
# There can be more than one on no candidate per instance
#
# Possible critera:
# - same IP(s) & no SID
# - same IP(s) & same SID
# - same instance
# - same instance & same IP
# - same DB SID
# - if no IP select same instance
# - ....
#
# ToDo: final selection may (has to) depend on instance type (SCS,ERS,DVEBMGS)
# and "combination type" ("single","multi")
#
# Arguments:
#   \%instPkgCandidates  Reference to candiate hash (read,write)
#           (write last index of %instPkgCandidates)
#   \@sapinstshrdres  (ref to) array containing instances sharing resources
#           (could be empty)
#   $multiple	if set to 1 tries to put each instance in separate package,
#               otherwise it tries to combine them
#
# Returns:
#   1 - package selection ok
#   0 - some kind of problem with selection (e.g. multiple packages possible)
#
#
#############################################################################

sub selectPkgCandiate {

    # current selection mechanism
    # select set of "same IP",
    # if more than one, see if there is one also one in "same instance"
    #   if so, select this/these
    #   if not, check if there is one or moe in "same SID"
    # if there is now only one left to select use that one,
    # otherwise check if one of the so far selected is also in "multiple IP"
    # if more than one -> errror, otherwise select that one

    # used indexes:
    #$select_on_IP_idx       = 0; # package selected based on IP
    #$select_on_multIP_idx   = 1; # package selected because it contains more multiple necessary IPs (future use)
    #$select_on_noSID_idx    = 2; # package selected because it doesn't contain a SID (but a matching IP)
    #$select_on_sameSID_idx  = 3; # package selected based on (same) SAP SID  (unrelated to IP)
    #$select_on_sameInst_idx = 4; # package selected based on (same) SAP instance (implies "sameSID", also unrelated to IP)
    #$select_on_SIDInst_idx  = 5; # package selected based on instance of the same SAP SID (but not the very same instance)
    #$select_on_sameDBSID_idx = 6; # DB SID is the same as given SID (may also be in sameSID).
    #$select_on_VHost_idx    = 7; # selection done based on SAP virtual hostname
    #$select_on_combined_idx = 8; # package selected on predicates applied to the previous sets
    #$select_shared_res = 9; # helper to store shared resource instances (are not packages)
    #$select_preferred = 10 # helper...
    #$select_on_nodelist =    # same nodelist
    #$select_on_pkgtype =     # selection on required package type (failover,...)
    #$select_final = 11; # final candidates (could be empty)

    my $multipkg = $_[2];
    my $result = 1;

    # SCS and ERS are needed later. Store them in array during first loop
    my @SCSinst = ();
    my @ERSinst = ();

    # uses an internal anonymous helper subroutine to do the intersection between 2 arrays

    if (($ENV{"deploysappkgs_SkipPkgSelect"} || 0) == 1) {
      printL "deploysappkgs_SkipPkgSelect is set. Skipping existing pkg selection\n";
      # shortcut everything by setting select_final to empty list
      foreach my $sapinst (keys %{$_[0]}) {
        @{$_[0]->{$sapinst}->[$select_final]} = ();
      }
      return 1;
    }
    my $intersect = sub {
        my @Artn = ();
        # return empty set if one (or both) of the args are empty
        return @Artn unless (scalar @{$_[0]} and scalar @{$_[1]});
        foreach my $elem (@{$_[0]}) {
            push @Artn, $elem if grep {$_ eq $elem } @{$_[1]};
        }
        return @Artn;
    };

    foreach my $sapinst (keys %{$_[0]}) {
        # print "$sapinst\n";
        # my $lclref = $_[0]->{$sapinst};
        # if a DVEBMGS is encountered, it is also considered a SCS (otherwise
        # it would have been filtered out already)
        $_[0]->{$sapinst}->[$select_on_combined_idx] = [];
        push @SCSinst,$sapinst if ($sapinst =~ /^(ASCS|SCS|DVEBMGS)\d\d$/);
        push @ERSinst,$sapinst if ($sapinst =~ /^ERS\d\d$/);
        # Start out with the $select_on_IP_idx array - if not empty.
        if (scalar @{$_[0]->{$sapinst}->[$select_on_IP_idx]}) {
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]}= @{$_[0]->{$sapinst}->[$select_on_IP_idx]};
        }
        elsif (scalar @{$_[0]->{$sapinst}->[$select_on_VHost_idx]}) {
            # no package with matching IP, try virt. host (vhost<->IP mapping may have changed,
            # but usually the IP for the vhost is also in the package)
            # Empty list can't be returned (scalar above)
            # This won't help for DB package as they don't have a specific vhost configured
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]}= @{$_[0]->{$sapinst}->[$select_on_VHost_idx]};
            next;
        }
        elsif ($sapinst =~ /^DB-\w+/) { # special "instance" name (DB-<SID>)
            # "sapinst" is the DB. use $select_on_sameDBSID_idx (maybe empty)
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @{$_[0]->{$sapinst}->[$select_on_sameDBSID_idx]};
            next;
        }
		elsif ($sapinst =~ /^REP-\w+/) { # special "instance" name (REP-<SID>)
            # "sapinst" is the replication. use $select_on_sameDBSID_idx (maybe empty)
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @{$_[0]->{$sapinst}->[$select_on_sameDBSID_idx]};
            next;
        }
        elsif ($sapinst =~ /^HDB(s|p|)-\w+/) { # special "instance" name (HDB-<SID>)
            # "sapinst" primary or secondary. use $select_on_sameDBSID_idx (maybe empty)
            my @tmpArr=();
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @{$_[0]->{$sapinst}->[$select_on_sameDBSID_idx]};
            @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_hdbrepl]);
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr; #if (scalar @tmpArr);
            # Intersecting with select_on_hdbrepl isn't just a narrowing down at possible options. Result empty means
            # that replication restrictions aren't meet by currently configured packages and none of these should be used
            # So the intersection should be assigned even if empty.
            #
            # new intersections for HDB scaleout (influencing also scaleup): intersect with (new) package type index
            # and nodelist index
            # WVE: on "empty" cluster is select_on_pkgtype undef or empty?
            @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_pkgtype]);
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr;
            if (defined $_[0]->{$sapinst}->[$select_on_nodelist]) {
            	@tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_nodelist]);
            	@{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr;
            }
            next;
        }
        else {
            # Otherwise use  $select_on_sameInst_idx & return (this maybe the empty list!)
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @{$_[0]->{$sapinst}->[$select_on_sameInst_idx]};
            next;
        }

        # intersect "IP list" with $select_on_sameInst_idx and return if result isn't empty
        # (intersect with $select_on_sameDBSID_idx when "instance" is DB)
        # otherwise intersect with $select_on_sameSID_idx and return if not empty
        # otherwise intersect with $select_on_noSID_idx (to rule out other SIDs)
        my @tmpArr=();
        @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_sameInst_idx]);
        #print "select_on_sameInst_idx - [ @tmpArr ] \n";
        if (scalar @tmpArr) {
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr;
            next;
        }
        @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_sameSID_idx]);
        #print "select_on_sameSID_idx - [ @tmpArr ] \n";
        if (scalar @tmpArr) {
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr;
            next;
        }
        @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_sameDBSID_idx]);
        #print "select_on_sameDBSID_idx - [ @tmpArr ] \n";
        if (scalar @tmpArr) {
            @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr;
            next;
        }
        # special HDB handling. Looking at (non-)replication
        #if ($sapinst =~ /^HDB-\w+/) {
        #  #print "select_on_sameDBSID_idx - [ @tmpArr ] \n";
        #  @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_hdbrepl]);
        #  if (scalar @tmpArr) {
        #    @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr;
        #    next;
        #  }
        #}

        @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_noSID_idx]);
        #print "select_on_noSID_idx - [ @tmpArr ] \n";
        @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr; # assign even if empty

    }

    # Another round to handle those yielding more than 1 package. These will combined with the
    # $select_on_multIP_idx packages
    # (this may lead to the question why a package support two IPs and 1 instance should be a better
    # candidate than one having 1 IP and 2 instances? Maybe it is a better idea to narrow down the sets
    # first by looking cross instance which they have in common).
    foreach my $sapinst (keys %{$_[0]}) {
        next if (scalar @{$_[0]->{$sapinst}->[$select_on_combined_idx]} < 2);
        my @tmpArr = ();
        @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$sapinst}->[$select_on_multIP_idx]);
        # use result only if it isn't empty
        @{$_[0]->{$sapinst}->[$select_on_combined_idx]} = @tmpArr if (scalar @tmpArr);
    }

    # loop through \@sapinstshrdres first and assign shared resource array (index 9)
    # print STDERR Data::Dumper->Dump( [ \@{$_[1]} ], [ "arrrefs" ] );
    for my $arrref (@{$_[1]}) {
        # arrref contains now (ref to) array of instances already sharing resource
        for my $sapinst (@{$arrref}) {
            if (defined $_[0]->{$sapinst}->[$select_shared_res]) {
                # sapinst shares different resources with different group of packages. would need to combine
                # these groups...  - but so far we're giving up
                sgeasy_print80 "$sapinst already shares resources with [ @($_[0]->{$sapinst}->[$select_shared_res])]";
                sgeasy_print80 "Conflicts with [ @{$arrref} ]\n";
                return 0;
            }
            else {
                # assign array
                @{$_[0]->{$sapinst}->[$select_shared_res]} = @{$arrref};
            }
        }
    }
    # Check $select_shared_res if the ERS share and SCS not, SCS have to be set
    # to share a (imaginary) resource (same for SCS sharing resource and ERS not)
    # TODO ENHANCEMENT
    # TODO there is already a SCSlist and ERSlist
    my @scslist = grep (/SCS\d\d$/, keys %{$_[0]});
    my @erslist = grep (/^ERS\d\d$/, keys %{$_[0]});
    my $scsshare = 0;
    my $ersshare = 0;
    if (scalar(@scslist) + scalar (@erslist) == 4) {
      printL "Full list of SCS/ERS pairs [ @scslist @erslist ] \n";
      # check if SCS share, by searching the 2nd in the shared list
      # of the first (this is symmetrical"")
      if (grep (/$scslist[1]/,@{$_[0]->{$scslist[0]}->[$select_shared_res]})) {
        printL "SCS sharing resource: $scslist[1],@{$_[0]->{$scslist[0]}->[$select_shared_res]} \n";
        $scsshare = 1;

      }
      if (grep (/$erslist[1]/,@{$_[0]->{$erslist[0]}->[$select_shared_res]})) {
        printL "ERS sharing resource: $erslist[1],@{$_[0]->{$erslist[0]}->[$select_shared_res]} \n";
        $ersshare = 1;
      }
      if ($scsshare && !$ersshare) {
        # add other ERS to share list
        printL "Adding other ERS to ERS shared_res list \n";
        push @{$_[0]->{$erslist[0]}->[$select_shared_res]},$erslist[1];
        @{$_[0]->{$erslist[1]}->[$select_shared_res]} = @{$_[0]->{$erslist[0]}->[$select_shared_res]};
      }
      if (!$scsshare && $ersshare) {
        # add other SCS to share list
        printL "Adding other SCS to SCS shared_res list \n";
        push @{$_[0]->{$scslist[0]}->[$select_shared_res]},$scslist[1];
        @{$_[0]->{$scslist[1]}->[$select_shared_res]} = @{$_[0]->{$scslist[0]}->[$select_shared_res]};
      }
    }

    # loop through arrays at index 9 and intersect index 8 arrays
    # based on sharing information
    foreach my $sapinst (keys %{$_[0]}) {
        #printL "x: $sapinst\n";
        if (scalar @{$_[0]->{$sapinst}->[$select_shared_res]} < 2) {
            # no sharing, copy array over
            @{$_[0]->{$sapinst}->[$select_final]} = @{$_[0]->{$sapinst}->[$select_on_combined_idx]};
            next;
        }
        # intersect with other instances sharing resources with this sapinst
        # message: if both combined_idx aren't empty and it'll lead to an empty list
        # if one is empty, assign the other one. if both are empty leave it empty
        # (not quite sure here, if the select_final instance of the otherinst should
        # also be checked)
        foreach my $otherinst (@{$instPkgCandidates{$sapinst}->[$select_shared_res]}) {
            next if $sapinst eq $otherinst; # don't intersect with self
            my $num1cand = scalar (@{$_[0]->{$sapinst}->[$select_on_combined_idx]});
            my $num2cand = scalar (@{$_[0]->{$otherinst}->[$select_on_combined_idx]});
            my @tmpArr = ();
            if (($num1cand == 0) && ($num2cand == 0)) {
              # set to empty
              @{$_[0]->{$sapinst}->[$select_final]} = ();
            }
            elsif (($num1cand > 0) && ($num2cand == 0)) {
              @{$_[0]->{$sapinst}->[$select_final]} = @{$_[0]->{$sapinst}->[$select_on_combined_idx]};
            }
            elsif (($num1cand == 0) && ($num2cand > 0)) {
              @{$_[0]->{$sapinst}->[$select_final]} = @{$_[0]->{$otherinst}->[$select_on_combined_idx]};
            }
            else {
              # both lists have members, intersect them (accept also empty list, but issue a message)
              @tmpArr = $intersect->($_[0]->{$sapinst}->[$select_on_combined_idx],$_[0]->{$otherinst}->[$select_on_combined_idx]);
              if (scalar (@tmpArr) == 0) {
                sgeasy_print80 ("No common existing package between $sapinst and $otherinst could be selected");
                sgeasy_print80 ("([ @{$_[0]->{$sapinst}->[$select_on_combined_idx]} ][ @{$_[0]->{$otherinst}->[$select_on_combined_idx]} ])");
              }
              @{$_[0]->{$sapinst}->[$select_final]} = @tmpArr;
            }
        }
    }

    # evaluate selection
    # printL "Existing packages selected for sapinstances\n";
    foreach my $sapinst (keys %{$_[0]}) {
        sgeasy_print80 "Existing package selection for $sapinst: [ @{$instPkgCandidates{$sapinst}->[$select_final]} ]";
        if (scalar @{$_[0]->{$sapinst}->[$select_final]} > 1) {
            sgeasy_print80 "ERROR: More than 1 package selected for $sapinst\n";
            # in some configs (Metrocluster/site aware?) this isn't an error as (almost?) identical package(s) maybe configured
            # for different sites
            $result = 0;
        }
    }
    # check here if SCS and ASCS have different packages and ERS must be combined
    # (or are destined for one package) - and vica versa.
    # (could also use the SCS and ERS-arrays..)
    my ($combined_SCS,$combined_ERS) = check_combined_SCS_ERS ($_[0]);
    if ($combined_SCS || $combined_ERS) {
        # check if selected packages for SCS and ERS are equal
        # combined ERS also implies that there are two SCS instances which are to be combined
        # maybe empty ([0] undefined)
        my $SCSpkg1 = @{$_[0]->{$SCSinst[0]}->[$select_final]}[0];
        my $SCSpkg2 = @{$_[0]->{$SCSinst[1]}->[$select_final]}[0];
        if (((defined $SCSpkg1) && (defined $SCSpkg2)) && ($SCSpkg1 ne $SCSpkg2)) {
            sgeasy_print80 "ERROR: Combined SCS instances [ @SCSinst ] have different package selection [ $SCSpkg1 $SCSpkg2 ]";
            return 0;
        }
        if ($combined_ERS) {
            my $ERSpkg1 = @{$_[0]->{$ERSinst[0]}->[$select_final]}[0];
            my $ERSpkg2 = @{$_[0]->{$ERSinst[1]}->[$select_final]}[0];
            if (((defined $ERSpkg1) && (defined $ERSpkg2)) && ($ERSpkg1 ne $ERSpkg2)) {
                sgeasy_print80 "ERROR: Combined ERS instances [ @ERSinst ] have different package selection [ $ERSpkg1 $ERSpkg2 ]";
                return 0;
            }
        }
    }
    # if combined_SCS and !$combined_ERS handled above
    # also: if !combined_SCS and $combined_ERS
    # all !combined_ isn't an interesting case here.

    return $result
}

#############################################################################
#
# sub check_combined_SCS_ERS
#
#
# Checks if SCS/ASCS need to be combined
# Done looking into \%instPkgCandidates $select_shared_res index
# (@sapinstshrdres could also be used)
#
# Returns: 2 element array values 0,1
#          1st index: SCS combined, 2nd index: ERS combined
#
# So far this decision ist just based on the fact if instances share
# a common resource (IP address)
#
#############################################################################

sub check_combined_SCS_ERS {
   my $combined_SCS=0;
   my $combined_ERS=0;
   foreach my $sapinst (keys %{$_[0]}) {
     if ($sapinst =~ /^(ASCS|SCS)\d\d$/) {
        # each SCS returns same result
        #$combined_SCS = 1 if (scalar @{$_[0]->{$sapinst}->[$select_shared_res]} > 1);
        # The check above may not be sufficient as a SCS can share its resources with
        # another instance than the other SCS
		# Better check if is "other SCS" shows ups in shared resource list (be aware that
		# the list might also contain the instance itself)
		printL "shared list $sapinst: @{$_[0]->{$sapinst}->[$select_shared_res]}\n";
		$combined_SCS = 1 if (grep (/^(ASCS|SCS)\d\d$/, grep(/!$sapinst/,@{$_[0]->{$sapinst}->[$select_shared_res]})) > 0);
     }
     if ($sapinst =~ /^ERS\d\d$/) {
        # each ERS returns same result
        $combined_ERS = 1 if (scalar @{$_[0]->{$sapinst}->[$select_shared_res]} > 1);
        # this check is sufficient as we made sure that only ERS shares with ERS (done in buildshrdres)
     }
   }
   return ($combined_SCS,$combined_ERS);
}

#############################################################################
#
# sub assignAllUnassignedInst
#
# Assign all yet assigned instances to either an existing package or newly
# to be created package
# Reads (and possibly) write just array at index
# $select_final = 10; - final candidates (could be empty)
#
#  \%instPkgCandidates - (ref to) package candidate arrays (read/write)
#  \%CandPkg2Inst      - (ref to) package to instance map (write)
#  $sapsid    - SAP SID
#  $packaging - 1 if multi
#  \%clusrsrchash       - Ref to cluster resource crossref (read)
#  \%sapgcinsthash      - instance hash build from sapgetconf output
#                           (now required to determine HANA scaleout/up)
#
#  Returns: true/false
#
#############################################################################
sub assignAllUnassignedInst
{
   my $sapsid = $_[2];
   my $multi = $_[3];

   my $sitename = "";  # store (HDB) sitename here. Use at the end of name building
                       # together with sapsid

   # Loop through all instance arrays again
   # If not empty search for an appropriate "partner" with non-empty pkg name

NSI: foreach my $sapinst (keys %{$_[0]}) {
        next if (scalar @{$_[0]->{$sapinst}->[$select_final]} > 0); # has already a candidate
        #printL "Unassigned: $sapinst\n";
        #print STDERR Data::Dumper->Dump( [ \@{$_[0]->{$sapinst}} ], [ "sapinst" ] );
        if ($multi) {
            # build package according to packages needing combination (despite multi)
            # based on shared resources, but start out with sapinst name alone (as
            # [$select_shared_res] array might be undefined in certain cases).
            my $pkgnameadd = "$sapinst";
            if ( exists($_[0]->{$sapinst}->[$select_shared_res]) && @{$_[0]->{$sapinst}->[$select_shared_res]}) {
              $pkgnameadd = join ('',@{$_[0]->{$sapinst}->[$select_shared_res]});
            }
			#printL "pkgnameadd: $pkgnameadd\n";
            # TODO: decide if NFS, DB, HDB, REP appear uppercase or lowercase in pkgname here.
            #       or lowercase everything except SID after the packagename has been built up entirely
            $pkgnameadd = "${1}NFS${2}" if ($pkgnameadd =~ /^(\w*)NFS-\|\S+\|(\S*)$/); # just add "NFS" (before replacing DB!)
			#printL "pkgnameadd: $pkgnameadd\n";
			$pkgnameadd = "${1}REP${2}"  if ($pkgnameadd =~ /^(\w*)REP-\w\w\w(\w*)$/); # just add "REP" not "REP-<SID>"
			if ($pkgnameadd =~ /^(\w*)HDB(p|s|)-\w\w\w(\w*)$/) {
				# just add "HDB" not "HDB-<SID>" for scaleup and scaleout, but store site for latter
				$pkgnameadd = "${1}HDB${2}";

				if (scalar(my @tmparr = split (",",$_[5]->{$sapinst}->[1])) > 1) {
					$sitename = (my @tmparr = split (",",$_[5]->{$sapinst}->[0]))[2];
					printL "sitename: $sitename\n";
				}

				# WVE SO: Need to at site above. But site not passed into the datastructure passed into this sub.
                          #         Refer back to sapgchash . This needs an additional parameter!!	($_[5])
			}
            $pkgnameadd = "${1}DB${2}"   if ($pkgnameadd =~ /^(\w*)DB-\w\w\w(\w*)$/); # just add "DB" not "DB-<SID>"
			#printL "pkgnameadd: $pkgnameadd\n";
            #my $pkgname = $sapinst.$sapsid;
            # WVE ENHANCEMENT
            # WVE check if package name is already in use by cluster (will cause conflict)

            # Check if package needs to be combined with other instance(s) based on resource
            # If so and one of them has already a name assigned, use that one. (could need
            # combination with more than one other inst which have different packages assigned.
            # This isn't checked!!)
            # Otherwise assign a own package name.
            foreach my $otherinst (@{$_[0]->{$sapinst}->[$select_shared_res]}) {
              next if $sapinst eq $otherinst;
              if (scalar @{$_[0]->{$otherinst}->[$select_final]} > 0) {
                printL "Using $otherinst package assignment for $sapinst\n";
                printL "[ @{$_[0]->{$otherinst}->[$select_final]} ] \n";
                @{$_[0]->{$sapinst}->[$select_final]} = @{$_[0]->{$otherinst}->[$select_final]};
                # so far we don't use the package name extension mechanism here
                next NSI; # assigned something, check next sapinst
              }
            }
            $sitename = "_".$sitename if ($sitename ne "");
            printL "Assigning package name ${pkgnameadd}${sapsid}${sitename} to $sapinst\n";
            push @{$_[0]->{$sapinst}->[$select_final]}, $pkgnameadd.$sapsid.$sitename ;
        }
        else { # combi
            # check preferred other instances list if they have already a package assigned
            my $pkgnamechanged;  # set to old package name if extended
            for my $otherinst (@{$_[0]->{$sapinst}->[$select_preferred]}) {
              if (scalar @{$_[0]->{$otherinst}->[$select_final]} > 0) {
                printL "$sapinst preferred $otherinst ([ @{$_[0]->{$otherinst}->[$select_final]} ])\n";
                my $curpkgname = @{$_[0]->{$otherinst}->[$select_final]}[0];

                if ($curpkgname =~ /^\*\w+$/) {
                   # newly created name (with '*') to be extended
                   my $pkgnameadd = $sapinst;
                   # TODO: Lower/uppercase issues as with "multi"
                   $pkgnameadd = "HDB"  if ($pkgnameadd =~ /^HDB(p|s|)-\w+$/); # just add "HDB" not "HDB-<SID>"
                   $pkgnameadd = "DB"   if ($pkgnameadd =~ /^DB-\w+$/);  # just add "DB" not "DB-<SID>"
				   $pkgnameadd = "REP"  if ($pkgnameadd =~ /^REP-\w+$/); # just add "REP" not "REP-<SID>"
                   $pkgnameadd = "NFS"  if ($pkgnameadd =~ /^NFS-\|\S+\|$/); # just add "NFS"
                   printL "curpkgname before subst: $curpkgname\n";
                   # replace '*' by '*$sapinst'
                   $pkgnamechanged = $curpkgname;
                   $curpkgname =~ s/^\*(\w+)$/\*${pkgnameadd}$1/;
                   printL "curpkgname after subst: $curpkgname\n";
                   # extend otherinst package name
                   @{$_[0]->{$otherinst}->[$select_final]}[0] = $curpkgname;
                   # Need to update the other sapinst using the old name ($pkgnamechanged)
                   # A bit clumsy but works
                   for my $sapinst2 (keys %{$_[0]}) {
                     @{$_[0]->{$sapinst2}->[$select_final]}[0] = $curpkgname
                       if ((defined @{$_[0]->{$sapinst2}->[$select_final]}[0])
                           && (@{$_[0]->{$sapinst2}->[$select_final]}[0] eq $pkgnamechanged));
                   }
                }
                @{$_[0]->{$sapinst}->[$select_final]} = @{$_[0]->{$otherinst}->[$select_final]};

                next NSI; # goto next sapinst
              }
            }
            # nothing assigned based on preferred instance
            # loop again to find one with a candidate

            foreach my $otherinst (keys %{$_[0]}) {
                next if (scalar @{$_[0]->{$otherinst}->[$select_final]} == 0); # doesn't have one, no candidate
                # Check if this instance has a package name assigned which was extended in a previous iteration
                # If so, replace it
                # (so far this code assumes that there isn't more than one package in the list. this is ok here
                # as there are no previous candidates and there's only one newly genereated package name assigned)
                #printL "$pkgnamechanged , @{$_[0]->{$otherinst}->[$select_final]}[0]\n";
                if (defined $pkgnamechanged) {
                    if ($pkgnamechanged eq @{$_[0]->{$otherinst}->[$select_final]}[0]) {
                        printL "Updating pkg name $pkgnamechanged of $otherinst to @{$_[0]->{$sapinst}->[$select_final]}[0]\n";
                        @{$_[0]->{$otherinst}->[$select_final]} = @{$_[0]->{$sapinst}->[$select_final]};
                    }
                    # already had a changed name generated and assigned to the current sapinst, so there's no need
                    # to look for another candidate package
                    next;
                }

                #printL "Trying: $otherinst\n"
                if (compatibleInst ($sapinst, $otherinst)) {
                    printL "Assigning $otherinst package [ @{$_[0]->{$otherinst}->[$select_final]} ] to $sapinst\n";
                    # Check if this one begins with a '*"
                    my $curpkgname = @{$_[0]->{$otherinst}->[$select_final]}[0];
                    if ($curpkgname =~ /^\*\w+$/) {
                        my $pkgnameadd = $sapinst;
                        # yet another add? TODO HDB
                        $pkgnameadd = "DB"  if ($pkgnameadd =~ /^DB-\w+$/); # just add "DB" not "DB-<SID>"
                        $pkgnameadd = "HDB"  if ($pkgnameadd =~ /^HDB(p|s|)-\w+$/); # just add "HDB" not "HDB-<SID>"
						$pkgnameadd = "REP"  if ($pkgnameadd =~ /^REP-\w+$/); # just add "REP" not "REP-<SID>"
                        $pkgnameadd = "NFS" if ($pkgnameadd =~ /^NFS-\|\S+\|$/); # just add "NFS"
                        printL "curpkgname before subst: $curpkgname\n";
                        # replace '*' by '*$sapinst'
                        $pkgnamechanged = $curpkgname;
                        $curpkgname =~ s/^\*(\w+)$/\*${pkgnameadd}$1/;
                        printL "curpkgname after subst: $curpkgname\n";
                        @{$_[0]->{$otherinst}->[$select_final]}[0] = $curpkgname;
                    }
                    @{$_[0]->{$sapinst}->[$select_final]} = @{$_[0]->{$otherinst}->[$select_final]};
                    last unless defined $pkgnamechanged;  # otherwise continue looping and replace old names
                }
            }
            if (scalar @{$_[0]->{$sapinst}->[$select_final]} == 0) {
                # nothing found, assign something "random"
                # "*" should be replaced by additional instance(s)
				# Add just "DB" or "NFS" (as above) as there might be also cases where
				# this is the first "instance"
                # TODO: below is the 3rd place dealing with replacements based on HDB,DB,.. prefixes...
				my $pkgnameadd = $sapinst;
                $pkgnameadd = "HDB"  if ($pkgnameadd =~ /^HDB(p|s|)-\w+$/); # just add "HDB" not "HDB-<SID>"
                $pkgnameadd = "DB"  if ($pkgnameadd =~ /^DB-\w+$/); # just add "DB" not "DB-<SID>"
				$pkgnameadd = "REP"  if ($pkgnameadd =~ /^REP-\w+$/); # just add "REP" not "REP-<SID>"
                $pkgnameadd = "NFS" if ($pkgnameadd =~ /^NFS-\|\S+\|$/); # just add "NFS"
                my $pkgname = "*".$pkgnameadd.$sapsid;
                printL "No instance found to combine with $sapinst. Setting up temporary package name $pkgname\n";
                push @{$_[0]->{$sapinst}->[$select_final]} , $pkgname ;
            }
        }
    }

    # remove '*' if still in
    foreach my $sapinst (sort keys %{$_[0]}) {
    	my $curpkgname = @{$_[0]->{$sapinst}->[$select_final]}[0];
        if ($curpkgname =~ /^\*(\w+)$/) {
            # printL "curpkgname: $curpkgname - $1\n";
            $curpkgname = $1; # s/^\*(\w+)$/\1/;  -- no need for subst
            # also move NFS or DB to the end (before the SID) in generated names
            # TODO HDB: what about HDB? (usually, at the end anyway as there is no other item configured in it)
            $curpkgname = $1.$2."NFS".$3 if ($curpkgname =~ /^(\w*)NFS(\w+)(\w{3})$/);
            $curpkgname = $1.$2."DB".$3  if ($curpkgname =~ /^(\w*)DB(\w+)(\w{3})$/);
            @{$_[0]->{$sapinst}->[$select_final]}[0] = $curpkgname;
        }
        $curpkgname = pkgname2lc ($curpkgname,$_[4]);
        if (defined $_[1]->{$curpkgname}) {
            push @{$_[1]->{$curpkgname}},$sapinst;
        }
        else {
            $_[1]->{$curpkgname} = [ $sapinst ];
        }
    }
}

#############################################################################
#
# sub pkgname2lc
#
# Convert newly created packagenames to lowercase except SID
#
#  $curpkgname  - name to be converted
#  \%clusrsrchash - Ref to cluster resource crossref for current pkgnames (read)
#
# Returns:
#   Either unconverted pkgname if pkg already exists or convert if newly created
#
#############################################################################
sub pkgname2lc {
  my $pkgname = shift;
  my $crh = shift;
  my $prefix;
  my $sid;



  return $pkgname if (grep {$_ eq $pkgname} keys %{$crh->{"PKG_NAME"}}); # exist. pkg

  # convert to lc - so far no other rearrangements (up to what looks like a SID)

  if ($pkgname =~ /^(\w+)([A-Z][A-Z0-9][A-Z0-9])(_\w+)$/) {
  	# seems to contains a site name
  	$pkgname = lc($1).$2.$3;
  }
  elsif ($pkgname =~ /^(\w+)([A-Z][A-Z0-9][A-Z0-9])$/){
  	$pkgname = lc($1).$2;
  }
  return $pkgname;
}


#############################################################################
#
# sub compatibleInst
#
# Check if instances are compatible to go into one package.
# ERS is considered to be compatible on with other ERS. All others
# are compatible with everything else (of course except ERS)
#
#  $inst1  - instances to
#  $inst2  - check compatibility
#
#  Returns: true/false
#
#############################################################################
sub compatibleInst {
  my ($inst1, $inst2) = @_;

  return 1 if (($inst1 !~ /^ERS\d\d$/) && ($inst2 !~ /^ERS\d\d$/));
  return 1 if (($inst1 =~ /^ERS\d\d$/) && ($inst2 =~ /^ERS\d\d$/));

  return 0; # one is a ERS, the other not
}

#############################################################################
#
#############################################################################

#############################################################################
#
# sub getclusternodes
#
# Get physical nodes of cluster.
#
# Calls cmviewcl -f line -l node
# and parses for
#   node:<node>|name=<node>
#
#  Returns: Array of nodes
#
# Could be combinated with getting package information
# using no "-l" option there
#
# Candiate for sap_functions.pm
#
#############################################################################

sub getclusternodes {

  my @nodearray=();
  my $cmd=qq[$ENV{SGSBIN}/cmviewcl -f line -s config -l node 2>/dev/null];

  open IN, "$cmd |" or die "Can't run $cmd: $?";
  while (<IN>) {
    chomp;
    if ( $_ =~ m#^node:(\S+)\|name=(\S+)$# ) {
      push @nodearray,$1;
      next;
    }
  }
  close IN or warn "Couldn't run $cmd: $?";

  return @nodearray;
}

#############################################################################
#
# subroutines dealing with creation of package configs
#
#############################################################################

#############################################################################
#
# sub compare_mods
#
# Compare module names from both configs and figure out if 1st config
# (usually the one from cmgetconf) has modules missing required by the 2nd conf
#
# Arguments:
#   \@cmhash     Reference of cmgetconf hash
#   \@saphash    Reference of sapgetconf hash
#
# Returns:
#  Array of modules missing on first hash (that is in 2nd, but not in 1st)
#
#############################################################################

sub compare_mods
{
   my $mod;
   my @nomatch = ();
   # loop through array of 2nd hash and check if it is in first hash
   for $mod (@{${$_[1]}{"module_name"}}) {
     if (grep {$_ eq $mod} @{${$_[0]}{"module_name"}}) {
     }
     else {
       push @nomatch, $mod;
     }
   }
   return @nomatch;
}

#############################################################################
#
# sub mod2keys
#
#  Get the "key attributes" of for a list of modules
# (be aware that there is no 1:1 relationship between module and (key) attribute(
#
# Arguments:
#   @mods     Array containing the modules
#
# Uses:
#  %SGKeyAttr hash
#
# Returns:
#  Array of key attributes
#
#############################################################################

sub mod2keys {
  my @themods = @_;
  my @thekeys = ();

  foreach my $keyattr (keys %SGKeyAttr) {
    push @thekeys,$keyattr if grep { $_ eq $SGKeyAttr{$keyattr} } @themods;
  }
  return @thekeys;
}

#############################################################################
#
# sub addmods2config
#
# add modules required by discovered attributes to config
# two subcases:
#  (1) pkg doesn't exist yet (indicated by 2nd parameter being 0)
#  (2) pkg does exist (indicated by !=0)
#  In the first case an initial pkg config doesn't need to be generated
#  to be included - compared to the second case.
#
# Arguments:
#  $pkg        package name
#  $pkgexists  flag indicating if packge already exists
#  @mods       modlues to be included
#
# Uses:
#  $cmgetconf_cmd
#  cmmakepkg commmand
#
# Returns:
#  Name of newly generated package config file
#
#############################################################################

#
# Returns: Filename (full pathc) containing new config
sub addmods2config
{
  my $pkgname = shift;
  my $pkgexists = shift;
  my @newmods = @_;  # rest are modules
  # my $pkgpath="/tmp"; -- global
  my $modoptlist="";
  my $cmd2exec="";
  my $oconf= "$pkgpath"."/"."$pkgname.$$.o.conf";  # orig conf
  my $iconf= "$pkgpath"."/"."$pkgname.$$.i.conf";  # intermediate conf w/ add. modules added

  printL "addmods2config: [ @newmods ] \n"; # should be LogEnterSub

  # overwrite pkgexists when deploysappkgs_SkipPkgSelect is set
  $pkgexists = 0 if (($ENV{"deploysappkgs_SkipPkgSelect"} || 0) == 1);

  # generate a '-m' option list. needs to prepend one '-m'
  $modoptlist=join(' -m ',@newmods);
  $modoptlist="-m ".$modoptlist;
  printL "opt list: $modoptlist \n";

  if ($pkgexists) {
    # generate initial package config
    $cmd2exec= qq[$cmd_cmgetconf $pkgname $oconf];
    # this might be the 2nd time where we get this config. perhaps we can
    # merge that with the initial case.
    printL "cmd2exec: $cmd2exec \n";
    `$cmd2exec` ; # or warn "LOG: couldn't get config of $pkgname";
    # sometimes (for reasons unknown)  it goes into "warn" part
    # but config has been created (commented)

    # check if orig conf file was generated
    unless (-f $oconf) {
      sgeasy_print80 "Original config file wasn't generated";
      sgeasy_print80 "($oconf)";
      return "";
    }

    # update this package with new modules to create intermediate package
    $cmd2exec= qq[$ENV{SGSBIN}/cmmakepkg -v 0 -i $oconf $modoptlist $iconf];
    printL "cmd2exec: $cmd2exec \n";
    `$cmd2exec` or warn "couldn't run cmmmakepkg";
    unlink $oconf; # cleanup
    # check if intermediate config was generated
    unless (-f $iconf) {
      sgeasy_print80 "Intermediate config file ($iconf) wasn't generated";
      return "";
    }
  }
  else
  {
    $cmd2exec= qq[$ENV{SGSBIN}/cmmakepkg -v 0 -n $pkgname $modoptlist $iconf];
    printL "cmd2exec: $cmd2exec \n";
    # generate a new intermediate package
    `$cmd2exec` or warn "couldn't run cmmmakepkg";
  }
  return $iconf;
}

#############################################################################
#
# sub AttrIsRelated
#
# check if an attribute is related to a parent/key attribute
#
# Arguments:
#  $keyattr        name of key attribute
#  $checkattr    name of attribute to be checked
#
# Uses:
#  %SGRelAttr  (hash of key attribute to related attribute arrays)
#
# Returns:
#  0  - if not related
#  1 - if related
# 2 - if it is the key attribute itself
#
#############################################################################

sub AttrIsRelated {
  my $keyattr = shift;
  my $checkattr = shift;

  return 0 unless defined $checkattr;

  if (defined ($SGRelAttr{$keyattr})) {
    my @attribs = @{$SGRelAttr{$keyattr}};
    return 2 if ($keyattr eq $checkattr);  # related to next occurence of itself (?)
    return grep {$_ eq $checkattr} @attribs; # grep should return 1 if in array
  }
  else {
    warn "$keyattr not a key attribute";
    return 0;
  }
  return 0;
}

#############################################################################
#
# sub printAttr
#
# print out attributes
#
# Arguments:
#  *FH        File handle to read attributes from
#  $attname  attribute name
#  $attrval    attribute value
#  $phase     string indicating at which "phase" the attribute was added (optional)
#
# Uses:
#  $cmgetconf_cmd
#  cmmakepkg commmand
#
# Returns:
#  Name of newly generated package config file
#
#############################################################################

sub printAttr {
    local *FH = shift;
    my $attrname = shift;
    my $attrval  = shift;
    my $phase    = shift;
    my $newattr  = shift;

    #printL "[$attrname] [$attrval]\n";
    #print FH "# -- $phase\n" if defined $phase;
    print FH "# -- deploysappkgs new attribute --\n" if ((defined $newattr) && ($newattr == 1));
#    print FH "# -- deploysappkgs new attribute --[$phase] \n" if ((defined $newattr) && ($newattr == 1));
    print FH "$attrname\t\t$attrval\n";
}
#############################################################################
#
# sub checkSAPSIDs
#
# check if SAP SIDs don't conflict
#
# Arguments
# \%cmgetconfhash
# \%sapgetconfhash
#
# 4 Cases for sgesap/sap_global/sap_system (which is always defined in 2nd hash):
#   1. not defined in first hash --> ok
#   2. defined in first hash, but empty --> ok
#   3. defined and equal --> ok
#   4. defined but not equal  --> not ok
#
# For a DB only config sap_system isn't defined in 2nd hash. In this case
# it also doesn't make sense to compare to
# "sgesap/db_global/db_system"
# as this might also be different.
#
# Returns: true if SIDs are compatible
#
#############################################################################

sub checkSAPSIDs
{
  my $pkgsapsid=${${$_[0]}{"sgesap/sap_global/sap_system"}}[0];
  my $sgcsapsid=${${$_[1]}{"sgesap/sap_global/sap_system"}}[0];

  #printL "SIDs found : $pkgsapsid , $sgcsapsid \n" ;

  return $pkgsapsid eq $sgcsapsid if (defined $pkgsapsid && $pkgsapsid ne "" && defined $sgcsapsid );
  return 1;
}

#############################################################################
#
# section with merging routines
# contains a generic merging driver routine and
# specific routines for merging attributes from specific modules
#
#############################################################################

#############################################################################
#
# sub mergeAttributes
#
# merge two attribute lists
#
# Arguments:
# FH - filehandle to write to
# \@cmgetconfarry -  ref  of array from cmgetconf
# \@sapgetconfarray -  ref of array from sapgetconf
# \@cmgetconfhash - ref to hash from cmgetconf
# \@sapgetconfhash -  ref to hash from sapgetconf
# $pkgname - name of package
#
# Curently assumes that related module attributes are group together in
# one sequence in first array, that is e.g all ip_subnet and their related attributes
# are in one sequence that isn't interrupted by unrelated attributes.
# This makes it easier to handle new key/parent attributes in one invocation
# of the subroutine responsible for these attributes and also skip over that
# section in the for loop after the group has been handled.
# (this also means that each existing key attribute the related routine is only
#   called once)
#
# Remarks:
# - a newly created dbinstance package doesn't contain any (uncommented)
#   related attributes
# - The output routine called by this and other merge routines is triggered
#   to indicate new attributes in the config file. However, only those
#   attributes are indicated as new which were discovered by sapgetconf
#   and were not in the config file yet. If an attribute came in by adding
#   a new module, these arent't marked as new. (such attribute could be
#   required ones with default values, or e.g. the hanfs service definition)
# - pkgname is (will be) just used during the check if an attribute is already
#   used in another package.
#
#############################################################################

sub mergeAttributes
{
  local *FH = shift;

  my $cmaref;  # cmgetconf array reference
  my $curParent;
  my $lastParent;
  my $cmaidx=0;
  my $nextidx=0;
  my %remainmodshash=();

  # first create an internal hash mod from array of modules reported
  # by sapgetconf (3rd hash) to make it easier
  # to identify modules unhandled by the "merging in" of attributes
  %remainmodshash = map { $_ => 1 } @{${$_[3]}{"module_name"}};

  {my @thekeys = keys %remainmodshash ;
   printL "Attribute of modules possibly merged in: [ @thekeys ]\n";};

  foreach $cmaref ( @{$_[0]} ) {
    if ($nextidx > $cmaidx) {
      $cmaidx++; # skip entry
      next;
    }
    # For some reason we have an array reference at the end, but
    # only indexes in them.
    last unless defined $cmaref->[0];
    #printL "cmaref0: $cmaref->[0]\n";

    if (defined $SGKeyAttr{$cmaref->[0]}) {
      $curParent=$cmaref->[0];

      printL "Key attribute found: $curParent\n";

      # dispatch special routine for handling the attributes
      # (code below to be replaced by a dispatch table)
      if ($curParent eq "ip_subnet") {
        $nextidx=mergePkgAttr_package_ip (*FH, $_[0],$_[2],$_[1],$_[3],
                                           $curParent, $cmaidx);
         # set module as handled (even if it might not be on the list of
         # additional modules)
         # (Not sure if this case never happens. Either a module was identified
         #  as an additional required module. Then there couldn't be related
         #  attributes in the original config. Or the module is already in (but
         #  maybe even w/o attributes configured) then it doesn't show up in
         #  the addmods list.
         #  It would make more sense to setup the list of all modules required
         #  from sapgetconf and remove them from the list once handled here.
         #  This way we would end up with a list of modules not handled by
         #  going through the list of the original attributes. This would also
         #  cover the case of a module already being configured but having no
         #  related attributes setup. )
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif (($curParent eq "vg")
              || ($curParent eq "vxvm_dg") ||  ($curParent eq "cvm_dg"))
      {
        $nextidx=mergePkgAttr_volume_group (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx, $_[4]);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "fs_name") {
        $nextidx=mergePkgAttr_fs_name (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "resource_name") {
        $nextidx=mergePkgAttr_resource_name (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
	  elsif ($curParent eq "generic_resource_name") {
        $nextidx=mergePkgAttr_generic_resource_name (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "service_name") {
        $nextidx=mergePkgAttr_service_name (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx, $_[4]);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "sgesap/sap_global/sap_system") {
        $nextidx=mergePkgAttr_sap_system (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        # delete required module only if there are no sap_instance attributes
        # returned from sapgetconf
        delete($remainmodshash{$SGKeyAttr{$curParent}})
          unless defined $_[3]->{"sgesap/stack/sap_instance"};
      }
      elsif ($curParent eq "sgesap/stack/sap_instance") {
        $nextidx=mergePkgAttr_sap_instance (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif (($curParent eq "sgesap/db_global/db_system")
             || ($curParent eq "sgesap/db_global/db_vendor")) {
        $nextidx=mergePkgAttr_db_system (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif (($curParent eq "sgesap/hdb_global/hdb_system")
             || ($curParent eq "sgesap/hdb_global/hdb_retry_count")
             || ($curParent eq "sgesap/hdbinstance/hdb_instance")
             || ($curParent eq "sgesap/hdbinstance_global/hdb_instance")) {
        $nextidx=mergePkgAttr_hdb_system (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif (($curParent eq "sgesap/hdbprimary/hdb_sync_time_tolerance")
             || ($curParent eq "sgesap/hdbprimary/hdb_sync_logpos_tolerance")
             || ($curParent eq "sgesap/hdbprimary/hdb_init_into_fullsync")) {
        $nextidx=mergePkgAttr_hdb_primary (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}})
      }
      elsif ($curParent eq "sgesap/hdbdistinstance/hdb_ip_subnet") {
        $nextidx=mergePkgAttr_hdb_ip (*FH, $_[0],$_[2],$_[1],$_[3],
                                           $curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "priority") {
        $nextidx=mergePkgAttr_priority (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
	  elsif ($curParent eq "sgesap/repdbinstance/repdb_system")  {
        $nextidx=mergePkgAttr_db_system (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "package_description") {
        $nextidx=mergePkgAttr_pkg_descr (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        #delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "node_name") {
        $nextidx=mergePkgAttr_node_name (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        #delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "script_log_file") {
        $nextidx=mergePkgAttr_script_logfile (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx, $_[4]);
      }
      elsif (($curParent eq "nfs/hanfs_export/FILE_LOCK_MIGRATION")
             || ($curParent eq "nfs/hanfs_export/XFS")) {
        $nextidx=mergePkgAttr_nfs(*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif (($curParent eq "tkit/nfs/nfs/TKIT_DIR")
             || ($curParent eq "tkit/nfs/nfs/XFS")) {
        $nextidx=mergePkgAttr_nfs(*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
      elsif ($curParent eq "dependency_name") {
        $nextidx=mergePkgAttr_dependency (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
        delete($remainmodshash{$SGKeyAttr{$curParent}});
      }
	  elsif ($curParent eq "vmdk_file_name") {
	    $nextidx=mergePkgAttr_vmfs (*FH, $_[0],$_[2],$_[1],$_[3],$curParent, $cmaidx);
		delete($remainmodshash{$SGKeyAttr{$curParent}});
	  }
      else {
        # shouldn't get here - when all keyattr are handled
        #print FH "# -- deploysappkgs key attribute $cmaref->[0] not handled --\n";
        printL "Warning: unhandled key attribute $cmaref->[0]\n";
        printAttr(*FH,$cmaref->[0],$cmaref->[1],"cmgetconf-X0X");
      }
      # take actions on these parent attrib.
      $lastParent=$curParent;
    }
    else {
      # unhandled attributes just passed through from cmgetconf
      printAttr(*FH,$cmaref->[0],$cmaref->[1],"cmgetconf-0");
    }
    $cmaidx++;
  }

  my @thekeys = keys %remainmodshash ;
  printL "Remaining modules to be merge in: [ @thekeys ]\n";

  # post process that remaining modules which weren't touched by
  # passing through the cmgetconf list
  my @remmodkeyattrs = mod2keys(@thekeys);
  printL "remodkeyattr [ @remmodkeyattrs ] \n";
  # avoid key attributes which are part of another key attribute's group
  # (in that case the first one is a related attribute of the latter)
  # to be processed twice by setting up a "visited hash"
  my %keyattrvisited = ();

  foreach my $keyattr (@remmodkeyattrs) {
    # module and therefore the attribute might be considered as unhandled,
    # but it could already be partly handled, e.g.
    # "sgesap/sap_global/sap_system", skip it in this case
    # ENHANCEMENT:
    # related attribute may be a key attribute (e.g. db_vendor/db_system).
    # In such a case remove related attribute for remmodkeyattr. Otherwise
    # it will be handled twice
    next if defined $_[2]->{$keyattr}; # was in cmgetconf hash (and should be already handled?)
    next if defined $keyattrvisited{$keyattr}; # already handled with another key attribute
    # now loop the values for this key attribute and all related attributes
    for my $i (0 .. $#{$_[3]->{$keyattr}}) {
      my $lkeyattrval=$_[3]->{$keyattr}->[$i];
      my @usedinpkg = checkUsedClusResources (\%clusrsrchash ,$keyattr, $lkeyattrval, $_[4]);
      if (scalar @usedinpkg) {
        print FH "# Warning: ". $lkeyattrval ." already used in package(s) [ @usedinpkg ]\n";
      }
      printAttr(*FH,$keyattr,$lkeyattrval,"sapgetconf-4",1);
      #printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-4",1);
      # Print out related attribute (assumes there is a 1:1 relationship to
      # key attrib the same number of related attribute is delivered per key attrib!!)
      foreach my $relattr (@{$SGRelAttr{$keyattr}}) {
        if (defined $_[3]->{"$relattr"}->[$i]) {
            my @usedinpkg = checkUsedClusResources (\%clusrsrchash ,$relattr, $_[3]->{"$relattr"}->[$i], $_[4]);
            if (scalar @usedinpkg) {
              print FH "# Warning: ". $_[3]->{"$relattr"}->[$i] ." already used in package(s) [ @usedinpkg ]\n";
            }
            printAttr (*FH, "$relattr",$_[3]->{"$relattr"}->[$i],"sapgetconf-4",1);
            # mark attribute as visited (don't really care if this is a key attribute)
            $keyattrvisited{"$relattr"} = 1;
        }
      }
    }
  }
}

#############################################################################
#
#############################################################################

#
# merge resource name attributes
#
# same args and returns as other "merge routines"
#
sub mergePkgAttr_resource_name {
  #
  # Algorithm:
  # - tbd
  # - tbd
  # - tbd
  #
  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  # Algorithm basically the same as with mergePkgAttr_service_name
  # However, be aware that resource_name might be identical (although
  # they won't go into the same package.

  printL "mergePkgAttr_resource_name called ($keyattr - $keyattrval)\n"; # LogEnterSub

  # print out resouce related attribute from cmgetconf
  do {
    $nextattrrel=0;   # assume that next attribute isn't related
    #print "[@{$curnvref}] , $curidx \n";
    printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    $curidx++;
    $curnvref = $aref->[$curidx];  # next
    $nextattrrel = AttrIsRelated($keyattr,$curnvref->[0]);
    # switch to next key attribute value
    $keyattrval = $curnvref->[1] if ($nextattrrel eq 2);
  } while ($nextattrrel > 0);

  # Finished with cmgetconf output.
  # Get additional entries from sapgetconf (if any).

  for my $i (0 .. $#{$_[3]->{$keyattr}}) {
    my $lkeyattrval=$_[3]->{$keyattr}->[$i];
    next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}}; # already in cmgetconf
    #print "Found new $keyattr: $lkeyattrval\n";
    printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-3",1);
    # Print out related attribute (assumes there is a 1:1 relationship to
    # key attrib the same number of related attribute is delivered per key attrib)
    foreach my $relattr (@{$SGRelAttr{$keyattr}}) {
      printAttr (*FH, "$relattr",$_[3]->{"$relattr"}->[$i],"sapgetconf-3",1)
        if defined $_[3]->{"$relattr"};
    }
  }
  return $curidx++
}


# merge generic resource name attributes
# (code is identical to mergePkgAttr_resource_name -
#  - check if that one can be used also for generic resources)
#
# same args and returns as other "merge routines"
#
sub mergePkgAttr_generic_resource_name {
  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  # Algorithm basically the same as with mergePkgAttr_service_name
  # However, be aware that resource_name might be identical (although
  # they won't go into the same package.

  printL "mergePkgAttr_generic_resource_name called ($keyattr - $keyattrval)\n"; # LogEnterSub

  # print out resouce related attribute from cmgetconf
  do {
    $nextattrrel=0;   # assume that next attribute isn't related
    #print "[@{$curnvref}] , $curidx \n";
    printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    $curidx++;
    $curnvref = $aref->[$curidx];  # next
    $nextattrrel = AttrIsRelated($keyattr,$curnvref->[0]);
    # switch to next key attribute value
    $keyattrval = $curnvref->[1] if ($nextattrrel eq 2);
  } while ($nextattrrel > 0);

  # Finished with cmgetconf output.
  # Get additional entries from sapgetconf (if any).

  for my $i (0 .. $#{$_[3]->{$keyattr}}) {
    my $lkeyattrval=$_[3]->{$keyattr}->[$i];
    next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}}; # already in cmgetconf
    #print "Found new $keyattr: $lkeyattrval\n";
    printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-3",1);
    # Print out related attribute (assumes there is a 1:1 relationship to
    # key attrib the same number of related attribute is delivered per key attrib)
    foreach my $relattr (@{$SGRelAttr{$keyattr}}) {
      printAttr (*FH, "$relattr",$_[3]->{"$relattr"}->[$i],"sapgetconf-3",1)
        if defined $_[3]->{"$relattr"}->[$i];
    }
  }
  return $curidx++
}

#############################################################################
#
#############################################################################
sub mergePkgAttr_service_name {
  #
  # Algorithm:
  # - loop through all services delivered by cmgetconf, print them out
  # - check services delivered by sapgetconf, print them out if the service
  #     isn't delivered by cmgetconf already (check is based on service_name).
  # So cmgetconf config takes precedence here, if a service with same name
  # is also delivered by sapgetconf. It is *not* checked if the attributes
  #  of such a service are different.
  # A service is also already considered configured if it has the same service_cmd
  #
  # There some flaw if nfs/hanfs is included:
  # This has nfs.monitor service hardcoded into the module (nfs_service on LX).
  # Needs to be  changed when coming across it. Keep in mind that this service is
  # included by including the module, not by this utiltiy.
  # Checks for "servicename nfs.monitor" if (global) nfsmodadded is set

  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_service_name called ($keyattr - $keyattrval)\n"; # LogEnterSub

  # Print out service related attribute from cmgetconf
  # There needs to be a check for service_name "nfs.monitor" if the nfs/hanfs module
  # got added during this run. Otherwise this might cause a error as other
  # modules having nfs configured might use the same service_name.
  do {
    $nextattrrel=0;   # assume that next attribute isn't related
    #print "[@{$curnvref}] , $curidx \n";
    # check of nfs.monitor
    if ($nfsmodadded && ($curnvref->[0] eq "service_name") &&
	     (($curnvref->[1] eq "nfs.monitor") || ($curnvref->[1] eq "nfs_service"))) {
      printAttr(*FH,$curnvref->[0],$_[6].$curnvref->[1],"cmgetconf-2");
    }
    else {
      printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    }
    $curidx++;
    $curnvref = $aref->[$curidx];  # next
    $nextattrrel = AttrIsRelated($keyattr,$curnvref->[0]);
    # switch to next key attribute value
    $keyattrval = $curnvref->[1] if ($nextattrrel eq 2);
  } while ($nextattrrel > 0);

  # Finished with cmgetconf output.
  # Get additional entries from sapgetconf (if any).

  for my $i (0 .. $#{$_[3]->{$keyattr}}) {
    my $lkeyattrval=$_[3]->{$keyattr}->[$i];
    next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}}; # already in cmgetconf
    # Skip also if they have the same service_cmd - relying on the fact that
    # there is just one monitor configured per package per monitor type in
    # modular SGeSAP. Doesn't rely on exact string match anymore but allows
	# a environment variable at the start of the path
    #next if grep {$_ eq  $_[3]->{"service_cmd"}->[$i]} @{$_[1]->{"service_cmd"}};
	next if grep {service_cmd_eq($_ ,$_[3]->{"service_cmd"}->[$i])} @{$_[1]->{"service_cmd"}};
    #print "Found new $keyattr: $lkeyattrval\n";
    printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-3",1);
    # Print out related attribute (assumes there is a 1:1 relationship to
    # key attrib the same number of related attribute is delivered per key attrib)
    foreach my $relattr (@{$SGRelAttr{$keyattr}}) {
      printAttr (*FH, "$relattr",$_[3]->{"$relattr"}->[$i],"sapgetconf-3",1)
        if defined $_[3]->{"$relattr"};
    }
  }
  return $curidx++
}

# Helper to establish equality between service_cmd
# String equality or $<envvar>/<rest> eq $ENV{<envvar>}/<rest>
# (substituting the env variable)
# Assumes that only one of the arguments starts with an environment variable
# if they are not string equal
# However, there can also be the case where both have (and start) with
# env variables and one of the is enclosed in curly braces (${<envvar})
# Comparions is done after removing curly braces in such a case.
sub service_cmd_eq {
  my $left  = shift;
  my $right = shift;

  return 1 if ($left eq $right); # string equal

  # switch right/left if left starts with a environment variable
  ($right, $left) = ($left, $right) if ($left =~ m#^\$(\w+)/#);

  printL "service_cmd_eq: $left - $right\n";

  if ($right =~ m#^\$(\w+)/(.+)$#) {
    printL "match ENV - $1 - $2\n";
    return 1 if ($left eq "$ENV{$1}/$2");
  }

  # subst. curly braces non-greedy global (with nothing)
  my $sleft  = $left;
  my $sright = $right;
  $sleft  =~ s/\$\{(\S+?)\}/\$$1/g;
  $sright =~ s/\$\{(\S+?)\}/\$$1/g;
  printL "match subst - $sleft - $sright\n";
  return 1 if ($sleft eq $sright); # string equal

  return 0;
}

#############################################################################
#
#############################################################################
sub mergePkgAttr_fs_name {
  #
  # Algorithm:
  # - ...
  # - ...
  # TODO ENHANCEMENT
  # TODO - check if new DG is already part of a CFS package
  # TODO    if so, skip it and related attribs
  #
  # vxvm FS are report as "dev/vx/dsk/<dg>/<lvol>"
  # by sapgetconf, but have the form <dg>/<lvol> in CFS ouput
  #
  #
  local *FH = shift;

  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $attrrel = 2;  # same key attribute
  my $idxsap = -1;

  # To be implemented
  printL "mergePkgAttr_fs_name called ($keyattr - $keyattrval)\n"; # LogEnterSub

  # loop through all whats in first array
  while ($attrrel > 0) {
    #print "[@{$curnvref}] , $curidx \n";
    printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    if ($attrrel eq 2) {
      $idxsap = -1;
      $keyattrval = $curnvref->[1];
      # check if this also shows up in list reported by sapgetconf
      for my $i (0 .. $#{$_[3]->{$keyattr}}) { # need to check for undef first?
        if ($keyattrval eq $_[3]->{$keyattr}->[$i]) {
          $idxsap = $i;
          printL "same key value $keyattrval at index $idxsap\n";
        }
      }
    }
    elsif ($idxsap >= 0) {
       # check for conflicting entries under the same attribute name
       if (defined $_[3]->{$curnvref->[0]}->[$idxsap]) {
         printL "conflicting values for $curnvref->[0] attribute of $keyattr $keyattrval\n"
            if ($curnvref->[1] !~ /("|)$_[3]->{$curnvref->[0]}->[$idxsap]("|)/)
       }
    }
    $curidx++;
    $curnvref = $aref->[$curidx];  # next entry
    $attrrel = AttrIsRelated($keyattr,$curnvref->[0]);

  }

  # Get unhandled key values from sapgetconf
  # It is legal to ommit attributes specified as "iterations 1" but whose
  # default value is the empty string (e.g. fs_server)
  for my $i (0 .. $#{$_[3]->{$keyattr}}) {
    my $lkeyattrval=$_[3]->{$keyattr}->[$i];
	# on Linux a simple "eq" isn't sufficient anymore as one of them could
	# be a /dev/mapper spec and the other not
    #next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}}; # skip if already handled
	next if grep {fs_name_eq($_,$lkeyattrval)} @{$_[1]->{$keyattr}}; # skip if already handled
    # TODO ENHANCEMENT
    # TODO check for CFS here...
    #print "Found new $keyattr: $lkeyattrval\n";
    printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-3",1);
    foreach my $lattr (@{$SGRelAttr{$keyattr}})
    {
      printAttr(*FH, $lattr, $_[3]->{$lattr}->[$i], "sapgetconf-3",1)
        if (defined ($_[3]->{$lattr}->[$i]));
    }
  }

  return $curidx  # TBD - not sure if we need "++" here
}

# Helper to establish equality between fs_name attributes
# String equality or /dev/vg/lvol eq /dev/mapper/vg-lvol
# Assumes that only one of the arguments contains /dev/mapper
# if they are not string equal
sub fs_name_eq {
  my $left  = shift;
  my $right = shift;
  #printL "fs_name_eq 1: $right - $ left\n";
  return 1 if ($left eq $right); # string equal

  # switch right/left if left contains mapper
  ($right, $left) = ($left, $right) if ($left =~ m#^/dev/mapper/#);

  printL "fs_name_eq: $left - $right\n";

  if ($right =~ m#^/dev/mapper/(\w+)-(\w+)#) {
    #printL "match - $1 - $2\n";
    return 1 if ($left eq "/dev/$1/$2");
  }
  return 0;
}

#############################################################################
#
#############################################################################
#
# merge volume group attributes
# supposed to all all three kind of VGs
#
# 0: FH (shift)
# 1: array from cmgetconf
# 2: hash from cmgetconf
# 3: array from sapgetconf
# 4: hash from sapgetconf
# [5: key attribute] - implicitly
# 6: current index into first (defined) array
# 7: packagename

sub mergePkgAttr_volume_group {
  #
  # Algorithm:
  # - collect all vg,vxmv_dg and cvm_dg in the cmgetconf array
  # - check sapgetconf hash if there are any not accounted for yet
  # TODO ENHANCEMENT
  # TODO - check if new cvm_dg is already configured in in a CFS package,
  # TODO   if so skip it

  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_volume_group called ($keyattr - $keyattrval)\n"; # LogEnterSub

  do {
    $nextattrrel=0;   # assume that next attribute isn't related
    printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    $curidx++;
    $curnvref = $aref->[$curidx];  # next
    $nextattrrel = AttrIsRelated($keyattr,$curnvref->[0]);
    # switch to next key attribute value
    $keyattrval = $curnvref->[1] if ($nextattrrel eq 2);
  } while ($nextattrrel > 0);

  # Got everything from cmgetconf output. Check sapgetconf output now
  # Need to loop through array linked to "vg" "vxvm_dg" and "cvm_dg" hashes
  # to find additional new VGs
  #
  # Get key attribute first ...
  if (defined ($_[3]->{$keyattr})) {
    foreach my $vgname (@{$_[3]->{$keyattr}}) {
      # already in existing package config?
      next if grep {$_ =~ "$vgname\$"} @{$_[1]->{$keyattr}};
      # does a match to end of string to accomodate "/dev" prefixes
      # reported by cmviewcl and cmgetconf (starting with PHSS_41523?)
      # TODO ENHANCEMENT
      # TODO check for CFS here...
      my @usedinpkg = checkUsedClusResources (\%clusrsrchash , $keyattr, $vgname, $_[6]);
      if (scalar @usedinpkg) {
        print FH "# Warning: $vgname already used in package(s) [ @usedinpkg ]\n";
      #  printAttr (*FH,"# ".$keyattr,$vgname,"sapgetconf-2");
      }
      printAttr(*FH,$keyattr,$vgname,"sapgetconf-2",1);
    }
  }
  # ... and then related attributes (which in this case are other type
  #  of vg or dg specs)
  foreach my $ovgattr (@{$SGRelAttr{$keyattr}}) {
    if (defined ($_[3]->{$keyattr})) {
      foreach my $vgname (@{$_[3]->{$ovgattr}}) {
        next if grep {$_ =~ "$vgname\$"} @{$_[1]->{$ovgattr}};
        # TODO ENHANCEMENT
        # TODO check for CFS here...
        my @usedinpkg = checkUsedClusResource (\%clusrsrchash , $keyattr, $vgname, $_[6]);
        if (scalar @usedinpkg) {
            print FH "# Warning: $vgname already used in package(s) [ @usedinpkg ]\n";
        #    printAttr (*FH,"# ".$keyattr,$vgname,"sapgetconf-2");
        }
        printAttr(*FH,$keyattr,$vgname,"sapgetconf-2",1);
      }
    }
  }
  # maybe the two cases above could be folded into one if just
  # had a combined list

  return $curidx++
}

#############################################################################
#
# mergePkgAttr_dependency
#
# Merge a dependency block (dependency_name,_condition,_location)
#
# dependency_name is is the parent attribute, but can't be taken as the
# key to see if a dependency has already be merged in (similar to service).
# A dependency is considered to be already merged in if _condition and
# _location are the same.
#
# Same set of parameters as other merge routines
#
# Keep in mind that sapgetconf itself doesn't report dependency attributes,
# but deploysapppkgs' handleCFS might add them to this output.
#
#############################################################################
sub mergePkgAttr_dependency {
  #
  # Algorithm:
  # - loop through all dependencies delivered by cmgetconf, print them out
  # - check dependencies "delivered" by sapgetconf, print them out if the service
  #   isn't delivered by cmgetconf already (this first check is based on service_name).
  # - check if condition match, which is package matches (left side) and up/down
  #   (right side) matches. Right side matches case insensitive. Skip on match.
  # cmgetconf config takes precedence here, if a dependency with same name
  # is also delivered by sapgetconf. It is *not* checked if the attributes
  # are different.

  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_dependency called ($keyattr - $keyattrval)\n"; # LogEnterSub

  # Print out dependency related attribute from cmgetconf
  # There needs to be a check for service_name "nfs.monitor" if the nfs/hanfs module
  # got added during this run. Otherwise this might cause a error as other
  # modules having nfs configured might use the same service_name.
  do {
    $nextattrrel=0;   # assume that next attribute isn't related
    printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    $curidx++;
    $curnvref = $aref->[$curidx];  # next
    $nextattrrel = AttrIsRelated($keyattr,$curnvref->[0]);
    # switch to next key attribute value
    $keyattrval = $curnvref->[1] if ($nextattrrel eq 2);
  } while ($nextattrrel > 0);

    # Finished with cmgetconf output.
  # Get additional entries from sapgetconf (if any).

  for my $i (0 .. $#{$_[3]->{$keyattr}}) {
    my $lkeyattrval=$_[3]->{$keyattr}->[$i];
    # skip if a dependency with the same name is already configured (cmgetconf)
    next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}};
    # Skip also if they have the same condition. Doesn't check location for this.
    # Just comparing strings for the condition might not be sufficient.
    # Better: compare left and righ parts of the equation, which requires too loop through
    #   all already configured conditions. Left part can also be upper or lowercase.
    #   or: match something like "\s*<leftpart>\s*=\s*<rightpart>\s*
    my ($condleft,$condright)=split(/\s*=\s*/,$_[3]->{"dependency_condition"}->[$i]);
    #printL "[$condleft], [$condright]\n";
    next if grep {$_ =~  /\s*${condleft}\s*=\s*(?i)${condright}\s*/} @{$_[1]->{"dependency_condition"}};
    #print "Found new $keyattr: $lkeyattrval\n";
    printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-3",1);
    # Print out related attribute (assumes there is a 1:1 relationship to
    # key attrib the same number of related attribute is delivered per key attrib)
    foreach my $relattr (@{$SGRelAttr{$keyattr}}) {
      printAttr (*FH, "$relattr",$_[3]->{"$relattr"}->[$i],"sapgetconf-3",1)
        if defined $_[3]->{"$relattr"};
    }
  }
  return $curidx++
}

#############################################################################
#
#############################################################################
#
# merge sap_global attributes
#
# same parameter and returns as mergePkgAttr_ip_subnet
# Args:
# 0: array from cmgetconf
# 1: hash from cmgetconf
# 2: array from sapgetconf
# 3: hash from sapgetconf
# [4: key attribute] - implicitly
# 5: current index into first (defined) array
#
# This routine assumes that compatility of sap_system attributes have been
# already checked with  checkSAPSIDs routine.

sub mergePkgAttr_sap_system {
  #
  # Algorithm:
  # - use sap_system from sapgetconf
  # - use attribs from cmgetconf so far then ignore those from sapgetconf
  # - check if sapgetconf delivers attribs not in cmgetconf
  #
  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_sap_system called ($keyattr - $keyattrval)\n"; # LogEnterSub

  if (defined $_[3]->{$keyattr}->[0]) {
    # use from sapgetconf, but doesn't really matter. occurence:1
    printAttr(*FH, $keyattr, $_[3]->{$keyattr}->[0], "sapgetconf-2");
  }
  else {
    printAttr(*FH, $keyattr, $keyattrval, "cmgetconf-1");
  }

  # loop through related attributes of cmgetconf
  foreach my $lattr (@{$SGRelAttr{$keyattr}})
  {
    printAttr(*FH, $lattr, $_[1]->{$lattr}->[0], "cmgetconf-1") if defined $_[1]->{$lattr};
  }

  # loop through related attributes of sapgetconf, ommitting those
  # which are already in cmgetconf
  foreach my $lattr (@{$SGRelAttr{$keyattr}})
  {
    if (defined ($_[3]->{$lattr}) && !defined ($_[1]->{$lattr})) {
      printAttr(*FH, $lattr, $_[3]->{$lattr}->[0], "sapgetconf-2",1);
    }
  }

  # finally advance index in array beyond group of related attributes
  do {
    $curidx++;
  } while (AttrIsRelated($keyattr,$aref->[$curidx]->[0]) > 0);
  return $curidx++
}

sub mergePkgAttr_db_system {
  #
  # Algorithm:
  # - use db_system or db_vendor from sapgetconf
  # - use attribs from sapgetconf so far then ignore those from cmgetconf
  # - check if cmgetconf delivers attribs not in sapgetconf
  # so this will update the related entries by the discovered attributes
  #
  # Doesn't check if DBSIDs are equal (maybe they shouldn't because they have
  # been changed)
  # Package conf by cmgetconf may have db_system or db_vendor empty
  # (one of them, not both) --> this isn't tested yet
  # Also "conflicts" aren't handled yet.
  #
  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_db_system called ($keyattr - $keyattrval)\n"; # LogEnterSub

  printAttr(*FH, $keyattr, $_[3]->{$keyattr}->[0], "sapgetconf-2");

  # loop through related attributes of sapgetconf
  foreach my $lattr (@{$SGRelAttr{$keyattr}})
  {
    printAttr(*FH, $lattr, $_[3]->{$lattr}->[0], "sapgetconf-2",1) if defined $_[3]->{$lattr};
  }

  # loop through related attributes of cmgetconf, ommitting those
  # which are already in cmgetconf
  # (which is the opposite logic mergePkgAttr_sap_system)
  foreach my $lattr (@{$SGRelAttr{$keyattr}})
  {
    if (defined ($_[1]->{$lattr}) && !defined ($_[3]->{$lattr})) {
      printAttr(*FH, $lattr, $_[1]->{$lattr}->[0], "cmgetconf-1");
    }
  }

  # finally advance index in array beyond group of related attributes
  do {
    $curidx++;
  } while (AttrIsRelated($keyattr,$aref->[$curidx]->[0]) > 0);
  return $curidx++
}

sub mergePkgAttr_hdb_system {
  #
  # Algorithm:
  # - a pkg configuration is considered initial if hdb_system from
  #   cmgetconf is empty
  # - use attributes from sapgetconf if initial otherwise
  # - use attributes from cmgetconf if not
  # - check if cmgetconf delivers attribs not in sapgetconf
  #
  # Package conf by cmgetconf may only have retry count
  #
  # v1 attribute sgesap/hdbinstance/hdb_instance *won't* be replaced
  # by v2 attribute sgesap/hdbinstance_global/hdb_instance (code commented)
  #
  # 0 == cmgetconf array    == @pkgcmconf
  # 1 == cmgetconf hash     == %pkgcmhash
  # 2 == sapgetconf array   == @pkgsapconf
  # 3 == sapgetconf hash    == %pkgsaphash

  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  my $srchash = $_[1];
  my $otherhash = $_[3];
  my $srcindicator = "cmgetconf-1";
  my $otherindicator = "sapgetconf-2";
  my $newindicator = 0;

  printL "mergePkgAttr_hdb_system called ([$keyattr] - [$keyattrval])\n"; # LogEnterSub

  # swap hashes if current srchash contains empty hdb_system attribute (meaning that
  # package doesn't have hdb attributes defined yet)
  # PRD was introduced in rev1452 of hdb_global.1 as default value for below attribute
  if ($_[1]->{"sgesap/hdb_global/hdb_system"}->[0] eq "" || $_[1]->{"sgesap/hdb_global/hdb_system"}->[0] eq "PRD") {
    $srchash = $_[3];
    $otherhash = $_[1];
    $srcindicator = "sapgetconf-1";
    $otherindicator = "cmgetconf-2";
    $newindicator = 1;
  }

  # key attribute
  printAttr(*FH, $keyattr, $srchash->{$keyattr}->[0], $srcindicator,$newindicator);

  # loop through related attributes of srchash
  # hdb_system (if made mandatory) maybe incomplete the
  #       first time in cmgetconf
  foreach my $lattr (@{$SGRelAttr{$keyattr}})
  {
  ##  printAttr(*FH, $lattr, $_[3]->{$lattr}->[0], "sapgetconf-2",1) if defined $_[3]->{$lattr};
    # make sure that a hdbinstance/hdb_instance attribute in srchash isn't printed if there's
    # a hdbinstance_global/hdb_instance in other hash
    #next if (($lattr eq "sgesap/hdbinstance/hdb_instance") && (defined ($otherhash->{"sgesap/hdbinstance_global/hdb_instance"}))
    printAttr(*FH, $lattr, $srchash->{$lattr}->[0], $srcindicator,$newindicator) if defined $srchash->{$lattr};
  }

  # loop through related attributes in other hash, ommitting those
  # which are already in srchash
  foreach my $lattr (@{$SGRelAttr{$keyattr}})
  {
    if (defined ($otherhash->{$lattr}) && !defined ($srchash->{$lattr})) {
      # as above: don't print hdbinstance if there's a hdbinstance_global in (this time) srchash
      #next if (($lattr eq "sgesap/hdbinstance/hdb_instance") && (defined ($srchash->{"sgesap/hdbinstance_global/hdb_instance"}))
      # don't add v2 attribute (from "otherhash" == sapgetconf in this case)
      # if v1 attribute (from package) is already there (from srchash in this case)
      next if (($lattr eq "sgesap/hdbinstance_global/hdb_instance") && defined ($srchash->{"sgesap/hdbinstance/hdb_instance"}));
      printAttr(*FH, $lattr, $otherhash->{$lattr}->[0], $otherindicator,!$newindicator);
    }
  }

  # finally advance index in array beyond group of related attributes
  do {
    $curidx++;
  } while (AttrIsRelated($keyattr,$aref->[$curidx]->[0]) > 0);
  return $curidx++
}
sub mergePkgAttr_hdb_primary {
  #
  # Algorithm:
  # - just take sync tolerance attributes from current pkg config if there
  # - if one or both are missing take them from sapgetconf output
  #
  # Package conf by cmgetconf may only have retry count
  #
  # 0 == cmgetconf array    == @pkgcmconf
  # 1 == cmgetconf hash     == %pkgcmhash

  # 2 == sapgetconf array   == @pkgsapconf
  # 3 == sapgetconf hash    == %pkgsaphash
  #
  # Remark: generic way of merging used here:
  # 1) all that is related to key from cmgetconf (incl. "key")
  # 2) anything remaining from sapgetconf


  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  my $srchash = $_[1];
  my $otherhash = $_[3];
  my $srcindicator = "cmgetconf-1";
  my $otherindicator = "sapgetconf-2";
  my $newindicator = 0;

  printL "mergePkgAttr_hdb_primary called ([$keyattr] - [$keyattrval])\n"; # LogEnterSub

    # key attribute
    printAttr(*FH, $keyattr, $srchash->{$keyattr}->[0], $srcindicator,$newindicator);

  # loop through related attributes of cmgetconf
  foreach my $lattr (@{$SGRelAttr{$keyattr}})
  {
    printAttr(*FH, $lattr, $srchash->{$lattr}->[0], $srcindicator,$newindicator) if defined $srchash->{$lattr};
  }

  # loop through related attributes in other hash (=sapgetconf), ommitting those
  # which are already in srchash
  foreach my $lattr (@{$SGRelAttr{$keyattr}})
  {
    if (defined ($otherhash->{$lattr}) && !defined ($srchash->{$lattr})) {
      printAttr(*FH, $lattr, $otherhash->{$lattr}->[0], $otherindicator,!$newindicator);
    }
  }

  # finally advance index in array beyond group of related attributes
  do {
    $curidx++;
  } while (AttrIsRelated($keyattr,$aref->[$curidx]->[0]) > 0);
  return $curidx++
}

sub mergePkgAttr_priority {
  # Algorithm:
  # - if priority from cmgetconf is != no_priority use this
  # - if it is no_priority and there is one in sapgetconf use the latter
  # - otherwise use cmgetconf

  #

  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_priority called ($keyattr - $keyattrval)\n"; # LogEnterSub

  if ($keyattrval ne "no_priority") {
    printAttr(*FH, $keyattr, $keyattrval, "cmgetconf-2");
  }
  elsif (defined ($_[3]->{$keyattr})) {
    printAttr(*FH, $keyattr, $_[3]->{$keyattr}->[0], "sapgetconf-1",1);
    # assumes that index 0 is defined if hash is defined
  }
  else {
    printAttr(*FH, $keyattr, $keyattrval, "cmgetconf-2");
  }

  return $curidx++
}

sub mergePkgAttr_repdb_system {
  # Algorithm:
  # - use repdb_system from sapgetconf
  # - use attribs from sapgetconf so far then ignore those from cmgetconf
  # - check if cmgetconf delivers attribs not in sapgetconf
  # so this will update the related entries by the discovered attributes
  #
  # Doesn't check if DBSIDs are equal (maybe they shouldn't because they have
  # been changed)
  # Package conf by cmgetconf may have repdb_system or repdb_role empty
  # (one of them, not both) --> this isn't tested yet
  # Also "conflicts" aren't handled yet.
  # TODO REP: Code below looks a bit rudimentary.
  #           Is it sufficient to handle a "merge in"

  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_repdb_system called ($keyattr - $keyattrval)\n"; # LogEnterSub

  printAttr(*FH, $keyattr, $_[3]->{$keyattr}->[0], "sapgetconf-2");

  return $curidx++
}

#############################################################################
#
#############################################################################
#
# same parameters...
#

sub mergePkgAttr_sap_instance {

  #
  # Algorithm:
  # - add instances from cmgetconf, check if the entry is also in sapgetconf
  #     and has additional attributes
  # - add all instances from sapgetconf not in cmgetconf
  # - tbd: what if these two config have conflicting information
  #     or one of the (esp. sapgetconf) contains additional attributes

  local *FH = shift;
  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;
  my $idxsap=-1;
  # not all of them maybe needed anymore...

  printL "mergePkgAttr_sap_instance called ($keyattr - $keyattrval)\n"; # LogEnterSub

  # loop through list of all sap_instances from cmgetconf
  for my $idxcm (0 .. $#{$_[1]->{$keyattr}}) {
    my $lkeyattrval=$_[1]->{$keyattr}->[$idxcm];
    if (grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}}) {
      # also in sapgetconf. maybe we need to add additional child attributes
      # get index into that array
      for my $idx (0 .. $#{$_[3]->{$keyattr}}) {
        if ($_[3]->{$keyattr}->[$idx] eq $lkeyattrval) {
          $idxsap=$idx;
          last;
        }
      }
      printL "same attribute value ($keyattr/$lkeyattrval) found in ($idxcm,$idxsap)\n";
    }
    printAttr (*FH, $keyattr,$lkeyattrval,"cmgetconf-1");
    # now loop through all related attributes with this index (or indexes)
    foreach my $lattr (@{$SGRelAttr{$keyattr}})
    {
      # Merge attributes. existing attributes from cmgetconf take precedence
      # Those from sapgetconf will be merged in.
      if (defined ($_[1]->{$lattr}->[$idxcm])) {
        printAttr(*FH, $lattr, $_[1]->{$lattr}->[$idxcm], "cmgetconf-1");
      }
      elsif (defined ($_[3]->{$lattr}->[$idxsap])) {
        printAttr(*FH, $lattr, $_[3]->{$lattr}->[$idxsap], "sapgetconf-2",1);
      }
    }
  }

  # All instances from sapgetconf handled.
  # Loop through list of all sap_instances from sapgetconf and look for remaining ones
  for my $i (0 .. $#{$_[3]->{$keyattr}}) {
    my $lkeyattrval=$_[3]->{$keyattr}->[$i];
    next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}}; # skip if already handled
    printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-3");
    foreach my $lattr (@{$SGRelAttr{$keyattr}})
    {
      printAttr(*FH, $lattr, $_[3]->{$lattr}->[$i], "sapgetconf-3",1)
        if (defined ($_[3]->{$lattr}->[$i]));
    }
  }

  # advance array index
  do {
    $curidx++;
  } while (AttrIsRelated($keyattr,$aref->[$curidx]->[0]) > 0);

  return $curidx++;
}

#############################################################################
#
# sub mergePkgAttr_package_ip
# sub mergePkgAttr_hdb_ip
#
# hdb_ip sub basically identical execept for the attribute names
# (WVE: merge into single sub)
#
#############################################################################
#
# merge the package_ip attributes
# cases (valid in a similar form for all attr.):
# 1. only pkgconf contains ip related entries
# 2. only sapconf contains ip related entries
# 3. both contain entries, but there are no new entries in 2nd one
# 4. ... there are new ip_addresses for already existing subnets in 2nd one
# 5  ... there are new subnets ("parent") in 2nd one
# case 2 (and maybe 5) not handled here
#
# Args:
# 0: Filehandle to write to
# 1: array from cmgetconf
# 2: hash from cmgetconf
# 3: array from sapgetconf
# 4: hash from sapgetconf
# [5: key attribute] - implicitly
# 6: current index into first (defined) array
#
# Returns the index of the next unrelated attribute
#
# NOTE: each address in hash has a corresponding subnet at same index, therefore
#       subnets may be "duplicate" in config
#
sub mergePkgAttr_package_ip {
  #
  # Algorithm:
  # - check if 1st array/2nd hash contains the key attribute (it should,
  #    otherwise this subroutine wouldn't have been triggered)
  # - if so, do for all of them:
  #   -- print it
  #   -- search children (in array)
  #   -- print them
  #   -- check if this key attribute value is also in 3rd array/4th hash
  #   -- if so, print all children which are not in 1st array/2nd hash
  # This code assumes that the output of cmgetconf has the following form
  # - all ip related entries from one block
  # - ip_subnet values occur only once (won't be the case for sapgetconf outp)
  # - all ip_address are group under the related ip_subnet

  local *FH = shift;

  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_package_ip called ($keyattr - $keyattrval)\n"; # LogEnterSub

  do {
    $nextattrrel=0;   # assume that next attribute isn't related
    printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    $curidx++;
    $curnvref = $aref->[$curidx];  # next
    $nextattrrel = AttrIsRelated($keyattr,$curnvref->[0]);
    # check for 2 (next key attr) or 0 (no related attr)
    if ($nextattrrel ne 1) {
      # Next (same name) key attribute. Check if something needs to get
      # merged in from sapgetconf for that.
      # Loop through subnet/ip_address pairs
      # (be aware that they are in different arrays behind a hash)
      for my $i (0 .. $#{$_[3]->{$keyattr}}) {
          next unless $keyattrval eq $_[3]->{$keyattr}->[$i];
          my $ipaddr=$_[3]->{"ip_address"}->[$i];
          # print found ip_address only if not in existing config
          printAttr(*FH,"ip_address",$ipaddr,"sapgetconf-2",1)
            unless grep {$_ eq $ipaddr} @{$_[1]->{"ip_address"}};
      }
      # switch to next key attribute value
      $keyattrval = $curnvref->[1] if ($nextattrrel eq 2);
    }
  } while ($nextattrrel > 0);

  # Finished with cmgetconf output and matching sapgetconf output
  # (basically ip_addresses belonging to the same subnet(s)).
  # There could be, however, additional key attribute values (in this
  # case new ip_subnets) which aren't in the orig. config yet.
  # Requires to loop over sapgetconf (param 3 and 4) ip_subnets again
  # and find those values not in cmgetconf (param 1 and 2) and their
  # related ip_addresses.

  for my $i (0 .. $#{$_[3]->{$keyattr}}) {
    my $lkeyattrval=$_[3]->{$keyattr}->[$i];
    next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}};
    printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-3",1);
    printAttr (*FH, "ip_address",$_[3]->{"ip_address"}->[$i],"sapgetconf-3",1);
    # Same ip_subnet can occur multiple times. cmcheck|applyconf can cope with that
    # However, we have to filter out physical hosts (where possible)
  }
  return $curidx;
}

sub mergePkgAttr_hdb_ip {

  local *FH = shift;

  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_hdb_ip called ($keyattr - $keyattrval)\n"; # LogEnterSub

  do {
    $nextattrrel=0;   # assume that next attribute isn't related
    printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    $curidx++;
    $curnvref = $aref->[$curidx];  # next
    $nextattrrel = AttrIsRelated($keyattr,$curnvref->[0]);
    # check for 2 (next key attr) or 0 (no related attr)
    if ($nextattrrel ne 1) {
      # Next (same name) key attribute. Check if something needs to get
      # merged in from sapgetconf for that.
      # Loop through subnet/ip_address pairs
      # (be aware that they are in different arrays behind a hash)
      for my $i (0 .. $#{$_[3]->{$keyattr}}) {
          next unless $keyattrval eq $_[3]->{$keyattr}->[$i];
          my $ipaddr=$_[3]->{"sgesap/hdbdistinstance/hdb_ip_address"}->[$i];
          # print found ip_address only if not in existing config
          printAttr(*FH,"sgesap/hdbdistinstance/hdb_ip_address",$ipaddr,"sapgetconf-2",1)
            unless grep {$_ eq $ipaddr} @{$_[1]->{"sgesap/hdbdistinstance/hdb_ip_address"}};
      }
      # switch to next key attribute value
      $keyattrval = $curnvref->[1] if ($nextattrrel eq 2);
    }
  } while ($nextattrrel > 0);

  # Finished with cmgetconf output and matching sapgetconf output
  # (basically ip_addresses belonging to the same subnet(s)).
  # There could be, however, additional key attribute values (in this
  # case new ip_subnets) which aren't in the orig. config yet.
  # Requires to loop over sapgetconf (param 3 and 4) ip_subnets again
  # and find those values not in cmgetconf (param 1 and 2) and their
  # related ip_addresses.

  for my $i (0 .. $#{$_[3]->{$keyattr}}) {
    my $lkeyattrval=$_[3]->{$keyattr}->[$i];
    next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}};
    printAttr (*FH, $keyattr,$lkeyattrval,"sapgetconf-3",1);
    printAttr (*FH, "sgesap/hdbdistinstance/hdb_ip_address",
                     $_[3]->{"sgesap/hdbdistinstance/hdb_ip_address"}->[$i],"sapgetconf-3",1);
    # Same ip_subnet can occur multiple times. cmcheck|applyconf can cope with that
    # However, we have to filter out physical hosts (where possible)
  }
  return $curidx;
}

#############################################################################
#
# Merge NFS attributes
#
# Arguments: as the others
#
# Either there is already one or more XFS attribute in the config file (that
# is it contains already an export) or the first hanfs attribute is
# FILE_LOCK_MIGRATION. In the first case it is just echoed into the new conf
# file, the second case triggers for additional XFS from sapgetconf
#
# Currently, this sub only handles only one attribute from cmgetconf per
# call. It could, however, handle all the XFS attibute(s) and FILE_LOCK_MIGRATION
# attribute in a single call.
#
#############################################################################

sub mergePkgAttr_nfs {
  local *FH = shift;

  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_nfs called ($keyattr - $keyattrval)\n"; # LogEnterSub

  if ($keyattr eq "nfs/hanfs_export/XFS") {
    # just print it from cmgetconf
    printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1");
    $curidx++;
  }
  if ($keyattr eq "nfs/hanfs_export/FILE_LOCK_MIGRATION") {
    # key attribute to look for remaining XFS in sapgetconf
    # loop through array at key nfs/hanfs_export/NFS
    my $relattr="nfs/hanfs_export/XFS";
    for my $i (0 .. $#{$_[3]->{$relattr}}) {
        my $lrelattrval=$_[3]->{$relattr}->[$i];
        # get the last part of the XFS string (the directory to export)
        # as the option might differ (sapgetconf can't reliably report
        # the options)
        my @parts = split(/\s+/, $lrelattrval);
        my $currNFS=$parts[$#parts];
        next if grep {$_ =~ /$currNFS/} @{$_[1]->{$relattr}};
        printAttr (*FH, $relattr,$lrelattrval,"sapgetconf-3",1);
    }
    # now print the key attribute from cmgetconf
    printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1");
    $curidx++;
  }

  # handle Linux related attributes
  if ($keyattr eq "tkit/nfs/nfs/TKIT_DIR") {
    # print from cmgetconf if it has a value
	if ($keyattrval ne "") {
      printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1");
	}
	else {
	  # hardcode SGCONF
	  printAttr (*FH, $keyattr,"\"\$SGCONF\"","deploysappkgs-0");
	}
    $curidx++;
  }
  if ($keyattr eq "tkit/nfs/nfs/XFS") {
  	my $relattr="tkit/nfs/nfs/XFS";  # kinda relates to itself
    # Loop through all XFS from cmgetconf. Might contain a empty XFS initially
    while ($keyattr eq $relattr) {
      printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1") unless($keyattrval eq "");
	  $curidx++;
	  $curnvref = $aref->[$curidx];
      $keyattr = $curnvref->[0];
      $keyattrval = $curnvref->[1];
	}
	# Now those from sapgetconf
    for my $i (0 .. $#{$_[3]->{$relattr}}) {
        my $lrelattrval=$_[3]->{$relattr}->[$i];
        # get the last part of the XFS string (the directory to export)
        # as the option might differ (sapgetconf can't reliably report
        # the options)
        my @parts = split(/\s+/, $lrelattrval);
        my $currNFS=$parts[$#parts];
        # mask a '*' in nfsxfs to avoid interpretation as part of a regex
        $currNFS =~ s/[*]/\[*\]/g;
        next if grep {$_ =~ /$currNFS/} @{$_[1]->{$relattr}};
        printAttr (*FH, $relattr,$lrelattrval,"sapgetconf-3",1);
    }
  }

  return $curidx;
}

#############################################################################
sub mergePkgAttr_pkg_descr {
  local *FH = shift;

  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  if ($keyattrval ne "Serviceguard Package") {
    printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1")
  }
  else {
    $keyattrval=$_[3]->{$keyattr}->[0];
    printAttr (*FH, $keyattr, $keyattrval,"sapgetconf-2",1);
  }
  $curidx++;
  return $curidx;
}

sub mergePkgAttr_script_logfile {
  local *FH = shift;

  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  # two cases for script_log_file:
  # 1. "initial" default containing $ variables (cmmakepkg)
  # 2. applied default not containing $ variables anymore (cmgetconf)
  my $initial_default_logfile_name="\$SGRUN/log/\$SG_PACKAGE.log";
  my $default_logfile_name="$ENV{SGRUN}/log/$_[6].log";
  # requires SGRUN env variable and $_[6]=pkgname

  printL "mergePkgAttr_script_logfile called ($keyattr - $keyattrval)\n"; # LogEnterSub

  #take attribute value from cmgetconf if it isn't the default
  # (keep in mind that $SG variables from pkg conf are expanded)
  #..else take it from sapgetconf - if reported
  if (($keyattrval ne $default_logfile_name) && ($keyattrval ne $initial_default_logfile_name)){
    printL "Non-default filename (default: $default_logfile_name)\n";
    printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1");
  }
  else {
    my $sgcattrval;
    $sgcattrval = $_[3]->{$keyattr}->[0] if (defined($_[3]->{$keyattr}));
    if ( defined($sgcattrval)) {
      if ($sgcattrval eq $keyattrval) {
        # identical, keep cmgetconf val
        #(may not be sufficent as env var get expanded when applied)
        printL "Identical values ($sgcattrval)\n";
        printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1");
      }
      else {
        # use the one from sapgetconf
        printL "Overriding package conf value\n";
        printAttr (*FH, $keyattr,$sgcattrval,"sapgetconf-2",1);
      }
    }
    else {
      # no attribute with that name defined in sapgetconf, use cmgetconf
      printL "No script logfile attribute reported by sapgetconf\n";
      printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1");
    }
  }
}

#############################################################################
#
# mergePkgAttr_node_name
#
# Special handling for node_name. Will replace '*' by a dynamically
# generated explicit node list.
# Assumes that either '*' or explicit list is configured in a package.
# The node list to replace '*' isn't delivered by sapgetconf, but generated
# from the cluster nodelist (@clusnodes) with local hostname as first one.
#
# Uses global @clusnodes and $host (latter from SAP_functions.pm)
#
# Applies only for packages containing HDB (later for SCS/ERS pairs)
#
# No check so far if all the nodes are configured to run HDB
#
# Restriction: This algorithm will only work if the cluster doesn't contain
# "extra" nodes (that is other than the HDB primary and secondary HDB nodes)
#
# WVE SO:
# - as HANA SU can run on a subset of (cluster-)nodes taking the whole list
#     might lead to a wrong config (already accomplished through the
#     SGeSAP_HDBREP_HOST mechanism)
# - as HANA SO sapgetconf reports a nodelist. This list should
#    -- replace "*" (for newly created packages) or
#    -- be merged into current list (for existing packages)
#############################################################################
sub mergePkgAttr_node_name {
  local *FH = shift;

  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_node_name called ($keyattr - $keyattrval)\n"; # LogEnterSub

  my $keyattr_initial = $keyattr;
  my $nodelist_change = 0;
  my @newnodelist = ();

  # WVE: use keyattr_inital instead of explicit "nodename"?

  # special handling for HDB if there is no explicit nodelist in sapgetconf output
  if (getshdbmodule ($_[3]) && ! defined $_[3]->{"node_name"}) {
    # For HDB this is (usually) a 2 node list (primary+secondary) with
    # localhost as 1st and other host in pair as 2nd.
    # If SGeSAP_HDBREP_HOST is defined in ENV, it is considered the "other" host
    if (defined ($ENV{"SGeSAP_HDBREP_HOST"})) {
      @newnodelist = ($host, $ENV{"SGeSAP_HDBREP_HOST"});
    }
    else {
      # Get local host in front of clusternodes list - which is usually
      # sufficient for 2 node clusters
      my $lnidx = first { $clusnodes[$_] eq $host } 0..$#clusnodes;
      # splices clusnodes at index of local hostname and print out elements
      # (splicing twice rotates array by lnidx elements
      @newnodelist = @clusnodes[$lnidx .. $#clusnodes,0 .. $lnidx-1];
    }

    if ($keyattrval eq "*") {
      $nodelist_change = 1; # replace default '*'
    }
    else {
      # Compare new with current list (as string). Overwrite current package node_name list!
      $nodelist_change = 1 if (join(" ",@newnodelist) ne join(" ",@{$_[1]->{"node_name"}}));
    }
    printL "$nodelist_change - @newnodelist\n";
  }

  map { printAttr (*FH, $keyattr,$_,"deploysappkgs-1",1) } @newnodelist if ($nodelist_change);

  # Get all node_name's (one chunk) from current or new pkg env and loop while
  # the attribute name stays the same
  # Advances to the end of the group. Also print them if no change indicated
  # If attribute value is '*' and there is a explicit sapgetconf nodelist, skip it
  while ($keyattr eq $keyattr_initial) {
  	  unless ($nodelist_change) {
  	  	printAttr (*FH, $keyattr,$keyattrval,"cmgetconf-1") unless (($keyattrval eq "*") && (defined $_[3]->{"node_name"})) ;
  	  }
      $curidx++;
      $curnvref = $aref->[$curidx];
      $keyattr = $curnvref->[0]; $keyattrval = $curnvref->[1];
  }
  # get the (possible) additional nodes (or replacement) from sapgetconf
  foreach my $sgcnode (@{$_[3]->{"node_name"}}) {
        next if grep {$_ eq $sgcnode} @{$_[1]->{"node_name"}};
        sgeasy_print80 ("WARNING: node $sgcnode not a cluster member") unless (grep {$_ eq $sgcnode} @clusnodes);
        printAttr (*FH, "node_name",$sgcnode,"sapgetconf-3",1);
  }

  return $curidx;
}

#############################################################################
#
# mergePkgAttr_vmfs
#
# Merging vmfs attributes is somewhat special as parent attribute reported
# by sapgetconf might not contain a valid value (one reflecting the name
# of a vmdk file on the VM host), but just a placeholder.
#
# As the check for new attributes can't be based on the parent attribute
# vmdk_file_name, the scsi_controller id is used instead. However, this
# might be not 100% reliable.
#
# The attributes reported from sapgetconf contains placeholders for
# vmdk_file_name and data_store. These needs to be edited manually before
# applying the generated package config.
#
#############################################################################
sub mergePkgAttr_vmfs {
  local *FH = shift;

  my $aref = $_[0];  # array of name/value arrays
  my $curidx = $_[5];
  my $curnvref = $aref->[$curidx];
  my $keyattr = $curnvref->[0];
  my $keyattrval = $curnvref->[1];
  my $nextattrrel = 0;

  printL "mergePkgAttr_vmfs ($keyattr - $keyattrval)\n"; # LogEnterSub

  do {
    $nextattrrel=0;   # assume that next attribute isn't related
    printAttr(*FH,$curnvref->[0],$curnvref->[1],"cmgetconf-1");
    $curidx++;
    $curnvref = $aref->[$curidx];  # next
    $nextattrrel = AttrIsRelated($keyattr,$curnvref->[0]);
    # switch to next key attribute value
    $keyattrval = $curnvref->[1] if ($nextattrrel eq 2);
  } while ($nextattrrel > 0);

  # Got everything from cmgetconf output. Check sapgetconf output now.
  # First figure out if number of (key) attributes in cmgetconf and
  # sapgetconf output differ. If more sapgetconf, then #sapgetconf-#cmgetconf
  # need to be added. Pick those which scsi_id isn't in cmgetconf already.

  # Get number of key attribute entries from cmgetconf and sapgetconf hash
  # Right now just for informational purposes
  my $num_cmgc_keyattrs = scalar @{$_[1]->{$keyattr}};
  my $num_sgc_keyattrs  = scalar @{$_[3]->{$keyattr}};
  printL "mergePkgAttr_vmfs $keyattr occurence: $num_cmgc_keyattrs - $num_sgc_keyattrs\n";

  # More entries reported from sapgetconf. Try figuring out which one to add based on SCSI id
  my $elemidx=0;
  foreach my $vmdkfile (@{$_[3]->{$keyattr}}) {
    printL "mergePkgAttr_vmfs: sgch $vmdkfile - ".$_[3]->{"scsi_controller"}->[$elemidx]."\n";
    # scsi id already in pkg -> skip
    if (grep (/$_[3]->{"scsi_controller"}->[$elemidx]/, @{$_[1]->{"scsi_controller"}})) {
	  $elemidx++;
	  next;
	}
	# this appears as if it needs to be added to pkg config
	# "vmdk_file_name" => [ "datastore_name", "scsi_controller", "disk_type" ],
	# This algorithm might end up adding more entries than num_sgc_keyattrs-num_cmgc_keyattrs
	printL "mergePkgAttr_vmfs: adding $keyattr - $vmdkfile\n";
	printAttr(*FH,$keyattr,$vmdkfile,"sapgetconf-2",1);
	printAttr(*FH,"datastore_name",$_[3]->{"datastore_name"}->[$elemidx],"sapgetconf-2",1);
	printAttr(*FH,"scsi_controller",$_[3]->{"scsi_controller"}->[$elemidx],"sapgetconf-2",1);
	printAttr(*FH,"disk_type",$_[3]->{"disk_type"}->[$elemidx],"sapgetconf-2",1);
	# WVE: could loop through related attribute array instead of hardcoding names (last 3 PrintAttr)
	# for my $attrname (@{$SGKeyAttr{$keyattr}}) {
	#   printAttr(*FH,$attrname,$_[3]->{$attrname}->[$elemidx],"sapgetconf-2",1);
	# }
  }
  return $curidx;
}

#############################################################################
#
# end of merging routines
#
#############################################################################


#############################################################################
#
# getshdbmodule
#
# Helper function for merge to check whether HDB module attributes
# will be (potentially added).
#
# Relies only on module list so far.
#
# Parameter:
# 0:  hash from sapgetconf (ref)
#############################################################################
sub getshdbmodule {
  return 1 if (grep {$_ =~ "^sgesap/hdb"} @{$_[0]->{"module_name"}});
  return 0;
}

#############################################################################
#
# sub checkforNewAttributes
#
# Arguments:
# \@pkgcmconf, \@pkgsapconf, \%pkgcmhash, \%pkgsaphash)
#
# Check if list of attributes returned by sapgetconf (pkgsap*) contains
# new attributes, that is attributes contained in pkgsap*, but not in pgkcm*
#
# Stops at first new attribute found.
#
# Doesn't take package_description and module_name into account as new attribute.
# ENHANCEMENT
# Might also to be discussed if all others should be taken into account
# (as it is done right now) or only parent/"key" attributes and their values.
#
# Returns:
# 0 - no new attributes found
# 1 - new attributes found
#
#############################################################################
sub checkforNewAttributes {

    my $attrname;
    my $attrval;

    # the code below is pretty much equivalent to that used in the merge
    # subroutines to check if an attribute needs to be added
    foreach $attrname (keys %{$_[3]}) {
      # ignore list (quite long, but seems it can't be written across multiple lines)
      # service_name is also ignored. check is done on service_cmd
      # fs_type alone also is not really useful to check if something has changed. a changed
      # attribute might match with that of another fs_name. Also <fs_type vxfs> does not
      # match with default <fs_type ""> (also having the meaning that fs_type is vxfs)
      next if ($attrname =~
        /(^module_|^package_description$|^fs_type$|^service_(restart|fail_fast_enabled|halt_timeout|name)$|^resource_polling_interval$|^sgesap\/sap_global\/(cleanup_policy|retry_count)|^#)/ );
      # if attribute isn't in pkgcmhash this is a new one
      unless (defined  $_[2]->{$attrname}) {
        printL "First new attribute name:  $attrname \n";
        return 1 ;
      }
      # attribute to be considered...
      # loop through array at attrname key
      foreach $attrval (@{${$_[3]}{$attrname}}) {
        #  next if grep {$_ eq $lkeyattrval} @{$_[1]->{$keyattr}}; # already in cmgetconf
        # grep below might need a more "relaxed" comparison (e.g. for VGs)
        # also one of them might be enclosed in doublequotes
        my $lclval = $attrval; # to avoid attrval from being modified
        $lclval =~ s/"//g; # doesn't explicity look for doublequote at begin and end
        $lclval =~ s/^\$/\\\$/g; # escape $ at the beginning (otherwise $something might be interpreted in grep again
        # special handling for VGs (they might be report back from cluster with a "/dev" prefix
        # match only end
        next if (($attrname eq "vg") && (grep {$_ =~ /${lclval}$/} @{$_[2]->{$attrname}}));
        #print "-- $attrname - $lclval -\n";
        #print "-- 3: @{${$_[3]}{$attrname}} -\n";
        #print "-- 2: @{${$_[2]}{$attrname}} -\n";
        unless (grep {$_  =~ /^("|)$lclval("|)$/} @{$_[2]->{$attrname}}) {
        #unless (grep {$_  =~ /^$lclval$/} @{$_[2]->{$attrname}}) {
          printL "First new attribute: $attrname - $attrval \n";
          return 1;
        }
      }
    }
    return 0;
}

#############################################################################
#
# Print some info to pkg config to identify what, when, where it (was) generated
#
#############################################################################
sub printheaderinfo {

  local *FH = shift;

  print FH "# -- $0\n";
  print FH "# -- ".substr($WHATSTRING,6,-1)."\n";
  print FH "# -- ".$host." - ".scalar localtime()."\n\n";
}

#############################################################################
#############################################################################
sub createpkgconfig {
# $sapsid
# $pkgname - package to create or extend
# $CandPkg2Inst{pkgname}) - list of entities/instance selected for this package
#
# $pkgpath not passed as parameter - global
# return 1 if pkg config needs to be changed, otherwise 0
#
# fills external @deployxtractions
#

  my $sapsid=shift;
  my $pkgname=shift;
  my $arrref=shift;

  my $pkgfilename=$pkgpath."/".$pkgname.".conf"; # pkgpath already contains sapsid
  my $updconffile;
  my $newattributes = 0; # flag indicating that new attributes have to be merged in

  my @sapinstlist = ();
  my @nfsxfslist = ();
  my $dbinst;
  my $hdbinst;
  my $repdbinst;

  my $rtnval=1;

  # iSCSI related
  my $pr_cntl_fname="sg/pr_cntl";
  my %PVsOfVGs=(); # hash holding PVs of configured LVs (key: LV - value: (ref to) array of PVs)

  #printL "$pkgname [ @{$arrref} ]\n";

  # get current package config
  # (when still pkgname is passed via cmdline. later: either selected existing package
  #  or package name made up of components to be clustered)
  # pkgcmconf and pkgcmhash could become local vars

  sgeasy_print80 ("Getting package config for $pkgname",1);
  # sgeasy_print80 ("(errors can be ignored - package may not exist)");
  # TODO ENHANCEMENT
  # TODO that problem (execting cmvgetconf with non-existing package) could
  # TODO   be worked around when checking the package config hash before...
  # TODO   In that case cmgetconf would only be executed for package which are
  # TODO   there and an error would trigger an exit.

  @pkgcmconf = () ;%pkgcmhash = (); # clear contents
  printL "cmd2exec: $cmd_cmgetconf $pkgname |\n";  # not "-f line" format
  # redirect error to suppress errormessage in case package doesn't exist
  open IN, "$cmd_cmgetconf $pkgname 2>/dev/null |"; # or die "Could not run $cmd_cmgetconf: $!\n";
  populateArray(*IN, \@pkgcmconf, \%pkgcmhash);
  close IN;
  #print Data::Dumper->Dump( [ \@pkgcmconf ], ["pkgcmconf"] );
  # reset pkgcm vars if there was stdout output for a non-successful exection
  # (like "Unable to send a datagram ...")
  if ($?) {  @pkgcmconf = () ;%pkgcmhash = (); }
  if (($ENV{"deploysappkgs_IgnPkgConfig"}||0) == 1) {
    # start over/fresh by ignoring current package configuration
    printL "Ignoring current package configuration for $pkgname\n";
    @pkgcmconf = () ;%pkgcmhash = ();
  }

  sgeasy_print80 ("Getting package config for $pkgname",1,"[SUCCESS]");

  # split list of instances in into real SAP instance, DB, REP and NFS
  # (Usually this recovers the SID again from these special instance names DB,REP and HDB.
  # Done this way because DBSID can be != SAPSID)
  foreach my $elem (@{$arrref}) {
    if ($elem =~ /^DB-(\w+)$/) {
        $dbinst = $1;
    }
	elsif ($elem =~ /^REP-(\w+)$/) {
        $repdbinst = $1;
    }
    if ($elem =~ /^HDB(p|s|)-(\w+)$/) {
        $hdbinst = $2.$1;  # TODO HDB: where to use hdbinst?
        # WVE: $1 is added to call in "forced mode". However this should only be done for SO!
    }
    elsif ($elem =~ /^NFS-\|(\S+)\|$/) {
        # only last part of host:path spec. Otherwise pattern matching will be confused with "*:"
        push @nfsxfslist,(split(':',$1))[-1];
    }
    else {
        push @sapinstlist,$elem;
    }
  }

  sgeasy_print80 ("Collecting instance information configured to be in package",1);

  # it has been made sure that at least one sapinstance was returned by previous sapgetconf cmd (to be tested !!)
  @pkgsapconf = ();  %pkgsaphash = ();
  my $pkgactions = ();  # ref to array or undef

  my $cmd2exec = "$cmd_sapgceasy";
  $cmd2exec = "$cmd2exec sapsid=$sapsid,sapinstance=" . join(",sapinstance=",@sapinstlist)  if (scalar @sapinstlist);
  sgeasy_print80 (join(',',@sapinstlist))  if (scalar @sapinstlist);
  # use sapinst instead of dbinst for DB. sometimes different, then sapinst give better results
  # TODO: use of "," and casading if's are a bit akward...
  $cmd2exec = ${cmd2exec}."," if (defined $dbinst) && (scalar @sapinstlist);
  $cmd2exec = $cmd2exec."dbinstance=".$sapsid  if (defined $dbinst);
  $cmd2exec = ${cmd2exec}."," if (defined $repdbinst) && ((defined $dbinst) || (scalar @sapinstlist));
  $cmd2exec = $cmd2exec."repdbinstance=".$sapsid  if (defined $repdbinst);
  $cmd2exec = ${cmd2exec}."," if (scalar @nfsxfslist && ((defined $dbinst) || (scalar @sapinstlist)));
  $cmd2exec = ${cmd2exec}."nfs=\"(" . join ("|",@nfsxfslist) .")\"" if (scalar @nfsxfslist);
  $cmd2exec = $cmd2exec."hdbinstance=".$hdbinst  if (defined $hdbinst);
  sgeasy_print80 ("DB $dbinst ($sapsid)") if (defined $dbinst);
  sgeasy_print80 ("HDB $hdbinst") if (defined $hdbinst);
  sgeasy_print80 ("REP $repdbinst" . ($dbinst?" ($dbinst)":"")) if (defined $repdbinst);
  sgeasy_print80 ("NFS ". join(',',@nfsxfslist))  if (scalar @nfsxfslist);
  printL "cmd2exec: $cmd2exec |\n";
  my $cmdx;
  # force sapgetconf to report generic_resource and dependency even in the absence of a
  # enqor package - by setting the environment variable SGeSAP_ENQOR_PKGNAME
  # $ENV{"SGeSAP_ENQOR_PKGNAME"} = getpkgname();
  my $enqpkgname = enqor::getpkgname(); # get name of existing or default (to be generated) enqor pkg

#  open IN, "export SG_PACKAGE=$pkgname; export SGeSAP_ENQOR_PKGNAME=$enqpkgname; $cmd2exec |"
#     or die "Could not run $cmd2exec: $!\n";

  $cmdx="export SG_PACKAGE=$pkgname; export SGeSAP_ENQOR_PKGNAME=$enqpkgname; $cmd2exec";

  if ( defined $hdbinst ) {
    # kludge - swap primary / secondary hdb instance
    # will only work for a two node configuration
    my $reppkgname;
    # get "other" package name for the dependency ... either through already
    # defined SGeSAP_HDBREP_PKGNAME or assigning the default name
    if (defined ($ENV{"SGeSAP_HDBREP_PKGNAME"})) {
      $reppkgname = $ENV{"SGeSAP_HDBREP_PKGNAME"};
    }
    else {
      $reppkgname = "hdbs${hdbinst}" if ( $pkgname =~ /^hdbp/ );
      $reppkgname = "hdbp${hdbinst}" if ( $pkgname =~ /^hdbs/ );
    }
    $cmdx="export SG_PACKAGE=$pkgname; export SGeSAP_HDBREP_PKGNAME=$reppkgname; $cmd2exec";

    if ($ip_opts) {
      # before exporting sanity check if $ip_opts contain valid IP addresses
      if (checkIPspec($ip_opts) == 0) {$cmdx="export START_IP_ADDRESS=$ip_opts;".$cmdx};
    }
  }

  printL "$cmdx\n";
  open IN, "$cmdx |" or die "Could not run $cmd2exec: $!\n";
  populateArray(*IN, \@pkgsapconf, \%pkgsaphash);
  close IN;
#  print Data::Dumper->Dump( [ \@pkgsapconf ], ["pkgsapconf"] );
#  print Data::Dumper->Dump( [ \%pkgsaphash ], ["pkgsaphash"] );

  # Filter out extra actions and delete them from hash
  # (can be left in array as these entries happen to be ignored)
  # ... push only if !localhdb
  $pkgactions = delete $pkgsaphash{"\@deploysappkgs:"};
  undef $pkgactions if ($localhdb); # pkg extra actions only for "main" call -  comment for testing
  push (@deployxtractions,@{$pkgactions}) if (defined $pkgactions);

  sgeasy_print80 ("Collecting instance information",1,"[SUCCESS]");

  # Figure out if attributes where reported which should belong to a CFS package
  # skip on linux
  unless ($^O eq "linux") {
    sgeasy_print80 ("Checking for CFS related attributes",1);
    handleCFS(\@pkgsapconf, \%pkgsaphash, \%clusrsrchash, $pkgname);
    sgeasy_print80 ("Checking for CFS related attributes",1,"[SUCCESS]");
  }

  # Handle generic resource. So far only Linux, but decision on this is based on sapgetconf
  # reporting a generic resource. This will also trigger the creation of the enqor MNP package
  # if it doesn't exist yet (usage of the sgesap generic resource in existing package won't
  # trigger such a creation!)
  if (defined ($pkgsaphash{"generic_resource_name"})) {
    if  (grep (!/^(sitecontroller|safesync)/, @{$pkgsaphash{"generic_resource_name"}})) {

  	  # WVE SO: the first check was too generic (pun intended). Now also HDB SO will report
  	  #         the sitelatch as generic resource (which will trigger creation of enqord MNP)
  	  #         A enqor package should only be created if there's not just the sitecontroller rsrc.
  	  # sapgetconf reporting that generic_resource is needed which implies tht
  	  # enqor package needs to be created if none exists
  	  # Relies on the assumption that sapgetconf is reporting only generic_resources requiring enqor MNP
  	  # Instead of creating it here in the middle of other SGeSAP package generation, a flag is set
  	  printL "generic_resource_name reported\n";
      enqor::set_pkggen_required() unless (defined(enqor::getfoundpkgname()));
    }
    if  (grep (/^sitecontroller/, @{$pkgsaphash{"generic_resource_name"}})) {
    	# WVE SO: sitecontroller resource reported. trigger the generation of sc package ??
    	printL "sitecontroller resource reported\n"
    }
  }

#  print Data::Dumper->Dump( [ \@pkgsapconf ], ["pkgsapconf"] );
#  print Data::Dumper->Dump( [ \%pkgsaphash ], ["pkgsaphash"] );

  # check if additional modules need to be added to package.
  # also covers the fact that package has to be generated from scratch when
  # package doesn't exist yet
  sgeasy_print80 ("Check modules required by instance(s) not in package $pkgname",1);
  @addmods = ();
  @addmods = compare_mods(\%pkgcmhash, \%pkgsaphash);

  # Additional check for new iSCSI related pr_cntl module. This check isn't executed
  # by sapgetconf (to keep it cluster and OS agnostic and avoid unnecessary checks).
  #
  # If a VG has a iSCSI PV it is assumed that pr_cntl is needed. It is *not* checked if all
  # PVs of a VG are on iSCSI and if the pkg has just one sort of VGs (either iSCSI or FC).
  #
  # First check if module (link) exists at all
  #

  if (-l ($ENV{SGCONF}||"/etc/cmcluster")."/modules/".$pr_cntl_fname) {
    # Module there. Need to check if needed in SGeSAP pkg
    sgeasy_print80 ("Checking if persistent reservation module needs to be added to package",1);
    printL "$pr_cntl_fname link found\n";
    if (! grep $pr_cntl_fname, @{${\%pkgcmhash}{"module_name"}} ) {
      my $vgname="";
      my $pvoniscsi; # set to first PV on iSCSI found, vgname contains the VG then
      # module not configured to pkg yet, check if necessary by looking at PVs of VGs in pkgsaphash
      sgeasy_print80 ("Checking if discovered volume groups have iSCSI physical volumes",1);
      printL "$pr_cntl_fname not in package config\n";
      # Build up LV/PVs hash. Called for every package to be created (alternative: call it once
      # unconditionally in global context)
      # Linux shortcut: pretend that PVsOfVGs stays empty, but module is mandatory in any case
      if ($^O eq "linux") {
      	push @addmods, $pr_cntl_fname;
        sgeasy_print80 ("Module $pr_cntl_fname added to list of required modules");
      }
      else {
      	%PVsOfVGs=getPVsOfVGs(); # HP-UX
      }


      VGLOOP: foreach $vgname (@{${\%pkgsaphash}{"vg"}}) {
        #printL "$vgname\n";
        foreach my $pvname (@{$PVsOfVGs{$vgname}}) {
          if (diskoniscsi($pvname)) {
            $pvoniscsi=$pvname;
            printL "PV on iSCSI - VG: $vgname, PV: $pvoniscsi\n";
            # having one is enough to trigger module inclusion
            last VGLOOP;
          }
        }
      }

      if (defined $pvoniscsi) {
        push @addmods, $pr_cntl_fname;
        sgeasy_print80 ("iSCSI physical volume(s) discovered");
        sgeasy_print80 ("Module $pr_cntl_fname added to list of required modules");
      }
      else {
        sgeasy_print80 ("No iSCSI physical volumes discovered");
      }
      sgeasy_print80 ("Checking if discovered volume groups have iSCSI physical volumes",1,"[SUCCESS]");
    }
    else {
      sgeasy_print80 ("Persistent reservation module already configured in package");
    }
    sgeasy_print80 ("Checking if persistent reservation module needs to be added to package",1,"[SUCCESS]");
  }
  else {
    printL "$pr_cntl_fname link not found (installed SG version doesn't support persistent reservation\n";
  }  # end of iSCSI check

  if (scalar @addmods != 0)
  {
    sgeasy_print80 ("[ @addmods ]");
    $nfsmodadded = grep(/^${nfsmodname}$/,@addmods); # number of matches
    printL "nfsmodadded: $nfsmodadded (\$SGCONF/modules/${nfsmodname})\n";
    # check for module existance. uses -e although this is usually a link and -l could be used)
    if ($nfsmodadded && !(-e ($ENV{SGCONF}||"/etc/cmcluster")."/modules/${nfsmodname}" )) {
        sgeasy_print80 ("ERROR: HANFS toolkit (module $nfsmodname) required but not installed");
        sgeasy_print80 ("Please install HANFS toolkit");
        sgeasy_print80 ("Skipping package update/creation");
        sgeasy_print80 ("Checking modules",1,"[FAILURE]");
        return 0;
    }
    # restricted to list of instances. Check if instance is in list. (??)
    # Partial specifications allowed (??)
    # @pkgcmconf being empty is considered as "package doesn't exist" case
    sgeasy_print80 ("Generating new package config by adding required modules",1);
    $updconffile = addmods2config($pkgname, scalar @pkgcmconf, @addmods);
    if ($updconffile eq "") {
      sgeasy_print80 ("Generating new package config",1,"[FAILURE]");
      return 0;
    }
    sgeasy_print80 ("Generating new package config",1,"[SUCCESS]");
    # as the config got updated with new modules this also means that sapgetconf
    # found attributes for these modules
    # (however, no new modules doesn't mean that there are no new attributes. there
    #  may be some for the already configured modules. needs to be found out later)
    $newattributes = 1;
  }
  else {
    sgeasy_print80 ("All required modules already in package");
  }
  sgeasy_print80 ("Checking modules",1,"[SUCCESS]");

  # at this stage either the package already contains all the required modules
  # or a new/updated config has been created. reread latter config (if defined)
  # (only to be set by addmods2config)
  if (defined $updconffile) {
    printL "Reading updated config file $updconffile\n";
    open IN, $updconffile or die "Could not open $updconffile: $!\n";
    # clear hash and array
    @pkgcmconf=();%pkgcmhash=();
    populateArray(*IN, \@pkgcmconf, \%pkgcmhash);
    close IN;
    unlink $updconffile; # cleanup
  }
  #print Data::Dumper->Dump( [ \@pkgcmconf ], ["pkgcmconf"] );
  #print STDERR Data::Dumper->Dump( [ \%pkgcmhash ], ["pkgcmhash-1"] );

  # check if SAP SIDs are compatible.
  # This usually only makes sense if you point the script to
  # a certain package (-n) option
  # If the selection is left to the script, it should always
  # be compatibel.
  die "Incompatible values for attribute sap_system (SID)"
         unless  checkSAPSIDs (\%pkgcmhash, \%pkgsaphash);


  # Now start walking through the first array and print it out
  # while trying to merge the 2nd in
  # Certain attribute name act as trigger for module attributes and parent relationship
  # (having a attr-xref would be helpful here)
  # Note: Even if a module is configured in a package it doesn't need to contain
  #    an attribute of it. So we won't have "output triggers" for all attributes
  #    of the 2nd array.
  #

  # Find out if config needs to be updated with new attributes. Otherwise merging can be skipped
  # Flag may already be set because new modules needed to be added to accomodate some new attributes.
  $newattributes = checkforNewAttributes(\@pkgcmconf, \@pkgsapconf, \%pkgcmhash, \%pkgsaphash, $pkgname)
    unless ($newattributes == 1);


#print STDERR Data::Dumper->Dump( [ \@pkgcmconf ], ["pkgcmconf"] );
#print STDERR Data::Dumper->Dump( [ \@pkgsapconf ], ["pkgsapconf"] );
#print STDERR Data::Dumper->Dump( [ \%pkgcmhash ], ["pkgcmhash-1"] );
#print STDERR Data::Dumper->Dump( [ \%pkgsaphash ], ["pkgsaphash-1"] );
  if ($newattributes) {
    sgeasy_print80 ("Merging attributes and writing config file",1);
    sgeasy_print80 ("New attribute(s) found");
    open PKGFILE, ">$pkgfilename" or die "Cant open $pkgfilename - $!";
    printheaderinfo (*PKGFILE);
    mergeAttributes(*PKGFILE,\@pkgcmconf, \@pkgsapconf, \%pkgcmhash, \%pkgsaphash, $pkgname);
    close PKGFILE;
    sgeasy_print80 ("Merging attributes",1,"[SUCCESS]");

    # Output for triggering SGMgr to pickup the generated config file
    # (has to follow the format: (File<blank>Name=<filespec><end of line>)
    #print "\nFile Name=$pkgfilename\n\n";
  }
  else {
    sgeasy_print80 ("Merging attributes",1,);
    sgeasy_print80 ("No new attribute(s) found");
    sgeasy_print80 ("Skipping package creation/merging for $pkgname");
    sgeasy_print80 ("Merging attributes",1,"[SUCCESS]");
    $rtnval = 0;
  }
  return $rtnval;
}

#############################################################################
#
# sub getRunningPkgs
#
# Arguments:
# \@modifiedPkgs  - list of existing packages needing modification (ref)
# \%clusrsrchash  - used cluster resource cross-reference (ref)
#
# Check if packages needing modification are up. And if they are up need
# to be stopped. Returns a list of packages to be stopped. The list is
# already in the correct stop-order
#
# - Just checking for modules might not be sufficient as the module could
#   be included w/o any attributes configured.
# - It might also be the case that just online reconfigurable general
#   SG attributes got changed/added. In this case halting wouldn't be
#   necessary.
#
#
# Returns:
# List of packages to be stopped
#
# TODO HDB: what to do here? (also necessary with repdb?)
#
#############################################################################
sub getRunningPkgs {

  # Get packages having
  # (1) nfs/hanfs
  # (2) sgesap/dbinstance
  # (3) sgesap/sapinstance
  # (4) none of above
  # and are on modified list.
  my @nfslist = ();
  my @dblist  = ();
  my @saplist = ();
  my @nonelist = ();

  for my $pkg (@{$_[0]}) {
    if (grep (/^${pkg}$/,@{$_[1]->{"MOD_NAME"}->{"nfs/hanfs"}})) {
      push (@nfslist,$pkg); next;
    }
    if (grep (/^${pkg}$/,@{$_[1]->{"MOD_NAME"}->{"sgesap/dbinstance"}})) {
      push (@dblist,$pkg); next;
    }
    if (grep (/^${pkg}$/,@{$_[1]->{"MOD_NAME"}->{"sgesap/sapinstance"}})) {
      push (@saplist,$pkg); next;
    }
    push (@nonelist,$pkg);
  }
  printL "getRunningPkgs: [ @nfslist ] [ @dblist ] [ @saplist ] [ @nonelist ]\n";
  return (@nonelist,@saplist,@dblist,@nfslist);
}

#############################################################################
#
# sub pkgHasTKmod
#
# Check if package has a module of the toolkit configured. Toolkit name
# is lowercase.
# Assumes that module starts with toolkit name.
# Looks into "MOD_NAME" part of the hash.
#
# Arguments:
# pkgname         - name of packages
# tkname          - name of the toolkit (lowercase)
# \%clusrsrchash  - used cluster resource cross-reference (ref)
#
# Returns: 1 if it has a module of that toolkit, 0 if not
#
#############################################################################
sub pkgHasTKmod {

  my $pkgname = shift;
  my $tkname = shift;

  printL "pkgHasTKmod: $pkgname, $tkname\n";

  foreach my $modname (keys %{$_[0]->{"MOD_NAME"}}) {
    #printL "$modname\n";
    next unless ( "$modname" =~ m#^${tkname}/# );
    if (grep (/^${pkgname}$/,@{$_[0]->{"MOD_NAME"}->{$modname}})) {
      printL "$pkgname has $tkname\n";
      return 1;
    }
  }
  return 0;
}

#############################################################################
#
# sub handleCFS
#
# Handle attributes related to CFS. The only attribute reported by sapgetconf
# so far is "cvm_dg". However, it also may report related "fs_name",
# "fs_directory", "fs_type"
# If such attributes are there, they will be removed from the config reported
# sapgetconf and dependencies added instead.
#
# To be verified: what does sapgetconf report in such a case as it relies
#   on mount command to get directory, filesystem and type and derive the DG
#   from that?
#
# Restrictions:
# - so far no "raw" cvm_dg are supported
# - current coding may also find cvm_dgs and related fs_name (=volume) in non-CFS
#   packages and set up a dependendy to it
#
# Also keep in mind that the index of the DG isn't necessarily the index
# of name,directory and type as each DG is only reported once, but can contain
# more than one FS.
#
# Arguments:
# \@pkgsapconf    - array and
# \%pkgsaphash    - hash of attributes returned by sapgetconf (ref, read,write)
# \%clusrsrchash  - used cluster resource cross-reference (ref, read)
# $pkgname        - current package name
#
# Returns: List of dependencies found
#
# Implementation:
# - Return empty list if no cvm_dg hash entry exist
# - For all "cvm_dg" attributes in pkgsaphash do
# -- Find related fs_name,fs_directory, fs_type attributes
# -- Find CFS package containing fs_directory
# -- Add this package to a dependency list (if found, otherwise warning)
# -- Remove cvm_dg and related fs_name,fs_directory, fs_type information
#      from hash and array
# - Remove cvm_dg hash entry
# - Return (with warning) if no dependencies could be found
# - Add dependency module to list of required modules
# - For all dependencies
# -- Add name, location (same_node) and condition (up) to array and hash
#      (name build from packagename+"UP")
#############################################################################
sub handleCFS {

  my %deppkghash = ();
  my @deppkglist = ();

  my $sgaidx=0;
  my $sgaref; # ref to array with index 0 attrib name and index 1 attrib value

  return () unless defined $_[1]->{"cvm_dg"};

  foreach my $dg (@{$_[1]->{"cvm_dg"}}) {
    printL "cvm_dg: $dg\n";
    my @usedinpkg = checkUsedClusResources ($_[2] ,"cvm_dg", $dg, $_[3]);
    next unless (scalar @usedinpkg);
    # maybe there should be a warning that a CVM DG should be configured
    # in package, but no dependency can be setup (or setup dependency to
    # SG-CFS-pkg?)
    sgeasy_print80("cmv_dg $dg used in package @usedinpkg");

    # Actually the dependencies have to be setup to the package containing
    # the FS/MP not the DG (in case of modular it may be the same).
    next unless defined $_[1]->{"fs_name"}; # no fs_name reported (why not?)
    my @dgfss = grep {$_ =~ m#^/dev/vx/.+/${dg}/#} @{$_[1]->{"fs_name"}};
    printL "[ @dgfss ]\n";
    for my $dgfs (@dgfss) {
      @usedinpkg = checkUsedClusResources ($_[2] ,"fs_name", $dgfs, $_[3]);
      next unless (scalar @usedinpkg);
      sgeasy_print80("fs_name $dgfs used in package @usedinpkg");
      # The following code assumes that the list just contains 1 package
      # it may be necessary to make sure that just the CFS package is returned
      $deppkghash{$usedinpkg[0]}=1;
    }

    # find the related cvm_dg entry in pkgsapconf - remove it
    $sgaidx=-1;
    #printL "".scalar @{$_[0]} ."\n";
    foreach $sgaref ( @{$_[0]} ) {
      $sgaidx++;
      next unless ($sgaref->[0] eq "cvm_dg");
      next unless ($sgaref->[1] eq $dg);
      # at position of correct entry, slice it there
      printL "cvm_dg $dg found at index $sgaidx\n";
      splice(@{$_[0]},$sgaidx,1);
      last;
    }
    # Start over finding related fs_name, fs_directory, fs_type
    # Make use of the fact they show up in that order each fs_name
    $sgaidx = -1;
    foreach $sgaref ( @{$_[0]} ) {
      $sgaidx++;
      next unless ($sgaref->[0] eq "fs_name");
      next unless ($sgaref->[1] =~ m#^/dev/vx/.+/${dg}/#);
      # at position of correct entry, slice it there
      printL "fs_name $sgaref->[1] found at index $sgaidx\n";
      splice(@{$_[0]},$sgaidx,3);
      last;
    }
    #printL "".scalar @{$_[0]} ."\n";

    # Go through array in hash "fs_name", find matching entry there
    # and remove it. Remove entries at same index in array of hashes
    # fs_directory and fs_type.
    $sgaidx=-1;
    foreach my $fsname (@{$_[1]->{"fs_name"}}) {
      $sgaidx++;
      next unless ($fsname =~ m#^/dev/vx/.+/${dg}/#);
      printL "$fsname found at index $sgaidx in hash\n";
      splice (@{$_[1]->{"fs_name"}},$sgaidx,1);
      splice (@{$_[1]->{"fs_directory"}},$sgaidx,1);
      splice (@{$_[1]->{"fs_type"}},$sgaidx,1);
    }
    # finally, delete cvm_dg hash
    delete $_[1]->{"cvm_dg"};
  }

  @deppkglist = keys %deppkghash;
  printL "deppkglist: [ @deppkglist ] \n";
  # do nothing if no dependencies where found
  return @deppkglist unless (scalar @deppkglist);

  # add dependency module (code assumes that module_names
  # appear at the end of sapgetconf output)
  push @{$_[0]}, ["module_name","sg/dependency" ];
  push @{${$_[1]}{"module_name"}}, "sg/dependency";
  #printL "".scalar @{$_[0]} ."\n";

  # splice in dependency attributes before module_name
  $sgaidx=0;
  $_[1]->{"dependency_name"} = [] unless defined $_[1]->{"dependency_name"};
  $_[1]->{"dependency_location"} = [] unless defined $_[1]->{"dependency_location"};
  $_[1]->{"dependency_condition"} = [] unless defined $_[1]->{"dependency_conditiion"};
  foreach $sgaref ( @{$_[0]} ) {
    last if ($sgaref->[0] eq "module_name");
    $sgaidx++;
  }
  foreach my $deppkg (@deppkglist) {
    splice(@{$_[0]},$sgaidx++,0,["dependency_name",$deppkg."_SNUPDEP"]);
    splice(@{$_[0]},$sgaidx++,0,["dependency_condition",$deppkg." = UP"]);
    splice(@{$_[0]},$sgaidx++,0,["dependency_location","same_node"]);
    push @{${$_[1]}{"dependency_name"}}, $deppkg."_SNUPDEP" ;
    push @{${$_[1]}{"dependency_condition"}}, $deppkg." = UP" ;
    push @{${$_[1]}{"dependency_location"}}, "same_node" ;
  }

  sgeasy_print80("All reported cvm_dg and related filesystem attributes where removed");
  sgeasy_print80("from package configuration and dependencies created instead");

  return @deppkglist;
}

#############################################################################
#
#  iSCSI related functions
#  OS dependent
#
#############################################################################

my $CMD_LVMADM=qq[/usr/sbin/lvmadm -l -F |]; # get list of PVs for all VGs, no parameter (HP-UX 11.31 only)
my $CMD_IOSCAN=qq[/usr/sbin/ioscan -kF ]; # param: legacy character disk (HP-UX only)
my $CMD_SCSIMGR=qq[/usr/sbin/scsimgr lun_map -pD ]; # param: agile character disk (HP-UX only?)

#############################################################################
#
# getPVsOfVGs
#
# create map VG -> List of PVs
#
# no parameter
#
# uses CMD_LVMADM
#
# output: hash of LVs (value: ref to array with PVs)
#
# Expected output of CMD_LVMADM (example):
# vg_name=/dev/vgsapSB2:pv_name=/dev/disk/disk18,/dev/disk/disk27
# vg_name=/dev/vgSB2DB:pv_name=/dev/disk/disk21,/dev/disk/disk29
# vg_name=/dev/vgSB2DBdata:pv_name=/dev/disk/disk23
# vg_name=/dev/vgSB2DBlog:pv_name=/dev/disk/disk25
# (line by line output w/o -F otherwise)
#
#############################################################################

sub getPVsOfVGs {
  my %PVperLV = ();
  my $pvon = 0;
  my $cmd = $CMD_LVMADM;

  printL "Executing $cmd \n";

  open (my $FH, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
  while (<$FH>) {
    #print $_;
    if (m#vg_name=/dev/(\w+):pv_name=(.+)$#) {
      #print "vg: $1\n";
      #print "pvlist: $2\n";
      my $curr_vg=$1;
      $PVperLV{$curr_vg}=[];
      @{$PVperLV{$curr_vg}}=split(/,/,$2);
    }
  }
  close ($FH) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

  return %PVperLV;
}

#############################################################################
#
# getchardiskdev
#
# Replace block diskname by character diskname
# Assumes one of the forms /dev/dsk|disk|cdisk
#
# Parameter: blockdisk pathname
#
# Returns: characterdisk pathname
#
#############################################################################

sub getchardiskdev {
  my $pvname = shift;
  $pvname =~ s#/dev/dsk/#/dev/rdsk/#;
  $pvname =~ s#/dev/disk/#/dev/rdisk/#;
  $pvname =~ s#/dev/cdisk/#/dev/rcdisk/#;
  return $pvname;
}

#############################################################################
#
# diskislegacy
#
# check if disk is legacy
#
# Parameter: diskname (character device)
#
# Returns: 1 if legacy, 0 otherwise
#
#############################################################################

sub diskislegacy {
  my $pvname = shift;
  return 1 if ($pvname =~ m#^/dev/rdsk/#);
  return 0;
}

#############################################################################
#
# diskoniscsi
#
# Check if disk is on iSCSI
#
# Parameter: diskname (character)
#
#
# Calls: ioscan -kF <chardev> for legacy
#        scscimgr lun_map -pD for agile
#
# If output contains "iscsi" it is considered to be on ISCSI, otherwise FC
#
#############################################################################

sub diskoniscsi {
  my $pvname = shift;
  my $cpvname = getchardiskdev($pvname);
  my $isiscsi = 0;
  my $cmd="";
  my $FH;

  $cmd = $CMD_SCSIMGR . " $cpvname |";
  $cmd = $CMD_IOSCAN . " $cpvname |" if (diskislegacy($cpvname)); # overwrite for legacy
  printL "Executing $cmd\n";

  open ($FH, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
  $isiscsi = 1 if (<$FH> =~ "iscsi"); # single line input only
  close ($FH) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

  return $isiscsi;
}
#############################################################################
#
# IsMCforHDBSO
#
# Check if theres is a Metrocluster with SITEs configured when there are
# HANA scaleout instances discovered.
# If there are scaleout instances but no MC the intances are being removed.
#
# Parameters:
#   %sapgcinsthash   ref. to instance hash (sapgetconf disovery)
#   %clusrsrchash    ref. to cluster resources hash (contains SITENAMES)
#
# Returns:
#   1    OK
#   0    config wrong
#
#############################################################################

sub IsMCforHDBSO {

    my $sapgchashref = shift;
    my $clusrsrchashref = shift;
    my $subrc = 1;

    foreach my $sgcinst (keys %{$sapgchashref}) {
    	printL "$sgcinst\n";
    	next unless ($sgcinst =~ /HDB(s|p)-/); # not a HANA instance
    	next unless (scalar (my @tmparr = split(",",$sapgchashref->{$sgcinst}->[1])) > 1);
    	# scaleout hana at this point, check cluster config for SITENAMES (this assumes that
    	# 2 sites are setup then)
    	unless (defined ($clusrsrchashref->{SITENAMES})) {
    		sgeasy_print80 ("$sgcinst requires SITEs to be defined in cluster config");
    		# delete $sapgchashref->{$sgcinst};  # WVE SO - decomment later
    		$subrc=0;
    	}
    }
	return $subrc;
}

#############################################################################
#
# ensureHDBConfigRestrictions
#
# Ensure that HDB config restrictions w.r.t. scaleup/out are met
#
# Parameters:
#   $sapsid          SID of HDB to be configured
#   %sapgcinsthash   ref. to instance hash (sapgetconf disovery)
#   %clusrsrchash    ref. to cluster resources hash
#   %cluspkgconfighash  ref to cluster pkg config hash (
#
# Returns:
#   1    OK
#   0    config wrong
#
#############################################################################

sub ensureHDBConfigRestrictions {

    my $hdbsid = shift;
	my $sapgchashref = shift;
    my $clusrsrchashref = shift;
	my $cluspkgconfighashref = shift;
    my $subrc = 1;
	my $rtntext = "";

	return ($subrc,"Restrictions ignored") if (($ENV{"deploysappkgs_ignore_hdbconfigrestrictions"}||0) == 1);

    foreach my $sgcinst (keys %{$sapgchashref}) {
    	printL "ensureHDBConfigRestrictions: $sgcinst\n";
    	next unless ($sgcinst =~ /HDB(s|p)-/); # not a HANA instance
		# 1. if scale-out check if there is already a HDB configured
		# 2. if scale-up check if there is already a HDB scaleout configured
		# if one of the conditions above apply it is considered a unsupported config

		# Technical implementation:
		# Figure out if current sapsid requires scale-up or out.
		# Look into clusrsrchashref for package(s) under sgesap/hdbprimary module ("MOD_NAME")
		# With that packages look up pkg SID(s) and if they are failover or multinode.
		#
		# Get a existing configured other SID pkg with a sgesap/hdbprimary module and its SID first
		my $has_scaleout_pkg = 0;
		my @hdbprim_otherpkgs = ();
	    foreach my $pkg (@{$clusrsrchashref->{"MOD_NAME"}->{"sgesap/hdbprimary"}}) {
		  next if ($cluspkgconfighashref->{$pkg}->[$hdbsid_idx] eq $hdbsid); # own SID
		  $has_scaleout_pkg = 1 if ($cluspkgconfighashref->{$pkg}->[$pkgtype_idx] eq "multi_node");
		  push @hdbprim_otherpkgs,$pkg;
		}
		printL "hdbprim_otherpkgs: @hdbprim_otherpkgs ($has_scaleout_pkg)\n";

		if (scalar (my @tmparr = split(",",$sapgchashref->{$sgcinst}->[1])) == 1) {
		  # wants to configure a scale-up pkg, not allowed if there is already a scaleout pkg
		  if ($has_scaleout_pkg) {
		    $subrc = 0;
			$rtntext = "Requested scale-up package configuration conflicts with scale-out packages (@hdbprim_otherpkgs)"
		  }
		}
		else {
		  # wants to configure a scale-out, not allowed if there is already another HANA pkg
		  if (scalar (@hdbprim_otherpkgs) > 0) {
		    $subrc = 0;
			$rtntext = "Requested scale-out package configuration conflicts with already configured HANA packages (@hdbprim_otherpkgs)"
		  }
		}
    }
	return ($subrc, $rtntext);
}
#############################################################################
#
# GetMCSiteNames
#
# Get Metrocluster Site IDs. Returns a liste of sites.
#
# HANA scaleout instances discovered.
# If there are scaleout instances but no MC the intances are being removed.
#
# Parameters:
#   %clusrsrchash    ref. to cluster resources hash (contains SITENAMES)
#   nodename         optional node to be matched to site. node's
#                    site appears first in list.
#
# Returns:
#   undef if no MC Sites are defined
#   empty list if (optional) node isn't in any site
#   List of site names (local site first or site of optional node first)
#
#############################################################################
sub GetMCSiteNames {
	my $clusrsrchash = shift;
	my $clusnode = shift || $host ;
	my @sitenames=();
	my $nodesite;

	printL "GetMCSiteNames\n";

	return undef unless defined ($clusrsrchash->{"SITENAMES"} ); # no SITEs in cluster

	foreach my $sitename (keys %{$clusrsrchash->{"SITENAMES"}}) {
		my $siteid = $clusrsrchash->{"SITENAMES"}->{$sitename};
		printL "GetMCSiteNamess: $sitename - $siteid\n";
		if (grep {$clusnode eq $_} @{$clusrsrchash->{"SITENODES"}->{$siteid}}) {
		  $nodesite = $sitename;
		  last;
		}
	}
	if (defined $nodesite) {
		$sitenames[0] = $nodesite;
		push @sitenames, grep ({$nodesite ne $_} keys %{$clusrsrchash->{"SITENAMES"}});
	}
	#printL "GetMCSiteNames: [ @sitenames ]\n";
	return @sitenames;
}

#############################################################################
#
# Some stuff related to Sybase toolkit on Linux
#
#############################################################################

#############################################################################
#
# IsTKConfigured (\%sapgcinsthash, \%clusrsrchash, $sgckeyprefix, $crsrckey)) {
#
# Match discovered instance IP address(es) to a maybe already configured package(s).
# Used as first steop to "assign" an instance to a matching package.
# IP address(es) are taken from discovered SAP config (sapgchash). Addresses to match to
# are looked up in clusip2pkgmap.
#
# Arguments:
#  \%sapgchash          Ref to hash of discovered instances  (read)
#  \%clusrsrchash       Ref to cluster resource crossref (read)
#  $sgckeyprefix        Prefix of the key into sapgchasg (read)
#  $crsrckey            Key into clusrsrchash (read)
#
# Return: Full key into sapgchash (or undef)
#
# Checks if key values of clusrsrchash{$crsrckey} (key1) have also a related key
# in sapgchash (key2). Related: key2 is a prefix of $sgckeyprefix.key2.
# Returns related key, otherwise undef.
#
# The code assumes that the key in sapgchash is equal or shorter than
# the key from clusrsrchash with $sgkeyprefix. It also assumes that there is
# just one (or no) key starting with DB- or REPDB-
#
#############################################################################
sub IsTKConfigured {
  my $sapgchashref = shift;
  my $clusrsrchashref = shift;
  my $sgckeyprefix = shift;
  my $crsrckey = shift;

  # Is crsrckey there?
  return undef unless defined ($clusrsrchashref->{$crsrckey});

  foreach my $instname (keys %{$clusrsrchashref->{$crsrckey}}) {
    printL " IsTKConfigured: $instname , [ ",@{$clusrsrchashref->{$crsrckey}->{$instname}}," ]\n";
    my $keywpref=$sgckeyprefix.$instname;
    #print "$keywpref\n";
    foreach my $sgkey (keys %{$sapgchashref}) {  # could also grep instead of loop
      #print "$sgkey\n";
      if ($keywpref =~ /^$sgkey/) {
        sgeasy_print80 ("$sgkey ($instname) is configured to [ @{$clusrsrchashref->{$crsrckey}->{$instname}} ]");
        return $sgkey;
      };
    }
  }
  return undef;
}

#
# Perl package scpkg (SG )
#
package scpkg;
# SG Metrocluster/HANA  sitecontroller package related code

my $hdbsid;
my $hdbinst; # optional (but better set it for sico_hdb_instance)
my $scpkg_default_name;
my $scpkg_found_name;

my $hdbsico_mod = "sgesap/hdbsico";

# initialize, param HDB SID & Instance (WVE SO: is instance optional?)
sub init {
	$hdbsid=shift;
	$hdbinst=shift;
	$scpkg_default_name="hdbc".$hdbsid;

  # reach out to filesystem to get hdbinst if undefined
  unless (defined $hdbinst) {
    my $instpath=(</usr/sap/${hdbsid}/HDB[0-9][0-9]>)[0]; # first (and only) HDB inst
    $hdbinst = (split("/",$instpath))[4];
  }
}

#indicate if sc pkg generation is required
my $scpkggen_required=0;
sub pkggen_required {
	return $scpkggen_required;
}

sub set_pkggen_required {
	$scpkggen_required = 1;
}

# return a name of a sc package. either the existing or default
sub getpkgname {
	return ($scpkg_found_name || $scpkg_default_name);
}

# module_name                     sgesap/hdbsico
# module_name                     sg/sc
# sgesap/hdbsico/sico_hdb_system                  DB1
# optional (however set by this script):
# sgesap/hdbsico/sico_hdb_instance          HDB01
#
# example for modules  and attribute
# package:hdbcDB1|module_name:sgesap/hdbsico|module_name=sgesap/hdbsico
# package:hdbcDB1|module_name:sg/sc|module_name=sg/sc
# package:hdbcDB1|sgesap/hdbsico/sico_hdb_system:DB1|sgesap/hdbsico/sico_hdb_system=DB1
# -- it is considered sufficient only to scan for the last attribute

sub searchpkg {
  my $linesArr = shift;  # Ref to Array with lines of cmviewcl output

  foreach (@{$linesArr}) {
    if ( $_ =~ m#^package:([\S-]+)\|sgesap/hdbsico/sico_hdb_system:(${hdbsid})\|sgesap/hdbsico/sico_hdb_system=(\2)$# ) {
    	$scpkg_found_name = $1;
    	return $scpkg_found_name; # found
    }
  }
  return (undef);	# not found
}

# generate_pkgconfig
#
# Parameter:
#   target-dir
#   sitelist	ref to array containing sites
#   pkglist1..n refs to list of packages containing critical pkg as 1st,
#                managed package as 2nd entry. There have to be as many
#                pkglist as there are sites (usually 2)
#
# Result:
#   A package config (<pkgname>.conf) in <targetdir>
#
# Generates a package config for the hdbsico package.
# Note that there is no merging with an already existing package.
# Either the package is there or it needs to be generated (which in
# the latter case should trigger the call of this subroutine)
# It's also not generated along with the other SGeSAP packages.
#
# Package name is generated using variable scpkg_default_name
#
# Calls
#   cmmakepkg -v 0 -n <pkgname> -m sgesap/hdbsico <pkgconf>
#
# The call should generated a correct package conf for the enqor MNP
#

sub generate_pkgconfig {
	my $targetdir = shift;
	my $siteslist = shift;  # ref to arry
	my $nextnodelist_pos = 0;
	my $pkgfilename = $targetdir."/".$scpkg_default_name.".conf";
	my $hdbsico_mod_link = $ENV{"SGCONF"}."/modules/".$hdbsico_mod;
	my $sg_sites_attribs = "";

	unless (-l $hdbsico_mod_link) {
		main::sgeasy_print80("Error: Link $hdbsico_mod_link doesn't exist");
		return (0);
	}

	# try to generate the site/critial_package/remote_managed_package attribute list
	foreach my $sitename (@{$siteslist}) {
		main::printL "scpkg generate_pkg: $sitename\n";
		$sg_sites_attribs .= "sc_site\t$sitename\n";
		foreach my $pkgname (@{$_[$nextnodelist_pos]}) {
			main::printL "scpkg generate_pkg:   $pkgname\n";
		}
		$sg_sites_attribs .= "critical_package\t".$_[$nextnodelist_pos]->[0]."\n";
		$sg_sites_attribs .= "remote_managed_package\t".$_[$nextnodelist_pos]->[1]."\n";
		$nextnodelist_pos++;
	}
	main::printL "sg_sites_attribs: $sg_sites_attribs";

    # make package config
	my $cmd2exec= qq[$ENV{SGSBIN}/cmmakepkg -v 0 -n $scpkg_default_name -m $hdbsico_mod];
	my $inscsvc = 0; # flag if service sitecontroller_service has been seen
	main::printL "cmd2exec: $cmd2exec\n";
	main::printL "pkg config file: $pkgfilename\n";
	#unlink ($pkgfilename); # delete existing conf (if there)
    open IN, "$cmd2exec |" or die "Could not run $cmd2exec: $!\n";
    open PKGFILE, ">$pkgfilename" or die "Cant open $pkgfilename - $!";
    while (<IN>) {
    	if (m#^(package_description)(.+)"(.+)"#) {
    		print PKGFILE "#deploysappkgs - modified:\n";
    		print PKGFILE $1.$2."\"$3 for HANA\"\n";
    		next;
    	}
    	# get the unset hdb_system attribute
    	if (m#^(sgesap/hdbsico/sico_hdb_system)(\s+)$#) {
    		print PKGFILE "#deploysappkgs - modified:\n";
    		print PKGFILE $1."\t$hdbsid\n";
        # add the optional sico_hdb_instance (not in generated config)
        defined ($hdbinst) && print PKGFILE "sgesap/hdbsico/sico_hdb_instance\t$hdbinst\n";
    		next;
    	}
    	# get the unset sc_site
    	if (m#^(sc_site)(\s+)$#) {
    		print PKGFILE "#deploysappkgs - sites and packages:\n";
    		print PKGFILE $sg_sites_attribs;
    		next;
    	}
		# for new service_halt_on_maintenance 12.00.40 feature
		# sitecontroller_service (assuming name is fixed) value
		# is changed from "no" to "yes". service name is usually
		# predefined in sc module.
		if (m#^(service_name)(\s+)(sitecontroller_service)#) {
		  # inner loop until attribute service_halt_on_maintenance
		  # or next service_name or non-service related attribute
		  print PKGFILE $_;
		  while (<IN>) {
			if (m#^(service_halt_on_maintenance)(\s+)(no)#) {
			  print PKGFILE "#deploysappkgs - modified:\n";
			  print PKGFILE $1.$2."yes\n";
			  # last;
			  # do not exit loop here, but on next service_name or unrelated attr
			}
			elsif (m#^(service_name)(\s+)#) {
			  # next service. exit loop
			  last;
			}
			elsif (m#^service_#) {
			  # stay in inner loop
			  print PKGFILE $_; next;
			}
			else {
			  # unrelated attribute. exit loop
			  last;
			}
		  }
		}

    	print PKGFILE $_;
    }
    close IN;
    close PKGFILE;

	return (1);
}


#
# Perl enqor package
#
package enqor;
# Linux SGeSAP enqor package related code


my $enqor_pkg_default_name = "enqor";
my $enqor_pkg_service_name = "sapenqor.mon";
# Path of monitor not checked. Should use at least monitors/sgesap/enqor.mon?
my $enqor_pkg_service_namefull = "\$SGCONF/monitors/sgesap/$enqor_pkg_service_name";

my $enqor_mod = "sgesap/enqor";


# service definition to be added to newly created enqor pkg
# OBSOLETE!! when using enqor module
my @enqor_service_attributes = (
"service_name               ${enqor_pkg_default_name}_mon",
"service_cmd                $enqor_pkg_service_namefull",
"service_restart            unlimited",
"service_fail_fast_enabled  no",
"service_halt_timeout       60"
);

# dependency attributes to be added to SGeSAP pkg (if not already in)
# hash keys are attribute names, hash values are attribute values
# dependency_condition needs to be prepended by enqor pkg name
# need to have ordered list, otherwise keys will not come in
# correct order
# OBSOLETE!! now reported by sapgetconf
my @dependency_attrnames =
       qw(dependency_name dependency_condition dependency_location);
my %dependency_attributes = (
	"$dependency_attrnames[0]" => "enqor_dep",
	"$dependency_attrnames[1]" => " = UP",
	"$dependency_attrnames[2]" => "same_node"
);

my $enqor_pkg_found_name; # set to pkg name if already configured one, else undef
sub getfoundpkgname {
	return $enqor_pkg_found_name;
}

#indicate if enqor MNP package generation is required
my $enqorMNPpkggen_required=0;
sub pkggen_required {
	return $enqorMNPpkggen_required;
}
sub set_pkggen_required {
	$enqorMNPpkggen_required = 1;
}

# return a name of a enqor package. either the existing or default
sub getpkgname {
	return ($enqor_pkg_found_name || $enqor_pkg_default_name);
}

# Check if an enqor package is already configure
# Must be be a multinode package and have service enqor configured
# Reads through "cmviewcl -f line -v -l package" output. Output
# is contained in an array which reference is the parameter.
# Returns 1 if configured, 0 otherwise. Writes also pkg name
# into variable enqor_pkg_found_name.
#
# Uses: search_enqorpkg from sap_functions.pm
#
sub checkif_enqorpkg_configured {
  my $linesArr = shift;
  $enqor_pkg_found_name = main::search_enqorpkg($linesArr, $enqor_pkg_service_name);
  return (1) if (defined($enqor_pkg_found_name));
  return (0);
}

# generate_pkgconfig
#
# Parameter:
#   target-dir
#
# Result:
#   A package config (<pkgname>.conf) in <targetdir>
#
# Generates a package config for the enqor MNP package
# Note that there is no merging with an already existing package.
# Either the package is there or it needs to be generated (which in
# the latter case should trigger the call of this subroutine)
# It's also not generated along with the other SGeSAP packages.
#
# Package name is generated using variable enqor_pkg_default_name
#
# Calls
#   cmmakepkg -v 0 -n <pkgname> -m sgesap/enqor <pkgconf>
#
# The call should generated a correct package conf for the enqor MNP
#

sub generate_pkgconfig {
	my $targetdir = shift;
	my $pkgfilename = $targetdir."/".$enqor_pkg_default_name.".conf";
	my $enqor_mod_link = $ENV{"SGCONF"}."/modules/".$enqor_mod;

	unless (-l $enqor_mod_link) {
		main::sgeasy_print80("Error: Link $enqor_mod_link doesn't exist");
		return (0);
	}

    # make package config
	my $cmd2exec= qq[$ENV{SGSBIN}/cmmakepkg -v 0 -n $enqor_pkg_default_name -m $enqor_mod $pkgfilename >/dev/null];
	main::printL "cmd2exec: $cmd2exec\n";
	unlink ($pkgfilename); # delete existing conf (if there)
	if (system($cmd2exec) != 0) {
		main::sgeasy_print80("Executing <$cmd2exec> failed");
		return (0);
	}
	# Indicate newly generated config file to SGMgr (if there is one)
    #print "\nFile Name=$pkgfilename\n\n";
	return (1);
}

# generate_pkgconfig
# (old, pre enqor module version)
#
# Parameter:
#   target-dir
#
# Result:
#   A package config (<pkgname>.conf) in <targetdir>
#
# Generates a package config for the enqor MNP package
# Note that there is no merging with an already existing package.
# Either the package is there or it needs to be generated (which in
# the latter case should trigger the call of this subroutine)
# It's also not generated along with the other SGeSAP packages.
#
# Package name is generated using variable enqor_pkg_default_name
#
# Calls
#   cmmakepkg -v 0 -n <pkgname> -m sg/multi_node -m sg/service
# Replaces package_description and appends service attributes
sub generate_pkgconfig_old {
	my $targetdir = shift;
	my $pkgfilename = $targetdir."/".$enqor_pkg_default_name.".conf";
	my $modoptlist = "-m sg/multi_node -m sg/service";

	# make package config
	my $cmd2exec= qq[$ENV{SGSBIN}/cmmakepkg -v 0 -n $enqor_pkg_default_name $modoptlist];
    main::printL "cmd2exec: $cmd2exec\n";
    main::printL "pkg config file: $pkgfilename\n";
    open IN, "$cmd2exec |" or die "Could not run $cmd2exec: $!\n";
    open PKGFILE, ">$pkgfilename" or die "Cant open $pkgfilename - $!";
    while (<IN>) {
    	if (/^package_description/) {
    		print PKGFILE "package_description \"SAP Enqueue Operation Resource MNP\"\n";
    		next;
    	}
    	print PKGFILE $_;
    }
    close IN;
    foreach my $service_attr (@enqor_service_attributes) {
    	print PKGFILE $service_attr."\n";
    }
    close PKGFILE;
    # Indicate newly generated config file to SGMgr (if there is one)
    #print "\nFile Name=$pkgfilename\n\n";
}

#
# Handle enqor dependency
# If there is already a required dependency report in pkg conf reported from cluster (pkgcm*))
# nothing needs to be done. If not, such a dependency needs to be added to sapgetconf output
# (pkgsap*)
#
# Dependency is a same_node, <enqorpkgname> = UP;
# enqor::handle_dependency(\@pkgcmconf, \@pkgsapconf, \%pkgcmhash, \%pkgsaphash, $pkgname)
#
# Return
#   0 - no extra handling required. enqor package does exist and dependency already configured
#   1 - extra handling necessary (pkg exists or needs to be created, dependency not configured yet)
#
# This subroutine will become OBSOLETE if sapgetconf takes over the task of delivering the
# dependency attributes. Then, the these attributes will be handled with the usual merging process.
#
sub handle_dependency {
	my $pkgcmconf = shift;
	my $pkgsapconf = shift;
	my $pkgcmhash = shift;
	my $pkgsaphash = shift;

	my $enqpkgname = getfoundpkgname();
	if ((defined ($enqpkgname)) && (defined ($pkgcmhash->{"dependency_condition"}))) {
		# there is already a enqor package and configured pkg has dependencies
		# check if one of these dependencies contains the required condition
		foreach my $cond (@{$pkgcmhash->{"dependency_condition"}}) {
			main::printL "cond: $cond\n";
			return 0 if ($cond =~ m/$enqpkgname\s*=\s*(UP|up)/);
			# should be "UP", but seem "up" is also allowed
			# no further check if this is same_node
		}
	}

  	# need to add a dependency to (either exising or to be created) enqor package
  	$enqpkgname = getpkgname();
  	#main::printL "enqorpkgname: $enqpkgname\n";
  	# prepend enqor pkgname (if not already there)
  	unless ($dependency_attributes{"dependency_condition"} =~ m/^$enqpkgname/) {
  	  $dependency_attributes{"dependency_condition"} = $enqpkgname.$dependency_attributes{"dependency_condition"};
  	}
  	# Add dependency module. This assumes that no other dependency is reported by sapgetconf
  	push @{$pkgsapconf}, ["module_name",   "sg/dependency"];
  	push @{${$pkgsaphash}{"module_name"}},  "sg/dependency";
  	# set enqor MNP pkg conf generation to required (if none exits)
  	$enqorMNPpkggen_required = 1 unless (defined($enqor_pkg_found_name));

  	foreach my $depattrname (@dependency_attrnames) {
  		main::printL "$depattrname - $dependency_attributes{$depattrname}\n";
  		# push onto name/value pair array
  		push @{$pkgsapconf}, [ $depattrname , $dependency_attributes{$depattrname} ];
  		# add to pkgsaphash
  		push @{${$pkgsaphash}{"$depattrname"}}, $dependency_attributes{$depattrname};
  	}
  	return 1;
}

#############################################################################
#
#  Package for HDB related inifile handling
#  (should go into sap_functions.pm at some point)
#
#############################################################################
package SGeSAP::HDB::inifile;

# each section is a hashkey of object
# each hashkey has reference to another hash with
# valuename as key and valuevalue as hash value

use strict;
use Carp;
use FileHandle;
use Data::Dumper;

sub new {

  my $class = shift;
  my $self = {};
  bless ($self,$class);
  return $self;
}

# fill object from file (or parse...?)
sub fill {
  my $self = shift;
  my $fh   = shift;

  #print "\$fh is of type ".ref($fh)."\n";

  my @lines = <$fh>;
  my $in_site_master = 0;
  my $section = "";
  foreach my $line (@lines) {
    if ($line =~ /^\[([\w_]+)\]$/) {
      $section = $1 ;
      $self->{$section} = {};
      next;
    }
    # else parse the name = value line
    # alternate match to pick up everything after the whitespace(s) after the equal
    # /^\s*(\S+)\s*=\s*(.+)$/
    # or /^\s*(\S+)\s*=\s*(\S.+)$/ (if won't rely on \s* after = being greedy)
#    if ($line =~ /^\s*(\S+)\s*=\s*(\S+)\s*$/) {
    if ($line =~ /^\s*(\S+)\s*=\s*(\S.*)$/) {
      #print "$1 $2\n";
      $self->{$section}->{$1} = $2;
    }
    # system_replication_site_masters" could also be an array
    # as value id is numeric

  }
  return (scalar @lines);
}

sub dump {
  # debug purpose only
  my $self = shift;

  # wont work: Data::Dumper->Dump( [ \@{$self} ] , [ "globini" ] );

  foreach my $sectionkey (keys %{$self}) {
    print "[$sectionkey]\n";
    foreach my $valuekey (keys %{$self->{$sectionkey}}) {
       print "$valuekey => $self->{$sectionkey}->{$valuekey}\n";
    }
  }
  return 0;
}

# return ref to section hash
# parameter: section id
# returns undef if section not in hash
sub getsection {
  my $self = shift;
  my $sectid = shift;

  return $self->{$sectid} if defined $self->{$sectid};
  return undef;
}

# return value
# paramters: section id, value id
# returns undef if value not in section hash
sub getvalue {
  my $self = shift;
  my $sectid = shift;
  my $valueid = shift;

  return $self->{$sectid}->{$valueid} if defined $self->{$sectid}->{$valueid} ;
  return undef;
}

package main;

#############################################################################
# Figure out remote site and hostname via hdbnsutil -sr_state command
# However, the command requires that the secondary connected at least once
# and that HDB is up at the time running this command.
#
#  Parameters:
#  $1 - HDB SID (read)
#
#  Returns:
#  List (localsite, remotesite, remotehost) (members may be undefined)
#
#############################################################################
sub hdbgetothersite {

  printL "Entering hdbgetothersite\n";

  my $HDBSID=shift;
  my $sidadm=lc($HDBSID)."adm";
  my $fh;
  my $mysitename;
  my $otherhost;
  my $othersite;

  my $hdbnsutil_cmd = qq[hdbnsutil -sr_state]; # assumed to be in path
  my $cmd2exec = qq[su - $sidadm -c "$hdbnsutil_cmd"  |];

  printL "\$cmd2exec: $cmd2exec \n";

  open ($fh, $cmd2exec) or warn "popen \$cmd=[$cmd2exec] failed: \$?=[$?] \$!=[$!]";
  while (<$fh>) {
    #print $_;
    $mysitename = $1 if (/site name: (.+)$/);
    if (/^(.+) -> \[(.+)\] (.+)$/) {
      printL "$1 - $2 - $3\n";
      # The following finds the "other" site/host. Either by comparing the hostname or comparing
      # sitenames. It assumes that there is just 1 other site (or none)
      if ("$1" ne "$3") {
  #   if ("$mysitename" ne "$2") {
        $otherhost = $3; $othersite=$2;
        last;
      }
    }
  }
  close($fh) or warn "pclose \$cmd=[$cmd2exec] failed: \$?=[$?] \$!=[$!]\n";
  return ($mysitename,$othersite,$otherhost);
}

#############################################################################
# hdbgetothersite_yao (yet another one)
# ... helper for hdbdeployremote_70_yao
#
# Figure out remote site by investigation global.ini on all clusternodes
#
#  Searches for matching secondary (if primary) or primary (if secondary)
#  Does this by investigating global.ini on all other clusternodes (there
#  may be none on some of the clusternodes).
#  Keep in mind that there may be even a tier-3 async replication on
#  a node in the cluster.
#
#  If the local node is a primary it tries to find the node which has
#  the local site id in the system_replication_site_masters section
#  e.g. 1 = sgxsap52:30001, site id is 1.
#  If the local node is secondary it tries to find the node which has the
#  site id in system_replication_site_masters as local site 1 (basically
#  reverse to above). Keep in mind that the nodename (target) specified
#  in that section doesn't have to be the nodename used in cluster.
#  (hananode may be different than localhost names)
#
#  Parameters:
#  $1 - HDB SID (read)
#  $2 - ref to array of clusternodes
#
#  Returns:
#  List (localsiteid, remotesiteid, remotehost) (members may be undefined)
#  or undef
#
#  Uses $host from sap_functions.pm for localhostname.
#
#############################################################################
sub hdbgetothersite_yao {

  printL "Entering hdbgetothersite_yao\n";

  my $HDBSID=shift;
  my $refclusnodes=shift;

  my $globinipath = "/usr/sap/${HDBSID}/SYS/global/hdb/custom/config/global.ini";

  # First get some information out of local global.ini
  my $fh;
  open ($fh, "<", $globinipath) || die "Can't open  $globinipath";
  my $local_globini = SGeSAP::HDB::inifile->new();
  my $local_numlines = $local_globini->fill($fh);
  close ($fh);
  return undef if ($local_numlines == 0);

  my $local_sitemasterssection = $local_globini->getsection("system_replication_site_masters");
  for my $site_id (keys %{$local_sitemasterssection}) {
    printL $site_id." ".$local_sitemasterssection->{$site_id}."\n";
  }

  my $localsite_id   = $local_globini->getvalue("system_replication","site_id");
  my $localsite_name = $local_globini->getvalue("system_replication","site_name");
  my $localreplmode  = $local_globini->getvalue("system_replication","actual_mode");
  printL "local: $localsite_id/$localsite_name/$localreplmode\n";

  for my $mic (@{$refclusnodes}) {
    next if ($mic eq $host); # no need to exec on localhost
    my $cmd2exec = qq[$ENV{SGSBIN}/cmexec $mic cat $globinipath];
    printL "\$cmd2exec=[$cmd2exec]\n";
    open IN, "$cmd2exec 2>&1|" or next;
    my $remote_globini = SGeSAP::HDB::inifile->new();
    my $remote_numlines = $remote_globini->fill(*IN);
    close IN;
    next if ($remote_numlines == 0);
    printL "Number of lines: $remote_numlines\n";
    my $remotesite_id   = $remote_globini->getvalue("system_replication","site_id");
    my $remotesite_name = $remote_globini->getvalue("system_replication","site_name");
    my $remotereplmode  = $remote_globini->getvalue("system_replication","actual_mode");
    printL "remote: $remotesite_id/$remotesite_name/$remotereplmode\n" if defined $remotesite_id;
    my $remote_sitemasterssection = $remote_globini->getsection("system_replication_site_masters");

    # Try to establish a relationship between these two ini files (based on site_ids)
    # if local is primary, then local site_id should show up in remote system_replication_site_masters
    # if local is *sync*, then "foreign" site_id in local system_replication_site_masters should
    # match remote site_id
    # ... and for primary other has to be *sync*, and for *sync* other has to be primary (to get
    # tier-3 replication sorted out)
    # There's no additional comparison of hostnames as these would need to be based on HANA nodes,
    # not on clusternodes
    if (($localreplmode eq "primary") && (defined $remote_sitemasterssection)) {
      for my $site_id (keys %{$remote_sitemasterssection}) {
        # printL $site_id." ".$remote_sitemasterssection->{$site_id}."\n";
        next unless ($site_id = $localsite_id);
        # found matching site_id, make sure other also has matching mode
        # won't return host from master section, but remote clusternode
        printL "remote/local site_id ($site_id) match found with node $mic\n";
        # TODO: First check remotereplmode before looping trough site_ids?
        if ($remotereplmode =~ /sync/) {
          sgeasy_print80 ("Remote host ($mic) with site_id $site_id and mode $remotereplmode found in cluster");
          return ($localsite_id,$remotesite_id,$mic)
        }
        else { last; } # give up on this loop. don't expect the same site_id appearing twice
      }
    }
    elsif (($localreplmode =~ /sync/) && (defined $local_sitemasterssection)){
      for my $site_id (keys %{$local_sitemasterssection}) {
        # printL $site_id." ".$remote_sitemasterssection->{$site_id}."\n";
        next unless ($site_id = $remotesite_id);
        # matching ids - similar checks as above
        printL "local/remote site_id ($site_id) match found\n";
        if ($remotereplmode eq "primary") {
          sgeasy_print80 ("Remote host ($mic) with site_id $site_id and mode $remotereplmode found in cluster");
          return ($localsite_id,$remotesite_id,$mic)
        }
        else { last; } # give up on this loop
      }
    }
  }
  return undef;
}

#############################################################################
#
#  Figure on which clusternodes deploy should also be run to cover replication
#  modes. Only those clusternodes being also replication site master will be
#  consideres and a site master not being in cluster will be ignored.
#  Reads global.ini (through package).
#
#  Parameters:
#  $1 - HDB SID (read)
#  $1 - command to run (w/o SID) (read)
#  $2 - name of the (local) HDB package to create a ref to the remote HDB packagae
#  $3 - (ref to) list of clusternodes (read)
#  $4 - (ref to) list of masters in cluster (write) (not used)
#
#  Returns:
#  Package name of remote HDB package if exist, otherwise empty string (or undef?).
#  (curently assumes that there is just one, but could be extended to a list)
#
#
#############################################################################
sub hdbdeployremote {

  printL "Entering hdbdeployremote\n";

  my $HDBSID=shift;
  my $cmd=shift;
  my $localpkgname=shift;
  my $refclusnodes=shift;

  printL "$HDBSID - $cmd - $localpkgname - @{$refclusnodes} \n";

  my $globinipath = "/usr/sap/${HDBSID}/SYS/global/hdb/custom/config/global.ini";

  my $fh;
  open ($fh, "<", $globinipath) || die "Can't open  $globinipath";

  my $globini = SGeSAP::HDB::inifile->new();
  my $numlines = $globini->fill($fh);
  close ($fh);
  return 0 if ($numlines == 0);

  my @hdbsitemaster = ();
  my $mysite_name;
  my $mysite_id;
  my $myreplmode;
  my $myreplsection;
  my $mysitemasterssection;
  my $clusnodes = join(" ",@{$refclusnodes});
  my $remotehdbpkg;

  printL "\@clusnodes: @{$refclusnodes} \n";
  printL "\$clusnodes:-$clusnodes-\n";
  #printL "Local node $ENV{HOSTNAME} is in cluster\n" if grep {/^$ENV{HOSTNAME}$/} @{refclusnodes};

  $mysitemasterssection = $globini->getsection("system_replication_site_masters");
  # can there be more than 2 and other indexes than 1 and 2 ??
  for my $site_id (keys %{$mysitemasterssection}) {
    printL $site_id." ".$mysitemasterssection->{$site_id}."\n";
  }

  # splitting masternode entries assumes <host>:<port> format
  my @replmasternodes = map {(split(":", $_))[0]} values %{$mysitemasterssection};
  printL "\@replmasternodes: @replmasternodes \n";

  # were replicationmaster found? might not be the case with new (SPS7)
  # global.ini on primary
  if (scalar(@replmasternodes) == 0) {
    sgeasy_print80("Warning: No replication masters found in global.ini",0);
    # try hdbnsutil -sr_state cmd and get the names of local & remote sitename
    # and remote hostname. The latter is then considered as replmasternode.
    my ($localsite,$othersite,$otherhost) = hdbgetothersite($HDBSID);
    if (defined($othersite) && defined($otherhost)) {
      # othersite and otherhost (usually those of secondary) found
      # Push otherhost on replmasternodes. No need to push local host.
      printL "$localsite - $othersite - $otherhost\n";
      push (@replmasternodes, $otherhost);
    }
    else {
      sgeasy_print80("ERROR: Remote site for local site $localsite not found",0);
    }
    printL "\@replmasternodes: @replmasternodes \n";
  }

  my @mastersincluster = grep { $clusnodes =~ /\b$_\b/} @replmasternodes;
  printL "\@mastersincluster: @mastersincluster \n";

  # prefix cmd with cwd if not absolute
  $cmd = $ENV{PWD}."/".$cmd unless (substr($cmd,0,1) eq "/") ;

  $cmd .= " $HDBSID localhdb";
  # prepend the export to determine package name in dependency
  $cmd = "export SGeSAP_HDBREP_PKGNAME=$localpkgname;".$cmd;
  $cmd = "export START_IP_ADDRESS=$ip_opts;".$cmd if ($ip_opts);

  # All structures setup now.
  # cmexec cmd on each masterincluster - except when master is localnode
  # In latter case just continue executing after this sub.
  for my $mic (@mastersincluster) {
    # no need to execute remote if this is the local node,
    # local execution will generate the file
    next if ($mic eq $ENV{HOSTNAME});
    my $cmd2exec = qq[$ENV{SGSBIN}/cmexec $mic "$cmd"];
    my @generatedFiles = ();
    sgeasy_print80("Executing deploy on $mic",1);
    printL "\$cmd2exec: $cmd2exec \n";
    open IN, "$cmd2exec 2>&1|" or die "Could not run $cmd2exec: $!\n";
    # report output of remote deploy with node prefix
    # check output for keywords "File Name" and "Local HDB Package"
    while (<IN>) {
      print "$mic: $_";
      push @generatedFiles,$1 if (/^File Name=(.+)$/);
      $remotehdbpkg=$1 if (/^Local HDB Package=(.+)$/);
    }
    close IN;
    sgeasy_print80("Executing deploy on $mic",1,$? ? "[FAILURE]":"[SUCCESS]");

    printL "Filenames generated: @generatedFiles \n";
    # So far this is 0 or 1 HDB related conf file expected
    # Copy it over to the local directory
    for my $conffile (@generatedFiles) {
      # usually filename is like /path/hdb*<SID>*
      # (but this relies on the naming convention !!)
      if ($conffile =~ m#(\S+)/${remotehdbpkg}.conf#) {
        sgeasy_print80 ("Copying $conffile from $mic");
        $cmd2exec="cmcp  $mic:$conffile $1";
        printL "\$cmd2exec: $cmd2exec \n";
        printL "mod pkg: $remotehdbpkg\n";
        push (@modifiedPkgs,$remotehdbpkg);
        `$cmd2exec` ; # or warn "$cmd2exec failed"; - warn although return 0 (?)
        # TODO HDB - add to generated filenames
      }
    }
  }
  printL "Leaving hdbdeployremote ($remotehdbpkg)\n" if (defined($remotehdbpkg));
  return $remotehdbpkg;
}

#############################################################################
#
#  Adapted hdbdeployremote for HANA Rev 70+.
#  Uses a "brute force" algorithm by running deploysappkgs on all other node
#  in the cluster until it gets a matching package from there. (which might
#  be a bit lengthy for large clusters)
#  New behaviour is necessary because
#  - global.ini of masternode may not contain a master section anymore
#  - with tier-3 secondary there may be more than 2 sitenames in -sr_state output
#  - nodenames used in HANA may not be the ones used in cluster
#
#  (makes hdbdeployremote and hdbgetothersite obsolete)
#
#  Parameters:
#  $1 - HDB SID (read)
#  $2 - command to run (w/o SID) (read)
#  $3 - name of the (local) HDB package to create a ref to the remote HDB packagae
#  $4 - (ref to) list of clusternodes (read)
#
#  Uses cmexec to run extended command ($2+HDBSID+"localhdb") remotely
#  Remote execution is only triggered by a HANA installation found locally.
#
#  Returns:
#  Package name of remote HDB package if exist, otherwise empty string (or undef?).
#  (curently assumes that there is just one, but could be extended to a list)
#
#############################################################################
sub hdbdeployremote_70 {

  printL "Entering hdbdeployremote_70\n";

  my $HDBSID=shift;
  my $cmd=shift;
  my $localpkgname=shift;
  my $refclusnodes=shift;

  printL "$HDBSID - $cmd - $localpkgname - @{$refclusnodes} \n";
  printL "\@clusnodes: @{$refclusnodes} \n";

  my $remotehdbpkg;

  # prepare cmd
  # prefix cmd with cwd if not absolute
  $cmd = $ENV{PWD}."/".$cmd unless (substr($cmd,0,1) eq "/") ;
  $cmd .= " $HDBSID localhdb";
  # prepend the export to determine package name in dependency
  $cmd = "export SGeSAP_HDBREP_PKGNAME=$localpkgname;".$cmd;
  $cmd = "export START_IP_ADDRESS=$ip_opts;".$cmd if ($ip_opts);

  # cmexec cmd on each clusternode - except when it is localnode
  # In latter case just continue executing after this sub.
  for my $mic (@{$refclusnodes}) {
    # no need to execute remote if this is the local node,
    # initial local execution will generate the file
    next if ($mic eq $ENV{HOSTNAME});
    my $cmd2exec = qq[$ENV{SGSBIN}/cmexec $mic "$cmd"];
    my @generatedFiles = ();
    sgeasy_print80("Executing deploy on $mic",1);
    printL "\$cmd2exec: $cmd2exec \n";
    open IN, "$cmd2exec 2>&1|" or die "Could not run $cmd2exec: $!\n";
    # report output of remote deploy with node prefix
    # check output for keywords "File Name" and "Local HDB Package"
    # (remark: deploy might not even be on that remote node)
    while (<IN>) {
      print "$mic: $_";
      push @generatedFiles,$1 if (/^File Name=(.+)$/);
      $remotehdbpkg=$1 if (/^Local HDB Package=(.+)$/);
    }
    close IN;
    sgeasy_print80("Executing deploy on $mic",1,$? ? "[FAILURE]":"[SUCCESS]");

    printL "Filenames generated: @generatedFiles\n";

    # check if remotehdbpkg is defined and doesn't match localhdbpkg
    # if latter is the case undef it again (undef $remotehdbpkg)
    next unless defined($remotehdbpkg);
    if ($remotehdbpkg eq $localpkgname) {
      # same pkg created on other node agin(why?)
      undef $remotehdbpkg;
      next;
    }

    # 1 HDB related conf file is now expected
    # Copy it over to the local directory and then eit
    # this sub as only one "other" is necessary
    # TODO - grep filename out of list instead of loop
    for my $conffile (@generatedFiles) {
      # usually filename is like /path/hdb*<SID>*
      # (but this relies on the naming convention !!)
      if ($conffile =~ m#(\S+)/${remotehdbpkg}.conf#) {
        sgeasy_print80 ("Copying $conffile from $mic");
        $cmd2exec="cmcp  $mic:$conffile $1";
        printL "\$cmd2exec: $cmd2exec \n";
        printL "mod pkg: $remotehdbpkg\n";
        push (@modifiedPkgs,$remotehdbpkg);
        `$cmd2exec` ; # or warn "$cmd2exec failed"; - warn although return 0 (?)
        printL "Leaving hdbdeployremote ($remotehdbpkg)\n" ;
        return $remotehdbpkg;
      }
    }
  }
  printL "Leaving hdbdeployremote (no remote pkg)\n";
  return $remotehdbpkg;
}

#############################################################################
#
#  Yet another adapted hdbdeployremote for HANA Rev 70+.
#
#  Parameters:
#  $1 - SAP SID (read)
#  $2 - HDB SID (read)
#  $3 - command to run (w/o SID) (read)
#  $4 - name of the (local) HDB package to create a ref to the remote HDB package
#  $5 - (ref to) list of clusternodes (read)
#
#  Uses cmexec to run extended command ($2+HDBSID+"localhdb") remotely
#  Remote execution is only triggered by a HANA installation and primary or *sync*
#  setup found locally.
#
#  Uses $host from sap_functions.pm for localhostname.
#
#  Writes global array HDBNODELIST.
#
#  Returns:
#  Package name of remote HDB package if exist, otherwise empty string (or undef?).
#  (curently assumes that there is just one, but could be extended to a list)
#
#############################################################################
sub hdbdeployremote_70_yao {

  printL "Entering hdbdeployremote_70_yao\n";

  my $SAPSID=shift;
  my $HDBSID=shift;
  my $cmd=shift;
  my $localpkgname=shift;
  my $refclusnodes=shift;

  printL "$SAPSID - $HDBSID - $cmd - $localpkgname - @{$refclusnodes} \n";
  printL "\@clusnodes: @{$refclusnodes} \n";

  my @remotehdbpkgs = ();

  # Prepare cmd. prefix with cwd if not absolute, add args and variables
  # Be aware that local environment is used with cmd executing remotely
  $cmd = $ENV{PWD}."/".$cmd unless (substr($cmd,0,1) eq "/") ;
  $cmd .= " $SAPSID localhdb";
  # prepend the export to determine package name in dependency
  $cmd = "export SGeSAP_HDBREP_PKGNAME=$localpkgname;".$cmd;
  $cmd = "export START_IP_ADDRESS=$ip_opts;".$cmd if ($ip_opts);

  # prepend local hostname as this is one part of the pair
  $cmd = "export SGeSAP_HDBREP_HOST=$host;".$cmd;

  my ($localsiteid,$othersiteid,$otherhost) = hdbgetothersite_yao($HDBSID,$refclusnodes);
  printL "$localsiteid/$othersiteid/$otherhost\n" if (defined $localsiteid);
  # WVE: This clever sub above even works in HANA SO on metrocluster with SG sites being defined.
  #      However, it could be optimized by not contacting the nodes in the local SG site at all,
  #      but just use the subset of clusternodes belonging to the other SG site.

  unless (defined ($otherhost)) {
    sgeasy_print80 ("Couldn't find a remote host to match local HANA replication setup");
    sgeasy_print80 ("Remote package creation not possible");
    return undef;
  }

  # run remote deploy on host figured out (no need for looping through clusternodes anymore)
  my $cmd2exec = qq[$ENV{SGSBIN}/cmexec $otherhost "$cmd"];
  my @generatedFiles = ();
  sgeasy_print80("Executing deploy on $otherhost",1);
  printL "\$cmd2exec: $cmd2exec \n";
  open IN, "$cmd2exec 2>&1|" or die "Could not run $cmd2exec: $!\n";
  # report output of remote deploy with node prefix
  # check output for keywords "File Name" and "Local HDB Package"
  # (remark: deploy might not even be on that remote node)
  while (<IN>) {
    print "$otherhost> $_";
    push @generatedFiles,$1 if (/^File Name=(.+)$/);
    push @remotehdbpkgs,$1 if (/^Local HDB Package=(.+)$/);
  }
  close IN;
  sgeasy_print80("Executing deploy on $otherhost",1,$? ? "[FAILURE]":"[SUCCESS]");

  printL "Filenames generated: @generatedFiles\n";
  printL "remotehdbpkgs: @remotehdbpkgs\n";

  # check if remotehdbpkg is defined and doesn't match localhdbpkg
  # if latter is the case undef it again (undef $remotehdbpkg)
  # WVE: remotehdbpkgs is a list, loop thru all?
  if ($remotehdbpkgs[0] eq $localpkgname) {
    # same pkg created on other node agin(why?)
    @remotehdbpkgs = ();
  }

  # 1 (or more) HDB related conf file is now expected
  # Copy it over to the local directory and then eit
  # this sub as only one "other" is necessary
  # WVE - grep filename out of list instead of loop
  for my $conffile (@generatedFiles) {
    # usually filename is like /path/hdb*<SID>*
    # (but this relies on the naming convention !!
    for my $remotehdbpkg (@remotehdbpkgs) {
    	if ($conffile =~ m#(\S+)/${remotehdbpkg}.conf#) {
    		sgeasy_print80 ("Copying $conffile from $otherhost");
            $cmd2exec="cmcp  $otherhost:$conffile $1";
            printL "\$cmd2exec: $cmd2exec \n";
            printL "mod pkg: $remotehdbpkg\n";
            push (@modifiedPkgs,$remotehdbpkg);
           `$cmd2exec` ; # or warn "$cmd2exec failed"; - warn although return 0 (?)
            #printL "Leaving hdbdeployremote ($remotehdbpkg,$otherhost)\n" ;
            # return ($remotehdbpkg,$otherhost); # HDB SO: more than one
        }
    }
  }
  printL "Leaving hdbdeployremote ($otherhost, @remotehdbpkgs)\n" ;
  return ($otherhost, @remotehdbpkgs); # or undef?
}

# create hash
# {pkg}->{tag}->{action} => [[attrname1,attrval1],...]
# from xtraction (extra actions)
sub xtractions_perpkg {
	my %pkgaction=();

	while ($_=shift @_) {
		my ($tag,$pkg,$action,$attrname,$attrval)=split((/\s+/,$_));
		#$pkgaction{$pkg}->{$tag}->{$action} = [$attrname,$attrval];
		push (@{$pkgaction{$pkg}->{$tag}->{$action}},[$attrname,$attrval])
	}
	#print STDERR Data::Dumper->Dump( [ \%pkgaction ], ["pkgaction"] );
	return \%pkgaction;
}

sub execute_del_pkgxtra {
	my $pkgactcmconf_ref = shift;
	my $linesdelactions = shift;
	my $starting = -1;
	my $ending = -1;
	my $firstfull = -1;
	my @delidxes = ();
	my $curridxconf = 0;
	my $curridxact = 0;

	#print Data::Dumper->Dump( [ \$pkgactcmconf_ref ], ["pkgactcmconf"] );

	foreach my $elem (@{$pkgactcmconf_ref}) {
		if (defined $linesdelactions->[$curridxact]->[1]) {
			# try exact match
			if ($elem->[0] eq $linesdelactions->[$curridxact]->[0] && $elem->[1] eq $linesdelactions->[$curridxact]->[1]) {
				printL "xtra exact match with: $elem->[0] $elem->[1] ($curridxconf/$curridxact)\n";
				$starting = $curridxconf unless ($starting >= 0);
				$firstfull = $curridxconf unless ($firstfull >= 0);
				$ending   = $curridxconf;
				$curridxact++; # match. that one is "used up"
			}
			# no match. if there was already a exact match with a previous line, this isn't a consecutive group
			# probably need to return here !!?? (WVE)
			elsif ($starting >= 0) {
				sgeasy_print80 ("No consecutive match with");
				sgeasy_print80 ("[$elem->[0] $elem->[1]] and [$linesdelactions->[$curridxact]->[0] $linesdelactions->[$curridxact]->[1]");
				# return 0 indicating no entire deletion of that group being possible
				return 0;
			}
			$curridxconf++; # next line from pkg conf
		}
		else {
			# try regexp match on attribute name
			if ($elem->[0] =~ m#$linesdelactions->[$curridxact]->[0]#) {
				printL "xtra regex match with: $elem->[0] $linesdelactions->[$curridxact]->[0] ($curridxconf/$curridxact)\n";
				# check if it has the same attribute name as the first exact match (if exist)
				if ($firstfull >= 0 && $elem->[0] eq $pkgactcmconf_ref->[$firstfull]->[0]){
				  printL "xtra regex matches first full: $elem->[0] $pkgactcmconf_ref->[$firstfull]->[0] ($curridxconf/$curridxact)\n";
				  # next elements from both array
				  $curridxact++;
				  $curridxconf++
				}
				else {
					$starting = $curridxconf unless ($starting >= 0);
					$ending   = $curridxconf;
					$curridxconf++
				}
			}
			elsif ($starting >= 0) {
				# next del line, with same line from pkg conf
				$curridxact++;
			}
			else {
				$curridxconf++;
			}
		}
		# check if there are still actions or if $curridxact is already beyond end of array
		# if beyond break the loop
		last if ($curridxact >= scalar(@{$linesdelactions}));
	}
	# WVE: delete or comment the lines to be removed from pkg config?
	sgeasy_print80 ("Linenumbers determined to be removed from package: $starting - $ending");
	#delete $pkgactcmconf_ref->[$starting .. $ending]; -> doesn't work
	if ($starting >= 0) {
		for my $idx ($starting .. $ending) {
		  delete $pkgactcmconf_ref->[$idx];
		}
		return 1;
	}
	return 0;
}

# execute hash created by previous sub
# (tag should be renamed to grpid)
#
# foreach pkg in hash
#    get pkg config and write into "lines"
#    foreach tag execute action(s)
#       (currently only def action is supported)
#       search from beginning until first match
#       from there try to match the next attributes
#       stop matching if there is a non-matching line or the last
#          attrname with an defined attrvalue is found (again)
sub execute_xtractions {
	my $pkgactions_ref=shift;
	my @modpkgs = ();

	foreach my $pkgname (keys %{$pkgactions_ref}) {
		my @pkgactcmconf = ();
		my %pkgactcmhash = ();
		my $confmodified = 0;
		#my $cmd_cmgetconf=qq[$ENV{"SGSBIN"}/cmgetconf -v 0 -p];  # packagename to be appended
		open IN, "$cmd_cmgetconf $pkgname 2>/dev/null |"; # or die "Could not run $cmd_cmgetconf: $!\n";
        populateArray(*IN, \@pkgactcmconf, \%pkgactcmhash);
        close IN;
        #print Data::Dumper->Dump( [ \@pkgactcmconf ], ["pkgactcmconf"] );
        # WVE: sort tags (numeric) below?
        foreach my $tag (keys %{$pkgactions_ref->{$pkgname}}) {
        	foreach my $action (keys %{$pkgactions_ref->{$pkgname}->{$tag}}) {
        		unless ($action eq "del") {
        			sgeasy_print80 ("Action $action ($pkgname/$tag)is unsupported - skipping it");
        			next;
        			# now these (del) actions per pkg & tag can be executed -> sep sub
        			# two refs: pkg cm lines array, array lines to be deleted
        		}
        		# @pkgactcmconf gets modified by next call (if true, next checked so far)
        		$confmodified ||= execute_del_pkgxtra (\@pkgactcmconf,$pkgactions_ref->{$pkgname}->{$tag}->{$action});
        		#print Data::Dumper->Dump( [ \@pkgactcmconf ], ["pkgactcmconf"] );
        	}
        }
        # generate each pkg
        if ($confmodified) {
	        my $pkgfilename=$pkgpath."/".$pkgname.".conf"; # pkgpath already contains sapsid
		    sgeasy_print80 ("Generating new configuration for $pkgname",1);
		    open PKGFILE, ">$pkgfilename" or die "Cant open $pkgfilename - $!";
		    printheaderinfo (*PKGFILE);
		    # seems when deleting in the middle of an array (as done in execute_del_pkgxtra) it leaves undefs
		    for my $elem (@pkgactcmconf) {
		    	printAttr (*PKGFILE, $elem->[0],$elem->[1],"cmgetconf-xtra") if (defined $elem);
		    }
		    close PKGFILE;
		    push (@modpkgs, $pkgname);
		    sgeasy_print80 ("Generating new configuration for $pkgname",1,"[SUCCESS]");
        }
	}
	return @modpkgs;
}

#############################################################################
#
# sub checkIPspec
#
#  Parameters:
# $ip_opts contains start of ip address for the primary, optional second argument
#          contains ip address start for secondary
#
# deploysappkgs called with -i=10.3.1.61 or -i=10.3.1.61,10.3.1.64 or -i=hostA
#
# Simple check if this is a valid ip address specification.
# split ip_opts arguments into an array and call isIPaddress() for a very basic
# validation check.
#############################################################################
sub checkIPspec {
  my $ip_opts=shift;


  # relies on nwrksupp utility - no fallback here
  if (! -x $nwrksupp) {
    die "$nwrksupp utility not found - can't resolve ip address/hostname";
  }

  printL "ip address(es)/host(s) specified ip_opts=[$ip_opts]\n";
  my @ip = split(/,/, $ip_opts);
  foreach my $ip (@ip) {
    `$nwrksupp ghbn $ip 2>/dev/null`;
    if ($? != 0) {
      LogVerbose("can't assign ip=[$ip]");
      die "$USAGE";
    }
  }
}


############################################################################
#
# sub isIPaddress
#
#  Parameters:
#  $1  one IP address
#
# Simple check if this is a valid ip address
#############################################################################
sub isIPaddress {
  my $str=shift;
  if($str =~ m/(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})/ && ($1<=255 && $2<=255 && $3<=255 && $4<=255 )) {
    printL "valid IP [$str]\n";
    return 1;
  } else {
    printL "invalid IP [$str]\n";
    return 0;
  }
}


#############################################################################
#
#  main - starts here...
#
#############################################################################


#$sgeasyTool->screen_and_log("Starting $theapp");

# Do not print out initial starting when reporting directly to a terminal
-t STDOUT || sgeasy_print80("Starting with \"@ARGV\"",1);

# get options
# -a|--apply  ==> apply package configs (cmapplyconf w/o confirmation)
# -o|--option ==> option(s)
# -i|--ip ==> ip_address(es)
# -c|--check  ==> check package configs (cmcheckconf) - not implemented yet
GetOptions (
  "apply"     => \$applypkgs,
  "options=s" => \$special_opts,
  "ip=s"      => \$ip_opts,
  "check"     => \$checkpkgs)  or die "$USAGE";
printL "applypkgs=$applypkgs,options=$special_opts,ip=$ip_opts\n";

unless (parse_args()) {
  -t STDOUT || sgeasy_print80("Starting",1,"[FAILURE]");
  exit 1;
}

# handle options
# smart_quorum -> enforce smartqorum for at least 1 pkg
if ($special_opts =~ /smart_quorum/) {
  printL "smart_quorum enforced (by setting SAPGETCONF_USE_SMARTQ=2)\n";
  $ENV{"SAPGETCONF_USE_SMARTQ"}=2;
}
# safesync special opts (configure safesync resource regardless of global.ini settings)
if ($special_opts =~ /safesync/) {
  printL "safesync enforced (by setting SAPGETCONF_USE_SAFESYNC=1)\n";
  $ENV{"SAPGETCONF_USE_SAFESYNC"}=1;
}
# ip address range was passed in
if ($ip_opts) {
  printL "ip address(es) specified ip_opts=[$ip_opts]\n";
}


-t STDOUT || sgeasy_print80("Starting",1,"[SUCCESS]");

if ($testmode) {
  sgeasy_print80("Testmode turned on");
  $cmd_cmviewclpkg=qq[cat /tmp/cmviewcl.dspt] if $testmode;
}

my $cmd2exec = $cmd_sapgcinst;
$cmd2exec =~ s/\$\w+/$sapsid/g;  # replace $THESID
$cmd2exec = "cat /tmp/sapgetconf.fline.dspt" if ($testmode);

my $deployfailure = 0;

# check if user calling deploy belong to root group, otherwise
# deploy may fail
sgeasy_print80("Check if user belongs to root group",1,);
unless (chkrootprivs()) {
  sgeasy_print80("Current user doesn't belong to root group");
  sgeasy_print80("Run deploysappkgs as root user");
  sgeasy_print80("Check if user belongs to root group",1,"[FAILURE]");
  exit 1;
}
sgeasy_print80("Check if user belongs to root group",1,"[SUCCESS]");

# get list of instances and their associated values as seen by sapgetconf
if (defined $sapsid) {
  sgeasy_print80("Getting configured SAP instances and DB for ".$sapsid,1,);
  printL "cmd2exec: $cmd2exec |\n";
  # Force sapgetconf to report HANA as a "ordinary" DB
  #(by setting the env var - no external overwrite so far)
  $ENV{"SAPGETCONF_HDBASDBINSTANCE"}=1;
  open IN, "$cmd2exec |" or die "Could not run $cmd2exec: $!\n";
  fillsapgcHash(*IN, \%sapgcinsthash);
  close IN;
  # get the DBSID separately from list of instances. maybe different from SAPSID
  $dbsid = getgcdbSID(\%sapgcinsthash) || $sapsid;
}
else {
  printL "No SAP SID defined\n";
  exit 1;
}

if (keys %sapgcinsthash == 0) {
  sgeasy_print80("No SAP instances retrieved",1,"[FAILURE]");
  exit 1;
}
sgeasy_print80("Reported instances and entities:");
sgeasy_print80(join(',',keys %sapgcinsthash));
sgeasy_print80("DB SID: $dbsid") if ($sapsid ne $dbsid);
if (!$multipkg && grep (/^HDB(p|s)-/, keys %sapgcinsthash)) {
  sgeasy_print80("Info: HANA DB with combined packages currently not supported");
  sgeasy_print80("Info: Using multiple packages option instead");
  $multipkg=1
}
sgeasy_print80("Getting configured SAP instances",1,"[SUCCESS]");


# getting and checking NFS information
# TODO ENHANCEMENT
# TODO adding a filter expression at the end (this should also
# TODO include dbtype dependent exports, however, dbtype isn't stored
# TODO anywhere yet)
#$cmd2exec = $cmd_sapgcnfs." \"(/usr/sap/trans|/sapmnt/$sapsid)\"";
# make it less restricted:
$cmd2exec = $cmd_sapgcnfs." \"(trans|$sapsid)\"";
#$cmd2exec =~ s/\$\w+/$sapsid/g;  # replace $THESID
$cmd2exec = "cat /tmp/sapgetconf-nfs.fline.dspt" if ($testmode);
sgeasy_print80("Getting configured NFS exports for ".$sapsid,1,);
printL "cmd2exec: $cmd2exec |\n";
open IN, "$cmd2exec |" or die "Could not run $cmd2exec: $!\n";
fillsapgcNFSHash(*IN, \%sapgcnfshash);
close IN;
{
    my @nfsexp = keys %sapgcnfshash;
    if (scalar @nfsexp) {
      sgeasy_print80 ("Related NFS exports found:");
      sgeasy_print80 ("[ @nfsexp ]");
    }
    else {
      sgeasy_print80 ("No related NFS exports found");
    }
}
sgeasy_print80 ("Getting configured NFS exports",1,"[SUCCESS]");

# setup package path
$pkgpath = $pkgpath.$sapsid;
`mkdir $pkgpath` unless (-d $pkgpath);

#print STDERR Data::Dumper->Dump( [ \%sapgcnfshash ], ["sapgcnfshash"] );
#print STDERR Data::Dumper->Dump( [ \%sapgcinsthash ], ["sapgcinsthash"] );
#print STDERR Data::Dumper->Dump( [ \@sapinstshrdres ], ["sapinstshrdres"] );

unless (checksapgcHash (\%sapgcinsthash)) {
  sgeasy_print80("SAP configuration not supported by $theapp",1,"[FAILURE]");
  exit 1;
}

# get cluster nodes (could be setup while reading "cmviewcl" output)
# WVE: mode after gettig cluster resource hash
sgeasy_print80("Getting clusternodes",1);
@clusnodes = getclusternodes();
if (scalar @clusnodes == 0)
{
  sgeasy_print80("Getting clusternodes",1,"[FAILURE]");
  exit 1;
}
sgeasy_print80("@clusnodes");
sgeasy_print80("Getting clusternodes",1,"[SUCCESS]");

# remove all instances whose virtual hostname is that of a member of the cluster
sgeasy_print80("Checking for instances configured to physical cluster host",1);
removePhysInst (\%sapgcinsthash,\@clusnodes);
sgeasy_print80("Checking for instances configured to physical cluster host",1,"[SUCCESS]");


sgeasy_print80("Getting package information from cluster",1,);
# read in cluster package config
printL "cmd2exec: $cmd_cmviewclpkg |\n";
open IN, "$cmd_cmviewclpkg |" or die "Could not run cmviewcl: $!\n";
chomp(@linescmviewcl = (<IN>));
close IN;
sgeasy_print80("Getting package information from cluster",1,"[SUCCESS]");

# build a hash on package (IPs, SIDs,... used per package)
populatepkgHash (\@linescmviewcl, \%cluspkgconfighash);
#print STDERR Data::Dumper->Dump( [ \%cluspkgconfighash ], ["cluspkgconfighash"] );

# build a "resource used" (IP,FS,VG) by package cross reference
collectUsedClusResources (\@linescmviewcl, \%clusrsrchash);
#print STDERR Data::Dumper->Dump( [ \%clusrsrchash ], ["clusrsrchash"] );


# Check for HANA scaleout config. If this doesn't run on a cluster with
# SITEs (Metrocluster) configured, the packages can't be setup.
# The function removes the HDB SO instances from hash if it isn't a MC
sgeasy_print80("Checking for HANA scalout and Metrocluster" ,1);
if (IsMCforHDBSO (\%sapgcinsthash, \%clusrsrchash)) {
  sgeasy_print80("Checking for HANA scalout and Metrocluster" ,1,"[SUCCESS]");
}
else {
  sgeasy_print80("Checking for HANA scalout and Metrocluster" ,1,"[FAILURE]");
}

# QXCR1001433284 -- ER: Support for HANA MCOS configurations
# Other HANA configuration restrictions like, it isn't allowed to configure:
# - an additional scale-up instance on a already configured scale-out
# - an additional scale-out instance on any already configured system (up or out)
#
my $print80txt="Checking HANA scaleup & scalout MCOS configuration";
sgeasy_print80("$print80txt" ,1);
my ($subrc, $rctext) = ensureHDBConfigRestrictions ($sapsid,\%sapgcinsthash, \%clusrsrchash,\%cluspkgconfighash);
if ($subrc) {
  sgeasy_print80("$print80txt",1,"[SUCCESS]");
}
else {
  sgeasy_print80("$rctext");
  sgeasy_print80("$print80txt",1,"[FAILURE]");
  exit 1;
}

# Need to cleanout Sybase ASE and/or repserver from sapgcinsthash
# if already taken care of by Sybase toolkit. Simlar to removePhysInst,
# looks into %clusrsrchash (TKIT_ASE, TKIT_RS) this time.
# Keys to consider in %sapgcinsthash are DB-<DBSID> and REPDB-<DBSID>
# clusrsrchash{"TKIT_ASE"}->{<DBSID>} is a (ref to)list of packages
# sapgcinstash{"DB-".<DBSID}}is a (ref to) an array (contents undef)
#
sgeasy_print80("Checking if DB and/or replication is configured in a toolkit" ,1);

my $sgckey;
$sgckey = IsTKConfigured (\%sapgcinsthash, \%clusrsrchash, "DB-", "TKIT_ASE");
if (defined $sgckey) {
  sgeasy_print80("Removing $sgckey from package lookup/assignment");
  delete $sapgcinsthash{$sgckey};
}

$sgckey = IsTKConfigured (\%sapgcinsthash, \%clusrsrchash, "REP-", "TKIT_RS");
if (defined $sgckey) {
  sgeasy_print80("Removing $sgckey from package lookup/assignment");
  delete $sapgcinsthash{$sgckey};
}
#print "sgckey: ", defined($sgckey)?"$sgckey":"undef"," \n";

sgeasy_print80("Checking if DB and/or replication is configured in a toolkit" ,1,"[SUCCESS]");
# end of "toolkit cleanout"

# check if enqor package is already defined (TODO: LX only)
# (check may be a bit early at this as it is unknown yet if a enqor MNP is needed)
sgeasy_print80("Checking if enqor multinode package is already configured" ,1,);
if (enqor::checkif_enqorpkg_configured(\@linescmviewcl)) {
	sgeasy_print80("enqor MNP already configured: ".enqor::getfoundpkgname());
}
else {
	sgeasy_print80("No enqor MNP configured");
}
sgeasy_print80("Checking if enqor multinode package is already configured" ,1,"[SUCCESS]");

# check with NFS XFS are already exported by a package
# and remove these
sgeasy_print80("Checking if found NFS exports are already used in a package" ,1,);
checkNFSexports(\%sapgcnfshash,  \%clusrsrchash);
{
    my @nfsexp = keys %sapgcnfshash;
    if (scalar @nfsexp) {
      sgeasy_print80 ("NFS exports that need to be assigned to a package:");
      sgeasy_print80 ("[ @nfsexp ]");
    }
    else {
      sgeasy_print80 ("No NFS exports remaining for assignment");
    }
}
sgeasy_print80("Checking NFS exports" ,1,"[SUCCESS]");
# At this point sapgcnfshash only contains the exports which aren't assigned
# to a package already. The rest of the exports need to be based on IP information
# and multi/combi.
# Merge NFS hash into instance hash
mergeNFS2Instance (\%sapgcnfshash, \%sapgcinsthash);
#print STDERR Data::Dumper->Dump( [ \%sapgcinsthash ], ["sapgcinsthash"] );

# code moved down to also include NFS related entities
sgeasy_print80("Checking resource sharing",1);
unless (buildshrdres(\%sapgcinsthash, \@sapinstshrdres))
{
  sgeasy_print80("Checking resource sharing",1,"[FAILURE]");
  exit 1;
}
sgeasy_print80("Checking resource sharing",1,"[SUCCESS]");

#print STDERR Data::Dumper->Dump( [ \%cluspkgconfighash ], ["cluspkgconfighash"] );

# Everything read into local datastructures
# Start package selection.
# mapPkgIP2Pkg and lookup* subs just prepare other datastructures
# and the final selection is done by selectPkgCandidate (to that extent the call
# of the others could be moved into that one

sgeasy_print80("Determining package to instance mapping",1);

mapPkgIP2Pkg (\%cluspkgconfighash, \%clusip2pkgmap);
lookupPkgCandidatesIP (\%sapgcinsthash, \%clusip2pkgmap, \%instPkgCandidates, \%clusrsrchash);
lookupEmptySIDPkgCandidatesIP (\%cluspkgconfighash,\%instPkgCandidates);
lookupPkgCandidatesSID ($sapsid,\%cluspkgconfighash,\%instPkgCandidates, $dbsid);
lookupPkgCandidatesVHost ($sapsid, \%sapgcinsthash, \%clusrsrchash, \%cluspkgconfighash, \%instPkgCandidates);
lookupPkgCandidatesHDBrepl (\%sapgcinsthash, \%clusip2pkgmap, \%instPkgCandidates, \%clusrsrchash);
lookupPkgCandidatesPkgType ($sapsid, \%sapgcinsthash, \%clusrsrchash, \%cluspkgconfighash, \%instPkgCandidates);
lookupPkgCandidatesNodelist ($sapsid, \%sapgcinsthash, \%clusrsrchash, \%cluspkgconfighash, \%instPkgCandidates);
lookupPreferredInst (\%instPkgCandidates);

#print STDERR Data::Dumper->Dump( [ \%sapgcinsthash ], ["sapgcinsthash"] );
#print STDERR Data::Dumper->Dump( [ \%clusip2pkgmap ], ["clusip2pkgmap"] );
#print STDERR Data::Dumper->Dump( [ \%clusrsrchash ], ["clusrsrchash"] );
#print STDERR Data::Dumper->Dump( [ \%instPkgCandidates ], ["instPkgCandidates"] );

unless (selectPkgCandiate (\%instPkgCandidates, \@sapinstshrdres, 0)) {
  sgeasy_print80("Package to instance mapping",1,"[FAILURE]");
  exit 1;
};

#print STDERR Data::Dumper->Dump( [ \%cluspkgconfighash ], ["cluspkgconfighash"] );
#print STDERR Data::Dumper->Dump( [ \%clusrsrchash ], ["clusrsrchash"] );
#print STDERR Data::Dumper->Dump( [ \%instPkgCandidates ], ["instPkgCandidates"] );

# At this stage we (may) have the (existing) package candidates for all or some of instances assigned.
# (these assignments where based on certain package resource as well as certain resources shared by instances)
# If there is a candidate, it is just one package.
# Now we have to assign the unassigned instance a (maybe new) package name. This has to be based on
# the "combination strategy" (combi/multi) and the kind of instance. Based on this strategy it
# might be that a currently unassigned instance maybe assigned additionally to an existing package
# (containing an already package instance). E.g.: a newly created ASCS maybe assigned to an already
# packaged SCS if "combi" packaging is requested.

#print STDERR Data::Dumper->Dump( [ \%CandPkg2Inst ], ["CandPkg2Inst"] );
assignAllUnassignedInst (\%instPkgCandidates, \%CandPkg2Inst, $sapsid, $multipkg,\%clusrsrchash, \%sapgcinsthash);

#print STDERR Data::Dumper->Dump( [ \%instPkgCandidates ], ["instPkgCandidates"] );
#print STDERR Data::Dumper->Dump( [ \%CandPkg2Inst ], ["CandPkg2Inst"] );

my $localhdbpkg;  # what's been selected/determined for the local HDB instance as package (or a comma sep. list for SO)
my @localhdbpkgs=(); # list as it can be more than one for the SO case
my $remotehdbpkg; # what's been reported as selected/determined for the remote HDB instance as package by remote call
my @remotehdbpkgs=(); #as above
my $remotehdbhost; # what's been reported as the host where remotehdbpkg was generated (supposed to be generated)
# WVE: actually, as there are no dependencies in SO case, these package names are not relevant for SO package generation.

foreach my $candpkg (keys %CandPkg2Inst) {
  sgeasy_print80 ("$candpkg <-> @{$CandPkg2Inst{$candpkg}}");
  # special for HDB:
  push @localhdbpkgs,$candpkg if (grep {/^HDB(p|s|)-/} @{$CandPkg2Inst{$candpkg}});
  # add sanity check here. $candpkg isn't allowed to be undefined or empty
  # otherwise some of the logic went wrong
  unless ((defined $candpkg) && ($candpkg ne "")) {
    sgeasy_print80 ("Instance(s) @{$CandPkg2Inst{$candpkg}} didn't get a packagename assigned");
    sgeasy_print80 ("Package to instance mapping",1,"[FAILURE]");
    exit 1;
  }
}

# reorg a 2 element localhdbpkgs so that p pkg has index 0
if (scalar @localhdbpkgs == 2) {
	foreach my $candpkg (keys %CandPkg2Inst) {
		# WVE SO: assume only HDBp|s goes into package (otherwise grep as above)
		$localhdbpkgs[0] = $candpkg if ($CandPkg2Inst{$candpkg}->[0] =~ /^HDBp-/);
		$localhdbpkgs[1] = $candpkg if ($CandPkg2Inst{$candpkg}->[0] =~ /^HDBs-/);
	}
}
$localhdbpkg = join (",",@localhdbpkgs) if defined $localhdbpkgs[0];
printL "localhdbpkg: $localhdbpkg\n" if defined $localhdbpkg;

# Explicitly print out select HDB package.
# If executed on the remote site this string is used to identify the remotely
# generated HDB package.
# Message below is consumed by deploysappkgs calling this from remote clusternode
#sgeasy_print80("Local HDB Package=$localhdbpkg") if (defined($localhdbpkg)); # WVE: with HDB SO this can be a list
map sgeasy_print80("Local HDB Package=$_"), @localhdbpkgs;
sgeasy_print80("Package to instance mapping",1,"[SUCCESS]");

# Package names to be used (existing and new) are known and
# the instances are assigned to those packages.
# For each package start merging them into a config

# If there is a package selected/defined for the local HDB instance divert
# into calling deploysappkgs also on other nodes.
# This will already create package(s), which will be copied to the local node.
# TODO: Might go even into createpkgconfig to the part where the local HDB config is
#       going to be created - and called before that.
# Usually, a $localhdbpkg is assigned if there is locally discovered HDB instance.
if (!$localhdb && defined ($localhdbpkg)) {
  printL "!localhdb and HDB ($localhdbpkg) found\n";
  sgeasy_print80("Generating corresponding remote HANA replication package for $localhdbpkg",1);
  # Now figure out which HDB replication site master nodes are in cluster
  # Unless it is this node where the script is currently running, call
  # cmexec <node> <thisscript> <currentmulti/combi> <currentSID> localhdb
  # <thisscript> = $0 = $PROGRAM_NAME
  # printL "$0 - $ARGV[0] - $ARGV[1] - $ARGV[2] \n";
  ($remotehdbhost,@remotehdbpkgs) = hdbdeployremote_70_yao ($ARGV[1],$dbsid, $0." ".$ARGV[0],$localhdbpkg,\@clusnodes);
  # put pkg-name into environment to be consumed by createpkgconfig
  # WVE: overwrite if already set? External setting of SGeSAP_HDBREP_PKGNAME
  #       to contain the remote package (and now also SGeSAP_HDBREP_HOST as the remote host)
  if (defined($remotehdbpkgs[0])) {
    # if there's a remote pkg there's also a remote host
    $remotehdbpkg = join (",",@remotehdbpkgs);

    $ENV{"SGeSAP_HDBREP_PKGNAME"} = $remotehdbpkg;
    $ENV{"START_IP_ADDRESS"} = $ip_opts if ($ip_opts);
    $ENV{"SGeSAP_HDBREP_HOST"} = $remotehdbhost;
    sgeasy_print80("Generating corresponding remote HANA replication package",1,"[SUCCESS]");
  }
  else {
    # No remote package for local hdb package configured
    # Removing the local hdb package (name stored in localhdbpkg) from the
    # list of "package candidates" to be configured.
    sgeasy_print80("No corresponding remote HDB Package found for $localhdbpkg");
    sgeasy_print80("Removing $localhdbpkg from the list of packages to be configured");
    delete $CandPkg2Inst{$localhdbpkg}; # WVE: list!!
    sgeasy_print80("Generating corresponding remote HANA replication package",1,"[FAILURE]");
  }
}

sgeasy_print80("Creating package configurations",1,);

foreach my $currpkgname (keys %CandPkg2Inst) {
  printL "currpkgname=[$currpkgname]\n";
  push (@modifiedPkgs, $currpkgname) if createpkgconfig($sapsid, $currpkgname, $CandPkg2Inst{$currpkgname});
}

# handle extra deploy actions
if (@deployxtractions) {
	sgeasy_print80("Handling extra package directives",1);
	my $deployxtractions_perpkg = xtractions_perpkg(@deployxtractions);
	#print STDERR Data::Dumper->Dump( [ \$deployxtractions_perpkg ], ["deployxtractions_perpkg"] );
	my @xtrapkgs = keys(%{$deployxtractions_perpkg});
	sgeasy_print80("Packages @xtrapkgs");
	my @xtra_modpkgs = execute_xtractions($deployxtractions_perpkg); # extra modified packages
	push (@modifiedPkgs,@xtra_modpkgs);
	if (scalar (@xtrapkgs) eq scalar (@xtra_modpkgs)) {
	  sgeasy_print80("Handling extra package directives",1,"[SUCCESS]");
	} else { sgeasy_print80("Handling extra package directives",1,"[FAILURE]");}
}

#
# generate sitecontroller package: hdbs<SID>..
# 0. get cluster sites (2 local and 2 remote hdb packages reflect a HANA SO setup)
# 1. look for existing pkg containing
#      - sc & hdbsico module
# module_name                     sgesap/hdbsico
# module_name                     sg/sc
#
#      - required SID
# sgesap/hdbsico/sico_hdb_system                  DB1
# sgesap/hdbsico/sico_hdb_instance                        HDB01
#
# 2. If exists
#      YES => use its name and get its configuration
#      NO  => generate a new pkg conig
# 3. Do it the usual way (if possible). will probably need new attribute lists and also new merge ops
# => no, its done by
#      - adding SID to empty sgesap/hdbsico/sico_hdb_system  and
#      - replacing empty site attribute by retrieved (HDB) sites and critical_/remote_managed pkg
#
#  example:
#  sc_site                 rack5
#  critical_package                        hdbpDB1_rack5
#  remote_managed_package                  hdbsDB1_rack5
#  sc_site                 rack7
#  critical_package                        hdbpDB1_rack7
#  remote_managed_package                  hdbsDB1_rack7
#
#  => this means that a remote_managed_package appears under the site it belongs to (which is the
#     then inactive site if the critical package of a site in active)

my @mcsitenames; # this list and local/remotehdbpkgs list are later used for hdbc generation
if ((scalar @localhdbpkgs == 2) && (scalar @remotehdbpkgs == 2)) {
	# SG Metrocluster with SITEs needed
	sgeasy_print80("Getting Serviceguard cluster sites for scaleout setup",1);
	@mcsitenames = GetMCSiteNames(\%clusrsrchash); # get sites, for localnode is first in list
	printL " [ @mcsitenames ]\n" if (defined $mcsitenames[0]);
	if (scalar @mcsitenames == 2) {
		sgeasy_print80("Sites retrieved: [ @mcsitenames ] ");
	    sgeasy_print80("Getting Serviceguard cluster sites for scaleout setup",1,"[SUCCESS]");
	    # Flagging generation of sitecontroller package
		scpkg::set_pkggen_required() # ... if ...
	}
	elsif (scalar @mcsitenames == 0) {
		sgeasy_print80("ERROR: Local node not member of a site");
	    sgeasy_print80("Getting Serviceguard cluster sites for scaleout setup",1,"[FAILURE]");
	}
	elsif (!defined $mcsitenames[0]) {
		sgeasy_print80("ERROR: No sites defined in cluster");
		sgeasy_print80("Getting Serviceguard cluster sites for scaleout setup",1,"[FAILURE]");
	}
	else {
		sgeasy_print80("ERROR: unexpected number of sites [ @mcsitenames ] ");
		sgeasy_print80("Getting Serviceguard cluster sites for scaleout setup",1,"[FAILURE]");
	}
}

# generate hdbsico package if required
if (scpkg::pkggen_required() || $ENV{"deploysappkgs_forcescpkggen"}) {
	sgeasy_print80("Creating HDB site controller (hdbsico) package",1,);
	scpkg::init ($sapsid);
	if (scpkg::generate_pkgconfig ($pkgpath,\@mcsitenames,\@localhdbpkgs,\@remotehdbpkgs)) {
	  sgeasy_print80("Creating HDB site controller (hdbsico) package",1,"[SUCCESS]");
      push (@modifiedPkgs, scpkg::getpkgname());
	}
	else {
	  sgeasy_print80("Creating HDB site controller (hdbsico) package",1,"[FAILURE]");
	}
}

# generate enqor MNP if required
if (enqor::pkggen_required()) {
	sgeasy_print80("Creating enqor MNP package",1,);
	if (enqor::generate_pkgconfig ($pkgpath)) {
	  sgeasy_print80("Creating enqor MNP package",1,"[SUCCESS]");
      push (@modifiedPkgs, enqor::getpkgname());
	}
	else {
	  sgeasy_print80("Creating enqor MNP package",1,"[FAILURE]");
	}
}
sgeasy_print80("Creating package configurations",1,"[SUCCESS]");

# Report modified/new packages back to SGMgr (or others)
# using the report of
# File Name=<pkgpath>/<pkgname>.conf
# to stdout mechanism
sgeasy_print80("List of generated package configuration files ",1,);
sgeasy_print80("");
my $modpkgfnames="";
foreach my $modPkg (@modifiedPkgs) {
  my $fname=$pkgpath."/".$modPkg.".conf";
  # Message below consumed by SGMgr and remote deploysappkgs
  sgeasy_print80("File Name=$fname\n");
  $modpkgfnames .= "-P $fname "; # also already include -P for cmapply
}
sgeasy_print80("List of generated package configuration files ",1,"[SUCCESS]");

if ($applypkgs || $checkpkgs) {
  # immediately apply list of modified packages
  # - build filelist (as string) - see above
  # - build cmapplyconf command (as string)
  # - execute command
  # - ...
  sgeasy_print80("Applying/checking generated package configuration files ",1,);
  sgeasy_print80("Executing command:");
  printL "Modified package config files: $modpkgfnames\n";
  my @cmdoutput;
  if ($checkpkgs) {
 	  my $cmd2exec = qq[$ENV{SGSBIN}/cmcheckconf $modpkgfnames];
	  sgeasy_print80 ("$cmd2exec");
	  @cmdoutput = `$cmd2exec`;
	  #printL "-$!-$?-\n";
  }
  else {
	  my $cmd2exec = qq[$ENV{SGSBIN}/cmapplyconf $modpkgfnames];
	  sgeasy_print80 ("$cmd2exec");
	  @cmdoutput = `$cmd2exec 2>&1 <<EOF
	y
	EOF` ;
	  #printL "-$!-$?-\n";
  }
  unless ($?) {
    sgeasy_print80("Applying/checking generated package configuration files ",1,"[SUCCESS]");
    printL "@cmdoutput"."\n";
  }
  else {
    sgeasy_print80("Applying/checking generated package configuration files ",1,"[FAILURE]");
    print "@cmdoutput"."\n";
    $deployfailure = 1;
  }
}
# running pkgs only if not localhdb
elsif (!$localhdb) {
  # find out packages to be modified but up here
  # (package may be new and therefore not in hash)
  # (move into get getRunningPkgs)
  sgeasy_print80("Checking which modified packages need to be halted ",1,);
  foreach my $modPkg (@modifiedPkgs) {
    push (@modifiedPkgsStatusUp,$modPkg)
      if ((defined  $cluspkgconfighash{$modPkg}) && ($cluspkgconfighash{$modPkg}->[$pkgstatus_idx] eq "up"));
  }
  printL "[ @modifiedPkgs ] , [ @modifiedPkgsStatusUp ] \n";

  (@modifiedPkgsStatusUp) = getRunningPkgs (\@modifiedPkgsStatusUp, \%clusrsrchash);
  #(@modifiedPkgsStatusUp) = getRunningPkgs (\@modifiedPkgs, \%clusrsrchash);

  # Another string for SGMgr easy deployment consumption
  # Comma separate list of package which are up and need modification
  if (scalar @modifiedPkgsStatusUp) {
    # consumed by SGMgr
    print "\nrunning packages = ". join(',',@modifiedPkgsStatusUp) ."\n\n";
    sgeasy_print80("Packages to be halted: [ @modifiedPkgsStatusUp ]");
  }
  else {
    sgeasy_print80("No Packages need to be halted");
  }
  sgeasy_print80("Checking which modified packages need to be halted ",1,"[SUCCESS]");
}

if ($deployfailure) {
  sgeasy_print80("Finished",1,"[FAILURE]");
  exit 1;
}

sgeasy_print80("Finished",1,"[SUCCESS]");
exit 0;
