#!/usr/bin/perl -w


#"@(#) SGeSAP/LX - 12.20.00 - $Rev: 1819 $ - Last modified 02/02/17"
#
# **********************************************************************
# *                                                                    *
# * SGeSAP Verification Support                                        *
# *                                                                    *
# * (c) Copyright 2016-2017 Hewlett Packard Enterprise Co.             *
# * (c) Copyright 2011-2015 Hewlett-Packard Enterprise Corp.           *
# *                                                                    *
# * Description: Provides verification of SAP components in an SGESAP  *
# *              environment.                                          *
# *                                                                    *
# * Parameters:  See pod usage section below                           *
# *                                                                    *
# **********************************************************************


#############################################################################
#
# BEGIN POD SECTION
#
#############################################################################

#---------------------------------------------------
# Prints usage text using pod/perldoc syntax.
# run:
#   perldoc sapverify
#   sapverify -help
#---------------------------------------------------

=pod

=head1 NAME

sapverify - SGeSAP extension to verify the failover capability of a SGeSAP Serviceguard package

=head1 USAGE

sapverify [-h] [-p packagename] [-v 0-5] [-t 45] [-i]
    -h              :   This help message
    -p sgpkg        :   Verify a SGeSAP package
    -v loglevel     :   Corresponds to Serviceguard log_level
    -t timeout      :   Sets the script timeout.
    -i              :   Show SGeSAP package info summary

=head1 DESCRIPTION

sapverify gets executed on behalf of the Serviceguard "cmcheckconf -P
sgpkg_config_file" or "cmapplyconf -P sgpkg_config_file" commands. It
will try to verify failover capabilities of a SGeSAP Serviceguard
package.

The Serviceguard package parameter "log_level" influences how much data
is logged. A value of log_level=5 will will report success and failures,
log_level=4 and less will only report errors.

The program sapverify itself is called by Serviceguard via the validate
section of the SGeSAP dispatcher scripts which are located in the
${SGCONF}/scripts/sgesap directory.

Source and set the Serviceguard environment variables ${SG...} by running
the following command:
    . /etc/cmcluster.conf

It verifies various SAP and SGeSAP related directories and files as well as
file parameters (permissions, owner...), that are required for failing over or
starting a SGeSAP package. Verification is run on all eligible nodes of the
SGeSAP package.

The file "sapverify.rules" contains the specification of SAP, Database and
SGeSAP related files used for verification.

If a file "debug_sapverify" or "debug_sapverify_${sgpkg}" exists
in the $SGRUN directory (see /etc/cmcluster.conf details),
any execution of sapverify is skipped (e.g. basically all sapverify checks
are disabled during cmcheckconf or cmapplyconf runs). The existance of
file "debug_sapverify" turns off sapverify checks for ALL modular SGeSAP packages
of a node, the file "debug_sapverify_${sgpkg}" turns off sapverify checks
only for the specific package ${sgpkg}.

The extended-sapverify-tests (currently Oracle datafile verification) via
cmcheckconf and cmapplyconf can be disabled by adding the environment variable
    export SAPVERIFY_EXTENDED=0
    echo "In customer_functions.sh SAPVERIFY_EXTENDED=${SAPVERIFY_EXTENDED}"
to the package customer_function.sh file.

If a package is down, any filesystems configured as part of the package
will not be accessible (e.g. The "package' file systems are unmounted). In this
case sapverify will skip verification of any of the "package" file systems. If
the  package is up and running, (e.g. the "package" filesystems are
mounted), sapverify will verify the "package" filesystems too.



=head1 EXAMPLES
        ${SGSBIN}/sapverify -p sapC71
        ${SGSBIN}/sapverify -v 5 -p sapC71
        ${SGSBIN}/sapverify -i


=head1 ENVIRONMENT


=head1 FILES



    ${SGSBIN}/sapverify
    ${SGSBIN}/sapverify.pl
        The actual script which contains the code to verify the various SAP
        configurations and systems in a SGeSAP environment.

    ${SGCONF)/sgesap/sapverify.rules
        A specification/list of SAP, Database and SGeSAP files used for verification.

    ${SGRUN}/debug_sapverify
    ${SGRUN}/debug_sapverify_${sgpkg}
        Skips sapverify verification steps.

    ${SGCONF}/scripts/dbinstance.sh
    ${SGCONF}/scripts/livecache.sh
    ${SGCONF}/scripts/sapinstance.sh
    ${SGCONF}/scripts/infra_pre.sh
    ${SGCONF}/scripts/infra_post.sh
    ${SGCONF}/scripts/sapexistance_pre.sh

        The Serviceguard cmcheckconf and cmapplyconf commands will trigger the "validate"
        section in the above scripts. In the "validate" section the sapverify script gets
        executed. Note: When running cmcheckconf or cmapplyconf on one of the cluster
        nodes, Serviceguard will automatically also run the same "validate" section on all
        other failover nodes of the package, thereby executing sapverify on those nodes.


=cut


my $USAGE = <<"END_USAGE";
Usage: \${SGSBIN}/sapverify -p sapC71 ...
    -h              :   This help message
    -p sgpkg        :   Verify a SGeSAP package
    -v loglevel     :   Corresponds to Serviceguard log_level
    -t timeout      :   Sets the script timeout.
    -i info         :   Show SGeSAP package info summary

END_USAGE

#############################################################################
#
# END POD USAGE
#
#############################################################################

#############################################################################
#
# BEGIN INIT SECTION
#
#############################################################################

#print "(********** ENTER sapverify **********\n";

#---------------------------------------------------
# load perl modules
#---------------------------------------------------

use strict;
use warnings;
use diagnostics;

# Three possible locations for SGeSAP SAP_functions
# /opt/cmcluster/sap/lib     HP-UX
# /opt/cmcluster/lib         SLES
# /usr/local/cmcluster/lib   RHEL
use lib qw(
    /opt/cmcluster/sap/lib
    /opt/cmcluster/lib
    /usr/local/cmcluster/lib
);

use IO::Handle;
use Socket;
use FileHandle;
use Carp;
use File::Find;
use File::stat;
use File::Basename;
use Sys::Hostname;
use Cwd;
use Getopt::Long;
use Digest::MD5;
use Data::Dumper;
use POSIX 'strftime';
use Pod::Usage;
use IPC::Open2;
use SAP_functions;


#---------------------------------------------------
# set autoflush on STDOUT
#---------------------------------------------------
$|=1;

#---------------------------------------------------
# Global variables
#---------------------------------------------------

# script arguments
my ($help,                   # print help text for sapverify
    $sgpkg,                  # verify a SGeSAP package
    $info_sgesap,            # show all packages of type sgesap in cluster
    $logtofile,              # log to file, 0|1  default=0
    $verboselevel,           # how muhc to log 0..5
    $timeo,                  # sets the script timeout
    $testmode,               # run in internal testmode
    $getremotelvm,           # get lvm storage config
    $userfiles,              # get user startup files
    $dstrbfiles,             # list all files that are local copies on each node
    $node,                   # cluster node to run command on
    $clucomp,                # compare files clusterwide
    $filelist,               # list of file names to verify
    $pf_for_verify,          # verify a SAP instance
    $filecount,              # total number of files verified
    $errorcount,             # number of verify errors
   );

# package hashes/records for ***single package*** in a cluster
my %vf = ();                    # hash/record - contains list of files with verification data
my %pkgenv = ();                # hash/record - contains cmgetpkgenv sgpkg output
my %pkgstate = ();              # hash/record - contains cmviewcl -f line -p $sgpkg output
my %pkgswitchingstate = ();     # hash/record - contains cmviewcl -v -f line -p $sgpkg|grep "^node:"
my %nodestate = ();             # hash/record - contains cmviewcl -l node -f line
my %dbstorage = ();             # hash/record - contains the storage volumes configured in db itself
my %mount = ();                 # hash/record - contains data on filesystems mounted by system
my %lvmtab = ();                # hash/record - contains data as in/etc/lvmtab
my %remotelvmtab = ();          # hash/record - contains data as in/etc/lvmtab
my %dstrb_file = ();            # hash/record - contains stats of dstrb file for local node
my %remote_dstrb_file = ();     # hash/record - contains stats of dstrb file for remote node
my %home_dir = ();              # hash/record - tmp lookup: contains home directory startup files of user
my %userenv = ();               # hash/record - tmp lookup: contains user env data
my %startup_file = ();          # hash/record - tmp: contains home directory startup files of user
my %remote_startup_file = ();   # hash/record - contains data from a remote startupfiles
my %root_mp=();                 # hash/record - of all package "root" mountpoints
my %sp_helpdevice = ();         # hash/record - contains sybase physical files
my %param_getvolsall = ();      # hash/record - contains maxdb volumes


# Initialize %mpsfs      maps mountpoints -> filesystems
# Initialize %mpsfst     maps mountpoints -> filesystem types
# Initialize %mpsfsd     maps mountpoints -> device
# Initialize %vxdgshared maps ...
my %mpsfs      = ();
my %mpsfst     = ();
my %mpsfsd     = ();
my %vxdgshared = ();
my %verify_vxfs = ();

my @mounts;
my @vxdglist;
my @vxdg;
my @printarr;
my @dstrb_file_list;
my @sql_output;
#tmp
my @sorted_mount;
my  @db2storage;                # array - contains db2 volumes

my $vxvm_installed;
my $dc;
my $fst;
my $mp;


# Initialize @profiles, @ersprofiles containing SAP profile names
my @profiles=();
my %addrs   = ();
my %subnets = ();
my $apiversion;
my $hdbuk;
my $sidadm;
my $fh;
my @hdbvalues;
my @ersprofiles;

# List of symbolic instance filters (-f <symb.inst.filter>) . So far only "SCS" is supported, rest is inofficial
# Specifications allow reg. expressions
my %InstanceFilters= (
scs => "(SCS|ASCS|ERS|DVEBMGS)",
WDP => "W",
ERS => "ERS",
HDB => "HDB",
  );

# package configuration for ***all packages*** in a cluster
my ( %package_name,                     # names of all packages in the cluster
     %package_style,                    # modular or legacy
     %package_modules,                  # list of all serviceguard modules of a package
     %package_uses_sgesap_module,       # package has a sgesap module configured
     %package_uses_nfs_module,          # package has a nfs module configuired
     %package_db_system,                # of type db_system
     %package_sap_system,               # of type sap_system
     %package_sap_ext_system,           # of type external system
     %package_lc_system,                # of type livecache system
     %package_repdb_system,             # a sysbase replication system
     %package_hdb_system,             # a sysbase replication system
     %package_db_vendor,                # oracle, db2, maxdb or sybase
     %package_sap_instance,             # DVEBMGS|D|ASCS|SCS|ERS|REP|J|JD|JC|MDS|MDIS|MDSS|W
     %package_sap_ext_instance,         # external instance
     %package_sap_infra_sw_type,        # infra structure instance
     %package_sap_replicated_instance,  # replicated instance
     %package_hdb_instance,             # HDB
    );

# setup_script_runtime() will set following runtime environment for sapverify
my (
    $debug,                 # Turn on/off debug traces
    $dumpdata,              # Turn on/off dumping of datastructures
    $extended,              # Turn on/off extended verify test (don't run cmexec)
    $verbose,               # Turn on/off verbose traces
    $tracesubs,             # Turn on/off sburoutine enter/leave traces
    $timeout,               # Script timeout
    $logging,               # Turn on logging to default log file
    $fork,                  # fork a child to provide escape on NFS hangs
    $debugflagfile,         # SGeSAP / sapverify specific debug flag file
    $warningmode,           # only log warnings instead of errors
   );

my $ERRORT="ERROR:";         # type of error to log - default is ERROR

# Serviceguard cmcluster.conf
my $SGSBIN;                 # path to the SG binaries
my $SGRUN;                  # path to the SG run environment
my $SGCONF;                 # path to the SG conf environment
my $SGROOT;                 # path to the SG root environment

# File I/O
my $RULESF;                 # name of file containing the verification rules
my $RULES_FH;               # filehandle for the rules file
my $LOGF;                   # name of the logfile
my $LOG_FH;                 # file handle for the logfile
my $READ_SQL_FH;            # Oracle sqlplus read filehandle
my $WRITE_SQL_FH;           # Oracle sqlplus write filehandle
my $C_FH;
my $CF;
my $ISQL_FH;                # file handle for sysbase isql
my $SAPSERVICESF="/usr/sap/sapservices";
my $SAPSERVICESFH;          # filehandle for /usr/sap/sapservices:

my $PackageOnThisNode;      # boolean - if package running on this node
my $PackageDown;            # boolean - if package is down/unowned
my $PackageUp;              # boolean - if package is down/unowned
my $cmcompare_master;       # boolean - when set will run cmcompare commands
my @switchnodelist;         # list of nodes package switching is enabled for - formated
my @switchnodes;            # list of nodes package switching is enabled for - unformated

#my $fasttimeout=9;          # fast timeout e.g. accessing NFS files
my $fasttimeout=3;          # fast timeout e.g. accessing NFS files
my $childpid;               # pid of the child after a fork
my %loc_chksum;             # save checksum of a file

my $sg_op_code="";          # Serviceguard op code = "validate" here
my $calledby="";            # SGeSAP script who called us
my $first_sgesap_ops="";     # The first SGeSAP operation sequenc found
my $sg_owner_node="";       # Serviceguard
my $sg_master_node="";      # Serviceguard
$timeout=120;               # set a default timeout valid for any script options
my $datestring;             # contains timestamp

my $oracle_up=0;            # is Oracle up
my $oracle_down=1;          # is Oracle down
my $repdb_t=0;              # is a Sybase replication instance

my @dbdatafiles;            # list of of database data files
my $file="";                # belongs to follow link
my $dir;                    # belongs to follow link

my $SGeSAP_ASEUSER="";      # Sybase ASE user
my $SGeSAP_ASEPASSWD="";    # Sybase ASE passwd

#---------------------------------------------------
# save args for later use
#---------------------------------------------------
my @SAVED_ARGV = @ARGV;
LogDebug("BEGIN SAVED_ARGV=[@SAVED_ARGV]");

#---------------------------------------------------
# get/set system specifics
#---------------------------------------------------
my $host = hostname();
$node=$host                 unless defined $node;
my $cwd  = getcwd();

# split $0 into basename, directory and suffix
my ($basename, $dirname, $suffix) = fileparse($0,qr/\.[^.]*/);

$filecount=0;           # number of files verified
$errorcount=0;          # number of verification errors found

#---------------------------------------------------
# Don't fork if any non "package verify" code is run
# and fork if any package verify code is run in
# sapverify. fork is used to be able to timeout
# NFS access hangs
#---------------------------------------------------
$fork=0;

#---------------------------------------------------
# get cmcluster environment from /etc/cmcluster.conf
# e.g. set $SGSBIN, $SGRUN.... depending if it is
# RHEL/SLES/HPUX.
#
# SAP_functions.pm has already initialized these
# so make them available here
#---------------------------------------------------
$SGSBIN   = getparam('SGSBIN');
$SGRUN    = getparam('SGRUN');
$SGCONF   = getparam('SGCONF');
$SGROOT   = getparam('SGROOT');

#---------------------------------------------------
# Process the script arguments and set appropriate
# global sapverify variables
#---------------------------------------------------
parse_args();

#---------------------------------------------------
# If action to be run is "package verify", check for
# debug flag file for early exit, check package is
# known ...
# verify action is set if $sgpkg is defined
#---------------------------------------------------

analyze_package_runtime_env($sgpkg)         if defined $sgpkg;

# --> execution continues in MAIN section below

#############################################################################
#
# END INIT SECTION
#
#############################################################################

#############################################################################
#
# BEGIN SECTION FOR GENERAL FUNCTIONS ( also see SAP_functions.pm)
#
#############################################################################

#===  FUNCTION  =============================================================
# NAME:         CountAndCountAndLogErrorF()
# PURPOSE:      Count number of errors and log to stderr and $LOG_FH
# PARAMETERS:   @_  = arguments to function being called
# RETURNS:      None
# COMMENTS:
#============================================================================

sub CountAndLogErrorF {

    $errorcount++;
    LogErrorF(@_);

}

#===  FUNCTION  =============================================================
# NAME:         LogErrorF log with file name prepended
# PURPOSE:
# PARAMETERS:   @_  = arguments to function being called
# RETURNS:      None
# COMMENTS:
#============================================================================

sub LogErrorF {

    my $tmp = join( "", "(sapverify):", " ", @_ );
    LogError($tmp);

}

#===  FUNCTION  =============================================================
# NAME:         LogInfoF log with file name prpended
# PURPOSE:
# PARAMETERS:   @_  = arguments to function being called
# RETURNS:      None
# COMMENTS:
#============================================================================

sub LogInfoF {

    my $tmp = join( "", "(sapverify): INFO:", " ", @_ );
    LogInfo($tmp);

}

#===  FUNCTION  =============================================================
# NAME:         Set how much get logged to the log file
# PURPOSE:
# PARAMETERS:
# RETURNS:
# COMMENTS:     Used if the -v 0..5 is set on the command line or log_level
#               set in the package confiration file
#               cmcheckconf/cmapplyconf is used if validate flag is set
#               else
#               -l flag is set
#
#============================================================================

sub setverboselevel {
    LogEnterSub(@_);
    my $loglevel = shift;

    my $verbose;
    # if verbose was set in customer_functions.sh....
    # environment override everything
    if ( defined $ENV{SAPVERIFY_VERBOSE} ) {
        setparam( "verbose", $ENV{SAPVERIFY_VERBOSE} );
        LogReturnSub();
        return ();
    }

    $verbose = getparam('verbose');
    return () if $verbose > 0;

    my $default=4;

    # requires that cmgetpkgenv was run: it fills  %pkgenv
    if ( $sg_op_code eq "validate" ) {
        if (defined $pkgenv{sg_log_level}) {
            $loglevel = $pkgenv{sg_log_level};
        } else {
            $loglevel=$default;
        }
    }

    # obsolete as loglevel can't be > 5 any more
       if  ( $loglevel == 0 ) {
         setparam( "verbose", 0);

    } elsif ( $loglevel == 1 ) {
         setparam( "verbose", 0);

    } elsif ( $loglevel == 2 ) {
         setparam( "verbose", 0);

    } elsif ( $loglevel == 3 ) {
         setparam( "verbose", 0);

    } elsif ( $loglevel == 4 ) {
         setparam( "verbose", 0);

    } elsif ( $loglevel == 5 ) {
         setparam( "verbose", 0);

    } elsif ( $loglevel == 6 ) {
         setparam( "verbose", 1);
         setparam( "tracesubs", 1);

    } elsif ( $loglevel == 7 ) {
         setparam( "verbose", 1);
         setparam( "tracesubs", 1);
         setparam( "dumpdata", 1);
        $dumpdata = getparam('dumpdata');
    } else {
        print("Info: loglevel=[$loglevel] unknown - setting verbose=[1]\n");
        setparam( "verbose", 1);
    }

    LogReturnSub();
    return ();
}

#===  FUNCTION  =============================================================
# NAME:         Set logging to the log file
# PURPOSE:
# PARAMETERS:
# RETURNS:
# COMMENTS:     Used if the -l 0|1 set on sapverify command line
#               0=don't log or 1=log to file
#============================================================================

sub setlogtofile {
    LogEnterSub(@_);
    my $logflag = shift;

    if ( ( $logflag == 0 ) or ( $logflag == 1 ) ) {
        LogDebug("logflag=[$logflag]");
        setparam( "logging", $logflag );
    }
    else {
        print("Info: logflag=[$logflag] unknown - using default logging=[0]\n");
        setparam( "logging", 0 );
    }

    LogReturnSub();
    return ();
}


#===  FUNCTION  =============================================================
# NAME:         get the first sgesap entry on the sg operation sequence
# PURPOSE:
# PARAMETERS:
# RETURNS:
# COMMENTS:
#============================================================================

sub get_first_sgesesap_op_sequence {
    LogEnterSub(@_);

    my ($base,$dir,$suffix);

    $base="";
    foreach my $ops (@{$pkgenv{sg_operation_sequence}}) {
            LogDebug "Checking ops=[$ops]";

            if ( $ops =~ /sgesap/ ) {
                LogDebug "Found first sgesap ops=[$ops]";
                ($base, $dir, $suffix) = fileparse($ops,qr/\.[^.]*/);
                last;
            }
    }

    LogReturnSub($base);
    return ($base);
}


#===  FUNCTION  =============================================================
# NAME:         get the name of the script who called us
# PURPOSE:      the script name will determine which component will be verified
# PARAMETERS:
# RETURNS:
# COMMENTS:
# /etc/cmcluster/scripts/sgesap/dbinstance.sh
# /etc/cmcluster/scripts/sgesap/sapinstance.sh
# ....
#============================================================================

sub called_by_script {
    LogEnterSub(@_);
    my $script = shift;

    my ($base,$dir,$suffix);

    # split $script into calledby, directory and suffix
    ($base, $dir, $suffix) = fileparse($script,qr/\.[^.]*/);

    LogDebug "sapverify is being called from called script base=[$base]";

    # $calledby is global var
    $calledby=$base;

    LogReturnSub();
    return ();
}



#===  FUNCTION  =============================================================
# NAME:         Change default timeout of script
# PURPOSE:
# PARAMETERS:
# RETURNS:
# COMMENTS:
#============================================================================

sub settimeout {
    my $timeout = shift;

        LogDebug("Setting timeout=[$timeout]");
        setparam( "timeout", $timeout );

}


#===  FUNCTION  =============================================================
# NAME:         Get the device stats for a file
# PURPOSE:
# PARAMETERS:
# RETURNS:
# COMMENTS:
#============================================================================

sub get_stat_dev {
    LogEnterSub(@_);
    my $path = shift;
    my $tmp_str;

    LogDebug "path $path";

    # if autofs force a mount by calling opendir
    if ( -d $path ) {
        LogDebug "path=[$path] is a directory";
        opendir( my $dh, $path ) or LogRuntimeError "Couldn't opendir $path \$!=[$!]";
        closedir $dh or LogRuntimeError "Close failed for [$path]";
    }

    my $info = stat($path);
    if ($info) {

        $tmp_str = sprintf "path=[%s] dev=[%s]", $path, $info->dev;
        LogDebug "$tmp_str";

        $tmp_str = sprintf "path=[%s] ino=[%s]", $path, $info->ino;
        LogDebug "$tmp_str";

        LogReturnSub( $info->dev );
        return ( $info->dev );
    }

}

#############################################################################
# this function create system storage records for all devices configured in the db.
# It creates two hashes from the output of the mount command:
#   mpsfs{}     maps mountpoints -> filesystems
#   mpsfst{}    maps mountpoints -> filesystem types
#   mpsfsd{}    maps mountpoints -> file system storage device numbers
#
#   Example:
#   %mpsfs= {/oracle/MO7 /dev/vgoramo7/lvoramo7}
#   %mpsfst={/oracle/MO7 vxfs}
#   %mpsfsd={'64893' => '/oracle/MO7'}
# Filters mountpoints using the expression provided as function argument.
# get the mount information to match it later to the instance
# directories.
#
# Arguments:
#   $mpfilter   mount point filter -> the mount we are trying to get the
#               filesystem and the filesystem type for
# lifted from sapgetconf.pl
#############################################################################
sub gen_system_storage_records {
    LogEnterSub( @_ );

    my $mpfilter = shift;
    my $rc;

    # VXVM only used by HPUX
    $vxvm_installed = 0;
    my $vxvm_used = 0;

    if ( -f </etc/vx/volboot> ) {
        # VXVM installed (same test used in SGMgrPI)
        $vxvm_installed = 1;
        LogVerbose("vxvm installed");
    }

    my $cmd = "mount -v";
    LogDebug("Execute cmd=[$cmd]");
    LogVerbose("Execute cmd=[$cmd]");

    @mounts = `$cmd` or LogRuntimeError "cmd=[$cmd], return code [$!]";
    $rc=$?;
    LogDebug("\@mounts=[@mounts]");

    if ( $rc != 0 ) {
        LogReturnSub($rc);
        return $rc;
    }

    #------------------------------------------------------------------
    # Create hashes: Split the mount data. Save only if a "mount point"
    # $mp matches the "mount point filter" $mpfilter
    #   $dc   just a placeholder
    #   $fs   filesystem
    #   $mp   mountpoint
    #   $fst  filesystem type
    #------------------------------------------------------------------

    my ($fs, $dc, $mp, $fst);
    for $_ (@mounts) {
        ( $fs, $dc, $mp, $dc, $fst ) = split( / +/, $_ );
        # only interested in "real" filesystems mounted by SG package
#        if ( $fst !~ /^(ext2|ext3|ext4|nfs|nfs4|autofs|vxfs)$/) next;
        if ( $mp =~ /$mpfilter/ ) {
            $mpsfs{$mp}  = $fs;
            $mpsfst{$mp} = $fst;
            if ( $vxvm_installed && $fs =~ /^\/dev\/vx/ ) {
                $vxvm_used = 1;
                LogVerbose("vmvx: $fs");
            }
        }
    }

    #---------------------------------------------------
    # HPUX: collect vxvm specific info
    #---------------------------------------------------
    if ($vxvm_used) {
        # executed a vxdg list to figure out if dg is "shared"
        my $cmd = "vxdg list";
        LogVerbose("Execute cmd=[$cmd]");
        @vxdglist = `$cmd` or LogRuntimeError "cmd=[$cmd], return code [$!]";
        LogVerbose("\@vxdglist=[@vxdglist]");

        shift @vxdglist;

        for $_ (@vxdglist) {
            @vxdg = split(/ +/);
            LogVerbose("NAME: $vxdg[0] - STATE: $vxdg[1]");
            $vxdgshared{ $vxdg[0] } = ( $vxdg[1] =~ /shared/ );
            @printarr = %vxdgshared;
            LogVerbose("\%vxdgshared=[@printarr]");
        }
    }

    #---------------------------------------------------
    # Initialize %mpsfsd
    # this maps mountpoints -> storage device numbers
    #---------------------------------------------------
    for $file ( keys %mpsfst ) {
        LogDebug "mpsfst loop file=[$file]";
#       my $dev = get_stat_dev($file);
       my $dev = call_func_with_timeout_sig( $fasttimeout, \&get_stat_dev, $file );
        $mpsfsd{$dev} = $file;
    }

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%mpsfs ],  ["mpsfs"] ));
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%mpsfst ], ["mpsfst"] ));
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%mpsfsd ], ["mpsfsd"] ));

    LogReturnSub(0);
    return 0;
}

#############################################################################
# this function create system storage records for all devices configured in the db.
# It creates two hashes from the output of the mount command:
#   mpsfs{}     maps mountpoints -> filesystems
#   mpsfst{}    maps mountpoints -> filesystem types
#   mpsfsd{}    maps mountpoints -> file system storage device numbers
#
#   Example:
#   %mpsfs= {/oracle/MO7 /dev/vgoramo7/lvoramo7}
#   %mpsfst={/oracle/MO7 vxfs}
#   %mpsfsd={'64893' => '/oracle/MO7'}
# Filters mountpoints using the expression provided as function argument.
# get the mount information to match it later to the instance
# directories.
#
# Arguments:
#   $mpfilter   mount point filter -> the mount we are trying to get the
#               filesystem and the filesystem type for
# lifted from sapgetconf.pl
#############################################################################
sub gen_db_storage_records {
    LogEnterSub( @_ );

    LogVerbose "Generating database storage records";
    sql_get_db_storage();

    # get device numbers for each file in @dbdatafiles
    # and save in dbstorage hash
    get_device_for_file();


    #---------------------------------------------------
    # for each device/data volume configured in the
    # database set the corresponding device info  in
    # the $storage structure for the db
    #---------------------------------------------------
    foreach my $file ( keys %{dbstorage} ) {
        LogDebug "file=[$file]";

        my $mountbypackage = is_fs_mounted_by_package($file);
        LogDebug "file=[$file] mountbypackage=[$mountbypackage] $PackageOnThisNode=[$PackageOnThisNode]";

        $dbstorage{$file}{mountbypackage}=$mountbypackage;

        # get the data device data for the storage file
        # Note: storage could be a file or filesystem
        # Note: if $file doesn't exist on system then
        # value for $stodev is undefined

        my $stodev = $dbstorage{$file}{dev} if defined $dbstorage{$file}{dev};
        LogDebug "stodev=[$stodev] file=[$file]";

        my $mountpoint = $mpsfsd{$stodev};
        LogDebug "$mountpoint=[$mpsfsd{$stodev}]";

        $dbstorage{$file}{fst} = $mpsfst{$mountpoint};
        LogDebug "fst=[$mpsfst{$mountpoint}]";

        # if hpux system save vxfs volumes for later verification
        if ( $dbstorage{$file}{fst} eq "vxfs" ) {
            $verify_vxfs{$mountpoint} = "vxfs";
        }

        # with the $stodev index into the hash for the mount file systems.
        # and save it as the storage mount point
        $dbstorage{$file}{mountpoint} = $mountpoint;
    }

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%dbstorage ], ["dbstorage"] ));
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%verify_vxfs ], ["verify_vxfs"] ));

    LogReturnSub();
}

#===  FUNCTION  =============================================================
# NAME:         get device entry of a file
# PURPOSE:
# PARAMETERS:
# RETURNS:
# COMMENTS:
#============================================================================
sub get_device_for_file {
    LogEnterSub(@_);

    foreach my $file ( @dbdatafiles ) {
        # Check if the database files exist on system
        if ( -e $file ) {
            # my $dev = get_stat_dev($file);
            my $dev = call_func_with_timeout_sig( $fasttimeout, \&get_stat_dev, $file );

            # save the storage device data for the file
            $dbstorage{$file}{dev}  = $dev;
        } else {
            CountAndLogErrorF "$ERRORT host=[$host] ", "$pkgenv{sgesap_db_vendor} database file=[$file] not found on system";
        }
    }

    LogReturnSub();
    return;
}
#############################################################################
# This functions creates two associative arrays from the output of the
# mount command:
#   mpsfs{}     maps mountpoints -> filesystems
#   mpsfst{}    maps mountpoints -> filesystem types
#
#   Example:
#   %mpsfs=[/oracle/MO7 /dev/vgoramo7/lvoramo7]
#   %mpsfst=[/oracle/MO7 vxfs]
#
# Filters mountpoints using the expression provided as function argument.
# get the mount information to match it later to the instance
# directories.
#
# Arguments:
#   $mpfilter   mount point filter -> the mount we are trying to get the
#               filesystem and the filesystem type for
# lifted from sapgetconf.pl
#############################################################################
sub gen_system_storage_records_hpux {
    LogEnterSub( @_ );

    my $mpfilter = shift;

    $vxvm_installed = 0;
    my $vxvm_used = 0;

    if ( -f </etc/vx/volboot> ) {
        # VXVM installed (same test used in SGMgrPI)
        $vxvm_installed = 1;
        LogVerbose("vxvm installed");
    }

    my $cmd = "mount -v";
    LogDebug("Execute cmd=[$cmd]");

    @mounts = `$cmd` or LogRuntimeError "cmd=[$cmd], return code [$!]";
    LogDebug("\@mounts=[@mounts]");

    #------------------------------------------------------------------
    # Create hashes: Split the mount data. Save only if a "mount point"
    # $mp matches the "mount point filter" $mpfilter
    #   $dc   just a placeholder
    #   $fs   filesystem
    #   $mp   mountpoint
    #   $fst  filesystem type
    #------------------------------------------------------------------

    my ($fs, $dc, $mp, $fst);
    for $_ (@mounts) {
        ( $fs, $dc, $mp, $dc, $fst ) = split( / +/, $_ );
        if ( $mp =~ /$mpfilter/ ) {
            $mpsfs{$mp}  = $fs;
            $mpsfst{$mp} = $fst;
            if ( $vxvm_installed && $fs =~ /^\/dev\/vx/ ) {
                $vxvm_used = 1;
                LogVerbose("vmvx: $fs");
            }
        }
    }

    #---------------------------------------------------
    #---------------------------------------------------
    if ($vxvm_used) {
        # executed a vxdg list to figure out if dg is "shared"
        my $cmd = "vxdg list";
        LogVerbose("Execute cmd=[$cmd]");
        @vxdglist = `$cmd` or LogRuntimeError "cmd=[$cmd], return code [$!]";
        LogVerbose("\@vxdglist=[@vxdglist]");

        shift @vxdglist;

        for $_ (@vxdglist) {
            @vxdg = split(/ +/);
            LogVerbose("NAME: $vxdg[0] - STATE: $vxdg[1]");
            $vxdgshared{ $vxdg[0] } = ( $vxdg[1] =~ /shared/ );
            @printarr = %vxdgshared;
            LogVerbose("\%vxdgshared=[@printarr]");
        }
    }

    #---------------------------------------------------
    # Initialize %mpsfsd      maps mountpoints -> storage device
    #---------------------------------------------------
    for $file ( keys %mpsfst ) {
        LogDebug "mpsfst loop file=[$file]";
        my $dev = get_stat_dev($file);
        $mpsfsd{$dev} = $file;
    }

    #---------------------------------------------------
    #---------------------------------------------------
    foreach my $file ( keys %{dbstorage} ) {
        LogDebug "file=[$file]";

        # get the device data for the storage file
        my $stodev = $dbstorage{$file}{dev};
        LogDebug "stodev=[$stodev] file=[$file]";

        my $mountpoint = $mpsfsd{$stodev};
        LogDebug "$mountpoint=[$mpsfsd{$stodev}";

        # with the $stodev index into the hash for the mount file systems.
        # and save it as the storage mount point
        $dbstorage{$file}{mountpoint} = $mountpoint;
    }

    #---------------------------------------------------
    #---------------------------------------------------
    # Now get the vxfs filesytemes that needto be verified
    foreach my $file ( keys %{dbstorage} ) {

        my $stodev = $dbstorage{$file}{dev};
        my $stofs  = $mpsfsd{$stodev};
        my $stofst = $mpsfst{$stofs};

        # physical device
        my $stopd = $mpsfs{$stofs};

        LogDebug "Storage for db file=[$file]:";
        LogDebug "        stodev=[$stodev] stofs=[$stofs] stofst=[$stofst] stopd=[$stopd]";

        if ( $stofst eq "vxfs" ) {
            $verify_vxfs{$stopd} = "vxfs";
        }
    }

    my $count_volumes = keys %{verify_vxfs};
    LogVerbose "Found [$count_volumes] volume(s) for verification";

    foreach my $file ( keys %{verify_vxfs} ) {
        LogVerbose "Volume(s) selected for verification of DB datafile(s): [$file]";
    }

    #---------------------------------------------------
    #---------------------------------------------------


    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%mpsfs ],  ["mpsfs"] ));
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%mpsfst ], ["mpsfst"] ));
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%mpsfsd ], ["mpsfsd"] ));
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%dbstorage ], ["dbstorage"] ));
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%verify_vxfs ], ["verify_vxfs"] ));

    LogReturnSub();
}


#############################################################################
#
# END SECTION FOR GENERAL FUNCTIONS
#
#############################################################################
#
# BEGIN SECTION FOR SAPVERIFY FUNCTIONS
#
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: parse script arguments
#   PARAMETERS: @ARGV script arguments
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#============================================================================
sub parse_args {
    LogEnterSub(@_);

    # if early trace enabeled with verbose=2
    LogExtra "Called parse_args ARGV=[@ARGV]";

    #---------------------------------------------------
    # Process the script arguments
    #---------------------------------------------------
    my $options_okay = GetOptions(
        "help"              => \$help,               # print help text
        "info"              => \$info_sgesap,        # show info on all packages of type sgesap
        "package|p=s"       => \$sgpkg,              # SG package to be verified
        "logtofile=s"       => \$logtofile,          # define non-default file to log to
        "verboselevel=s"    => \$verboselevel,       # how verbose is output
        "calledby=s"        => \$calledby,           # the name of script/module calling us
        "timeout=s"         => \$timeo,              # controls timeout of script
        "xxxxxxxx"          => \$testmode,           # run in internal testmode
        "sg_op_code=s"      => \$sg_op_code,         # SG operation: validate
        "getremotelvm|g"    => \$getremotelvm,       # get/dump the /etc/lvmtab table
        "userfiles=s"       => \$userfiles,          # list user startup files for comparison
#        "dstrbfiles=s{,}"   => \@dstrbfiles,         # list all files that are local copies on each node for package
        "dstrbfiles=s@"     => \$dstrbfiles,         # list all files that are local copies on each node for package
        "profile|pf=s"      => \$pf_for_verify,      # used for external SGeSAP instances;
                                                     # verify SAP based on profile (<sid>_<instance>_<vip>)

        "filelist=s"        => \$filelist,           # file with file names to verify
    );

    #---------------------------------------------------
    # an argument is leftover, unknown...
    #---------------------------------------------------
    if (@ARGV) {
        print("Error: Unknown argument\n");
        warn $USAGE;
        exit 2;
    }

    usage_perldoc() if !$options_okay;
    usage_perldoc() if defined $help;
    return (0);
}

#===  FUNCTION  =============================================================
#      PURPOSE: init verify of sgesap package
#   PARAMETERS: $sgpkg  name of SG package
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#============================================================================
sub analyze_package_runtime_env {
    LogEnterSub(@_);
    LogExtra "analyze_package_runtime_env() [@_]";

    my $sgpkg = shift;

    #---------------------------------------------------
    # An initial sanitiy check (script will exit if error
    # occurred):
    # - find and set package $sgpkg
    # - get / set if in Serviceguard validate mode
    #---------------------------------------------------
    check_package_operation_env();

    #---------------------------------------------------
    # The following checks if execution of sapverify is
    # is to be postponed/skipped if debug flag file
    # exists
    #---------------------------------------------------
    check_sapverify_debug_file();

    #---------------------------------------------------
    # The subroutine setparam exports the $params{sgpkg}
    # to the to SAP_functions.pm so it can write the
    # SG package specific log file...  when subrotuine
    # setup_script_runtime() is called
    #
    # sub setup_script_runtime() finally sets up the
    # the runtime environment for sapverify.
    #---------------------------------------------------
    setparam( "sgpkg", $sgpkg );

    #---------------------------------------------------
    # sub setup_script_runtime() finally sets up the
    # the runtime environment for sapverify and package.
    #---------------------------------------------------
    setup_script_runtime();

    # Now Logfile should be open
#    LogVerbose "Script was called SAVED_ARGV=[@SAVED_ARGV]";

    #---------------------------------------------------
    # import variables defined in SAP_functions.pm and
    # and that are used here
    #---------------------------------------------------
    $timeout  = getparam('timeout');
    $fork     = getparam('fork');
    $RULESF   = getparam('RULESF');
    $testmode = getparam('testmode');
    $dumpdata = getparam('dumpdata');
    $extended = getparam('extended');
    $warningmode = getparam('warningmode');
    $ERRORT="WARNING:"           if  $warningmode;


    LogReturnSub(0);
    return (0);
}

#===  FUNCTION  =============================================================
#      PURPOSE: init and run verify of sgesap package
#   PARAMETERS: $sgpkg  name of SG package
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#============================================================================
sub run_verify_sgesap_package {
    LogEnterSub(@_);
    my $sgpkg = shift;

#    #---------------------------------------------------
#    # An initial sanitiy check (script will exit if error
#    # occurred):
#    # - check and set name of package $sgpkg
#    # - check and set SG validate mode
#    #---------------------------------------------------
#    check_package_operation_env();
#
#    #---------------------------------------------------
#    # The following checks if execution of sapverify is
#    # is to be postponed/skipped
#    #---------------------------------------------------
#    check_sapverify_debug_file();
#
#    #---------------------------------------------------
#    # The subroutine setparam exports the $params{sgpkg}
#    # to the to SAP_functions.pm so it can write the
#    # SG package specific log file...  when subrotuine
#    # setup_script_runtime() is called
#    #
#    # sub setup_script_runtime() finally sets up the
#    # the runtime environment for sapverify.
#    #---------------------------------------------------
#    setparam( "sgpkg", $sgpkg );
#    setup_script_runtime();
#
#    #---------------------------------------------------
#    # import variables defined in SAP_functions.pm and
#    # and that are used here
#    #---------------------------------------------------
#    $timeout  = getparam('timeout');
#    $fork     = getparam('fork');
#    $RULESF   = getparam('RULESF');
#    $testmode = getparam('testmode');
#    $dumpdata = getparam('dumpdata');
#    $extended = getparam('extended');
#
    #---------------------------------------------------
    # The real verify occurs here
    #---------------------------------------------------
    verify_sgesap_package($sgpkg);

    LogReturnSub(0);
    return (0);
}

#===  FUNCTION  =============================================================
#      PURPOSE: Verify a SGeSAP package
#   PARAMETERS: $sgpkg  Is name of Serviceguard package
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: gets executed either as standalone program:
#               sapverify.pl -p $sgpkg with a package name.
#                   or
#               indirect via cmcheckconf or cmapplyconf
#============================================================================
sub verify_sgesap_package {
    LogEnterSub(@_);
    my $sgpkg = shift;

    LogVerbose("");
    LogVerbose("Verifying SGeSAP package sgpkg=[$sgpkg]");

    if ( $sg_op_code eq "validate" ) {

        #---------------------------------------------------
        # if $sg_op_code is - set sapverify is being called
        # via cmcheckconf or cmapplyconf.
        #
        # Get the Serviceguard package configuration as well
        # as various states of the package. Note: The package
        # data is based on configuration provided via the
        # cmcheckconf/cmapplyconf -P sgpkg.conf call
        #---------------------------------------------------
        getpkgenv($sgpkg);
        LogExtra( Data::Dumper->Dump( [ \%pkgenv ], ["pkgenv"] ) );

        #---------------------------------------------------
        # Check if being called the first time. if not return
        #---------------------------------------------------
        $first_sgesap_ops = get_first_sgesesap_op_sequence($sgpkg);
        LogVerbose("sapverify ops calledby=[$calledby]");
        if ( "$first_sgesap_ops" ne "$calledby" ) {
             LogExtra("sapverify completed with first_sgesap_ops=[$first_sgesap_ops]");
             # execute sapverify only for first
             # sgesap module in package. for all
             # others just exit with success
             exit(0);
#            LogReturnSub(0);
#            return (0);
        }

        setverboselevel();
        getpkgstate($sgpkg);

        is_package_supported($sgpkg);
        is_package_on_this_node($sgpkg);
        is_package_down($sgpkg);
        get_switchnodelist($sgpkg);
        cmget_cmcompare_action($sgpkg);
    }
    else {
        #---------------------------------------------------
        # sapverify is being called as standalone program
        #
        # Get the Serviceguard package configuration as well as
        # various states of the package. Note: The package
        # data is based on configuration provided by the
        # cmgetpkgenv interface (e.g. package has previously
        # been configured and applied)
        #---------------------------------------------------
        cmgetpkgenv($sgpkg);
        LogExtra( Data::Dumper->Dump( [ \%pkgenv ], ["pkgenv"] ) );

        #---------------------------------------------------
        # Check if being called the first time. if not return
        #---------------------------------------------------
#        $first_sgesap_ops = get_first_sgesesap_op_sequence($sgpkg);
#        if ( "$first_sgesap_ops" ne "$calledby" ) {
#            LogReturnSub(0);
#            return (0);
#        }
#
        cmviewcl_p_package($sgpkg);

        is_package_supported($sgpkg);
        is_package_on_this_node($sgpkg);
        is_package_down($sgpkg);
        cmget_switchnodelist($sgpkg);
        cmget_cmcompare_action($sgpkg);
    }

    #---------------------------------------------------
    # setup the runtime environments
    #---------------------------------------------------
    setup_sap_system_runtime()      if defined $pkgenv{sgesap_sap_system};
    setup_db_system_runtime()       if defined $pkgenv{sgesap_db_system};
    setup_lc_system_runtime()       if defined $pkgenv{sgesap_lc_system};

    # if operation_sequence $SGCONF/scripts/sgesap/repdbinstance.sh
    # exists then it a sybase rep db instance
#    setup_repdb_system_runtime()  if defined $pkgenv{sgesap_repdb_system};
    setup_repdb_system_runtime();

    setup_hdb_system_runtime()    if defined $pkgenv{sgesap_hdb_system}[0];

    #---------------------------------------------------
    # init a list of files to verify with sg_fs_directory
    # package mount points. This forces verification of
    # package mount points on all nodes.
    #---------------------------------------------------
    verifylist_init("files_package");

    #---------------------------------------------------
    # append list of files to verify to %vf hash
    #---------------------------------------------------
    create_verifylist_sap_system()      if defined $pkgenv{sgesap_sap_system};
    create_verifylist_db_system()       if defined $pkgenv{sgesap_db_system};
    create_verifylist_lc_system()       if defined $pkgenv{sgesap_lc_system};
#    create_verifylist_repdb_system()    if defined $pkgenv{sgesap_repdb_system};
    create_verifylist_repdb_system()  if $repdb_t==1;
    create_verifylist_hdb_system()      if defined $pkgenv{sgesap_hdb_system}[0];

    #---------------------------------------------------
    # verify files according to %vf hash
    #---------------------------------------------------
    if ( defined $pkgenv{sgesap_sap_system}    or
         defined $pkgenv{sgesap_db_system}     or
         defined $pkgenv{sgesap_repdb_system}  or
         defined $pkgenv{sgesap_hdb_system}[0]  or
         defined $pkgenv{sgesap_lc_system} ) {

            verify_the_list();
    }


    #---------------------------------------------------
    # verify halib has been configured in SAP profile
    #---------------------------------------------------
    verify_for_halib()      if defined $pkgenv{sgesap_hdb_system}[0];

    #---------------------------------------------------
    # The following use cmexec commands for comparing
    # data from the different cluster nodes. Functions
    # will be called if $extended is enabled
    #---------------------------------------------------

    if ( $extended ) {
        if ( $cmcompare_master ) {
            LogVerbose "Extended-DB-sapverify-tests enabled";

            # extended verify for db
            if ( exists $pkgenv{sgesap_db_system} ) {
                verify_db_storage();
                LogReturnSub(0);
                return (0);
            }

            # extended verify for livecache
            # note: must set sgesap_db_vendor and sgesap_db_system!
            if ( exists $pkgenv{sgesap_lc_system}[0] ) {
                $pkgenv{sgesap_db_vendor}="livecache";
                if ( exists $pkgenv{sgesap_lc_system}[0] ) {
                    $pkgenv{sgesap_db_system}=$pkgenv{sgesap_lc_system}[0];
                } else {
                    LogVerbose "No liveCache system SID found";
                    LogReturnSub(1);
                    return (1);
                }
                verify_db_storage();
            } #sgesap_lc_system

        } #cmcompare_master
    } #extended

    LogReturnSub(0);
    return (0);
}

#===  FUNCTION  =============================================================
#      PURPOSE: Setup the runtime environment for type sap_system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: A package could contain multiple SAP instances (MDS01, MDIS02)
#               For each instance (MDS01...) create a mapping for
#                   instance type=MDS,
#                   instance nr=01 and
#                   instance vhost="virtual hostname"
#               and save them in the %pkgenv record. If the package only
#               contains one virtual hostname but multiple instances use that
#               for all the instance vips
#============================================================================
sub setup_sap_system_runtime {
    LogEnterSub(@_);
    my $i = 0;

    LogVerbose 'Setup found sgesap_sap_system=[' . "$pkgenv{sgesap_sap_system}" . ']';
    # sap system contains one or more sap instances
    if ( exists($pkgenv{sgesap_sap_instance}) && @{ $pkgenv{sgesap_sap_instance} } ) {

        LogDebug 'Setup: @{$pkgenv{sgesap_sap_instance}}=[' . "@{$pkgenv{sgesap_sap_instance}}" . "]";

        foreach my $instance ( @{ $pkgenv{sgesap_sap_instance} } ) {
            LogDebug "instance=[$instance]";

            #  get the SAP instance type and SAP instance number
            my ( $itype, $inr ) = $instance =~ /^([A-Z]+)([0-9][0-9])$/;

            # add them to the %pkgenv hash
            $pkgenv{sgesap_sap_instance_itype}{$instance} = $itype;
            $pkgenv{sgesap_sap_instance_inr}{$instance}   = $inr;

            if ( defined $pkgenv{sgesap_sap_virtual_hostname}[$i] ) {
                $pkgenv{sgesap_sap_instance_ivip}{$instance} = $pkgenv{sgesap_sap_virtual_hostname}[$i];
            }
            else {

                # instance doesn't have a vip configured. fallback and
                # set instance vip to the first ip address of package.
                # use nwrksupp to lookup the hostname
#                my $cmd = "$SGSBIN/nwrksupp ghba  $pkgenv{sg_ip_address}[0] | cut -d \";\" -f 2";
                my $cmd = "$SGSBIN/nwrksupp ghba  $pkgenv{sg_ip_address}[0] | cut -d \\; -f 2";
                LogDebug("Execute cmd=[$cmd]");

                $pkgenv{sgesap_sap_instance_ivip}{$instance} = `$cmd 2>&1` or LogRuntimeError "cmd=[$cmd]";

                chomp($pkgenv{sgesap_sap_instance_ivip}{$instance});
                LogDebug "Found instance_vip [$pkgenv{sgesap_sap_instance_ivip}{$instance}]";

                if (! defined  $pkgenv{sgesap_sap_instance_ivip}{$instance}) {
                    CountAndLogErrorF "$ERRORT Couldn't determine a virtual hostname for SAP instance=[$instance]";
                    LogErrorF " Consider setting SG package config parameter ip_address or sap_virtual_hostname";
                    timestamp_and_exit(2);
                }
            }
            LogVerbose "Setup found instance_type [$pkgenv{sgesap_sap_instance_itype}{$instance}] ",
                              "instance_nr [$pkgenv{sgesap_sap_instance_inr}{$instance}] ",
                              "instance_vip [$pkgenv{sgesap_sap_instance_ivip}{$instance}]";
            $i++;
        }
    }

    #---------------------------------------------------
    # contains one or more sap external instances
    #---------------------------------------------------
    #
    if ( exists($pkgenv{sgesap_sap_ext_instance}) && @{ $pkgenv{sgesap_sap_ext_instance} } ) {

#        LogVerbose 'Skip setup verify: @{$pkgenv{sgesap_sap_ext_instance}}=[' . "@{$pkgenv{sgesap_sap_ext_instance}}" . "]";

    }

    #---------------------------------------------------
    # contains one or more sap infra instances
    #---------------------------------------------------
    if ( exists($pkgenv{sgesap_sap_infra_sw_type}) && @{ $pkgenv{sgesap_sap_infra_sw_type} } ) {

#         LogVerbose 'Skip setup verify: @{$pkgenv{sgesap_sap_infra_sw_type}}=[' . "@{$pkgenv{sgesap_sap_infra_sw_type}}" . "]";

    }

    LogReturnSub(0);
    return (0);
}

#===  FUNCTION  =============================================================
#      PURPOSE: Setup the runtime environment for type db_system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub setup_db_system_runtime {
    LogEnterSub(@_);
    my $sgpkg;

    # Nothing to do

    LogReturnSub(0);
    return (0);

}

#===  FUNCTION  =============================================================
#      PURPOSE: Setup the runtime environment for lc_system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub setup_lc_system_runtime {
    LogEnterSub(@_);
    my $sgpkg;

    # Nothing to do

    LogReturnSub(0);
    return (0);

}
#===  FUNCTION  =============================================================
#      PURPOSE: Setup the runtime environment for repdb_system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub setup_repdb_system_runtime {
    LogEnterSub(@_);
    my $sgpkg;

    if (defined $pkgenv{sgesap_repdb_system}) {
        LogVerbose "Found sgesap_repdb_system repdb_t=[$repdb_t]";

        LogReturnSub(0);
        return (0);
    }

    foreach my $ops (@{$pkgenv{sg_operation_sequence}}) {
            LogDebug "Checking ops=[$ops]";

            if ( $ops =~ /repdbinstance.sh/ ) {
                $repdb_t=1;              # is a Sybase replication instance
                LogVerbose "Found repdbinstance sequence repdb_t=[$repdb_t]";
                last;
            }
    }

    LogReturnSub(0);
    return (0);

}

#===  FUNCTION  =============================================================
#      PURPOSE: Setup the runtime environment for type hdb_system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: A package could contain multiple SAP instances (HDB01, HDB00)
#               For each instance (HDB00...) create a mapping for
#                   instance type=HDB
#                   instance nr=00 and
#                   instance vhost="virtual hostname"
#               and save them in the %pkgenv record. If the package only
#               contains one virtual hostname but multiple instances use that
#               for all the instance vips
#============================================================================
sub setup_hdb_system_runtime {
    LogEnterSub(@_);
    my $i = 0;

    LogVerbose 'Setup found sgesap_hdb_system=[' . "$pkgenv{sgesap_hdb_system}[0]" . ']';

    # sap system contains one or more sap instances
    if ( exists($pkgenv{sgesap_hdb_instance}) && @{ $pkgenv{sgesap_hdb_instance} } ) {

        LogDebug 'Setup: @{$pkgenv{sgesap_hdb_instance}}=[' . "@{$pkgenv{sgesap_hdb_instance}}" . "]";

        foreach my $instance ( @{ $pkgenv{sgesap_hdb_instance} } ) {
            LogDebug "instance=[$instance]";

            #  get the SAP instance type and SAP instance number
            my ( $itype, $inr ) = $instance =~ /^([A-Z]+)([0-9][0-9])$/;

            # add them to the %pkgenv hash
            $pkgenv{sgesap_hdb_instance_itype}{$instance} = $itype;
            $pkgenv{sgesap_hdb_instance_inr}{$instance}   = $inr;
            #TODO
            $pkgenv{sgesap_hdb_instance_ivip}{$instance} = "XXXXXX";
            $pkgenv{sgesap_hdb_instance_ivip}{$instance} = "";

            my $vhost;

            if ( -e $SAPSERVICESF) {
                open $SAPSERVICESFH, '<', $SAPSERVICESF or warn $!;
            } else {
                # report error and return.
                CountAndLogErrorF "File [$SAPSERVICESF] not found";
                LogReturnSub(0);
                return (0);
            }
            my @sapservices = <$SAPSERVICESFH>;
            close($SAPSERVICESFH) or LogRuntimeError "Close failed on : \$!=[$!]";

            LogDebug "array sapservices=[@sapservices]";

            # search for correct instance and save vhost
            foreach (@sapservices) {
                if ( $_ =~ m#pf=\S+_${instance}_(\S+) # ) {
                    $vhost = $1;
                    $pkgenv{sgesap_hdb_instance_ivip}{$instance} = $vhost;
                    last;
                }
            }

            if ($vhost) {
                LogVerbose "Found vhost=[$vhost] for instance=[$instance] in $SAPSERVICESF";
            } else {
                CountAndLogErrorF "No vhost found for instance=[$instance] in $SAPSERVICESF";
            }


#            if ( defined $pkgenv{sgesap_hdb_virtual_hostname}[$i] ) {
#                $pkgenv{sgesap_hdb_instance_ivip}{$instance} = $pkgenv{sgesap_hdb_virtual_hostname}[$i];
#            }
#            else {

#                # instance doesn't have a vip configured. fallback and
#                # set instance vip to the first ip address of package.
#                # use nwrksupp to lookup the hostname
#                my $cmd = "$SGSBIN/nwrksupp ghba  $pkgenv{sg_ip_address}[0] | cut -d \\; -f 2";
#                LogDebug("Execute cmd=[$cmd]");
#
##               $pkgenv{sgesap_hdb_instance_ivip}{$instance} = `$cmd 2>&1` or LogRuntimeError "cmd=[$cmd]";
#                $pkgenv{sgesap_hdb_instance_ivip}{$instance} = `$cmd 2>&1`;
#
#                if ( "$pkgenv{sgesap_hdb_instance_ivip}{$instance}"  ne  "" ) {
#                    LogVerbose "Found hostname=[$pkgenv{sgesap_hdb_instance_ivip}{$instance}] for sg_ip_address=$pkgenv{sg_ip_address}[0]";
#                } else {
#                    LogInfoF "Error determining hostname for $pkgenv{sg_ip_address}[0] - setting to XXXXXX";
#                    $pkgenv{sgesap_hdb_instance_ivip}{$instance}="XXXXXX"
#
#                }
#
#                chomp($pkgenv{sgesap_hdb_instance_ivip}{$instance});
#                LogDebug "Found instance_vip [$pkgenv{sgesap_hdb_instance_ivip}{$instance}]";
#
#                if (! defined  $pkgenv{sgesap_hdb_instance_ivip}{$instance}) {
#                    CountAndLogErrorF "$ERRORT Couldn't determine a virtual hostname for SAP instance=[$instance]";
#                    LogErrorF " Consider setting SG package config parameter ip_address or sap_virtual_hostname";
#                    timestamp_and_exit(2);
#                }

#            }
            LogVerbose "Setup found instance_type [$pkgenv{sgesap_hdb_instance_itype}{$instance}] ",
                              "instance_nr [$pkgenv{sgesap_hdb_instance_inr}{$instance}] ",
                              "instance_vip [$pkgenv{sgesap_hdb_instance_ivip}{$instance}]";
            $i++;
        }
    }

    LogReturnSub(0);
    return (0);
}

#===  FUNCTION  =============================================================
#      PURPOSE: Setup the runtime environment for type sap_ext system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub setup_sap_ext_runtime {
    LogEnterSub(@_);
    my $sgpkg;

    # Nothing to do

    LogReturnSub(0);
    return (0);

}

#===  FUNCTION  =============================================================
#      PURPOSE: Inits the verify list with the filesystems mounted via the
#               package -> sg_fs_directory array of config file
#   PARAMETERS: $component   the verification component type = files_package
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Inits the verify list hash with the filesystems mounted via
#               the package config definitions. This allows the verification
#               of mount points on nodes not running the package
#============================================================================

sub verifylist_init {
    LogEnterSub(@_);
    my ( $component ) = @_;

    LogVerbose "Verify list create: for sg_fs_directory (Package) mounted filesystems";
    foreach $file (@{$pkgenv{sg_fs_directory}}) {
            LogDebug "Adding sg_fs_directory filename=[$file] to vf hash";

            # save verification config in hash
            $vf{'file'}{$file}       = $file;
#            $vf{'alternate'}{$file}  = "";
            $vf{'required'}{$file}   = "m";
            $vf{'adm'}{$file}        = "";
            $vf{'perms'}{$file}      = "";
            $vf{'owner'}{$file}      = "";
            $vf{'group'}{$file}      = "";
            $vf{'vcomponent'}{$file} = $component;

            # don't cluster compare mount directories as
            # they will have different owners unmounted
            $vf{'clucompare'}{$file} = "noclucompare";

            LogVerbose "Verify list init:   filename=[$file]";
        }
    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Adds files to  verifylist, substituing variables in rules
#               file with current values for an SAP instance
#   PARAMETERS: $rules_file     Name of the config file
#               $component      a verify component in the rules files
#               $SID            the current SAP SID
#               $INSTANCE       the current SAP instance
#               $NR             the current SAP instance nr
#               $VIP            the current virtual IP address
#               $FILESPEC       match rules_file line with FILESPEC - sap_infra_sw_type
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Reads the sapverify rules file and replaces any <SID>, <sid>
#               <INST>... with the real values of the current SGeSAP system
#               type e.g <SID> gets replaced with C11...
#============================================================================

sub verifylist_add {
    LogEnterSub(@_);
    my ( $RULESF, $component, $SID, $INST, $NR, $VIP, $FILESPEC ) = @_;

    # Note: INST,NR VIP not always defined depending on sapsystem, dbsystem....
    # set it or later su_and_get_path() will hiccup
    $INST=""    if ! defined $INST;
    $NR=""      if ! defined $NR;

    # so just log the component and SID
    LogVerbose "Verify list create: for vcomponent=[$component] SID=[$SID]";
#    LogDebug "Verify list create: for vcomponent=[$component] SID=[$SID] INST=[$INST] NR=[$NR] VIP=[$VIP] FILESPEC=[$FILESPEC]";

    my $sid = lc($SID);

    # for cluster wide file compares will trigger special case
    # if filename contains physical IP
    my $file_has_PIP;

    open( $RULES_FH, "<", "$RULESF" ) or LogRuntimeError "Couldn't open RULESF=[$RULESF]: \$!=[$!]";

    while (<$RULES_FH>) {

        ## skip uninteresting lines
        chomp;              # no newline
        s/#.*//;            # no comments
        s/^\s+//;           # no leading whitespace
        s/\s+$//;           # no trailing whitespace
        next unless length; # anything left?

        $file_has_PIP=0;
        #---------------------------------------------------
        # interpolate variables in rules specification with
        # package config values. These variables have an
        # SAP style syntax "<variable>"
        #---------------------------------------------------
        if ( $component =~ /^(sap_system|DVEBMGS|D|ASCS|SCS|ERS|REP|J|JD|JC|MDS|MDIS|MDSS|W)$/ ) {
            $_ =~ s/<SID>/$SID/g;
            $_ =~ s/<sid>/$sid/g;
            $_ =~ s/<INST>/$INST/g;
            $_ =~ s/<NR>/$NR/g;
            $_ =~ s/<VIP>/$VIP/g;
            $_ =~ s/<hostname>/$host/g;
            $_ =~ s/<HOSTNAME>/$host/g;

            # map "long" names
            $_ =~ s/<SGeSAP_SAP_SYSTEM>/$SID/g;
            $_ =~ s/<sgesap_sap_system>/$sid/g;
            $_ =~ s/<SGeSAP_SAP_INSTANCE_NAME>/$INST/g;
            $_ =~ s/<SGeSAP_SAP_INSTANCE_NR>/$NR/g;
            $_ =~ s/<SGeSAP_SAP_VIRTUAL_HOSTNAME>/$VIP/g;

        }
        elsif ( $component =~ /^(db_system|oracle|oracle12|db2|maxdb|sybase)$/ ) {
            $_ =~ s/<DBSID>/$SID/g;
            $_ =~ s/<dbsid>/$sid/g;
            $_ =~ s/<hostname>/$host/g;
            $_ =~ s/<HOSTNAME>/$host/g;

            # map "long" names
            $_ =~ s/<SGeSAP_DB_SYSTEM>/$SID/g;
            $_ =~ s/<sgesap_db_system>/$sid/g;

        }
        elsif ( $component =~ /^(lc_system)$/ ) {
            $_ =~ s/<LCSID>/$SID/g;
            $_ =~ s/<lcsid>/$sid/g;
            $_ =~ s/<hostname>/$host/g;
            $_ =~ s/<HOSTNAME>/$host/g;

            # map "long" names
            $_ =~ s/<SGeSAP_LC_SYSTEM>/$SID/g;
            $_ =~ s/<sgesap_lc_system>/$sid/g;
        }
        elsif ( $component =~ /^(repdb_system)$/ ) {
            $_ =~ s/<SGROOT>/$SGROOT/g;
        }
        elsif ( $component =~ /^(hdb_system)$/ ) {
            $_ =~ s/<SID>/$SID/g;
            $_ =~ s/<sid>/$sid/g;
            $_ =~ s/<INST>/$INST/g;
            $_ =~ s/<NR>/$NR/g;
            if ( $_ =~ /<IP>/ ) {
                $file_has_PIP=1;
            }
            $_ =~ s/<IP>/$VIP/g;
            # map "long" names
            $_ =~ s/<SGeSAP_HDB_SYSTEM>/$SID/g;
            $_ =~ s/<sgesap_hdb_system>/$sid/g;
            $_ =~ s/<SGeSAP_HDB_INSTANCE_NAME>/$INST/g;
            $_ =~ s/<SGeSAP_HDB_INSTANCE_NR>/$NR/g;
            $_ =~ s/<SGeSAP_HDB_VIRTUAL_HOSTNAME>/$VIP/g;

        } elsif ( $component =~ /^(sap_infra_sw_type)$/ ) {
            $_ =~ s/<SID>/$SID/g;
            $_ =~ s/<sid>/$sid/g;
            $_ =~ s/<SGeSAP_SAP_SYSTEM>/$SID/g;
            $_ =~ s/<sgesap_sap_system>/$sid/g;
            $_ =~ s/<sap_infra_sw_type>/$FILESPEC/g;

        } elsif ( $component =~ /^(files_sgesap|files_system)$/ ) {
            $_ =~ s/<SGCONF>/$SGCONF/g;
            $_ =~ s/<SGSBIN>/$SGSBIN/g;

        } else {
            LogDebug "component=[$component] is unknown";
        }

        # finally read a line with verification info
        LogDebug "$_";
        my ( $file, $required, $adm, $perms, $owner, $group, $vcomponent ) =
            split( /\s+/, $_ );

        # clear empty double quotes
        $required =~ s/""//g if defined $required;
        $adm =~ s/""//g      if defined $adm;
        $perms =~ s/""//g    if defined $perms;
        $owner =~ s/""//g    if defined $owner;
        $group =~ s/""//g    if defined $group;

        if (!defined $vcomponent) {
         LogInfoF "Unknown vcomponent=[$_]";
         next;
        }

        # sanity check for known verify components in the rules file
        if ( $vcomponent !~ /^(sap_system|DVEBMGS|D|ASCS|SCS|ERS|REP|HDB|J|JD|JC|MDS|MDIS|MDSS|W|db_system|oracle|oracle12|db2|maxdb|sybase|lc_system|repdb_system|hdb_system|sap_ext_instance|sap_infra_sw_type|files_sgesap|files_system|files_package)$/) {
            LogInfoF "Unknown vcomponent=[$vcomponent] for entry [$file] in [$RULESF]";
            next;
        }

        if ( $component eq $vcomponent ) {
            #---------------------------------------------------
            # if $component matches sap_infra_sw_type; it will
            # contain only one type. so get that one and discard
            # the others
            #---------------------------------------------------
            if ( $component eq "sap_infra_sw_type" ) {
                next unless ( $file eq $FILESPEC);
            }
            #---------------------------------------------------
            # Filename only interpolation checks follow
            #---------------------------------------------------
            my $home;

            #---------------------------------------------------
            # match tilde home syntax for file: ~sidadm/
            # match criteria: leading "~" and trailing "/"
            # (?=/) don't use trailing "/" for getpwnam lookup
            #---------------------------------------------------
            if ( $file =~ m#~(.*?(?=/))# ) {
                my $user = $1;

                if ( defined $home_dir{$user} ) {
                    $home = $home_dir{$user};
                } else {
                    $home = ( getpwnam $user )[7];
                    $home_dir{$user} = $home;
                }

                if ( !defined $home ) {

                    # log error and work on next line of rules file
                    CountAndLogErrorF "$ERRORT Can't find passwd entry for user (or user home) for [$user] and filename=[$file]";
                    next;
                }
                else {
                    $file =~ s#~(.*?(?=/))#$home#g;
                }
            }

            #---------------------------------------------------
            # check for files that don't use any absolute path:
            # if $file doesn't contain any "/" (e.g no absolute
            # path) then run a su - "sidadm" and and see if a
            # $path lookup can find the file.
            # might not always succeed since bin directory might
            # be mounted on other node
            #---------------------------------------------------
            if ( $file !~ m#/# ) {
                LogDebug "trying to find path for filename=[$file]";

#                # $INST and $NR not always defined e.g. a DB user
#                $INST="" if ! defined $INST;
#                $NR="" if ! defined $NR;

                if ( my $which = eval{ su_and_get_path( $adm, $file, "$SID", "$INST", "$NR"  ) } ) {
                    LogDebug "Found path which=[$which] for filename=[$file]";
                    $file = $which;
                } else {
                    # log error and continue with next line of rules file
                    LogInfoF "host=[$host] Can't find a PATH to filename=[$file] in environment of user [$adm]";
                    next;
                }
            }

            #---------------------------------------------------
            # check user environment:
            # try to interpolate any application specific
            # variables: <ORACLE_HOME> -> /home/oracle/....
            # by checking the user environment
            #---------------------------------------------------
            if ( $file =~ m#<(.*?(?=>))# ) {
                my $envmatch = $1;

                LogDebug "";
                LogDebug "Trying to find environment for filename=[$file]";

                if ( defined $home_dir{$adm} ) {
                    $home = $home_dir{$adm};
                } else {
                    $home = ( getpwnam $adm )[7];
                    $home_dir{$adm} = $home;
                }
                if ( !defined $home ) {

                    # log error and work on next line of rules file
                    LogInfoF "Can't find passwd entry for user (or user home) for [$adm] and filename=[$file]";
                    next;
                }
                else {
                    LogDebug "Got home directory home=[$home] for adm=[$adm]";
                }

                # if home directory is part of a failover fileystem
                # then we can't get any environment values
                if ( is_fs_mounted_by_package( $home ) ) {
                    if (  ($PackageDown)  or  (!$PackageOnThisNode) ) {
                        LogVerbose "Info: Skipping verify for this file - Can't interpolate user environment variable for rules filename=[$file]";
                        LogVerbose " Supporting Info: Home directory home=[$home] for adm=[$adm] is mounted by (Package) and not available on this node";
                        next;
                    }
                }
                    LogDebug "Interpolation required for [$envmatch] from file = [$file] ";
#                    if ( my $envvar =  su_and_get_env( $adm, $envmatch )  ) {

                my $envvar;
                if ( defined $userenv{$envmatch} ) {
                    $envvar = $userenv{$envmatch};
                } else {
                    $envvar = su_and_get_env( $adm, $envmatch);
                    $userenv{$envmatch} = $envvar;
                }

                if ( defined $envvar ) {
                        LogDebug "File filename=[$file] will be expanded with envvar=[$envvar] ";
                        $file =~ s#<(.*?>)#$envvar#g;
                        LogDebug "File filename=[$file] now expanded with envvar=[$envvar] ";
                    }
                    else {

                        # log error and work on next line of rules file
                        LogInfoF "Couldn't determine environment variable [$1] for user=[$adm]";
                        LogInfoF " Supporting Info: Verification line=[$file] from rules filename=[$RULESF]";
                        next;
                    }
            }
            #---------------------------------------------------
            # Done checking user environment
            #---------------------------------------------------

            # save verification config in hash
            $vf{'file'}{$file}       = $file;
#            $vf{'alternate'}{$file}  = $alternate;
            $vf{'required'}{$file}   = $required;
            $vf{'adm'}{$file}        = $adm;
            $vf{'perms'}{$file}      = $perms;
            $vf{'owner'}{$file}      = $owner;
            $vf{'group'}{$file}      = $group;
            $vf{'vcomponent'}{$file} = $vcomponent;
#            $vf{'clucompare'}{$file} = $clucompare;

            if ( $file_has_PIP ) {
                $vf{'pip'}{$file}=1;
                $file_has_PIP=0;
            }
            LogVerbose "Verify list add:    filename=[$file]";
        }
    }
    close($RULES_FH) or LogRuntimeError "Couldn't close failed RULESF=[$RULESF]: \$!=[$!]";

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: get a parameter out of an user environment
#   PARAMETERS: $adm    the user
#               $param  the environment parameter
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Try to find ORACLE_HOME for example
#============================================================================
sub su_and_get_env {
    LogEnterSub(@_);
    my ( $adm, $param ) = @_;
    my $foundenv;

    my $cmd = qq[su - $adm -c "env" |];
    LogDebug("Execute cmd=[$cmd]");

    open my $fh, "$cmd" or LogRuntimeError "Could not run cmd=[$cmd]: \$!=[$!]";

    while (<$fh>) {
        if ( $_ =~ m#^$param=(\S+)$# ) {
            $foundenv = $1;
            last;
        }
    }
    close($fh) or LogRuntimeError "Close failed on cmd=[$cmd]: \$!=[$!]";

    if ( !defined $foundenv ) {
        LogVerbose( "Environment \$param=[$param] not defined" );
        LogReturnSub();
        return;
    }

    LogDebug( "Environment \$param=[$param] has value of \$foundenv=[$foundenv]");
    LogReturnSub($foundenv);
    return ($foundenv);

}

#===  FUNCTION  =============================================================
#      PURPOSE: su and try to get the complete path of a command
#   PARAMETERS: $adm    the user
#               $param  file without absolute path
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: try to find a binary in the user path by calling the type...
#               command and/or checking some default sap directories
#============================================================================
sub su_and_get_path {
    LogEnterSub(@_);
    my ( $adm, $param, $sid, $inst, $nr ) = @_;
    my ( $rest, $path, $cmd );

    #---------------------------------------------------
    # determine platform so path can be completed
    #---------------------------------------------------
    $cmd="uname -m";
    my $platform        =  `$cmd` or croak "Error: cmd=[$cmd] failed";
    chomp $platform;

    my @sap_binary_dirs = (
        "/usr/sap/$sid/$inst$nr/exe",         "/usr/sap/$sid/SYS/exe/run",
        "/usr/sap/$sid/SYS/exe/runU",         "/usr/sap/$sid/SYS/exe/ctrun",
        "/usr/sap/$sid/SYS/exe/uc/$platform", "/usr/sap/$sid/SYS/exe/nuc/$platform",
        "/usr/sap/hostctrl/exe",              "/usr/sap/$sid/sapwebdisp/sapwebdisp",
    );


    #---------------------------------------------------
    #---------------------------------------------------

    #NOTE: 2>/dev/null required or else HPUX interactively queries TERM type
    #    my $cmd = qq[su - $adm -c "which $param" 2>/dev/null];
    $cmd = qq[su - $adm -c "which $param" 2>/dev/null];
    LogDebug("Execute cmd=[$cmd]");

    my @results = `$cmd`;
    my $rc      = $?;

    LogDebug "which $param result: rc=[$rc], results=[@results]";

    # example string returned:
    #   "mount is $SGSBIN/mount"
    #   "ls is a tracked alias for /usr/bin/ls"

    if ( $rc == 0 ) {

        for (@results) {
            chomp;
            if ( ( $^O eq "hpux" ) && ($_ =~ m#($param) is.* ([^ ]+\1)$# )) {
                $path = $2;
                LogDebug "path=[$path]";
                last;
              # match absolute path and ignore any noise in output
            } elsif ((  $^O eq "linux" ) && ($_ =~ m#(^/[^ ]+$param)$#)) {
                $path = $_;
                LogDebug "path=[$path]";
                last;
            }
        }
    }
    else {

        #    if ( !defined($path) ) {
        for my $dir (@sap_binary_dirs) {
            LogDebug "scanning default sap dirs: $dir";
            if ( -e "$dir/$param" ) {
                $path = "$dir/$param";
                LogDebug "default sap dir: found path=[$path]";
                last;
            }
        }
    }


    LogReturnSub($path);
    return defined $path ? $path : undef;

}


#===  FUNCTION  =============================================================
#      PURPOSE: Verify permissions, owner.... against against rule
#   PARAMETERS: $vcomponent     component to verfiy in the vf hash
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub verify_files_for_component {
    LogEnterSub(@_);
    my $vcomponent = shift;

    my $package_down;
    my $package_mountedfs;
    my $packageonthisnode;

    LogVerbose "Running verification for vcomponent=[$vcomponent]";

    foreach my $filename ( keys %{ $vf{'vcomponent'} } ) {

        $filecount++;

        if ( $vcomponent ne "ALL" ) {
            next unless ( $vf{'vcomponent'}{$filename} eq $vcomponent );
        }

        LogVerbose "";
        LogVerbose "Verification for filename=[$filename]";

        #---------------------------------------------------
        # if package is unowned == package is down, then the
        # fileystems configured with package are not mounted.
        # NFS and local file sytems should be accessible
        # though
        # if filename is configured in the sg_fs_directory
        # of the package configuration test if the mount
        # directory exists
        #---------------------------------------------------

        $package_down      = $PackageDown;
        $package_mountedfs = is_fs_mounted_by_package($filename);

        # if filename doesn't reside on a package mounted fs it
        #  is must be a local copy on each node
        if ( $package_mountedfs ) {
            $vf{'distrb'}{$filename} = 0;
        } else {
            $vf{'distrb'}{$filename} = 1;
        }

        LogDebug "package_down=[$package_down], package_mountedfs=[$package_mountedfs]";
        if ( $package_down && $package_mountedfs ) {
            LogVerbose "Skipping verify for this file - "
                . "Package is unowned/down and file is from an unmounted (Package) filesystem";
            next;
        }

        LogDebug "PackageOnThisNode=[$PackageOnThisNode], package_mountedfs=[$package_mountedfs]";
        if ( !$PackageOnThisNode && $package_mountedfs ) {
            LogVerbose "Skipping verify for this file - "
                . "Package not running on this node and file is from a (Package) filesystem";
            next;
        }

        my $found = call_func_with_timeout_sig( $fasttimeout, \&check_for_file, $filename );

        LogDebug "filename=[$filename]  $vf{'vcomponent'}{$filename}  vcomponent=[$vcomponent] found=[$found]";

        if ($found) {
            LogVerbose "Running local file verify - permissions, owner...";
            call_func_with_timeout_sig( $fasttimeout, \&VerifyFile, $filename );

        }
        else {

            if ( $vf{'required'}{$filename} eq "o" ) {
                LogVerbose "Verify-Info: filename=[$filename] not found but file is optional";
            }
            else {

                # filename wasn't found
                CountAndLogErrorF "$ERRORT host=[$host] ", "filename=[$filename] not found";
                check_for_missing_mount($filename);
            }
        }

    }
    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: Verify permissions, owner.... against against rule
#   PARAMETERS: $vcomponent     component to verfiy in the vf hash
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub get_distributed_files {
    LogEnterSub(@_);
    my $vcomponent;
    # vcomponenet , $sgpkg is global
    ($vcomponent, $sgpkg) = @_;

    my $package_down;
    my $package_mountedfs;
    my $packageonthisnode;
    my ($f, $l, $s);

    LogVerbose "Getting distributed files in vcomponent=[$vcomponent] for sgpkg=[$sgpkg]";

    # init based on $sgpkg
    analyze_package_runtime_env($sgpkg)    if defined $sgpkg;
    cmgetpkgenv($sgpkg);
    cmviewcl_p_package($sgpkg);
    is_package_supported($sgpkg);
    is_package_on_this_node($sgpkg);
    is_package_down($sgpkg);
    cmget_switchnodelist($sgpkg);
    cmget_cmcompare_action($sgpkg);

    #---------------------------------------------------
    # setup the runtime environments
    #---------------------------------------------------
    setup_sap_system_runtime()    if defined $pkgenv{sgesap_sap_system};
    setup_db_system_runtime()     if defined $pkgenv{sgesap_db_system};
    setup_lc_system_runtime()     if defined $pkgenv{sgesap_lc_system};

    # if operation_sequence $SGCONF/scripts/sgesap/repdbinstance.sh
    # exists then it a sybase rep db instance
#    setup_repdb_system_runtime()  if defined $pkgenv{sgesap_repdb_system};
    setup_repdb_system_runtime();

    setup_hdb_system_runtime()    if defined $pkgenv{sgesap_hdb_system}[0];

    #---------------------------------------------------
    # init a list of files to verify with sg_fs_directory
    # package mount points. This forces verification of
    # package mount points on all nodes.
    #---------------------------------------------------
    verifylist_init("files_package");

    #---------------------------------------------------
    # append list of files to verify to %vf hash
    #---------------------------------------------------
    create_verifylist_sap_system()    if defined $pkgenv{sgesap_sap_system};
    create_verifylist_db_system()     if defined $pkgenv{sgesap_db_system};
    create_verifylist_lc_system()     if defined $pkgenv{sgesap_lc_system};
#    create_verifylist_repdb_system()    if defined $pkgenv{sgesap_repdb_system};
    create_verifylist_repdb_system()  if $repdb_t==1;
    create_verifylist_hdb_system()    if defined $pkgenv{sgesap_hdb_system}[0];

    create_dstrb_files_hash($vcomponent) ;

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: List stats on files of type "distributed"
#   PARAMETERS: @$dstrb_files contains list of dstrb files
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: $dstrb_files contains an array reference to list of files
#               that are distributed over the cluster nodes. dstb_file e.g. each
#               node contains a copy of the file. Prints out a string "DSTRB,
#               filename, file was found, file execution mode, uid, gid,
#               softlink" for each file. DSTRB is used to identify file line
#               if debug statements are turned on in the output stream
#
#      EXAMPLE:
#               This function is called by get_remote_dstrb_files()
#
#               $SGSBIN/cmexec -n vmlx-0-73 \
#               $SGSBIN/sapverify.pl \
#               -dstrb $SGCONF/monitors/sgesap/sapms.mon \
#               -dstrb $SGCONF/scripts/sgesap/sapinstance.sh \
#               -dsrtb /xxx
#
# DSTRB,/usr/local/cmcluster/conf/monitors/sgesap/sapms.mon,1,0555,1,1,f,,
# DSTRB,/usr/local/cmcluster/conf/scripts/sgesap/sapinstance.sh,1,0555,2,2,f,,
# DSTRB,/xxx,0,,,,,,

#============================================================================
sub list_distributed_files {
    LogEnterSub(@_);
    my ($dstrb_files)=@_;

    my ($f, $l, $s, $found, $act_mode, $uid, $gid, $ftype);
    foreach my $filename ( @$dstrb_files ) {

        # print defaults even if file not found
        $found=0;       # file was found
        $act_mode='';   # file mode bits
        $uid='';        # uid for file
        $gid='';        # gid for file
        $ftype='';      # f=file, d=directory
        $l='';          # soft link

        # use timer - access might time out for bad NFS mounts....
        my $found = call_func_with_timeout_sig( $fasttimeout, \&check_for_file, $filename );

        if ($found) {
            $f = $filename;
            $l = "" if !defined( $l = readlink($f) );
            $s = lstat($f);
            $uid=$s->uid;
            $gid=$s->gid;

            $ftype='f';
            $ftype='d' if (-d $filename);
            $act_mode = sprintf( "0%o", $s->mode & 0777 );
        }

        #---------------------------------------------------
        # Prepend "DSTRB" to output string so output can be
        # matched even if debugging statements schould be in
        # the stream
        #---------------------------------------------------
        printf( "DSTRB,%s,%s,%s,%s,%s,%s,%s,\n",$filename,$found,$act_mode, $uid, $gid, $ftype, $l) if defined $dstrbfiles;
    }
    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: create a hash containing those files which are distributed /
#               are copies amongst cluster nodes
#   PARAMETERS: vcomponent specify rule verify component to use - usually ALL
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#               distributed files are not configured in a SG  package e.g.
#               they reside as a copy on each node
#============================================================================
sub create_dstrb_files_hash {
    LogEnterSub(@_);
    my ($vcomponent) = @_;

    my $package_down;
    my $package_mountedfs;
    my $packageonthisnode;
    my ($f, $l, $s, $ftype);

    LogVerbose "Analyzing SGeSAP rules for cluster distributed-files";
    foreach my $filename ( keys %{ $vf{'vcomponent'} } ) {


        if ( $vcomponent ne "ALL" ) {
            next unless ( $vf{'vcomponent'}{$filename} eq $vcomponent );
        }

        $package_mountedfs = is_fs_mounted_by_package($filename);

        # save info if filename is distributed
        if ( $package_mountedfs ) {
            $vf{'dstrb'}{$filename} = 0;
        } else {
            $vf{'dstrb'}{$filename} = 1;
        }

        # here: only verify files which are distributed over the cluster nodes
        next unless   $vf{'dstrb'}{"$filename"} == 1;

        LogVerbose "Filename=[$filename] is a cluster distributed-file";

        #increment filecount of files that were verified
        $filecount++;

        my $found = call_func_with_timeout_sig( $fasttimeout, \&check_for_file, $filename );

        if ($found) {
            $f = $filename;
            $l = "" if !defined( $l = readlink($f) );
            $s = lstat($f);

            my $ftype='f';

            $ftype='d' if (-d $filename);

            #---------------------------------------------------
            # Prepend "DISTRB" to output string so output can be
            # matched even if debugging statements schould be in
            # the stream
            #---------------------------------------------------

            my $act_mode = sprintf( "0%o", $s->mode & 0777 );
            # only print if $dstrbfiles is definded, e.g called for a remote node
            printf( "DSTRB,%s,%s,%s,%s,%s,%s,%s\n",$filename,$found,$act_mode, $s->uid, $s->gid, $ftype, $l) if defined $dstrbfiles;
            $dstrb_file{$filename}{'found'} = $found;
            $dstrb_file{$filename}{'mode'} = $act_mode;
            $dstrb_file{$filename}{'uid'} =  $s->uid;
            $dstrb_file{$filename}{'gid'} =  $s->gid;
            $dstrb_file{$filename}{'ftype'} =  $ftype;
            $dstrb_file{$filename}{'link'} = $l;

        }
        else {
            # filename was not found on the system
            printf( "DSTRB,%s,%s,,,,,\n", $filename,$found) if defined $dstrbfiles;
            $dstrb_file{$filename}{'found'} = $found;
        }
    }

    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%dstrb_file ], ["dstrb_file"] ));

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: create distributed files hash for a remote node / $sgpkg
#   PARAMETERS: remotenode: node to run sapverify on
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: $sgpkg is a global
#============================================================================
sub get_remote_dstrb_files {
    LogEnterSub(@_);
    my ($remotenode) = @_;

    my ($MATCH, $filename, $found, $mode, $uid, $gid, $link, $ftype);
    my $cmd;
    my $fh;
    my $exit_status;
    my $line;

    # create sapverify string with all distributed files: -dstrb f1 -dstrb f2 ...
    @dstrb_file_list =" -dstrb " . join(' -dstrb ',keys %dstrb_file);

    LogVerbose "Getting distributed files/dirs from remote cluster node=[$remotenode]";
    $cmd = "$SGSBIN/cmexec -n $remotenode -t $timeout $SGSBIN/sapverify.pl @dstrb_file_list 2>&1 |";
    LogVerbose("Execute cmd=[$cmd]");

    # $cmd can get fairly long - just print the first 132 chars of the $cmd string
    # LogVerbose("Execute \$cmd=[" .  substr ($cmd, 0, 132 ) . "....]");

    open( $fh, $cmd ) or LogRuntimeError "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
    while (<$fh>) {
        $line = $_;
        chomp $line;
        LogDebug("Raw DSTRB returned=[$line]");
#        ( $MATCH, $filename, $found, $mode, $uid, $gid, $link ) = split( /,/, $line );
        ( $MATCH, $filename, $found, $mode, $uid, $gid, $ftype, $link ) = split( /,/, $line );

        if (  (defined $MATCH) && ($MATCH =~ m#^DSTRB$#) ) {

            $remote_dstrb_file{$filename}{'found'} = $found;
            $remote_dstrb_file{$filename}{'mode'} = $mode;
            $remote_dstrb_file{$filename}{'uid'} =  $uid;
            $remote_dstrb_file{$filename}{'gid'} =  $gid;
            $remote_dstrb_file{$filename}{'ftype'} = $ftype;
            $remote_dstrb_file{$filename}{'link'} = $link;
        }
    }
    close($fh) or LogRuntimeError "Close failed for [$cmd]: \$?=[$?] \$!=[$!] line=[$line]";

    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%remote_dstrb_file ], ["remote_dstrb_file"] ));

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Verify distributed files (e.g. a copy of files residing on
#               each cluster node
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub verify_dstrb_files {
    LogEnterSub(@_);

    # scan and create the distributed files hash for local node
    create_dstrb_files_hash("ALL");

    LogVerbose "Running distributed file verification for switchnodes=[@switchnodes]";
    foreach my $remotenode (@switchnodes) {

        if ( "$host" ne "$remotenode" ) {

            get_remote_dstrb_files($remotenode);
            twonode_dstrb_file_compare($host, $remotenode);
        }
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Compare local distributed files hash against remote
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: logs and error if a difference was found
#               if files are available on local and remote node they might
#               still have different permissions.... Compare local with
#               remote settings
#============================================================================
sub twonode_dstrb_file_compare {
    LogEnterSub(@_);
    my ( $localnode, $remotenode) = @_;

    foreach my $file ( keys %{dstrb_file} ) {

        if ( $dstrb_file{$file}{found} == 0  ) {
            # check if file is optional or required
            if ( (defined $vf{'required'}{$file}) and ($vf{'required'}{$file} eq "o") ) {
                LogInfo "(sapverify): INFO: host=[$localnode] distibuted file=[$file] not found but file is optional";
            }

            # Don't log any error for the local file as this should have
            # already been found/flagged earlier in the sapverify stream.
            LogDebug"On local node=[$localnode] distributed file=[$file] missing";

            # if local file is missing we can't compare gid, uid... so check if
            # remote also missing and log. then just get next file in loop (no comparison)
            if ( (defined $remote_dstrb_file{$file} ) && ($remote_dstrb_file{$file}{found} == 0)  ) {
                if ( (defined $vf{'required'}{$file}) and ($vf{'required'}{$file} eq "o") ) {
                    LogInfo "(sapverify): INFO: host=[$remotenode] distibuted file=[$file] not found, but file is optional";
                }
                else {
                    CountAndLogErrorF "$ERRORT remote host=[$remotenode] distributed file=[$file] missing";
                }
            }
            next;
        }


        if ( (defined $remote_dstrb_file{$file} ) && ($remote_dstrb_file{$file}{found} == 1)  ) {

            if ( "$dstrb_file{$file}{uid}" ne "$remote_dstrb_file{$file}{uid}" ) {
                LogInfoF "host=[$remotenode] distributed file=[$file] "
                    . "local uid=[$dstrb_file{$file}{uid}] doesn't match remote uid=[$remote_dstrb_file{$file}{uid}]";
            }
            if ( "$dstrb_file{$file}{gid}" ne "$remote_dstrb_file{$file}{gid}" ) {
                LogInfoF "host=[$remotenode] distributed file=[$file] "
                    . "local gid=[$dstrb_file{$file}{gid}] doesn't match remote gid=[$remote_dstrb_file{$file}{gid}]";
            }
            if ( "$dstrb_file{$file}{mode}" ne "$remote_dstrb_file{$file}{mode}" ) {
                LogInfoF "host=[$remotenode] distributed file=[$file] "
                    . "local mode=[$dstrb_file{$file}{mode}] doesn't match remote mode=[$remote_dstrb_file{$file}{mode}]";
            }
            if ( "$dstrb_file{$file}{link}" ne "$remote_dstrb_file{$file}{link}" ) {
                LogInfoF "host=[$remotenode] distributed file=[$file] "
                    . "local link=[$dstrb_file{$file}{link}] doesn't match remote link=[$remote_dstrb_file{$file}{link}]";
            }

            LogVerbose "SUCCESS: host=[$remotenode] verification for distributed file=[$file] passed";
        }
        else {

            # check if file is optional or required
            if ( (defined $vf{'required'}{$file}) and ($vf{'required'}{$file} eq "o") ) {
                LogVerbose "Verify-Info: host=[$remotenode] distibuted file=[$file] not found, but file is optional";

            } elsif ( (defined $vf{'pip'}{$file}) and ($vf{'pip'}{$file} == 1) ) {
                # special case for HANA: after installation of a primary/secondary,
                # the SAP profiles contain a physical hostname and not a virtual.
                # therefore no local vs remote compare
                LogVerbose "(sapverify): INFO: host=[$remotenode] distibuted file=[$file] skipping dstrb verification as filename contains a physical hostname";

            } else {
                CountAndLogErrorF "$ERRORT remote host=[$remotenode] distributed file=[$file] missing";
            }

        }
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Compare local sg mount directory against remote
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: logs and error if a difference was found
#============================================================================
sub twonode_dstrb_dir_compare {
    LogEnterSub(@_);
    my ( $localnode, $remotenode) = @_;

    foreach my $file ( keys %{dstrb_file} ) {

        LogDebug "Verification for distributed mount directory =[$file]";
        if ( (defined $remote_dstrb_file{$file} ) && ($remote_dstrb_file{$file}{found} == 1)  ) {

            if ( "$dstrb_file{$file}{ftype}" ne "d" ) {
                LogInfoF "host=[$localnode] file=[$file] "
                    . "ftype=[$dstrb_file{$file}{ftype}] is not a directory for mounting ";
            } elsif ( "$dstrb_file{$file}{ftype}" ne "$remote_dstrb_file{$file}{ftype}" ) {
                LogInfoF "host=[$remotenode] distributed file=[$file] "
                    . "ftype=[$dstrb_file{$file}{ftype}] doesn't match remote ftype=[$remote_dstrb_file{$file}{ftype}]";
            } else {

                LogVerbose "SUCCESS: host=[$remotenode] package mount point=[$file] exists";
            }
        }
        else {
                CountAndLogErrorF "$ERRORT remote host=[$remotenode] package mount point=[$file] missing";
        }
    }


    $dumpdata && LogVerbose( Data::Dumper->Dump( [ \%pkgenv ], ["pkgenv"] ) );

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: Verify permissions, owner.... against specs
#   PARAMETERS: vcomponent  the compenet from rules file to verify
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub verify_files_for_component_noclu {
    LogEnterSub(@_);
    my $vcomponent = shift;

    foreach my $filename ( keys %{ $vf{'vcomponent'} } ) {

        next unless ( $vf{'vcomponent'}{$filename} eq $vcomponent );

        # if package is unowned == package is down, then the fileystems
        # configured with package are not mounted. NFS and local
        # file sytems should be accessible though

        LogVerbose "";
        LogVerbose "Running verification for filename=[$filename]";

        $filecount++;

        my $found = check_for_file($filename);
        LogDebug "filename=[$filename]  $vf{'vcomponent'}{$filename}  vcomponent=[$vcomponent] found=[$found]";

        if ($found) {
            VerifyFile($filename);
        }
        else {

            # filename wasn't found
            CountAndLogErrorF "$ERRORT host=[$host]", "filename=[$filename] not found";
        #    check_for_missing_mount($filename);
        }

    }
    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE: Verify permissions, owner.... against rules specs
#   PARAMETERS: $filename   name of file to verify
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub VerifyFile {
    LogEnterSub(@_);
    my ($filename) = @_;
    my $file       = $filename;
    my $verify_err = 0;

    #---------------------------------------------------
    #  open directory so it gets mounted if autofs
    #  is configured
    #---------------------------------------------------
    if ( -d $file ) {
        LogDebug "filename=[$file] is a directory";
        opendir( my $dh, $file ) || die "cannot open $file dir: $!";
    }


    #---------------------------------------------------
    # if $file is a link to a filename from the rules list,
    # then skip checks as the real file will explicitly
    # checked later - avoids ping-pong errors if owner
    # is different for link and $file and both are
    # on the rules verify list
    #---------------------------------------------------
    if ( defined( my $link = readlink($file) ) ) {
        LogVerbose "Filename=[$file] is a symbolic link to link=[$link]";
        if ( exists $vf{'file'}{$link}  ){
            LogVerbose "Skipping  check - found link=[$link] on verify list - will be verified then";
            LogVerbose "Passed - file verify" if $verify_err == 0;
            LogReturnSub();
            return;
        }
    }

    my ( $gid, $uid );
    LogDebug "filename=[$file]";
    my $info = stat($file);
#    my $info = lstat($file);
    if ($info) {

        #---------------------------------------------------
        # check file modes: actual mode against target mode
        # for permissions.
        #---------------------------------------------------
        my $act_mode = sprintf( "0%o", $info->mode & 07777 );
        my $tar_mode = $vf{'perms'}{$filename};

        if ( ( $tar_mode ne "" ) && ( $act_mode ne "$tar_mode" ) ) {
            LogInfoF "host=[$host] ",
                "filename=[$file] - current mode=[$act_mode] doesn't match suggested=[$tar_mode]";
            $verify_err = 1;
        }
        else {
            LogVerbose "Skipping mode verification  - mode attribute not set in rule" if $tar_mode eq "";
        }

        #---------------------------------------------------
        # check file owner
        #---------------------------------------------------
        if ( $vf{'owner'}{$filename} ne "" ) {
            $uid = getpwnam( $vf{'owner'}{$filename} );

            # check if uid exists
            if ( !defined($uid) ) {
                CountAndLogErrorF "$ERRORT host=[$host] ",
                    "filename=[$filename] - Can't find user=[$vf{'owner'}{$filename}] on the system";
                $verify_err = 1;
            }
            else {

                # check uid of file
                if ( $info->uid != $uid ) {
                    my $tmp_str = sprintf "filename=[%s] - current uid=[%s] doesn't match suggested=[%s] for user=[%s]",
                        $file, $info->uid, $uid, $vf{'owner'}{$filename};
                    LogInfoF "host=[$host] ", "$tmp_str";
                    $verify_err = 1;
                }

            }
        }
        else {
            LogVerbose "Skipping owner verification  - owner attribute not set in rule";
        }

        #---------------------------------------------------
        # check group
        #---------------------------------------------------
        if ( $vf{'group'}{$filename} ne "" ) {

            $gid = getgrnam( $vf{'group'}{$filename} );

            # check if gid exists
            if ( !defined($gid) ) {
                LogInfoF "host=[$host] ",
                    "filename=[$filename] - Can't find group=[$vf{'group'}{$filename}] on the system";
                $verify_err = 1;
            }
            else {

                #check gid of file
                if ( $info->gid != $gid ) {
                    my $tmp_str = sprintf "filename=[%s] - current gid [%s] doesn't match suggested=[%s] for group=[%s]",
                        $file, $info->gid, $gid, $vf{'group'}{$filename};
                    LogInfoF "host=[$host] ", "$tmp_str";
                    $verify_err = 1;
                }
            }
        }
        else {
            LogVerbose "Skipping group verification  - group attribute not set in rule";
        }

        #---------------------------------------------------
        # Done checking file attributes
        #---------------------------------------------------
    }
    else {

        #---------------------------------------------------
        # Couldn't check attibutes on file
        #---------------------------------------------------
        LogInfoF "host=[$host] ", "Couldn't stat filename=[$file]: $!";
    }

    LogVerbose "Passed - file verify" if $verify_err == 0;
    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Create checksums  for a list of files
#   PARAMETERS: $vcomponent     - only create checksums for a specific component
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: For a list of files run a chcksum on files flagged with
#               clucompare. Uses perl checksum. Not in use. Using cmcompare
#               instead
#============================================================================

sub ClucompFileListChksum {
    LogEnterSub(@_);
    my $vcomponent = shift;

    foreach my $file ( keys %{ $vf{'vcomponent'} } ) {

        # only verify files from verify component $vcomponent
        next unless ( $vf{'clucompare'}{$file} eq "clucompare" );

        open( FILE, $file ) or LogRuntimeError "ClucompFileListChksum(): Can't open '$file': $!";
        binmode(FILE);

        my $chksum = Digest::MD5->new->addfile(*FILE)->hexdigest;
        LogVerbose "$file $chksum";
        $loc_chksum{$file} = $chksum;
        close(FILE) or LogRuntimeError "Close failed for [$file]: \$!=[$!]";

    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: This function checks if cmcompare() should be executed
#               on the current node
#   PARAMETERS: $sgpkg  Name of the serviceguard package
#      RETURNS: 1 if cmcompare is to be run
#               0 if cmcompare will not be run
#  DESCRIPTION: None
#     COMMENTS:
#               Scenario1: an administrator executes cmcheckconf or cmapplyconf
#               on one cluster node. The cluster software will submit a
#               sapverify process on each node eligible node for running the
#               package.
#               Scenario2: sapverify regulary get executed via a SGeSAP
#               package monitor
#
#               cmcompare generates file checksums for all nodes in the
#               argument:  "cmcompare -n node1 -n node2 file" and returns
#               if the file differs/doesn't differ on the nodes in the
#               argument.
#
#               thus cmcompare should only be run from one (cmcompare master)
#               sapverify process and not all sapverify processes running
#               in the cluster so that file checksums are not generated
#               multiple times.
#
#               This function tries to determine the master for this task.
#               In the context of sapverify running on each node, cmcompare
#               should only get executed on the current node if:
#               1) The package is in state running: in this case cmcompare
#                  only gets executed on current node if it is running the
#                  package.
#                  or
#               2) The package is in state unowned/down and the commands
#                  cmcheckconf/ cmapplyconf are getting executed on this node
#
#============================================================================

sub cmget_cmcompare_action {
    LogEnterSub(@_);
    my ($sgpkg) = @_;
    my @results;
    my $found_cmcompare_master;
    my $CMD;

    # node will not run cmcompare commands
    $cmcompare_master = 0;

    # master node is the node running the package
    if ($PackageOnThisNode) {
        $cmcompare_master = 1;
#        LogVerbose "cmcompare_master=[$cmcompare_master] - package running on this node";
        LogVerbose "Running as cmcompare master - package running on this node";
        LogReturnSub(1);
        return (1);
    }

    # package is down/unowned - try to determine a cmcompare master
    if ( ( $sg_op_code eq "validate" ) && is_package_down($sgpkg) ) {

        # "master" node is the node running cmapplyconf/cmcheckconf
        $CMD = "UNIX95=1 ps -exo comm" if ( $^O eq "hpux" );
        $CMD = "UNIX95=1 ps -eo cmd"   if ( $^O eq "linux" );
        LogDebug "Running CMD=[$CMD]";

        @results = `$CMD` or LogRuntimeError "CMD=[$CMD] \$!=[$!]";
        LogDebug "host=[$host]", "[@results]";

        $found_cmcompare_master = grep ( /(cmapplyconf|cmcheckconf)\s+.*$sgpkg/, @results );
        LogDebug "Found_cmcompare_master=[$found_cmcompare_master]";

        if ($found_cmcompare_master) {
            $cmcompare_master = 1;
#            LogVerbose "cmcompare_master=[$cmcompare_master] - package unowned and cmapplyconf/cmcheckconf running on this node";
            LogVerbose "Running as cmcompare master - package unowned and cmapplyconf/cmcheckconf running on this node";
            LogReturnSub(1);
            return (1);
        }
    }

#    LogVerbose "cmcompare_master=[$cmcompare_master] - package not running on this node";
    LogVerbose "Running as cmcompare slave - package not running on this node";
    LogReturnSub(0);
    return (0);

}


#===  FUNCTION  =============================================================
#      PURPOSE: Run a cmcompare for one file and for all eligible nodes of a
#               package
#   PARAMETERS: $file   name of file to compare
#      RETURNS: 1       failure - found a difference
#               0       success - no difference found
#  DESCRIPTION: None
#     COMMENTS: Run the cmcompare command on one file
#               this is obsolete as cmcompare not available any more
#============================================================================

sub cmcompare_file {
    LogEnterSub(@_);
    my ($file, $rules_required_flag) = @_;

    my @results;
    my $line;
    my $CMD;

    # only verify files from component clucompare (cluster compare) set
    if ( $vf{'clucompare'}{$file} eq "clucompare" ) {

        #---------------------------------------------------
        #  For directories: run "cmcompare -m content" so if
        #  a autofs it gets mounted on all cluster nodes
        #  (else permissions, owner...) will be wrong
        #---------------------------------------------------
        if ( -d $file ) {
            LogDebug "cmcompare filename=[$file] is a directory";

            # run cmcompare command; needs 2>&1 for stderr
            $CMD = "$SGSBIN/cmcompare @switchnodelist -v -m content $file 2>&1";
            @results = `$CMD` or LogRuntimeError "Executing CMD=[$CMD] \$!=[$!]";

        }
        LogDebug "cmcompare filename=[$file]";

        # run cmcompare command; needs 2>&1 for stderr
        #        $CMD = "$SGSBIN/cmcompare @switchnodelist -v -m owner -m perm $file 2>&1";
        $CMD = "$SGSBIN/cmcompare @switchnodelist -v -m owner -m perm -m content $file 2>&1";
        LogVerbose "Running CMD=[$CMD]";

        @results = `$CMD` or LogRuntimeError "Executing CMD=[$CMD] \$!=[$!]";
        my $rc = $?;

        LogDebug "cmcompare result: rc=[$rc], results=[@results]";

        #---------------------------------------------------
        # NOTE: cmcompare with content has returncode of 2
        # for "found difference" or "same"
        # so grep for "is different across nodes"
        # or          "is the same across nodes"
        #---------------------------------------------------
        $rc = grep ( /is different across nodes/, @results );
        LogDebug "grep result: rc=[$rc], results=[@results]";
        if ( $rc != 0 ) {

            #---------------------------------------------------
            # Just log a warning to make the user aware the file
            # is different on the cluster nodes. primitive way of
            # generating attention, since file compare content
            # and logic not implemented yet
            #---------------------------------------------------
            if ( $rules_required_flag eq "w" ) {
                LogVerbose " Informational: cluster cmcompare found a file difference for filename=[$file] - this is a warning and not an error";
                LogVerbose " Supporting Info: File content different on the cluster nodes but content may be valid - please verify file manually";
                LogReturnSub(0);
                return (0);
            }
            #---------------------------------------------------
            # Log the error
            #---------------------------------------------------
            LogInfoF "host=[$host] Found cmcompare difference for filename=[$file]";
            LogDebug "Supporting Info: rc=[$rc] switchnodelist=[@switchnodelist]";
            for (@results) {
                chomp($_);
                # skip misleading cmcompare "files have different timestamps" message
                # as we are only  checking owner, perm and content
                next if ($_ =~ /has timestamp/);
                LogInfoF " Supporting Info: cmcompare: $_";
            }
            LogReturnSub(1);
            return (1);
        }
        else {

            #---------------------------------------------------
            # Log success
            #---------------------------------------------------
            LogVerbose "Passed - same on all nodes";
        }
    }
    else {

        LogVerbose "Skipping cluster file compare - clucompare attribute not set in rule";
    }
    LogReturnSub(0);
    return (0);
}


#===  FUNCTION  =============================================================
#      PURPOSE: List SGeSAP related info on SG packages
#   PARAMETERS: None
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: This option provides an overview of SGeSAP related packages
#               that might of of interest when running sapverify
#============================================================================

sub ShowSGeSAPPackages {
    LogEnterSub(@_);

    # tmp array to hold results
    my @tmp=();

    #---------------------------------------------------
    # call setup_script_runtime()
    # checks if $testmode has been set via environment
    #---------------------------------------------------
    setup_script_runtime();
    $testmode = getparam('testmode');

    # get info for all packages on the system
    cmviewcl_l_package();

    #---------------------------------------------------
    # print SGeSAP related package info
    #---------------------------------------------------
    @tmp=();
    LogInfo "";
    LogInfo "ServiceGuard packages with ***Package style legacy*** configured:";
    foreach ( keys %package_style) {
         push( @tmp, $_) if $package_style{$_} eq "legacy";
    }
    LogInfo "     [@tmp]";
    LogInfo "";

    @tmp=();
    LogInfo "ServiceGuard packages with ***Package style modular*** configured:";
    foreach ( keys %package_style) {
         push( @tmp, $_) if $package_style{$_} eq "modular";
    }
    LogInfo "     [@tmp]";
    LogInfo "";

    @tmp=();
    LogInfo "ServiceGuard packages with a ***SGeSAP module*** configured:";
    foreach ( keys %package_uses_sgesap_module) {
#        LogInfo "SG package=[$_] has a [$package_uses_sgesap_module{$_}] module configured";
         push( @tmp, $_);
    }
    LogInfo "     [@tmp]";
    LogInfo "";

    @tmp=();
    LogInfo "ServiceGuard packages with a ***NFS module*** configured:";
    foreach ( keys %package_uses_nfs_module) {
         push( @tmp, $_);
    }
    LogInfo "     [@tmp]";
    LogInfo "";

    LogInfo "ServiceGuard packages of type ***SGeSAP DB System***";
    foreach ( sort keys %package_db_system) {
        my $tmp_str = sprintf "     package: %-16s has DBSID [%s] configured", $_, @{$package_db_system{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages of type ***SGeSAP SAP System***";
    foreach ( sort keys %package_sap_system) {
        my $tmp_str = sprintf "     package: %-16s has SID [%s] configured", $_, @{$package_sap_system{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages of type ***SGeSAP LC System***";
    foreach ( sort keys %package_lc_system) {
        my $tmp_str = sprintf "     package: %-16s has SID [%s] configured", $_, @{$package_lc_system{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages of type ***SGeSAP REPDB System***";
    foreach ( sort keys %package_repdb_system) {
        my $tmp_str = sprintf "     package: %-16s has SID [%s] configured", $_, @{$package_repdb_system{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages of type ***SGeSAP HDB System***";
    foreach ( sort keys %package_hdb_system) {
        my $tmp_str = sprintf "     package: %-16s has SID [%s] configured", $_, @{$package_hdb_system{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages of type ***SGeSAP External System***";
    foreach ( sort keys %package_sap_ext_system) {
        my $tmp_str = sprintf "     package: %-16s has SID [%s] configured", $_, @{$package_sap_ext_system{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages with an ***SAP instance*** configured";
    foreach ( sort keys %package_sap_instance) {
        my $tmp_str = sprintf "     package: %-16s has [%s] configured", $_, @{$package_sap_instance{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages with an ***SAP DB vendor/instance*** configured";
    foreach ( sort keys %package_db_vendor) {
        my $tmp_str = sprintf "     package: %-16s has [%s] configured", $_, @{$package_db_vendor{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages with an ***SAP replicated instance*** configured";
    foreach ( sort keys %package_sap_replicated_instance) {
        my $tmp_str = sprintf "     package: %-16s has [%s] configured", $_, @{$package_sap_replicated_instance{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages with an ***HDB instance*** configured";
    foreach ( sort keys %package_hdb_instance) {
        my $tmp_str = sprintf "     package: %-16s has [%s] configured", $_, @{$package_hdb_instance{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";


    LogInfo "ServiceGuard packages with an ***SAP external instance*** configured";
    foreach ( sort keys %package_sap_ext_instance) {
        my $tmp_str = sprintf "     package: %-16s has [%s] configured", $_, @{$package_sap_ext_instance{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";

    LogInfo "ServiceGuard packages with an ***SAP infra_sw_type instance*** configured";
    foreach ( sort keys %package_sap_infra_sw_type) {
        my $tmp_str = sprintf "     package: %-16s has [%s] configured", $_, @{$package_sap_infra_sw_type{$_}};
        LogInfo "$tmp_str";
    }
    LogInfo "";


    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
# PURPOSE:      Get package config data for all packages in cluster
# PARAMETERS:   None
# RETURNS:      None
# COMMENTS:     Some pattern matches:
#============================================================================
sub cmviewcl_l_package {
    LogEnterSub(@_);

    my @all_cmviewcl;
    my $cmd;

    if ($testmode) {
         $cmd = "$ENV{'SAPVERIFY_CMD_cmviewcl-v-f_line-l_package'}";
    }
    else {
        $cmd = "$SGSBIN/cmviewcl -v -f line -l package |";
    }
    LogDebug "READING $cmd";
    open my $in, "$cmd" or LogRuntimeError "Could not run cmd=[$cmd]: \$?=[$?]";

    while (<$in>) {
        push( @all_cmviewcl, $_ );
    }

    close $in or LogRuntimeError "Close failed for cmd=[$cmd]: \$?=[$?]";

    foreach my $line (@all_cmviewcl) {
        chomp $line;

        #        LogDebug $line;

        #---------------------------------------------------
        # get the Serviceguard package name
        #---------------------------------------------------
        # Match: package:db71|name=db71
        #   ([^\|]+)    match anything that doesn't contain "|", multiple times
        #   \|?         non greedy - stop at first occurance of |
        #   name=(\1)$  match string found in the first match
        #
#        if ( $line =~ m#^package:([^\|]+)\|?name=([^=]+)$# ) {
        if ( $line =~ m#^package:([^\|]+)\|?name=(\1)$# ) {
            LogDebug "Matched package=[$1] name=[$2]";
            $package_name{$1} = $2;
        }

        #---------------------------------------------------
        # type/style: legacy or modular
        #---------------------------------------------------
        # Match: package:mdisMO7|style=modular
        if ( $line =~ m#^package:(\S+)\|style=(\S+)$# ) {
            LogDebug "Matched: package=[$1] style=[$2]";
            $package_style{$1} = $2;
        }

        #---------------------------------------------------
        # get sgesap modules
        #---------------------------------------------------
        # package:infra1|modu'le_name:sgesap/sap_global|module_name=sgesap/sap_global
        # package:WDP|module_name:sgesap/stack|module_name=sgesap/stack
        if ( $line =~ m#^package:(\S+)\|\S+:\S+\|module_name=(sgesap)/.*$# ) {
            LogDebug "Matched: package=[$1] module_name=[$2]";
            if ( $2 eq "sgesap" ) { $package_uses_sgesap_module{$1} = $2; }
        }

        #---------------------------------------------------
        # get nfs modules
        #---------------------------------------------------
        # Linux:
        # package:MSAPNFS|module_name:tkit/nfs/nfs|module_name=tkit/nfs/nfs
        # HPUX:
        # package:sapNFS|module_name:nfs/hanfs_export|module_name=nfs/hanfs_export
        # package:sapNFS|module_name:nfs/hanfs_flm|module_name=nfs/hanfs_flm
        # package:sapNFS|module_name:nfs/hanfs|module_name=nfs/hanfs

#        if ( $line =~ m#^package:(\S+)\|\S+:\S+\|module_name=(sgesap|nfs)/.*$# ) {
#        if ( $line =~ m#^package:(\S+)\|\S+:\S+\|module_name=(tkit/nfs/nfs)$# ) {
        if ( $line =~ m#^package:(\S+)\|\S+:\S+\|module_name=.*(nfs)/.*$# ) {
            LogDebug "Matched: package=[$1] module_name=[$2]";
            if ( $2 eq "nfs" )    { $package_uses_nfs_module{$1}    = $2; }
#            if ( $2 eq "tkit/nfs/nfs" )    { $package_uses_nfs_module{$1}    = $2; }
        }

        #---------------------------------------------------
        # Get SG modules (sg, sgesap, nfs...) info and save them in an
        # array for the package
        #---------------------------------------------------
        # package:infra1|modu'le_name:sgesap/sap_global|module_name=sg/...
        # package:WDP|module_name:sgesap/stack|module_name=sgesap/....
        if ( $line =~ m#^package:(\S+)\|\S+\|module_name=(\S+)$# ) {
            LogDebug "Matched: package=[$1] module_name=[$2]";
            push( @{ $package_modules{$1} }, $2 );
        }

        #---------------------------------------------------
        # get various SGeSAP _system definitions
        # A SAP system has a SAP SID
        #---------------------------------------------------

        # Match: package:mdbMO7|sgesap/db_global/db_system=MO7
        # package:testoradb2|sgesap/oracledb_global/db_system=P71
        # Note: P71 is part of oracledb_global... and not db_global

        if ( $line =~ m#^package:(\S+)\|\S+/db_system=(\S+)$# ) {
            LogDebug "Matched: package=[$1] db_system=[$2]";
            push( @{ $package_db_system{$1} }, $2 );
        }

        # match: package:mdssMM7|sgesap/sap_global/sap_system=MM7
        if ( $line =~ m#^package:(\S+)\|\S+/sap_system=(\S+)$# ) {
            LogDebug "Matched: package=[$1] sap_system=[$2]";
            push( @{ $package_sap_system{$1} }, $2 );
        }

        # package:SMDA|sgesap/sapextinstance/sap_ext_instance:SMDA97|sgesap/sapextinstance/sap_ext_system=DAA
        # package:SMDA|sgesap/sap_global/sap_system=C71
        # Note: sap_ext_system is "member" of a global sap_system...
        if ( $line =~ m#^package:(\S+)\|\S+\|\S+sap_ext_system=(\S+)$# ) {
            LogDebug "Matched: package=[$1] sap_ext_system=[$2]";
            push( @{ $package_sap_ext_system{$1} }, $2 );
        }

        # Match: package:MsapLC5|sgesap/livecache/lc_system:LC5|sgesap/livecache/lc_system=LC5
        # lc_system is not part of any ..._global definition
        if ( $line =~ m#^package:(\S+)\|\S+\|\S+/lc_system=(\S+)$# ) {
            LogDebug "Matched: package=[$1] lc_system=[$2]";
            push( @{ $package_lc_system{$1} }, $2 );
        }

        # Match: package:repdb|sgesap/repdbinstance/repdb_system=SYR
        if ( $line =~ m#^package:(\S+)\|\S+/repdb_system=(\S+)$# ) {
            LogDebug "Matched: package=[$1] repdb_system=[$2]";
            push( @{ $package_repdb_system{$1} }, $2 );
        }

        # Match: package:hdbHA1p|sgesap/hdb_global/hdb_system=HA1
        if ( $line =~ m#^package:(\S+)\|\S+/hdb_system=(\S+)$# ) {
            LogDebug "Matched: package=[$1] hdb_system=[$2]";
            push( @{ $package_hdb_system{$1} }, $2 );
        }

        #---------------------------------------------------
        # get the various SGeSAP  "instance" definitions
        # db_instances, sap_instances, sap_ext_instances
        # A instance belongs to one of the SAP systems
        #---------------------------------------------------

        # match: package:mcsMO7|sgesap/db_global/db_vendor=oracle
        if ( $line =~ m#^package:(\S+)\|sgesap/db_global/db_vendor=(\S+)$# ) {
            LogDebug "Matched: package=[$1] db_vendor=[$2]";
            push( @{ $package_db_vendor{$1} }, $2 );
        }

        # package:mcsMO7|sgesap/stack/sap_instance:MDS01|sgesap/stack/sap_instance=MDS01
        if ( $line =~ m#^package:(\S+)\|\S+\|\S+sap_instance=(\S+)$# ) {
            LogDebug "Matched: package=[$1] sap_instance=[$2]";
            push( @{ $package_sap_instance{$1} }, $2 );
        }

        # package:extinst-AP4|sgesap/sapextinstance/sap_ext_instance:DVEBMGS20|sgesap/sapextinstance/sap_ext_instance=DVEBMGS20
        if ( $line =~ m#^package:(\S+)\|\S+\|\S+sap_ext_instance=(\S+)$# ) {
            LogDebug "Matched: package=[$1] sap_ext_instance=[$2]";
            push( @{ $package_sap_ext_instance{$1} }, $2 );
        }

        # package:infra1|sgesap/sapinfra/sap_infra_sw_type:saposcol|sgesap/sapinfra/sap_infra_sw_type=saposcol
        if ( $line =~ m#^package:(\S+)\|\S+\|\S+sap_infra_sw_type=(\S+)$# ) {
            LogDebug "Matched: package=[$1] sap_infra_sw_type=[$2]";
            push( @{ $package_sap_infra_sw_type{$1} }, $2 );
        }

        # package:Mers16VM7|sgesap/stack/sap_instance:ERS16|sgesap/stack/sap_replicated_instance=SCS15
        if ( $line =~ m#^package:(\S+)\|\S+\|\S+sap_replicated_instance=(\S+)$# ) {
            LogDebug "Matched: package=[$1] sap_replicated_instance=[$2]";
            push( @{ $package_sap_replicated_instance{$1} }, $2 );
        }

        # package:hdbHA1p|sgesap/hdbsystem/hdb_instance:HDB00|sgesap/hdbsystem/hdb_instance=HDB00
        if ( $line =~ m#^package:(\S+)\|\S+\|\S+hdb_instance=(\S+)$# ) {
            LogDebug "Matched: package=[$1] hdb_instance=[$2]";
            push( @{ $package_hdb_instance{$1} }, $2 );
        }


    }

    #---------------------------------------------------
    $dumpdata && print Data::Dumper->Dump( [ \%package_name, \%package_style ], [ "package_name", "package_style" ] );

    #---------------------------------------------------
    $dumpdata && print Data::Dumper->Dump( [ \%package_uses_sgesap_module ], ["package_sgesap_module_name"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_uses_nfs_module ], ["package_sgesap_nfs_module"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_modules ], ["package_modules"] );

    #---------------------------------------------------
    $dumpdata && print Data::Dumper->Dump( [ \%package_db_system ], ["package_db_system"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_sap_system ], ["package_sap_system"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_sap_ext_system ], ["package_sap_ext_system"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_lc_system ], ["package_lc_system"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_repdb_system ], ["package_repdb_system"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_hdb_system ], ["package_hdb_system"] );

    #---------------------------------------------------
    $dumpdata && print Data::Dumper->Dump( [ \%package_db_vendor ], ["package_db_vendor"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_sap_instance ], ["package_sap_instance"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_sap_ext_instance ], ["package_sap_ext_instance"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_sap_infra_sw_type ], ["package_sap_infra_sw_type"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_sap_replicated_instance ], ["package_sap_replicated_instance"] );

    $dumpdata && print Data::Dumper->Dump( [ \%package_hdb_instance ], ["package_hdb_instance"] );

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Get the environment of package based on output of
#               "cmgetpkgenv package"
#   PARAMETERS: $sgpkg  A Serviceguard package
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Save output of cmgetpkgenv in hash %pkgenv
#               Typical output from "cmgetpkgenv sgpkg":
#               SG_PACKAGE_NAME="mdisMO7"
#               SG_PACKAGE_DESCRIPTION="Serviceguard Package"
#               SG_MODULE_NAME[0]="sg/basic"
#               SG_MODULE_VERSION[0]="1"
#               SG_MODULE_NAME[1]="sgesap/mdminstance"
#
#               Note: Convert to lc before adding to hash
#               Note: This subroutine uses cmgetpkgenv and therefore this
#               only works for a defined package
#============================================================================

sub cmgetpkgenv {
    LogEnterSub(@_);
    my $sgpkg = shift;
    my $cmd;
    my $exit_status;
    my $line;

    if ($testmode) {
        $cmd = "$ENV{'SAPVERIFY_CMD_cmgetpkgenv'}_$sgpkg";
    }
    else {
        $cmd = "$SGSBIN/cmgetpkgenv $sgpkg 2>&1 |";
    }
    open my $fh, "$cmd" or LogRuntimeError "Could not run [$cmd]: $!";

    while (<$fh>) {
        $line = $_;
        chomp $line;

        if ( $line =~ m#^(\S+)\[(\S+)\]="(\S*)"$# ) {

            # save package environment in %pkgenv
            # LogDebug "Found cmgetpkgenv array 1=[$1] 2=[$2] 3=[$3]";
            $pkgenv{ lc($1) }[$2] = $3;
        }
        elsif ( $line =~ m#^(\S+)=\"(\S*)\"$# ) {

            # LogDebug "Found cmgetpkgenv variable 1=[$1] 2=[$2]";
            $pkgenv{ lc($1) } = $2;
        }
    }

    #    close $fh or LogRuntimeError "Close failed for [$cmd]: \$!=[$!]";
    close $fh;
    if ($?) {
        $exit_status = $? >> 8;
        LogRuntimeError "Close for cmd=[$cmd] failed: \$!=[$!] \$exit_status=[$exit_status] $line";
    }

    $dumpdata && LogVerbose( Data::Dumper->Dump( [ \%pkgenv ], ["pkgenv"] ) );

    LogReturnSub();
    return;
}



#===  FUNCTION  =============================================================
#      PURPOSE: Get the environment of package provided in context of
#               cmcheckconf or cmapplyconf for a package conf file
#      PURPOSE: Setup the environment based on output of getpkgenv
#   PARAMETERS: $sgpkg  A Serviceguard package
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Used in SG validate environment. The environment context is
#               defined ina a $SG_ENV_FILE,  generated by cmapplyconf
#               or cmcheckconf
#
#               Save output  in hash %pkgenv
#               Typical output from "cmgetpkgenv sgpkg":
#               SG_PACKAGE_NAME="mdisMO7"
#               SG_PACKAGE_DESCRIPTION="Serviceguard Package"
#               SG_MODULE_NAME[0]="sg/basic"
#               SG_MODULE_VERSION[0]="1"
#               SG_MODULE_NAME[1]="sgesap/mdminstance"
#
#               Note: Convert to lc before adding to hash
#               Note: SG_NODE_NAME[0]="*" contains wildcard "*" instead of
#               a list of cluster nodes
#============================================================================

sub getpkgenv {
    LogEnterSub(@_);
    my $sgpkg = shift;
    my $cmd;
    my $exit_status;
    my $line;

    if ($testmode) {
        $cmd = "$ENV{'SAPVERIFY_CMD_cmgetpkgenv'}_$sgpkg";
    }
    else {

        # The environment provided by cmcheckconf/cmapplyconf should
        # contain a file with environment of the package conf
        LogRuntimeError "Environment varable \$sg_op_code not defined" if !defined $sg_op_code;
        LogRuntimeError "In sg_op_code=[$sg_op_code] but environment varable \$SG_ENV_FILE not defined" if !defined $ENV{'SG_ENV_FILE'};
        $cmd = "$ENV{'SG_ENV_FILE'}";
    }

    open my $fh, "$cmd" or LogRuntimeError "Could not run [$cmd]: $!";

    while (<$fh>) {
        $line = $_;
        chomp $line;

        if ( $line =~ m#^(\S+)\[(\S+)\]="(\S*)"$# ) {

            # LogDebug "Found cmgetpkgenv array 1=[$1] 2=[$2] 3=[$3]";
            $pkgenv{ lc($1) }[$2] = $3;
        }
        elsif ( $line =~ m#^(\S+)=\"(\S*)\"$# ) {

            # LogDebug "Found cmgetpkgenv variable 1=[$1] 2=[$2]";
            $pkgenv{ lc($1) } = $2;
        }
    }

    #    close $fh or LogRuntimeError "Close failed for [$cmd]: \$!=[$!]";;
    close $fh;
    if ($?) {
        $exit_status = $? >> 8;
        LogRuntimeError "Close for cmd=[$cmd] failed: \$!=[$!] \$exit_status=[$exit_status] $line";
    }
    $dumpdata && LogVerbose( Data::Dumper->Dump( [ \%pkgenv ], ["pkgenv"] ) );

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: quick check if package if package name is known
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#               If run in cmcheckconf of cmapplyconf context, the environement
#               variable will have been set by Serviceguard. If not set we
#               are being run from command line interface;
#============================================================================

sub check_package_operation_env {
    LogEnterSub(@_);

    if ( defined $ENV{'op_code'} ) {

        #---------------------------------------------------
        # sapverify is being executed via cmcheckconf or
        # cmapplyconf.
        #---------------------------------------------------

        $sg_op_code = $ENV{'op_code'} if defined $ENV{'op_code'};
        LogDebug "sg_op_code=[$sg_op_code]";

    }
    else {

        $sg_op_code = "";

        #---------------------------------------------------
        # sapverify is getting executed on command line
        # sanity check if package exists at all
        #---------------------------------------------------
        LogDebug "op_code=[not defined]";

        if ( defined $sgpkg ) {
            `$SGSBIN/cmviewcl -p $sgpkg 2>/dev/null`;
            if ($?) {
                LogRuntimeError "Couldn't find a Serviceguard package with name [$sgpkg] on the system";
                exit(2);
            }
        }
        else {
            LogRuntimeError "No package \$sgpkg defined";
            exit(2);
        }

    }

    LogExtra "check_package_operation_env() sg_op_code=[$sg_op_code] sgpkg=[$sgpkg]";
    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: check for existance of debug flag file for sapverify
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#               The following checks if execution of sapverify is
#               is to be postponed (and/or sapverify specific
#               variables are to be configured).
#
#               SGeSAP convention is to skip any SGeSAP startup
#               code if a "debug-flag-file" is found. This allows
#               the debugging of a the Serviceguard package
#               configuration.
#
#               sapverify follows the same convention.
#
#               Skip execution of any sapverify code if a
#               "debug-flag-file" was found and and the file has
#               a size of 0. Skip the sapverify code as early as
#               possible so the check is here.
#
#               $gf: global debugflagfile for all packages.
#               $sf: sgpkg package specific debugflagfile for on package
#               Note: $gf overrides $pf
#
#               If file has size > 0 it is assumed it contains
#               some custom sapverify configuration commands - e.g.
#               timeout parameter can be adjusted to customer
#               needs...
#               Note: this has nothing to todo with debugging of this
#               script
#============================================================================

sub check_sapverify_debug_file {
    LogEnterSub(@_);

    my $debugflagfile = "$SGRUN/debug_sapverify";
    LogExtra "check_sapverify_debug_file()[$debugflagfile]";

    my $gf = "$debugflagfile";
    my $sf = "${debugflagfile}_${sgpkg}";

    if ( -z $gf ) {
        print "Found [$gf] - Skipping sapverify execution...\n";
        exit(0);
    }

    if ( -z $sf ) {
        print "Found [$sf] - Skipping sapverify execution...\n";
        exit(0);
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: get the state of a package based on cmviewcl output
#               (nonverbose output)
#   PARAMETERS: $sgpkg  A Serviceguard package name
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Fill in global %pkgstate with Serviceguard package state info
#               Example of a nonverbose output: cmviewcl -f line -p mdisMO7
#                   name=mdisMO7
#                   style=modular
#                   type=failover
#                   status=down
#                   state=halted
#                   autorun=disabled
#                   owner=unowned
#============================================================================

sub cmviewcl_p_package {
    LogEnterSub(@_);
    my ($sgpkg) = @_;

    my $cmd;
    my $exit_status;
    my $line;

    if ($testmode) {
        $cmd = "$ENV{'SAPVERIFY_CMD_cmviewcl-f_line-p'}_$sgpkg";
    }
    else {
        # redirect stderr in shell so we can catch error string if cmd fails
        $cmd = "$SGSBIN/cmviewcl -f line -p $sgpkg 2>&1 |";
    }

    open my $fh, "$cmd" or LogRuntimeError "Open for [$cmd] failed: \$!=[$!]";

    while (<$fh>) {

        # $line can contain error string if close fails
        $line = $_;
        chomp $line;

        if ( $line =~ m#^(\S+)=(\S*)$# ) {
            LogDebug "Found cmviewcl param 1=[$1] 2=[$2]";
            $pkgstate{$sgpkg}{ ($1) } = $2;
        }
    }

    close $fh;
    if ($?) {
        $exit_status = $? >> 8;
        # $line contains error string
        LogRuntimeError "Close for cmd=[$cmd] failed: \$!=[$!] \$exit_status=[$exit_status] $line";
    }

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%pkgstate ], ["pkgstate"] ) );

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE: get the state cluster nodes based on cmviewcl output
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Fills in global %nodestate with Serviceguard node states
#               Example of a nonverbose output: cmviewcl -f line -p mdisMO7
#               node:hpvm-0-43|name=hpvm-0-43
#               node:hpvm-0-43|status=up
#               node:hpvm-0-43|state=running
#               node:hpvm-0-43|initial_incarnation=1
#               node:hpvm-0-43|id=1
#               node:hpvm-0-44|name=hpvm-0-44
#               node:hpvm-0-44|status=up
#               node:hpvm-0-44|state=running
#               node:hpvm-0-44|initial_incarnation=1
#               node:hpvm-0-44|id=2
#============================================================================

sub cmviewcl_l_node_f_line {
    LogEnterSub(@_);

    my $cmd;
    my $exit_status;
    my $line;

    if ($testmode) {
        $cmd = "$ENV{'SAPVERIFY_CMD_cmviewcl-l_node-f_line'}";
    }
    else {
        # redirect stderr in shell so we can catch error string if cmd fails
        $cmd = "$SGSBIN/cmviewcl -l node -f line 2>&1 |";
    }

    open my $fh, "$cmd" or LogRuntimeError "Open for [$cmd] failed: \$!=[$!]";

    while (<$fh>) {

        # $line can contain error string if close fails
        $line = $_;
        chomp $line;

        if ( $_ =~ m#^node:(\S+)\|(\S+)=(\S+)$# ) {
            LogDebug "Matched node=[$1] name=[$2] [$3]";
            $nodestate{($1)}{$2}=$3;
        }
    }

    close $fh;
    if ($?) {
        $exit_status = $? >> 8;
        # $line contains error string
        LogRuntimeError "Close for cmd=[$cmd] failed: \$!=[$!] \$exit_status=[$exit_status] $line";
    }
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%nodestate ], ["nodestate"] ) );

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE: Fill in global %pkgstate with Serviceguard package state info.
#   PARAMETERS: $sgpkg  A Serviceguard package name
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Run in contect of cmappyconf or cmcheckconf -P conf_file.
#               Basically the same output as with cmviewcl_p_package($sgpkg)
#               should be produced for the package state info.
#               Possbile states: unconfigured (=brand new), running or unowned
#
#               Note: cmviewcl uses different state names, fill in with cmviewcl
#               names so the code here has the same understanding
#               Note: $sg_owner_node was set thru paramater $ENV{'owner_node'}
#               in validate mode
#               Note: used in SG validate environment
#============================================================================

sub getpkgstate {
    LogEnterSub(@_);
    my ($sgpkg) = @_;

    $pkgstate{$sgpkg}{'name'}    = "$sgpkg";
    $pkgstate{$sgpkg}{'style'}   = "";
    $pkgstate{$sgpkg}{'type'}    = "";
    $pkgstate{$sgpkg}{'status'}  = "";
    $pkgstate{$sgpkg}{'state'}   = "";
    $pkgstate{$sgpkg}{'autorun'} = "";
    $pkgstate{$sgpkg}{'owner'}   = "";

    if (exists($pkgenv{sg_module_version}) && @{ $pkgenv{sg_module_version} }) {
      $pkgstate{$sgpkg}{'style'} = "modular"
    }

    # The owner_node will be set to "" if in "validate" environment
    # and the package is down
    if ( defined $ENV{'owner_node'} ) {

        if ( $ENV{'owner_node'} eq "" ) {
            $pkgstate{$sgpkg}{'owner'} = "unowned";
        }
        else {
            $pkgstate{$sgpkg}{'owner'} = "$ENV{'owner_node'}";
        }

    }

    $dumpdata && LogVerbose( Data::Dumper->Dump( [ \%pkgstate ], [" pkgstate "] ) );

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: get the state of a package based on cmviewcl output
#               (nonverbose output)
#   PARAMETERS: None
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#               Example of a nonverbose output: cmviewcl -f line -p mdisMO7
#                   name=mdisMO7
#                   style=modular
#                   type=failover
#                   status=down
#                   state=halted
#                   autorun=disabled
#                   owner=unowned
#============================================================================

sub get_node_switching_params {
    LogEnterSub(@_);
    my ($sgpkg) = @_;
    my $cmd;

    if ($testmode) {
        $cmd = "$ENV{'SAPVERIFY_CMD_cmviewcl-v-f_line-p'}_$sgpkg";
    }
    else {
        $cmd = "$SGSBIN/cmviewcl -v -f line -p $sgpkg |";
    }
    LogDebug "Running cmd=[$cmd]";
    open my $fh, "$cmd" or LogRuntimeError "Could not run [$cmd]: $!";

    while (<$fh>) {
        chomp;

        # only match the state info - prefixed with node:......
        if ( $_ =~ m#^node:(\S+)\|(\S+)=(\S+)$# ) {
            LogDebug "Matched package=[$1] name=[$1]";
            $pkgswitchingstate{ ($1) }{$2} = $3;
        }
    }
    close $fh or LogRuntimeError "Close failed for [$cmd]: \$!=[$!]";

    $dumpdata && LogVerbose( Data::Dumper->Dump( [ \%pkgswitchingstate ], ["pkgswitchingstate"] ) );

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: get a list of nodes the package could switch to
#   PARAMETERS: $sgpkg  serviceguard package
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Fills in global array @switchnodelist with nodes the package
#               could switch to in case of failure
#               Note: Used in SG validate environment, cmapplyconf/cmcheckconf
#               environment
#============================================================================
sub get_switchnodelist {
    LogEnterSub(@_);
    my ($sgpkg) = @_;

    my $foundstar = 0;

    #---------------------------------------------------
    #Package config file: The switchnode list could
    #contain a "*" (=star), instead of individual node
    #names. So if $foundstar, expand with names of  the
    #cluster nodes.
    #---------------------------------------------------

    foreach my $node ( @{ $pkgenv{sg_node_name} } ) {

        if ( $node eq "*" ) {

            # node matched wildcard *; get real nodes names
            LogDebug "wildcard node=[$node] matched";
            $foundstar = 1;
            last;
        }
        $pkgswitchingstate{'$node'}{'switching'} = "enabled";
        # unformated
        push( @switchnodes, "$node" );
        # formated
        push( @switchnodelist, "-n $node " );
    }

    if ($foundstar) {
        cmviewcl_l_node_f_line();
        @switchnodes=();
        @switchnodelist=();
        for my $node ( keys %nodestate ) {
            LogDebug "Checking nodestate node=[$node]";
            if ( $nodestate{$node}{'status'} eq "up" ) {
                push( @switchnodes, "$node" );
                push( @switchnodelist, "-n $node " );
            }
            else {
                LogVerbose "Info: node=[$node] not added to switchnodelist - status=[$nodestate{$node}{'status'}]";
            }
        }
    }
    LogVerbose "switchnodelist=[@switchnodelist]";

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE: get a list of nodes the package could switch to
#   PARAMETERS: $sgpkg  serviceguard package
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Fills in global array @switchnodelist with nodes the package
#               could switch to in case of failure. Used for an already
#               defined package
#============================================================================
sub cmget_switchnodelist {
    LogEnterSub(@_);
    my ($sgpkg) = @_;

    get_node_switching_params($sgpkg);

    for my $node ( keys %pkgswitchingstate ) {
        # $pkgswitchingstate{$node}{'switching'} doesn't exist if
        # cluster hasn't been started
        if ( (exists $pkgswitchingstate{$node}{'switching'}) &&  ($pkgswitchingstate{$node}{'switching'} eq "enabled" ) ) {
            push( @switchnodes, "$node" );
            push( @switchnodelist, "-n $node " );
        }
    }

    LogVerbose "switchnodelist=[@switchnodelist]";
    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: In the context of the executing package determines if this is
#               a RemoteClusterNode or a LocalClusterNode
#   PARAMETERS: $sgpackage  name of the serviceguard package
#      RETURNS: Returns 1 if package is running on this node, 0 if package is
#               not running on this node
#  DESCRIPTION: None
#     COMMENTS: Sets global $Packageonthisnode
#============================================================================
sub is_package_on_this_node {
    LogEnterSub(@_);
    my ($sgpkg) = @_;

    LogDebug "\$pkgstate{\$sgpkg}{'owner'}=[$pkgstate{$sgpkg}{'owner'}] host=[$host]";
    if ( "$pkgstate{$sgpkg}{'owner'}" eq "$host" ) {
        $PackageOnThisNode = 1;
        LogVerbose "Package IS running on this node PackageOnThisNode=[$PackageOnThisNode]";
    }
    else {
        $PackageOnThisNode = 0;
        LogVerbose "Package is NOT running on this node PackageOnThisNode=[$PackageOnThisNode]";
    }

#    LogVerbose "PackageOnThisNode=[$PackageOnThisNode]";
    LogReturnSub($PackageOnThisNode);
    return $PackageOnThisNode;
}

#===  FUNCTION  =============================================================
#      PURPOSE: If a package is unowned it is down
#   PARAMETERS: $sgpack is name of the serviceguard package
#      RETURNS:
#  DESCRIPTION: None
#     COMMENTS: None
#               See "man 1 cmviewcl" and "man 5 cmviewcl":
#
#               Status of the cluster, node, package, etc.
#               Status is either up, down, starting, halting,
#               detached, or unknown.  detached status applies
#               to packages only.  When a package is in
#               detached status, Serviceguard does not monitor
#               this package and its components.  unknown is a
#               special status that indicates insufficient
#               information was collected to exactly determine
#               whether the item is up or down, e.g., when some
#               of the nodes in the cluster cannot be reached
#               over the network.
#============================================================================
sub is_package_down {
    LogEnterSub(@_);
    my ($sgpkg) = @_;
    my $rc;

    LogDebug "\$pkgstate{\$sgpkg}{'owner'} = [$pkgstate{$sgpkg}{'owner'}]";

    if ( "$pkgstate{$sgpkg}{'owner'}" eq "unowned" ) {
        $PackageDown = 1;
        $PackageUp=0;
#        LogInfoF "Package is DOWN: PackageUp=[$PackageUp] - NOTE: If an error is found will exit on first verify error";
#        LogInfoF "Not implemented yet"
    } else {
        $PackageDown = 0;
        $PackageUp=1;
#        LogInfoF "Package is UP: PackageUp=[$PackageUp] - NOTE: If an error is found will continue with verification";
    }

#    LogVerbose "PackageUp=[$PackageUp]";
    LogReturnSub($PackageDown);
    return $PackageDown;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Check if package is supported e.g. is modular
#   PARAMETERS: $sgpack is name of the serviceguard package
#      RETURNS:
#  DESCRIPTION: None
#     COMMENTS: Right now only checks if package is modular
#============================================================================
sub is_package_supported {
    LogEnterSub(@_);
    my ($sgpkg) = @_;

    if ( "$pkgstate{$sgpkg}{'style'}" ne "modular" ) {
        LogRuntimeError "Package [$sgpkg] style=[$pkgstate{$sgpkg}{'style'}] not supported by sapverify";
        timestamp_and_exit(2);
    }
    else {
        LogDebug "\$pkgstate{\$sgpkg}{'style'} = [$pkgstate{$sgpkg}{'style'}]";
    }

    LogReturnSub($PackageDown);
    return $PackageDown;

}

#===  FUNCTION  =============================================================
# PURPOSE:      Checks if file exists
# PARAMETERS:   $filename   Name of file
# RETURNS:      1           File exists
#               0           File doesn't exist
# COMMENTS:     None
#============================================================================

sub check_for_file {
    LogEnterSub(@_);
    my ($filename)=@_;

    LogDebug "check_for_file File filename=[$filename]";

    my $rc;
    if ( -e $filename ) {
        LogDebug "File filename=[$filename] exists";
        $rc=1;
    } else {
        LogDebug "File filename=[$filename] doesn't exist";
        $rc=0;
    }

    LogReturnSub($rc);
    return $rc;
}
#===  FUNCTION  =============================================================
# PURPOSE:      Checks if directory is empty
# PARAMETERS:   $path   Name of directory path
# RETURNS:      1       Directory is empty
#               0       Directory is not empty
# COMMENTS:     Skip . and ..
#============================================================================

sub is_dir_empty {
    LogEnterSub(@_);
    my ($path)=@_;
    my $dh;

    opendir($dh, $path) or LogRuntimeError "Couldn't opendir $path \$!=[$!]";
    while (my $file = readdir($dh)) {
        # skip . and ..
        next if ($file =~ /^\.\.?$/);
        # found a file in the path
        closedir $dh;
        LogReturnSub(0);
        return(0);
    }
    closedir $dh or LogRuntimeError "Close failed for [$path]";
    LogReturnSub(1);
    return(1);
}


#===  FUNCTION  =============================================================
#      PURPOSE: This function scans a directory tree to find the "deepest"/
#               "lowest" directory available that could be used as a file
#               system mount point.
#   PARAMETERS: An absolute path
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: The lowest directory available in the path is an indication a
#               A previous verify subroutine call to xxxx  didn't find the
#               of an misconfigured or missing mount in the SG configuration.
#               file $path,  even though it must be there for a successfull
#               exectution of the SG package. The subruotine will
#               try to detemine if perhaps a file system wasn't mounted
#               by checking for directories in the pth of the file
#               walk path  and search for lowest level directory available on
#               system
#============================================================================
sub FindDeepestDir {
#   recursive - don't log subroutine calling
#    LogEnterSub(@_);

    my $path = shift;

    if ( $path eq "/" ) {

        LogDebug "At top \/ -  couldn't find a file system mount point";
        return $path;
    }

    if ( -d $path ) {
        LogDebug "Info: A file system mount is missing under directory=[$path]";
        return $path;
    }
    else {
        FindDeepestDir( dirname($path) );
    }

#    LogReturnSub();
#    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Compare $dir with the mount points in @sg_fs_directory
#   PARAMETERS:
#      RETURNS: 0   $dir is not mounted by/within package
#               1   $dir is mounted by/within package
#  DESCRIPTION: None
#     COMMENTS: The mount points listed in the @sg_fs_directory are mounted
#               exclusively on the node where the package gets executed. On
#               all other cluster nodes (= RemoteClusterNodes) these
#               directories are not accessible. So in the case of the
#               verification of filesystem on a RemoteClusterNode don't
#               report these filesystems (the PackageMounted) as an error
#
#		walk path tree and find lowest level directory on system
#		this could be a missing mount point
#============================================================================
sub is_fs_mounted_by_package {

    LogEnterSub(@_);
    my $dir   = shift;
    my $found = 0;
    my $sg_mp;

    LogDebug "Verify sgesap_db_directory=@{$pkgenv{sg_fs_directory}}";
    foreach $sg_mp (@{$pkgenv{sg_fs_directory}}) {

#        if ( "$dir" lt "$sg_mp" ) {
#            CountAndLogErrorF "************ dir=[$dir] lt sg_mp=[$sg_mp] found = 0";
#            $found = 0;
##            last;
#        }
#        elsif ( "$dir" eq "$sg_mp"  ) {
#            CountAndLogErrorF "************ dir=[$dir] eq sg_mp=[$sg_mp] found = 1";
#            $found = 1;
##            last;
#        }
#        elsif ( "$dir" gt "$sg_mp" ) {
#            CountAndLogErrorF "************ dir=[$dir] gt sg_mp=[$sg_mp] found = 1";
#            $found = 1;
#        }

        if ( $dir =~ m#$sg_mp# ) {
             LogDebug "dir=[$dir] is  a subtree of sg_mp=[$sg_mp] found = 1; last";
            $found=1;
            last;
        } else {
             LogDebug "dir=[$dir] not a subtree of sg_mp=[$sg_mp] found = 0";
        }

    }

    LogDebug "Directory dir=[$dir] is a subtree of the sg_fs_directory array mounted by package" if $found;

    LogReturnSub($found);
    return $found;
}

#===  FUNCTION  =============================================================
#      PURPOSE: test if a filesystem mount might be missing in the path of
#               the file
#   PARAMETERS: $file   an absolute path/filename
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Condition before entering: $file wasn't found.
#============================================================================
sub check_for_missing_mount {
    LogEnterSub(@_);
    my $file = shift;
    my $rc;
    my $mountpoint_path;
    my $mountbypackage;
    my $DirName;
    my $isempty;

    # get the directory part of $file
    $DirName = dirname($file);

    # find longest valid path to the file
    $mountpoint_path = FindDeepestDir($file);

    #---------------------------------------------------
    # check if path/directory is empty
    # empty directories are potential mount points :-)
    #---------------------------------------------------
    $isempty = is_dir_empty($mountpoint_path);

    #---------------------------------------------------
    # Try to find where the mount point is missing. Example:
    # Start Condition:
    #   $filename=/usr/sap/MO7/file was found missing.
    #   $DirName gets set to "/usr/sap/MO7"
    # Depending on the returned mount point pathi $mountpoint_path ...
    #---------------------------------------------------

    if ( "$mountpoint_path" eq "$DirName" ) {
        # if a valid path is: $mountpoint_path="/usr/sap/MO7"
        if ( !$isempty ) {
            LogInfoF " Supporting Info: Possibly missing file in mountpoint_path=[$mountpoint_path]";
        }
        else {
            LogInfoF " Supporting Info: Possibly missing file system mount under mountpoint_path=[$mountpoint_path]";
        }
    }
    else {
        # if a valid path is: $mountpoint_path="/usr/sap" or "/usr" or "/"
        if ( !$isempty ) {
            LogInfoF " Supporting Info: Possibly missing mount directory under mountpoint_path=[$mountpoint_path]";
        }
        else {
            LogInfoF " Supporting Info: Possibly missing file system mount under mountpoint_path=[$mountpoint_path]";
        }
    }

    #---------------------------------------------------
    # if directory is empty then presumably it is a
    # mount point
    #---------------------------------------------------
    return unless $isempty;

    #---------------------------------------------------
    # try to determine where the missing file system is
    # mounted from:
    #   1) a filesystem mounted by a package or
    #   2) a filesystem mounted by localnode or NFSshared
    #---------------------------------------------------
    $mountbypackage = is_fs_mounted_by_package($mountpoint_path);

    LogDebug "filename=[$file] mountpoint_path[$mountpoint_path] mountbypackage=[$mountbypackage] $PackageOnThisNode=[$PackageOnThisNode]";

    if ($PackageOnThisNode) {
        #  this is the node where the package is running on

        if ( !$mountbypackage ) {
            # not mounted via package
            LogInfoF " Supporting Info: Possibly missing (Package, NFS or local system) mount under directory mountpoint_path=[$mountpoint_path]";
        }
        else {
            # mounted via package
            LogInfoF " Supporting Info: Possibly missing (Package) mount under directory mountpoint_path=[$mountpoint_path] and package [$pkgenv{sg_package_name}]";
        }
    }
    else {
        #  this is a node where the package isn't running on
        if ( !$mountbypackage ) {
            # not mounted via package
            LogInfoF " Supporting Info: Possibly missing (NFS or local system) mount under mountpoint_path=[$mountpoint_path]";
        }
    }
    LogReturnSub();
    return;
}

#############################################################################
# SGeSAP SYSTEMS
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: This function creates a verify list for any db_system type
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#===========================================================================d=
sub create_verifylist_db_system {
    LogEnterSub(@_);

    LogVerbose "Verify-System: sgesap_db_vendor=$pkgenv{sgesap_db_vendor}";
    if ( ($pkgenv{sgesap_db_vendor} eq "oracle") || ($pkgenv{sgesap_db_vendor} eq "oracle12") )  {
        create_verifylist_db_system_oracle();
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "maxdb" ) {
        create_verifylist_db_system_maxdb();
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "db2" ) {
        create_verifylist_db_system_db2();
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "sybase" ) {
        create_verifylist_db_system_sybase();
    }
    else {

        LogInfoF "host=[$host]",
              "sgesap_db_vendor=[$pkgenv{sgesap_db_vendor}] not supported";

    }
    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: This function creates and adds a verify list (list of files to
#               verify) for an sgesap_sap_system to the verify hash
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub create_verifylist_sap_system {
    LogEnterSub(@_);

    my ( $sid, $instance, $itype, $inr, $vip );

    LogDebug "Adding verification files for sgesap_sap_system=[". "$pkgenv{sgesap_sap_system}" . "]";

    #---------------------------------------------------
    # contains one or more sap instances
    #---------------------------------------------------
    if ( exists($pkgenv{sgesap_sap_instance}) && @{ $pkgenv{sgesap_sap_instance} } ) {

        $sid = uc( $pkgenv{sgesap_sap_system} );

        foreach my $instance ( @{ $pkgenv{sgesap_sap_instance} } ) {
            $itype = $pkgenv{sgesap_sap_instance_itype}{$instance};
            $inr   = $pkgenv{sgesap_sap_instance_inr}{$instance};
            $vip  = $pkgenv{sgesap_sap_instance_ivip}{$instance};
            LogDebug " sid=[$sid] instance=[$instance] itype[$itype] inr=[$inr] VIP=[$vip]";

            verifylist_add( $RULESF, "sap_system",  $sid, $itype, $inr, $vip );
            verifylist_add( $RULESF, $itype,        $sid, $itype, $inr, $vip );
            verifylist_add( $RULESF, "files_sgesap",$sid, $itype, $inr, $vip );
            verifylist_add( $RULESF, "files_system",$sid, $itype, $inr, $vip );
        }

    }

    #---------------------------------------------------
    # contains one or more sap external systems/instances
    #---------------------------------------------------
    if ( exists($pkgenv{sgesap_sap_ext_instance}) && @{ $pkgenv{sgesap_sap_ext_instance} } ) {
        LogDebug 'Verify: @{$pkgenv{sgesap_sap_ext_instance}}=[' . "@{$pkgenv{sgesap_sap_ext_instance}}" . "]";

        # use timeout as verify_ext_sap_instance will execute remote shells
#        call_func_with_timeout_sig( $timeout, \&verify_ext_sap_instance);
    }

    #---------------------------------------------------
    # contains one or more sap infra instances
    #---------------------------------------------------
    if ( exists($pkgenv{sgesap_sap_infra_sw_type}) && @{ $pkgenv{sgesap_sap_infra_sw_type} } ) {

        LogDebug 'Verify: @{$pkgenv{sgesap_sap_infra_sw_type}}=[' . "@{$pkgenv{sgesap_sap_infra_sw_type}}" . "]";
        verify_sap_infra_sw_type();

    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: This function creates a verify list for an sgesap_lc_system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub  create_verifylist_lc_system {
    LogEnterSub(@_);

    my ($SID, $sid, $VIP);
    my $i = 0;

    LogVerbose "Verify list create: for \@sgesap_lc_system=[@{ $pkgenv{sgesap_lc_system} }]";
    foreach my $inst ( @{ $pkgenv{sgesap_lc_system} } ) {
        $SID = $inst;
        verifylist_add( $RULESF, "lc_system", $SID);
        verifylist_add( $RULESF, "files_sgesap",$SID);
        verifylist_add( $RULESF, "files_system",$SID);
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: This function creates a verify list for an sgesap_repdb_system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub  create_verifylist_repdb_system {
    LogEnterSub(@_);

    my ($SID, $sid);

    $SID="";

    if ( defined $pkgenv{sgesap_repdb_system} ) {
        $SID = uc( $pkgenv{sgesap_repdb_system} );
    }
    elsif  (  defined $pkgenv{sgesap_db_system} ) {
        $SID = uc( $pkgenv{sgesap_db_system} );
    } else {
        LogReturnSub(0);
        return 0
    }

    LogVerbose "Verify list create: for repdb system SID=[$SID]";

    verifylist_add( $RULESF, "repdb_system",$SID);
#    verifylist_add( $RULESF, "files_sgesap",$SID);
#    verifylist_add( $RULESF, "files_system",$SID);


    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: This function creates a verify list for an sgesap_hdb_system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub  create_verifylist_hdb_system {
    LogEnterSub(@_);

    my ($sid, $instance, $itype, $inr, $vip );

    $sid="";

    if ( defined $pkgenv{sgesap_hdb_system}[0] ) {
        $sid = uc( $pkgenv{sgesap_hdb_system}[0] );
    } else {
        LogReturnSub(0);
        return 0
    }

    LogVerbose "Verify list create: for hdb system sid=[$sid]";

#    verifylist_add( $RULESF, "hdb_system",$sid);
#    setparam( "debug", 1 );


    LogDebug "Adding verification files for sgesap_hdb_system=[". "$pkgenv{sgesap_hdb_system}[0]" . "]";

    #---------------------------------------------------
    # contains one or more sap instances
    #---------------------------------------------------
    if ( exists($pkgenv{sgesap_hdb_instance}) && @{ $pkgenv{sgesap_hdb_instance} } ) {

        foreach my $instance ( @{ $pkgenv{sgesap_hdb_instance} } ) {
            LogDebug " instance=[$instance]";
            $itype = $pkgenv{sgesap_hdb_instance_itype}{$instance};
            $inr   = $pkgenv{sgesap_hdb_instance_inr}{$instance};
            $vip  = $pkgenv{sgesap_hdb_instance_ivip}{$instance};
            LogVerbose " sid=[$sid] instance=[$instance] itype[$itype] inr=[$inr] VIP=[$vip]";

            verifylist_add( $RULESF, "hdb_system",  $sid, $itype, $inr, $vip );
#            verifylist_add( $RULESF, $itype,        $sid, $itype, $inr, $vip );
#            verifylist_add( $RULESF, "files_sgesap",$sid, $itype, $inr, $vip );
##            verifylist_add( $RULESF, "files_system",$sid, $itype, $inr, $vip );
        }
    }


    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: A verify list has been built for all SGeSAP systems; use
#               this list to verify all files
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#============================================================================
sub verify_the_list {
    LogEnterSub(@_);

    $dumpdata && LogVerbose ( Data::Dumper->Dump([\%vf],["vf"]) );

    LogVerbose "Running verify_files_for_component()";
    verify_files_for_component("ALL");

    $dumpdata && LogVerbose ( Data::Dumper->Dump([\%vf],["vf"]) );

    # only verify distributed files and login
    # files for non-external instances

    if (    ( exists($pkgenv{sgesap_sap_instance})     && @{ $pkgenv{sgesap_sap_instance} } )    ||
            ( exists($pkgenv{package_db_vendor})       && @{ $pkgenv{package_db_vendor} } )      ||
            ( exists($pkgenv{sap_replicated_instance}) && @{ $pkgenv{sap_replicated_instance} } )||
            ( $repdb_t==1 )                                                                       ||
            ( exists($pkgenv{sgesap_hdb_instance})     && @{ $pkgenv{sgesap_hdb_instance} } )    ||
            ( exists($pkgenv{sgesap_lc_system})        && @{ $pkgenv{sgesap_lc_system} } )         ) {

        # only the master should verify distributed files
        if ( ($extended) && ($cmcompare_master )) {
            LogVerbose "Running verify_dstrb_files()";
            verify_dstrb_files("ALL");
        }

        # only the master should verify login files
        if (($extended ) && ( $cmcompare_master )) {
            LogVerbose "Running verify_user_login_startup_files()";
            verify_user_login_startup_files();
        }
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: This function verifies an instance base on SAP profile. Used
#               for verifying a SGeSAP external instance
#   PARAMETERS: $pf is an SAP profile name
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#============================================================================
sub run_pf_based_verify {
    LogEnterSub(@_);
    my ( $pf ) = @_;
    my ( $sid, $instance, $itype, $inr, $vip );

    #---------------------------------------------------
    # call setup_script_runtime()
    # checks if $testmode has been set via environment
    #---------------------------------------------------
    setup_script_runtime();

    #---------------------------------------------------
    # import variables defined in SAP_functions.pm and
    # and that are used here
    #---------------------------------------------------
    $timeout  = getparam('timeout');
    $fork     = getparam('fork');
    $RULESF   = getparam('RULESF');
    $testmode = getparam('testmode');
    $dumpdata = getparam('dumpdata');



    LogVerbose "Verify instance for profile \$pf=[$pf]";

    ( $sid, $instance, $vip ) = split /_/, $pf;
    ( $itype, $inr ) = $instance =~ /^([A-Z]+)([0-9][0-9])$/;

    verifylist_add( $RULESF, "sap_system", $sid, $itype, $inr, $vip );
    verifylist_add( $RULESF, $itype,             $sid, $itype, $inr, $vip );

    verify_files_for_component_noclu("sap_system");
    verify_files_for_component_noclu("$itype");

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE: This function verifies an external SAP instance
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: A typical external system might contain:
#               SGeSAP_SAP_SYSTEM="C71"
#               SGeSAP_REM_COMM="remsh"
#               SGeSAP_SAP_EXT_INSTANCE[0]="DVEBMGS41"
#               SGeSAP_SAP_EXT_SYSTEM[0]="N73"
#               SGeSAP_SAP_EXT_HOST[0]="sgsap030"
#               SGeSAP_SAP_EXT_TREAT[0]="yynnn"
#               SGeSAP_SAP_EXT_INSTANCE[1]="SMDA96"
#               SGeSAP_SAP_EXT_SYSTEM[1]="DAB"
#               SGeSAP_SAP_EXT_HOST[1]="sgsap030"
#               SGeSAP_SAP_EXT_TREAT[1]="yynnn"
#============================================================================
sub verify_ext_sap_instance {
    LogEnterSub(@_);
    my ( $sid, $instance, $itype, $inr, $vip, $cmd, $rem_comm, $sidadm );
    my $exit_status;
    my $line;
    my $i = 0;

    LogVerbose "Verify-System: \@sgesap_sap_ext_instance=[@{ $pkgenv{sgesap_sap_ext_instance} }]";

    foreach $instance ( @{ $pkgenv{sgesap_sap_ext_instance} } ) {

        $sid      = $pkgenv{sgesap_sap_ext_system}[$i];
        $vip      = $pkgenv{sgesap_sap_ext_host}[$i];
        $rem_comm = $pkgenv{sgesap_rem_comm};
        $sidadm   = lc($sid) . "adm";


        $rem_comm="remsh" if !defined $rem_comm;

        $cmd = "$rem_comm $vip -l $sidadm -n \"/opt/cmcluster/sap/bin/sapverify -pf ${sid}_${instance}_${vip}\" 2>&1 |";

        LogVerbose "Executing command cmd=[$cmd]";

        open my $fh, "$cmd" or LogRuntimeError "Could not run [$cmd]: $!";

        while (<$fh>) {
            $line = $_;
            chomp $line;
            if ( $line =~ m#^Verify-Error# ) {
                LogInfoF "Warning for external instance ${sid}_${instance}_${vip}: [$line]";
            }

        }

        close $fh;
        if ($?) {
            $exit_status = $? >> 8;
            LogRuntimeError "Close for cmd=[$cmd] failed: \$!=[$!] \$exit_status=[$exit_status] $line";
        }

        $i++;
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: This function verifies an sap_infra_sw_type instance
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: A typical sap_infra_sw_type might contain:
#               SGeSAP_SAP_INFRA_SW_TYPE[0]="sapwebdisp"
#               SGeSAP_SAP_INFRA_SW_TREAT[0]="startnstop"
#               SGeSAP_SAP_INFRA_SW_PARAMS[0]="-shm_attach_mode 6"
#               SGeSAP_SAP_SYSTEM="N73"
#
#============================================================================
sub  verify_sap_infra_sw_type {
    LogEnterSub(@_);

   my ($SID, $sid, $VIP);

    LogVerbose "Verify-System: \@{\$pkgenv{sgesap_sap_infra_sw_type}}=[@{ $pkgenv{sgesap_sap_infra_sw_type} }]";
    foreach my $infra ( @{ $pkgenv{sgesap_sap_infra_sw_type} } ) {

        $SID = uc( $pkgenv{sgesap_sap_system} );

        # args: $RULESF, $component, $SID, $INST, $NR, $VIP, $FILESPEC ) = @_;
        verifylist_add($RULESF, "sap_infra_sw_type", $SID, "", "", "", $infra);
    }

#    verify_files_for_component("sap_infra_sw_type");

    LogReturnSub();
    return;
}

#############################################################################
# DB_SYSTEM
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: This function verifies "live" db data volumes against
#               against actual configuration
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#===========================================================================d=
sub verify_db_storage {
    LogEnterSub(@_);

	# apparently no db_system configured so just return
    if ( ! defined $pkgenv{sgesap_db_system} ) {
        LogReturnSub(0);
        return 0
    } else {
        LogVerbose "Running extended db_system verify";
    }

    if ( $PackageDown ) {
        LogInfoF "Package is down - can't query database storage - returning";
        LogReturnSub(1);
        return 1;
    }

    if ( !is_db_up() ){
        LogInfoF "Database is down - can't query database storage - returning";
        LogReturnSub(1);
        return 1;
    }

    if  ($pkgenv{sgesap_db_vendor} eq "sybase" ) {
        if ( ( defined $pkgenv{sgesap_aseuser} ) && ( defined $pkgenv{sgesap_asepasswd} ) )   {
            $SGeSAP_ASEUSER="$pkgenv{sgesap_aseuser}[0]";
            $SGeSAP_ASEPASSWD="$pkgenv{sgesap_asepasswd}[0]";
        } else {
            LogInfoF "No DB access - skipping verify of Sybase database storage";
            return 1;
        }
    }

    if ( $cmcompare_master ) {
        LogVerbose "On master (= node is running DB) - start verify of DB storage configuration";
    } else {
        LogInfoF "Not on master (= node not running DB) - skipping verify of DB storage configuration";
        LogReturnSub(1);
        return 1;
    }

    #---------------------------------------------------
    # get mountpoints and filesystems of the system and
	# save data in %mpsfs %mpsfst %mpsfsd
    #---------------------------------------------------
    if ( gen_system_storage_records("/") != 0) {
        LogReturnSub(1);
        return 1;
    }

    #---------------------------------------------------
    # Initialize %dbstorage - contains one storage record
    # for each device configured in the db
    #---------------------------------------------------
    gen_db_storage_records();

    #---------------------------------------------------
    # HPUX :Initialize %verify_vxfs contains list of vxfs
    # volumes to verify
    #---------------------------------------------------
    if ( $^O eq "hpux" ) {
        verify_lvm_hpux();
    }

    #---------------------------------------------------
    # verify all physical devices configured in the
    # running database are mounted with package.
    #---------------------------------------------------
    verify_db_storage_records();

    LogReturnSub(0);
    return 0;
}


#===  FUNCTION  =============================================================
#      PURPOSE: Checks if db can be accessed with an sql command
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: An initial check if database can be accessed at all with sql
#     COMMENTS: None
#===========================================================================d=
sub sql_is_db_up {
    LogEnterSub(@_);

    my $rc;
    my $SID;
    LogVerbose "Checking if sgesap_db_vendor=$pkgenv{sgesap_db_vendor} database is up";

    if ( $pkgenv{sgesap_db_vendor} eq "oracle" ) {
         sql_is_oracle_up($SID);
         $rc = $?;
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "maxdb" || $pkgenv{sgesap_db_vendor} eq "livecache" ) {
         sql_is_maxdb_up($SID);
         $rc = $?;
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "db2" ) {
         sql_is_db2_up($SID);
         $rc = $?;
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "sybase" ) {
         sql_is_sybase_up($SID);
         $rc = $?;
    }
    else {

        LogInfoF "host=[$host] ",
              "sgesap_db_vendor=[$pkgenv{sgesap_db_vendor}] not supported";
    }

    LogVerbose "Database is rc=[$rc]" if ($rc==1);
    LogReturnSub($rc);
    return $rc;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Checks if db processes are running
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION:
#     COMMENTS: None
#===========================================================================d=
sub is_db_up {
    LogEnterSub(@_);

    my $rc;
    LogDebug "Checking if [$pkgenv{sgesap_db_vendor}] database is up/down";

    if ( $pkgenv{sgesap_db_vendor} eq "oracle" ) {
         $rc=is_oracle_up();
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "maxdb"  || $pkgenv{sgesap_db_vendor} eq "livecache" ) {
         $rc=is_maxdb_up();
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "db2" ) {
         $rc=is_db2_up();
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "sybase" ) {
         $rc=is_sybase_up();
    }
    else {
        LogInfoF "host=[$host] ",
              "sgesap_db_vendor=[$pkgenv{sgesap_db_vendor}] not supported";
    }

    if ($rc == 1) {
        LogVerbose "Database [$pkgenv{sgesap_db_vendor}] is up";
    } else {

        LogVerbose "Database [$pkgenv{sgesap_db_vendor}] is down";
    }
    LogReturnSub($rc);
    return $rc;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Query database for configured storage
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION:
#     COMMENTS: None
#===========================================================================d=
sub sql_get_db_storage {
    LogEnterSub(@_);

    my $rc;
    LogVerbose "Query database $pkgenv{sgesap_db_vendor} for configured storage";

    if ( $pkgenv{sgesap_db_vendor} eq "oracle" ) {
        sql_get_oracle_db_storage();
        $rc = $?;
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "maxdb" || $pkgenv{sgesap_db_vendor} eq "livecache" ) {
        sql_get_maxdb_db_storage();
        $rc = $?;
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "db2" ) {
        sql_get_db2_db_storage();
        $rc = $?;
    }
    elsif ( $pkgenv{sgesap_db_vendor} eq "sybase" ) {
        sql_get_sybase_db_storage();
        $rc = $?;
    }
    else {

        LogInfoF "host=[$host]",
              "sgesap_db_vendor=[$pkgenv{sgesap_db_vendor}] not supported";
    }

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \@dbdatafiles ], ["dbdatafiles"] ));

    LogReturnSub($rc);
    return $rc;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Verify db storage
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub verify_db_storage_records {
    LogEnterSub(@_);

    #--------------------------------------------------------
    # assumption: a database file should always be mounted
    # in the package and not configured as distributed files
    #--------------------------------------------------------
    foreach my $file ( keys %{dbstorage} ) {
        if ( !$dbstorage{$file}{mountbypackage} ) {
            LogInfoF "WARNING: host=[$host] Database file=[$file] is configured in the running database and is not being mounted in the package";
        } else {
            LogVerbose "SUCCESS: Database file=[$file] is configured in the running database and is part of package vg";
        }
    }

    #--------------------------------------------------------
    # if a database does have distributed files configured,
    # compare these on all remote nodes
    #--------------------------------------------------------
    LogVerbose "Checking for distributed database files";
    create_dstrb_files_hash_db ();

    # if any distributed files found
    if ( keys %{dstrb_file} ) {
        LogVerbose "Running distributed database file verification for switchnodes=[@switchnodes]";
        foreach my $remotenode (@switchnodes) {
            if ( "$host" ne "$remotenode" ) {
                get_remote_dstrb_files($remotenode);
                twonode_dstrb_file_compare($host, $remotenode);
            }
        }
    } else {
        LogVerbose "No distributed files found for verifcation";
    }

    #--------------------------------------------------------
    # verify the package mount points for the db exist on all
    # remote nodes. Reset the hash with the previous
    # distributed file data and fill it with package mount
    # points
    #--------------------------------------------------------
    LogVerbose "Checking database mount points on all remote cluster nodes";
    %dstrb_file = ();            # reset the hash
    %remote_dstrb_file = ();     # reset the hash

    get_package_root_mountpoints();
    create_dstrb_files_hash_mp();
    LogDebug "Running cluster database root-mount-point verification for switchnodes=[@switchnodes]";
    foreach my $remotenode (@switchnodes) {
        if ( "$host" ne "$remotenode" ) {
            get_remote_dstrb_files($remotenode);
            twonode_dstrb_dir_compare($host, $remotenode);
        }
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Find the lowest level = root mount points in @sg_fs_directory
#   PARAMETERS:
#      RETURNS:
#  DESCRIPTION: None
#     COMMENTS: The mount points listed in the @sg_fs_directory are mounted
#               exclusively on the node where the package gets executed. On
#               all other cluster nodes (= RemoteClusterNodes). These dirs
#               are mounted on the node running the package and unmounted
#               on the other cluster nodes. This function determines the
#               root mount points and ignoring all subtree mounts.
#
#      EXAMPLE:
#               @{$pkgenv{sg_fs_directory}=qw(/c /a/b /a/c /a/b/c /a);
#               @sorted_dirs=[/a /a/b /a/b/c /a/c /c]"
#               %root_mps="/a /c"
#============================================================================

sub get_package_root_mountpoints {
    LogEnterSub(@_);
    my $dir;

    my $file;                   # contains a root mountpoint
    my $sg_mp;                  # one package mountpoint
                                # sort the package mountpoints for speedup
    my @sorted_dirs = sort @{$pkgenv{sg_fs_directory}};

    foreach $dir (@{$pkgenv{sg_fs_directory}}) {
        $file = $dir;

        foreach $sg_mp (@sorted_dirs) {

            if ( $dir =~ m#^$sg_mp# ) {
                LogDebug "dir=[$dir] is  a subtree of sg_mp=[$sg_mp] found = 1; last";
                $file=$sg_mp;
                last;
            } else {
                LogDebug "dir=[$dir] not a subtree of sg_mp=[$sg_mp] found = 0";
            }
        }

        $dstrb_file{$file}{'found'} = 1;
        LogDebug "Determined parent-mount-point=[$file] for package (sub)-mount=[$dir]";
    }

    foreach my $file ( keys %{dstrb_file} ) {
        if ($dstrb_file{$file}{'found'} == 1) {
            LogVerbose "Determined [$file] is a package parent-mount-point in the cluster";
        }

    }
    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%dstrb_file ], ["dstrb_file"] ));

    LogReturnSub();
    return ;
}

#===  FUNCTION  =============================================================
#      PURPOSE: create a hash containing cluster mount points used by package
#               are copies amongst cluster nodes
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#               distributed files are not configured in a SG  package e.g.
#               they reside as a copy on each node
#============================================================================
sub create_dstrb_files_hash_mp {
    LogEnterSub(@_);

    my ($f, $l, $s, $ftype);

    # scan for any distributed files and save in
    # $dstrb_file hash so we can compare with remote
    # versions later on
    foreach my $file ( keys %{dstrb_file} ) {
            my $found = call_func_with_timeout_sig( $fasttimeout, \&check_for_file, $file );
            my $ftype='f';

            if ($found) {
                $ftype='d' if (-d $file);
                $f = $file;
                $l = "" if !defined( $l = readlink($f) );
                $s = lstat($f);

                my $act_mode = sprintf( "0%o", $s->mode & 0777 );
                $dstrb_file{$file}{'found'} = $found;
                $dstrb_file{$file}{'mode'} = $act_mode;
                $dstrb_file{$file}{'uid'} =  $s->uid;
                $dstrb_file{$file}{'gid'} =  $s->gid;
                $dstrb_file{$file}{'link'} = $l;
                $dstrb_file{$file}{'ftype'} = $ftype;
#            printf( "DSTRB,%s,%s,%s,%s,%s,%s\n",$filename,$found,$act_mode, $s->uid, $s->gid, $l) if defined $dstrbfiles;
                $dstrb_file{$file}{'mountp'} = $mpsfs{"$file"} if ( exists $mpsfs{"$file"});

            }
            else {
                # file was not found on the system
#                printf( "DSTRB,%s,%s\n", $file,$found);
                $dstrb_file{$file}{'found'} = $found;
            }
    }

    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%dstrb_file ], ["dstrb_file"] ));

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: create a hash containing those files which are distributed /
#               are copies amongst cluster nodes
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#               distributed files are not configured in a SG  package e.g.
#               they reside as a copy on each node
#============================================================================
sub create_dstrb_files_hash_db {
    LogEnterSub(@_);

    my ($f, $l, $s, $ftype);

    # scan for any distributed files and save in
    # $dstrb_file hash so we can compare with remote
    # versions later on
    foreach my $file ( keys %{dbstorage} ) {
        if ( !$dbstorage{$file}{mountbypackage} ) {
            my $found = call_func_with_timeout_sig( $fasttimeout, \&check_for_file, $file );
            my $ftype='f';

            if ($found) {
                $ftype='d' if (-d $file);
                $f = $file;
                $l = "" if !defined( $l = readlink($f) );
                $s = lstat($f);

                my $act_mode = sprintf( "0%o", $s->mode & 0777 );
                $dstrb_file{$file}{'found'} = $found;
                $dstrb_file{$file}{'mode'} = $act_mode;
                $dstrb_file{$file}{'uid'} =  $s->uid;
                $dstrb_file{$file}{'gid'} =  $s->gid;
                $dstrb_file{$file}{'link'} = $l;
                $dstrb_file{$file}{'ftype'} = $ftype;
#            printf( "DSTRB,%s,%s,%s,%s,%s,%s\n",$filename,$found,$act_mode, $s->uid, $s->gid, $l) if defined $dstrbfiles;
                $dstrb_file{$file}{'mountp'} = $mpsfs{"$file"} if ( exists $mpsfs{"$file"});

            }
            else {
                # file was not found on the system
#                printf( "DSTRB,%s,%s\n", $file,$found);
                $dstrb_file{$file}{'found'} = $found;
            }
        }
    }

    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%dstrb_file ], ["dstrb_file"] ));

    LogReturnSub();
    return;
}

#############################################################################
# SGeSAP SYSTEM / DB_SYSTEM / ORACLE
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: This function creates a verify list for an oracle system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub create_verifylist_db_system_oracle {
    LogEnterSub(@_);

    my ($SID, $sid);

    LogVerbose "Verify list create: for sgesap_db_system=$pkgenv{sgesap_db_system}";
    LogVerbose "Verify list create: for sgesap_db_vendor=$pkgenv{sgesap_db_vendor}";

    $SID = uc( $pkgenv{sgesap_db_system} );
    $sid = lc($SID);

    verifylist_add($RULESF, "db_system", $SID );

    if ( is_oracle12($SID) ){
      verifylist_add($RULESF, "oracle12", $SID );
    } else {
      verifylist_add($RULESF, "oracle", $SID );
    }

    verifylist_add( $RULESF, "files_sgesap",$SID);
    verifylist_add( $RULESF, "files_system",$SID);

    LogReturnSub();
    return;
}

#
#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#         Determine if this is an Oracle 12 environment. Also see
#         sapverify.rules file
#         it is oracle12 configuration if:
#         true: if ora${sid} doesn't exist (and ${sid}adm exists)
#         or
#         true: as the ${sid}adm user check owner of an oracle exe
#               ls -l `which tnsping`
#               ... oracle oinstall ... /oracle/O2A/121/bin/tnsping
#
#         Determing oracle12 is not that simple if all filesytems are unmounted
#         and package is down - thus several checks
###############################################################################
sub is_oracle12{
  LogEnterSub(@_);

	my $SID = shift;
  my $home;

  # first check - check non-existence of $orauser
  my $sid = lc($SID);
  my $sapadm="${sid}adm";
  my $orauser="ora${sid}";

  if ( defined $home_dir{$orauser} ) {
      $home = $home_dir{$orauser};
  } else {
      $home = ( getpwnam $orauser )[7];
  }
  if ( !defined $home ) {
      LogDebug "Can't find a orauser=[$orauser] - this is an Oracle12 environment";
      return 1;
  }
  else {
      # version unclear as oracle12 might still have an orauser configured
      # continue checking
      LogDebug "Found orauser=[$orauser] - trying to determine Oracle11/12 version";
  }


  # second check - use Oracle exe to determine owner
  my $param="tnsping";
  my $adm=$sapadm;
  my $cmd;
  my $path;

    $cmd = qq[su - $adm -c "which $param" 2>/dev/null];
    LogDebug("Execute cmd=[$cmd]");

    my @results = `$cmd`;
    my $rc      = $?;

    LogDebug "which $param result: rc=[$rc], results=[@results]";


    if ( $rc == 0 ) {

        for (@results) {
            chomp;
            if ( ( $^O eq "hpux" ) && ($_ =~ m#(/.*$param)$# )) {
                $path = $1;
                LogDebug "path=[$path]";
                last;
              # match absolute path and ignore any noise in output
            } elsif ((  $^O eq "linux" ) && ($_ =~ m#(^/[^ ]+$param)$#)) {
                $path = $_;
                LogDebug "path=[$path]";
                last;
            }
        }
    }
    else {

        # if a path to tnsping not found - will use Pre-Oracle12 as default
        LogDebug "no path to Oracle exe - setting default Oracle11 (or lower) environment";
    }


  if ( defined $path ) {
    my $file=$path;
    my ( $gid, $uid );
    # if Oracle12 configuration then owner oracle and not ora${sid}
    $uid = getpwnam( "oracle");
    my $info = stat($file);

    if ( $info->uid == $uid ) {
      my $tmp_str = sprintf "filename=[%s] uid=[%s] matches uid=[%s] for user=[%s] - this is an Oracle12 environment",
         $file, $info->uid, $uid, "oracle";
         LogDebug "$tmp_str";
         return 1;
    }
  }


  # third check - if exists grep oracle version in startdb.log of $sapadm user
  my $filename="startdb.log";
  my $startdblog_fh;
  $home = ( getpwnam $sapadm )[7];
  if ( defined $home ) {

    my $file="${home}/${filename}";
    if ( -e "$file" ) {
        LogDebug "File file=[$file] exists";
        open( $startdblog_fh, "<", "$file" ) or confess "Error: Can't open [$file]: $!";
        while (my $line = <$startdblog_fh>) {
            if ( $line =~ m#^TNS Ping.*Version (\d+)\.# ) {
              my $version=$1;
              LogDebug "Found Oracle version=[$version] in $file";
              if ($version >= 12) { return 1};
              last;
            }
        }
    } else {
        LogDebug "File file=[$file] not found";
    }
  }

  # return Oracle 11 environment as default
  return 0;
}


#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#        this subroutine executes a sequence of commands for writing to sqlplus
#        and reading the results. the strategy is to:
#       	setup an sql connection
#       	write some sql commands
#       	read and save the result into array @oracle_output
#       	close the sql connection
#       	return the array to a higher level (function, variabel)
###############################################################################
sub sub_execute_oracle_command{
    LogEnterSub(@_);

	my $oracle_command = shift;
	my @oracle_output =();

	sub_setup_oracle_cli();
	sub_write_oracle_cli($oracle_command);
	@oracle_output = sub_read_oracle_cli();
	sub_close_oracle_cli();

    LogReturnSub();
	return @oracle_output;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               setup the sqlplus or svrmgrl connection. basically it connects
#               to the db  via sqlplus /nolog and creates an input and output
#               filehandle
###############################################################################
sub sub_setup_oracle_cli {
    LogEnterSub(@_);


    my $sid = $pkgenv{sgesap_db_system};
    my $dbadm = "ora" . lc($sid);

    $READ_SQL_FH  = FileHandle->new;
    $WRITE_SQL_FH = FileHandle->new;
    my $pid;
    my $cmd;

    $cmd = qq[su - $dbadm -c "sqlplus '/as sysdba'" 2>&1];

    LogDebug "Executing command [$cmd]";

    #--------------------------------------------------------
    # open2() does not return an error-value.  instead is raises
    # an exception (calls "die()"), which can be trapped in an
    # "eval" block as below
    #--------------------------------------------------------
    eval { $pid = open2( $READ_SQL_FH, $WRITE_SQL_FH, $cmd ); };
    if ($@) {
        croak "open2 failed: error=$@";
    }

    #--------------------------------------------------------
    # turn on autoflush, i.e., turn off buffering of I/O
    #--------------------------------------------------------
    $READ_SQL_FH->autoflush;
    $WRITE_SQL_FH->autoflush;

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               write an sqlcommand and log it into a logfile
###############################################################################
sub sub_write_oracle_cli {
    LogEnterSub(@_);

    my $cmd = shift;

    LogDebug "$cmd";
    print $WRITE_SQL_FH "$cmd\n";

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               read the result of the sql command and log it into the LOG
#               file. also return the result in array @output

###############################################################################
sub sub_read_oracle_cli {
    LogEnterSub(@_);

    my $success = 0;
    my @output  = ();

#    LogVerbose "================================= Begin output";
    while (<$READ_SQL_FH>) {
        LogDebug "$_";
        push( @output, ($_) );
    }
#    LogVerbose "================================= End output";

    LogReturnSub();
    return @output;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               close the input and output filehandle
###############################################################################
sub sub_close_oracle_cli {
    LogEnterSub(@_);

    $READ_SQL_FH->close;
    $WRITE_SQL_FH->close;

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE: query oracle db for actual configured data files
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################

sub sql_get_oracle_db_storage {
    LogEnterSub(@_);

    # setup the Oracle select names from file query
    my $oracle_command = <<'EOF';
    select name from v$datafile;
    select name from v$tempfile;
    select member from v$logfile;
    exit
EOF

    # Run the Oracle query
    LogDebug "sql select name from v\$database, v\$tempfile, v\$logfile ...";
    my @oracle_output = sub_execute_oracle_command($oracle_command);
    LogDebug "Select command returned oracle_output=[@oracle_output]";

    my $count_datafile=0;
    foreach my $file (@oracle_output) {

        # a filename starts with a "/' in the oracle sql output
        if ( $file =~ m#^/# ) {
            chomp $file;
            $count_datafile++;
            LogVerbose "sql select returned datafile=[$file]";

            # my $dev = get_stat_dev($file);
            my $dev = call_func_with_timeout_sig( $fasttimeout, \&get_stat_dev, $file );

            # save the storage device data for the file
            $dbstorage{$file}{path} = $file;
            $dbstorage{$file}{dev}  = $dev;
        }
    }

    if ( $count_datafile == 0 ) {
        LogInfoF "Could not find any Oracle datafiles";
        LogInfoF "Supporting Info: SQL returned oracle_output=[@oracle_output]";
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: check if oracle is down or up with sqlplus
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub sql_is_oracle_up {
    LogEnterSub(@_);
    my ($SID) = @_;

    # setup the Oracle select names from file query
    my $oracle_command = <<'EOF';
    set linesize 240;
    select * from v$instance;
    exit;
EOF

    $oracle_up=0;
    $oracle_down=1;
    # Run the Oracle query
    LogDebug "sql select * from v\$intance...";
    my @oracle_output = sub_execute_oracle_command($oracle_command);
    LogDebug "Select command returned oracle_output=[@oracle_output]";

    foreach my $line (@oracle_output) {

        # sql select * from v\$intance generates error if down
        if ( $line =~ m#ORACLE not available# ) {
            chomp $line;
            $oracle_up=0;
            $oracle_down=1;
            LogInfoF "$line";
            last;
        }

        #
        if ( $line =~ m#$SID# ) {
            $oracle_up=1;
            $oracle_down=0;
            last;
        }

    }
    LogDebug "Oracle output =[@oracle_output]";
    LogDebug "oracle_up=[$oracle_up] oracle_down=[$oracle_down]";
    LogReturnSub($oracle_up);
    return $oracle_up;
}

#===  FUNCTION  =============================================================
#      PURPOSE: check if db is down or up with ps
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub is_oracle_up {
    LogEnterSub(@_);

    my $dbsid = $pkgenv{sgesap_db_system};
    my $dbuser = "ora" .lc($dbsid);
    my $SID = uc($dbsid);
    my $db_down=0;
    my $db_up=1;
    my $pscmd;

#    $pscmd = "pgrep -u $dbuser -f ora_...._$SID 2>&1"   if ( $^O eq "linux" );
    $pscmd = "pgrep -u $dbuser -f ora_...._$SID 2>&1";
    LogDebug "Running pscmd=[$pscmd]";

    my @results = qx($pscmd 2>&1);
    my $rc=$?;
    LogDebug "host=[$host] rc=[$rc] results=[@results]";

    # remap return code
    if ( $rc != 0 ) {
        LogVerbose "Ran pscmd=[$pscmd]";
        $db_down=1; $db_up=0;
#        LogVerbose "host=[$host] dbsid=[$dbsid] rc=[" ,$rc == -1 ? $rc : $rc>>8, "] results=[@results] db_up=[$db_up]";
        LogVerbose "host=[$host] dbsid=[$dbsid] rc=[$rc] results=[@results]";
    }
    LogDebug "host=[$host] db_up=[$db_up]";

    LogReturnSub($db_up);
    return $db_up;
}

#===  FUNCTION  =============================================================
#      PURPOSE: create a lvmtab hash
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: Scans /etc/lvmtab and creates  volume group ids, physical
#               volumes....
#               Note: This subroutine also gets executed on remote cluster
#               notes and result is parsed on local node: therefore if
#               $getremotelvm is set it is being executed on remote node
#               via cmexec
#============================================================================
sub create_lvmtab {
    LogEnterSub(@_);
    my ($getremotelvm) = @_;

    my $vg_name;
    my $pv_name;

    my $cmd;
    my $fh;
    my ( $pvcpuid, $pvid, $vgcpuid, $vgid );

    $cmd="strings /etc/lvmtab";
    my @lvm = `$cmd` or croak "cmd=[$cmd] failed";

    foreach my $line (@lvm) {
        LogDebug "$line";

        # if lvmtab line starts with a "^/dev"
        if ( $line =~ m#^/dev/# ) {
            chomp $line;
            if ( $line =~ m#^/dev/vg# ) {
                $vg_name = $line;
                next;
            }
            $pv_name = $line;

            ( $pvcpuid, $pvid, $vgcpuid, $vgid ) = get_vg_ids( $vg_name, $pv_name );
            if ("$vgid") {
                $lvmtab{$vg_name}{$vgid}{vgcpuid} = $vgcpuid;
                $lvmtab{$vg_name}{$vgid}{pv_name} = $pv_name;
                $lvmtab{$vg_name}{$vgid}{pvid}    = $pvid;
#                LogVerbose "Host [$host] lvmtab: vg_name[$vg_name] pv_name=[$pv_name] vgcpuid=[$vgcpuid] vgid=[$vgid] pvid=[$pvid]";
                LogDebug "Host [$host] lvmtab: vg_name[$vg_name] vgcpuid=[$vgcpuid] vgid=[$vgid] pv_name=[$pv_name] pvid=[$pvid]";

                # if running strings /etc/lvmtab on remote node prepentd REMOTELVMTAB
                print "REMOTELVMTAB,$vg_name,$pv_name,$vgcpuid,$vgid,$pvid\n" if $getremotelvm;
            }
        }
    }

    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%lvmtab ], ["lvmtab"] ));

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: create remote lvmtab from remote node
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub create_remotelvmtab {
    LogEnterSub(@_);
    my ($remotenode) = @_;

    my $cmd;
    my $fh;

    %remotelvmtab = ();    # hash/record - contains data from a remote /etc/lvmtab
    LogVerbose "Getting lvmtab for remote cluster node=[$remotenode]";
    $cmd = "$SGSBIN/cmexec -n $remotenode -t $timeout $SGSBIN/sapverify.pl -g 2>&1 |";
    LogVerbose("Execute \$cmd=[$cmd]");
    open( $fh, $cmd ) or LogRuntimeError "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

    while (<$fh>) {
        chomp;
        my ( $REMOTELVMTAB, $vg_name, $pv_name, $vgcpuid, $vgid, $pvid ) = split( /,/, $_ );
        LogDebug ("Host [$remotenode] lvmtab: vg_name=[$vg_name] vgcpuid=[$vgcpuid] vgid=[$vgid] pv_name=[$pv_name] pvid=[$pvid]");

        $remotelvmtab{$vg_name}{$vgid}{"vgcpuid"} = $vgcpuid;
        $remotelvmtab{$vg_name}{$vgid}{"pv_name"} = $pv_name;
        $remotelvmtab{$vg_name}{$vgid}{"pvid"}    = $pvid;
    }
    close($fh);
    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%remotelvmtab ], ["remotelvmtab"] ));

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: get the volume group ids...
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub get_vg_ids {
    LogEnterSub(@_);
    my ( $vg_name, $pv_name ) = @_;

    my $cmd;
    my $fh;
    my ( $pvcpuid, $pvid, $vgcpuid, $vgid );
    my @results;

    $cmd = "xd -An -j8200 -N16 -tx $pv_name 2>&1 |";
    LogDebug "Execute \$cmd=[$cmd]";

    open( $fh, $cmd ) or warn "open failed";
    while (<$fh>) {

        LogDebug "line=[$_]";
        next if ( grep( /^$/, $_ ) );
        $_ =~ s/^\s+//g;
        chomp;
        (@results) = split( /\s+/, $_ );

    }
    close($fh);

    #    if ("$vgid") {
    #        LogErrorF "Found vg_name[$vg_name] pv_name=[$pv_name] vgcpuid=[$vgcpuid] vgid=[$vgid] pvid=[$pvid]";
    #    }

    LogReturnSub();
    if (wantarray) {
        return @results ? @results : (undef);
    }

}


#===  FUNCTION  =============================================================
#      PURPOSE: print the /etc/lvmtab with vgids
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub twonode_vgid_compare {
    LogEnterSub(@_);
    my ($localnode, $remotenode, $vg_name) = @_;

    for my $vgid ( keys %{ $lvmtab{$vg_name} } ) {

#        LogInfoF    "Host=[$localnode] lvmtab: "
        LogVerbose    "Host=[$localnode] lvmtab: "
                    . "vg_name=[$vg_name] "
                    . "vgid=[$vgid] "
                    . "pvid=[$lvmtab{$vg_name}{$vgid}{pvid}] "
                    . "vgcpuid=[$lvmtab{$vg_name}{$vgid}{vgcpuid}] "
                    . "pv_name=[$lvmtab{$vg_name}{$vgid}{pv_name}] ";


        if ( defined $remotelvmtab{$vg_name}{$vgid} ) {

            #---------------------------------------------------
            # local and remote hash keys are the same (=vgid).
            # therefore volumes are correctly configured
            #---------------------------------------------------
#            LogInfoF "Host=[$remotenode] lvmtab: "
            LogVerbose "Host=[$remotenode] lvmtab: "
                     . "vg_name=[$vg_name] "
                     . "vgid=[$vgid] "
                     . "pvid=[$remotelvmtab{$vg_name}{$vgid}{pvid}] "
                     . "vgcpuid=[$remotelvmtab{$vg_name}{$vgid}{vgcpuid}] "
                     . "pv_name=[$remotelvmtab{$vg_name}{$vgid}{pv_name}] ";

#            LogInfoF "Verification of for vgid=[$vgid] for vg_name=[$vg_name] passed";
            LogVerbose "Verification for vg_name=[$vg_name] and vgid=[$vgid] passed";

        }
        else {
            LogInfoF "vg_name=[$vg_name] pv_name=[$lvmtab{$vg_name}{$vgid}{pv_name}] vgid=[$vgid] not found";
            LogInfoF "Supporting Info: Couldn't find a corresponding entry in /etc/lvmtab on remotenode=[$remotenode]";

        }
    }

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: clusterwide verify sap specific .??* files used during login
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#============================================================================

sub verify_user_login_startup_files {
    LogEnterSub(@_);

    my $adm = "";
    my ($SID, $sid);

    LogVerbose "Verifying user login/startup files";

    if ( exists $pkgenv{sgesap_db_system} ) {
        $sid = lc( $pkgenv{sgesap_db_system} );

        if ( $pkgenv{sgesap_db_vendor} eq "oracle" ) {
            $adm = "ora" . "$sid";
        }
        elsif ( $pkgenv{sgesap_db_vendor} eq "oracle12" ) {
            $adm = "$sid" . "adm";
        }
        elsif ( $pkgenv{sgesap_db_vendor} eq "maxdb" ) {
            $adm = "sqd" . "$sid";
        }
        elsif ( $pkgenv{sgesap_db_vendor} eq "db2" ) {
            $adm = "$sid" . "adm";
        }
        elsif ( $pkgenv{sgesap_db_vendor} eq "sybase" ) {
            $adm = "syb" . "$sid";
        }
        else {
            LogInfoF "host=[$host]", "sgesap_db_vendor=[$pkgenv{sgesap_db_vendor}] not supported";
        }

        LogVerbose "sgesap_db_vendor=$pkgenv{sgesap_db_vendor} sid=[$sid] adm=[$adm]";
        verify_login_startup_files($adm) if defined $adm;
    }
    if ( exists $pkgenv{sgesap_sap_system} ) {
        $sid = lc( $pkgenv{sgesap_sap_system} );
        $adm = "$sid" . "adm";
        LogVerbose "sgesap_sap_system=$pkgenv{sgesap_sap_system} sid=[$sid] adm=[$adm]";
        verify_login_startup_files($adm) if defined $adm;
    }

    if ( exists $pkgenv{sgesap_lc_system} ) {
        if (defined $pkgenv{sgesap_lc_system}[0]) {
            $sid = lc( $pkgenv{sgesap_lc_system}[0] );
            $adm = "$sid" . "adm";
            LogVerbose "sgesap_lc_system=[$pkgenv{sgesap_lc_system}[0]] sid=[$sid] adm=[$adm]";
            verify_login_startup_files($adm) if defined $adm;
        }
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Verify data files used by Oracle against configured volumes
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub verify_login_startup_files {
    LogEnterSub(@_);
    my ($user) = @_;

    my $exit_status;
    my $line;
    my $SID;
    my $cmd;
    my $fh;
    my $home;


    if ( defined $home_dir{$user} ) {
        $home = $home_dir{$user};
    }
    else {
        $home = ( getpwnam $user )[7];
        $home_dir{$user} = $home;
    }

    # if home directory is part of a failover fileystem
    # then we can't get any environment values
    if ( is_fs_mounted_by_package($home) ) {
        if ($PackageDown) {
            LogVerbose "Skipping login/startup file comparison";
            LogVerbose "Supporting Info: Home directory home=[$home] is mounted by (Package) filesystem and package is down";
        }
        else {
            LogVerbose " Checking login/startup files  home=[$home] is mounted by (Package) filesystem and package is up";
            LogVerbose "Running login file verification for switchnodes=[@switchnodes]";
            foreach my $remotenode (@switchnodes) {

                # if remotenode doesn't match this node
                if ( "$host" ne "$remotenode" ) {
                    check_startup_files_w_hostname( $host, $remotenode, $home );
                }
            }
        }
        LogReturnSub();
        return;

    }

    LogVerbose "Verifying login startup files for user=[$user] and login directory home=[$home]";

    #---------------------------------------------------
    # get the login/startup files on the local node
    #---------------------------------------------------
    get_startup_files($host, $home);

    LogVerbose "Running login startup verification for switchnodes=[@switchnodes]";
    foreach my $remotenode (@switchnodes) {

        # if remotenode doesn't match this node
        if ( "$host" ne "$remotenode" ) {

            #---------------------------------------------------
            # get the login/startup files on the remote node
            #---------------------------------------------------
            get_remote_startup_files( $remotenode, $home );

            #---------------------------------------------------
            # Compare volume group for local/remote lvmtab
#            #---------------------------------------------------
#            foreach my $file ( keys %{startup_file} ) {
#                LogVerbose "       startup_file=[$file]";
#            }
#            foreach my $file ( keys %{remote_startup_file} ) {
#                LogVerbose "remote_startup_file=[$file]";
#            }

#            foreach my $file ( keys %{startup_file} ) {
#                LogVerbose "Verifying startup_file=[$file]";

                twonode_startupfiles_compare($host,$remotenode, $home);
#            }
        }
    }
    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: Search for startup files containing hostname and check
#               if all hosts are specified
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#============================================================================
sub check_startup_files_w_hostname {
    LogEnterSub(@_);
    my ($localnode,$remotenode,$user) = @_;

    my $cmd;
    my $fh;
    my ($f, $l, $s);
    my ($file, $remotfile);


    my $startupfiles="$user"."/.??*";

    LogVerbose "Collecting startup files=[$startupfiles] with hostname";
    while (<${startupfiles}>) {
        LogVerbose "Checking file=[$_]";
        #---------------------------------------------------
        # get file attirbutes
        # set link to "" if file not a link
        #---------------------------------------------------
        $f = $_;
        $l = "" if !defined( $l = readlink($f) );
        $s = lstat($f);

        $file=$f;
        my $remotefile=$file;

        if ( $file =~ m#^(\S+)\.(\w+)_(\S+)\.(sh|csh)$# ) {
            LogDebug "Rule 1: matched [$file]";
            if ( $3 eq "$localnode" ) {
                LogDebug "Rule 2: matched localnode=[$localnode]";
                $remotefile = "$1\.$2_$remotenode\.$4";
            }
            LogVerbose "file=[$file] remotefile[$remotefile]";
            if (!-x $remotefile) {
                    LogInfoF "A node-specific-login file=[$remotefile] not found on remote host=[$remotenode]";
                    LogInfoF " Supporting Info: For consistency create node-specific files on all nodes or on none";
            }
            $filecount++;
        }
    }

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: generates a list of SAPs startup files .??*
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#============================================================================
sub get_startup_files {
    LogEnterSub(@_);
    my ($localnode,$user) = @_;

    my $cmd;
    my $fh;
    my ($f, $l, $s);

#    setparam( "debug", 1 );


    my $startupfiles="$user"."/.??*";
    LogVerbose "Collecting startup files =[$startupfiles]";
    while (<${startupfiles}>) {
        #---------------------------------------------------
        # get file attirbutes
        # set link to "" if file not a link
        #---------------------------------------------------
        $f = $_;
        $l = "" if !defined( $l = readlink($f) );
        $s = lstat($f);

        #---------------------------------------------------
        # use relativ name instead of full name
        #---------------------------------------------------
        my ($b) = basename($f);
        LogDebug "basename=[$b]";
        my $file = $b;

        #---------------------------------------------------
        # only collect SAP relevent startup files
        # $1 any whitespace string followed by underscore
        # $2 only hostname allowed
        #---------------------------------------------------
        if ( $file =~ m#^\.(\w+)_(\S+)\.(sh|csh)$# ) {
            LogDebug "Rule 1: matched [$file]";
            if ( $2 eq "$localnode" ) {
                LogDebug "Rule 2: matched localnode=[$localnode]";
            } else {
                next;
            }
        #---------------------------------------------------
        # match explicit: .(cshrc|login|profile)
        #---------------------------------------------------
        } elsif ( $file =~ m#^\.(cshrc|login|profile)$# ) {
            LogDebug "Rule3: file matched [$file]";
        #---------------------------------------------------
        # match .xxxxx.(sh|csh)
        #---------------------------------------------------
        } elsif ( $file =~ m#^\.(\w+)\.(sh|csh)$# ) {
            LogDebug "Rule 4: file matched [$file]";
        #---------------------------------------------------
        # discard
        #---------------------------------------------------
        } else {
            LogDebug "Rule 5: discarding file [$file]";
            next;
        }

        LogVerbose "Login file add:     filename=[$file]";

        # save mode in human readable form
        my $act_mode = sprintf( "0%o", $s->mode & 07777 );

        $startup_file{$f}{mode} = $act_mode;
        $startup_file{$f}{uid} = $s->uid;
        $startup_file{$f}{gid} = $s->gid;
        $startup_file{$f}{'link'} = $l;

        # if running on remote node prepend REMOTE
        printf "REMOTE,%04o,%s,%s,%s,%s\n", $s->mode & 07777, $s->uid, $s->gid, $f, $l if $userfiles;

    }

    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%startup_file ], ["startup_file"] ));


#setparam( "debug", 0 );

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: get remote startup files
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub get_remote_startup_files {
    LogEnterSub(@_);
    my ($remotenode, $user) = @_;

    my $cmd;
    my $fh;

    %remote_startup_file = ();    # hash/record - contains data from a remote startupfiles

    LogVerbose "Execution on remote cluster node=[$remotenode]";
    $cmd = "$SGSBIN/cmexec -n $remotenode -t $timeout $SGSBIN/sapverify.pl -u $user 2>&1 |";
    LogVerbose("Execute \$cmd=[$cmd]");
    open( $fh, $cmd ) or LogRuntimeError "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

    while (<$fh>) {
        chomp;
        LogDebug("Raw REMOTE=[$_]");

        my ( $REMOTE, $mode, $uid, $gid, $file, $link ) = split( /,/, $_ );
        if (  (defined $REMOTE) && ($REMOTE =~ m#^REMOTE$#) ) {
            $remote_startup_file{$file}{mode} = $mode ;
            $remote_startup_file{$file}{uid} = $uid;
            $remote_startup_file{$file}{gid} = $gid;
            $remote_startup_file{$file}{'link'} = $link;
        }
    }
    close($fh) or LogRuntimeError "Close failed for cmd=[$cmd]: \$?=[$?] \$!=[$!]";

    $dumpdata && LogVerbose  ( Data::Dumper->Dump( [ \%remote_startup_file ], ["remote_startup_file"] ));

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: Compare a hash with local startup files again remote startup
#               files
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub twonode_startupfiles_compare {
    LogEnterSub(@_);
    my ( $localnode, $remotenode, $home ) = @_;
    my $remotefile;
    my $package_mountedfs ;
    my $contains_nodename;


    foreach my $file ( keys %{startup_file} ) {
#        LogVerbose "Checking if file [$file]";

        #---------------------------------------------------
        # if startup is part of a package mounted file system
        #---------------------------------------------------
#        LogVerbose "Checking if file [$file] mounted via package";
        $package_mountedfs = is_fs_mounted_by_package($file);

        if ( $package_mountedfs ) {
            LogVerbose "Skipping startup file compare - "
                . "file [$file] is from an (Package) filesystem";
            next;
        }

        # first set remotefile to local file
        $remotefile = $file;


        #---------------------------------------------------
        # if file contains a local hostname replace
        # it with remote hostname for remotefile
        # to able to compare:
        # .sapenv_hpvm-0-43.sh vs .sapenv_hpvm-0-44.sh
        #---------------------------------------------------
#        if ( $file =~ m#^\.(\w+)_(\S+)\.(sh|csh)$# ) {
#            LogDebug "Rule 1: matched [$file]";
#            if ( $2 eq "$localnode" ) {
#                LogDebug "Rule 2: matched localnode=[$localnode]";
#                $remotefile = "\.$1_$remotenode\.$3";
#            }
#        }


        $contains_nodename=0;
        if ( $file =~ m#^(\S+)\.(\w+)_(\S+)\.(sh|csh)$# ) {
            LogDebug "Rule 1: matched [$file]";
            if ( $3 eq "$localnode" ) {
                LogDebug "Rule 2: matched localnode=[$localnode]";
                $remotefile = "$1\.$2_$remotenode\.$4";
                $contains_nodename=1;
            }
        }

#        LogVerbose "file=[$file] remotefile[$remotefile]";

        LogDebug " Local node=[$localnode] startup file: "
            . "mode=[$startup_file{$file}{mode}] "
            . "uid=[$startup_file{$file}{uid}] "
            . "gid=[$startup_file{$file}{gid}] "
            . "file=[$file] "
            . "link=[$startup_file{$file}{link}] ";

        if ( defined $remote_startup_file{$remotefile} ) {

            # if global errorcount is increment then a error was found
            my $fileerrorcount=$errorcount;
            LogDebug "Remote node=[$remotenode] startup file: "
                . "mode=[$remote_startup_file{$remotefile}{mode}] "
                . "uid=[$remote_startup_file{$remotefile}{uid}] "
                . "gid=[$remote_startup_file{$remotefile}{gid}] "
                . "file=[$remotefile] "
                . "link=[$remote_startup_file{$remotefile}{link}] ";

            if ( "$startup_file{$file}{uid}" ne "$remote_startup_file{$remotefile}{uid}" ) {
                LogInfoF "login file=[$file] "
                    . "uid=[$startup_file{$file}{uid}] doesn't match remote host=[$remotenode] "
                    . "file=[$remotefile] uid=[$remote_startup_file{$remotefile}{uid}]";
            }
            if ( "$startup_file{$file}{gid}" ne "$remote_startup_file{$remotefile}{gid}" ) {
                LogInfoF "login file=[$file] "
                    . "gid=[$startup_file{$file}{gid}] doesn't match remote host=[$remotenode] "
                    . "file=[$remotefile] gid=[$remote_startup_file{$remotefile}{gid}]";
            }
            if ( "$startup_file{$file}{mode}" ne "$remote_startup_file{$remotefile}{mode}" ) {
                LogInfoF "login file=[$file] "
                    . "mode=[$startup_file{$file}{mode}] doesn't match remote host=[$remotenode] "
                    . "file=[$remotefile] mode=[$remote_startup_file{$remotefile}{mode}]";
            }
            if ( "$startup_file{$file}{link}" ne "$remote_startup_file{$remotefile}{link}" ) {
                LogInfoF "login file=[$file] "
                    . "link=[$startup_file{$file}{link}] doesn't match remote host=[$remotenode] "
                    . "file=[$remotefile] link=[$remote_startup_file{$remotefile}{link}]";
            }

            LogVerbose "SUCCESS: host=[$remotenode] Verification for login file=[$file] passed" if $fileerrorcount==$errorcount ;
        }
        else {
            LogInfoF "remote host=[$remotenode] node-specific-login file=[$remotefile] not configured on all nodes";
            #LogInfoF " Supporting Info: node specific file=[$file] exists on host=[$localnode]" if $contains_nodename==1;
            LogInfoF " Supporting Info: For consistency create node-specific files on all nodes or on none";
        }
        $filecount++;
    }

    LogReturnSub();
    return;
}


#############################################################################
# SGeSAP SYSTEM / DB_SYSTEM / DB2
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: This function creates verify list for a db2 system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub create_verifylist_db_system_db2 {
    LogEnterSub(@_);

    my ($SID, $sid);

    LogVerbose "Verify list create: for sgesap_db_system=$pkgenv{sgesap_db_system}";
    LogVerbose "Verify list create: for sgesap_db_vendor=$pkgenv{sgesap_db_vendor}";

    $SID = uc( $pkgenv{sgesap_db_system} );
    $sid = lc($SID);

    verifylist_add($RULESF, "db_system", $SID );
    verifylist_add($RULESF, "db2", $SID );
    verifylist_add( $RULESF, "files_sgesap",$SID);
    verifylist_add( $RULESF, "files_system",$SID);

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub verify_db_storage_db2 {
    LogEnterSub(@_);

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: check if db is down or up with sqlplus
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub sql_is_db2_up {
    LogEnterSub(@_);
    my ($SID) = @_;

    my $db_down=1;
    my $db_up=0;

    # NOTE: don't indent
    my $sql_command = <<"EOF";
list active databases
terminate
EOF

    # Run the sql query
    LogDebug "list active databases ...";
    @sql_output = sub_execute_sql_command($sql_command);
    LogDebug "SQL command returned sql_output=[@sql_output]";

    foreach (@sql_output) {
        chomp;
        LogDebug "db2 list active databases  line=[$_]";
        if ( $_ =~ m#^\s*Database name\s*=\s*(\S+)\s*$# ){
            LogDebug "found database [$1]";
            if ($1 =~ m#$SID# ) {
                LogDebug "Database [$SID] is active";
                $db_down=0; $db_up=1;
                last;
            }
        }
    }

    LogReturnSub($db_up);
    return $db_up;
}

#===  FUNCTION  =============================================================
#      PURPOSE: check if db is down or up with ps
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub is_db2_up {
    LogEnterSub(@_);

    my $dbsid = $pkgenv{sgesap_db_system};
    my $dbuser = "sdb";
    my $SID   = uc($dbsid);

    my $db_down=1;
    my $db_up=0;

    my $rc = sql_is_db2_up($SID);
    if ( $rc == 1 ) {
        $db_down=0; $db_up=1;
    }

    LogReturnSub($db_up);
    return $db_up;

}

#===  FUNCTION  =============================================================
#      PURPOSE: query db for configured "files"
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS:
#
# su - db2crm sqllib/bin/db2 list active databases
#
# Database name                              = CRM
# Applications connected currently           = 17
# Database path                              = /db2/CRM/db2crm/NODE0000/SQL00001/
###############################################################################

sub sql_get_db2_db_storage {
    LogEnterSub(@_);


    my $DBSID;
    my $dbpath;
    my $line;

    #--------------------------------------------------------
    # get db2 automatic storage paths
    #--------------------------------------------------------
    # NOTE: don't indent
    my $sql_command = <<"EOF";
get snapshot for all databases global
terminate
EOF

#    # Run the sql query
    LogDebug "get snapshot for all databases global ...";
    @sql_output = sub_execute_sql_command($sql_command);
    LogDebug "SQL command returned sql_output=[@sql_output]";

    foreach (@sql_output) {

        chomp;
        LogDebug "db2 snapshot line=[$_]";


        if ( $_ =~ m#^Database name\s*=\s*(\S+)\s*$# ){
            $DBSID=$1;
            LogDebug "found database name DBSID=[$DBSID]";
            next;
        }
        if ( $_ =~ m#^Database path\s*=\s*(\S+)\s*$# ){
            $dbpath=$1;
            LogDebug "found database name dbpath=[$dbpath]";
            push ( @db2storage , $dbpath);
            next;
        }
        if ( $_ =~ m#^Automatic storage path\s*=\s*(\S+)\s*$# ){
            $dbpath=$1;
            LogDebug "found database name dbpath=[$dbpath]";
            push ( @db2storage , $dbpath);
            next;
        }

    }

    #--------------------------------------------------------
    #--------------------------------------------------------
    # NOTE: don't indent
    $sql_command = <<"EOF";
list database directory
terminate
EOF

#    # Run the sql query
    LogDebug "list database directory ...";
    @sql_output = sub_execute_sql_command($sql_command);
    LogDebug "SQL command returned sql_output=[@sql_output]";

    foreach (@sql_output) {
        chomp;
        LogDebug "db2 list database dir line=[$_]";
        if ( $_ =~ m#^\s*Local database directory\s*=\s*(\S+)\s*$# ){
            push ( @db2storage , $1);
            next;
        }
    }

    #--------------------------------------------------------
    #--------------------------------------------------------

    # NOTE: don't indent
    $sql_command = <<"EOF";
get snapshot for tablespaces on $DBSID
terminate
EOF

#    # Run the sql query
    LogDebug "get snapshot for tablespaces on ...";
    @sql_output = sub_execute_sql_command($sql_command);
    LogDebug "SQL command returned sql_output=[@sql_output]";

    foreach (@sql_output) {
        chomp;
        LogDebug "db2 get snapshot for tablespaces line=[$_]";
        if ( $_ =~ m#^\s*Container Name\s*=\s*(\S+)\s*$# ){
            push ( @db2storage , $1);
            next;
        }
    }

    #--------------------------------------------------------
    # get path to log file
    #--------------------------------------------------------

    # NOTE: don't indent
    $sql_command = <<"EOF";
get db cfg for $DBSID
terminate
EOF

#    # Run the sql query
    LogDebug "get db cfg for [$DBSID] ...";
    @sql_output = sub_execute_sql_command($sql_command);
    LogDebug "SQL command returned sql_output=[@sql_output]";

    foreach (@sql_output) {
        chomp;
        LogDebug "db2 get db cfg for line=[$_]";
        if ( $_ =~ m#^\s*Path to log files\s*=\s*(\S+)\s*$# ){
            push ( @db2storage , $1);
            next;
        }
    }

    #--------------------------------------------------------
    #--------------------------------------------------------
    @dbdatafiles = @db2storage;

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \@db2storage ], ["db2storage"] ));


    LogReturnSub();
    return;
}

#############################################################################
# SGeSAP SYSTEM / DB_SYSTEM / MAXDB
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: This function verifies
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub create_verifylist_db_system_maxdb {
    LogEnterSub(@_);

    my ($SID, $sid);

    LogVerbose "Verify list create: for sgesap_db_system=$pkgenv{sgesap_db_system}";
    LogVerbose "Verify list create: for sgesap_db_vendor=$pkgenv{sgesap_db_vendor}";

    $SID = uc( $pkgenv{sgesap_db_system} );
    $sid = lc($SID);

    verifylist_add($RULESF, "db_system", $SID );
    verifylist_add($RULESF, "maxdb", $SID );
    verifylist_add( $RULESF, "files_sgesap",$SID);
    verifylist_add( $RULESF, "files_system",$SID);

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub verify_db_storage_maxdb {
    LogEnterSub(@_);

    LogReturnSub();
    return;
}


#===  FUNCTION  =============================================================
#      PURPOSE: check if db is down or up with sqlplus
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub sql_is_maxdb_up {
    LogEnterSub(@_);
    my ($SID) = @_;

    # not used for now
    my $db_down=1;
    my $db_up=0;

    LogReturnSub($db_up);
    return $db_up;
}


#===  FUNCTION  =============================================================
#      PURPOSE: check if db is down or up with ps
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub is_maxdb_up {
    LogEnterSub(@_);

    my $dbsid = $pkgenv{sgesap_db_system};
    my $dbuser = "sdb";
    my $SID   = uc($dbsid);
    my $db_down=0;
    my $db_up=1;

#    my $pscmd = "pgrep -u $dbuser -f \"kernel $SID\""   if ( $^O eq "linux" );
    my $pscmd = "pgrep -u $dbuser -f \"kernel $SID\"";
    LogDebug "Running pscmd=[$pscmd]";

    my @results = qx($pscmd 2>&1);
    my $rc=$?;
    LogDebug "host=[$host] rc=[$rc] results=[@results]";

    # remap return code
    if ( $rc != 0 ) {
        LogVerbose "Ran pscmd=[$pscmd]";
        $db_down=1; $db_up=0;
        LogVerbose "host=[$host] dbsid=[$dbsid] rc=[$rc] results=[@results]";
    }
    LogDebug "host=[$host] db_up=[$db_up]";

    LogReturnSub($db_up);
    return $db_up;
}

#===  FUNCTION  =============================================================
#      PURPOSE: query db for configured "files"
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#
# a maxdb param_getvolsall entry looks somewhat like this:
# su - sqdsx1 "dbmcli -U c -d SX1 param_getvolsall"
# LOG_VOLUME_NAME_001              819200     F  /sapdb/SX1/saplog/DISKL001
# DATA_VOLUME_NAME_0001            563200     F  /sapdb/SX1/sapdata/DISKD0001
# DATA_VOLUME_NAME_0002            563200     F  /sapdb/SX1/sapdata/DISKD0002
# DATA_VOLUME_NAME_0003            563200     F  /sapdb/SX1/sapdata/DISKD0003
# DATA_VOLUME_NAME_0004            563200     F  /sapdb/SX1/sapdata/DISKD0004
# DATA_VOLUME_NAME_0005            563200     F  /sapdb/SX1/sapdata/DISKD0005
# DATA_VOLUME_NAME_0006            563200     F  /sapdb/SX1/sapdata/DISKD0006
###############################################################################

sub sql_get_maxdb_db_storage {
    LogEnterSub(@_);
    my $param_getvolsall = \%param_getvolsall;

    # NOTE: don't indent
    my $sql_command = <<"EOF";
param_getvolsall
exit
EOF

#    # Run the sql query
    LogDebug "maxdb/livecache param_getvolsall...";
    @sql_output = sub_execute_sql_command($sql_command);
    LogDebug "SQL command returned sql_output=[@sql_output]";

    foreach (@sql_output) {

        ## skip uninteresting lines
        next unless ($_ =~ /^(MLOG_VOLUME|DATA_VOLUME|LOG_VOLUME|LogVolume|DataVolume)/);
        chomp;                          # remove newline
        LogDebug "maxdb volume line=[$_]";
        my @line =  split(/ +/, $_);
        LogDebug "MaxDB volume=[@line]";

        $param_getvolsall->{"$line[3]"}={
                                'vname'    => $line[0],
                                'size'     => $line[1],
                                'type'     => $line[2],
                                'fname'    => $line[3],
                                }
    }

    foreach my $file ( keys %param_getvolsall) {
        LogVerbose "sql select returned volume file=[$file]";
        push (@dbdatafiles, $file);
    }

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%param_getvolsall ], ["param_getvolsall"] ));

    LogReturnSub();
    return;
}

#############################################################################
# SGeSAP SYSTEM / DB_SYSTEM / SYBASE
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: This function creates verify list for a sybase system
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================
sub create_verifylist_db_system_sybase {
    LogEnterSub(@_);

    my ($SID, $sid);

    LogVerbose "Verify list create: for sgesap_db_system=$pkgenv{sgesap_db_system}";
    LogVerbose "Verify list create: for sgesap_db_vendor=$pkgenv{sgesap_db_vendor}";

    $SID = uc( $pkgenv{sgesap_db_system} );
    $sid = lc($SID);

    verifylist_add($RULESF, "db_system", $SID );
    verifylist_add($RULESF, "sybase", $SID );
    verifylist_add( $RULESF, "files_sgesap",$SID);
    verifylist_add( $RULESF, "files_system",$SID);

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#        this subroutine executes a sequence of commands for writing to sqlplus
#        and reading the results. the strategy is to:
#       	setup an sql connection
#       	write some sql commands
#       	read and save the result into array @sql_output
#       	close the sql connection
#       	return the array to a higher level (function, variabel)
###############################################################################
sub sub_execute_sql_command{
    LogEnterSub(@_);

	my $sql_command = shift;
	my @sql_output =();

	sub_setup_oracle_cli()  if ( $pkgenv{sgesap_db_vendor} eq "oracle" );
	sub_setup_db2_cli()     if ( $pkgenv{sgesap_db_vendor} eq "db2" );
	sub_setup_maxdb_cli()   if ( $pkgenv{sgesap_db_vendor} eq "maxdb" ||   $pkgenv{sgesap_db_vendor} eq "livecache" );
	sub_setup_sybase_cli()  if ( $pkgenv{sgesap_db_vendor} eq "sybase" );

	sub_write_sql_cli("$sql_command");
	@sql_output = sub_read_sql_cli();
	sub_close_sql_cli();


    LogReturnSub();
	return @sql_output;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               setup a db connection.
###############################################################################
sub sub_setup_db2_cli {
    LogEnterSub(@_);

    my $dbsid = $pkgenv{sgesap_db_system};
    my $DBSID = uc($dbsid);
    my $dbadm = "db2" . lc($dbsid);

    $READ_SQL_FH  = FileHandle->new;
    $WRITE_SQL_FH = FileHandle->new;
    my $pid;
    my $cmd;

    $cmd = qq[su - $dbadm -c "sqllib/bin/db2" 2>&1];
    LogDebug "Executing command [$cmd]";

    #--------------------------------------------------------
    # open2() does not return an error-value.  instead is raises
    # an exception (calls "die()"), which can be trapped in an
    # "eval" block as below
    #--------------------------------------------------------
    eval { $pid = open2( $READ_SQL_FH, $WRITE_SQL_FH, $cmd ); };
    if ($@) {
        croak "open2 failed: error=$@";
    }

    #--------------------------------------------------------
    # turn on autoflush, i.e., turn off buffering of I/O
    #--------------------------------------------------------
    $READ_SQL_FH->autoflush;
    $WRITE_SQL_FH->autoflush;

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               setup a db connection.
###############################################################################
sub sub_setup_maxdb_cli {
    LogEnterSub(@_);

    my $dbsid = $pkgenv{sgesap_db_system};
    my $DBSID = uc($dbsid);
    my $dbadm = "sqd" . lc($dbsid);
    my $dbuser = "c";

    $dbuser =  $pkgenv{sgesap_maxdb_userkey} if exists $pkgenv{sgesap_maxdb_userkey};

    if  ($pkgenv{sgesap_db_vendor} eq "livecache" ) {
        $dbadm = lc($dbsid) . "adm";
        $dbuser = $pkgenv{sgesap_lc_userkey}[0];
    }

    $READ_SQL_FH  = FileHandle->new;
    $WRITE_SQL_FH = FileHandle->new;
    my $pid;
    my $cmd;

    $cmd = qq[su - $dbadm -c "dbmcli -U $dbuser -d $DBSID" 2>&1];
    LogDebug "Executing command [$cmd]";

    #--------------------------------------------------------
    # open2() does not return an error-value.  instead is raises
    # an exception (calls "die()"), which can be trapped in an
    # "eval" block as below
    #--------------------------------------------------------
    eval { $pid = open2( $READ_SQL_FH, $WRITE_SQL_FH, $cmd ); };
    if ($@) {
        croak "open2 failed: error=$@";
    }

    #--------------------------------------------------------
    # turn on autoflush, i.e., turn off buffering of I/O
    #--------------------------------------------------------
    $READ_SQL_FH->autoflush;
    $WRITE_SQL_FH->autoflush;

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               setup a db connection.
###############################################################################
sub sub_setup_sybase_cli {
    LogEnterSub(@_);

    my $dbsid = $pkgenv{sgesap_db_system};
    my $dbadm = lc($dbsid) . "adm";

    $READ_SQL_FH  = FileHandle->new;
    $WRITE_SQL_FH = FileHandle->new;
    my $pid;
    my $cmd;

    $cmd = qq[su - $dbadm -c "isql -X -l 10 -U$SGeSAP_ASEUSER -S$dbsid -D$dbsid -s'|' -w 1024 -b" 2>&1];
    LogDebug "Executing command [$cmd]";

    #--------------------------------------------------------
    # open2() does not return an error-value.  instead is raises
    # an exception (calls "die()"), which can be trapped in an
    # "eval" block as below
    #--------------------------------------------------------
    eval { $pid = open2( $READ_SQL_FH, $WRITE_SQL_FH, $cmd ); };
    if ($@) {
        croak "open2 failed: error=$@";
    }

    #--------------------------------------------------------
    # turn on autoflush, i.e., turn off buffering of I/O
    #--------------------------------------------------------
    $READ_SQL_FH->autoflush;
    $WRITE_SQL_FH->autoflush;

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               write an sqlcommand and log it into a logfile
###############################################################################
sub sub_write_sql_cli {
    LogEnterSub(@_);

    my $cmd = shift;

    LogDebug "$cmd";
    print $WRITE_SQL_FH "$cmd\n";

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               read the result of the sql command and log it into the LOG
#               file. also return the result in array @output

###############################################################################
sub sub_read_sql_cli {
    LogEnterSub(@_);

    my $success = 0;
    my @output  = ();

    LogDebug "================================= Begin output";
    while (<$READ_SQL_FH>) {
        LogDebug "$_";
        push( @output, ($_) );
    }
    LogDebug "================================= End output";

    LogReturnSub();
    return @output;
}

#===  FUNCTION  =============================================================
#      PURPOSE:
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#               close the input and output filehandle
###############################################################################
sub sub_close_sql_cli {
    LogEnterSub(@_);

    $READ_SQL_FH->close;
    $WRITE_SQL_FH->close;

    LogReturnSub();
    return;
}
#===  FUNCTION  =============================================================
#      PURPOSE: check if db is down or up with sqlplus
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub sql_is_sybase_up {
    LogEnterSub(@_);
    my ($SID) = @_;

    # not used for now
    my $db_down=1;
    my $db_up=0;

}

#===  FUNCTION  =============================================================
#      PURPOSE: check if db is down or up with ps
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub is_sybase_up {
    LogEnterSub(@_);

    my $dbsid = $pkgenv{sgesap_db_system};
    my $dbuser = "syb" .lc($dbsid);
    my $SID   = uc($dbsid);
    my $db_down=0;
    my $db_up=1;
    my $pscmd;

    $pscmd = "/usr/bin/pgrep -u $dbuser -f dataserver";
    LogDebug "Running pscmd=[$pscmd]";

    my @results = qx($pscmd 2>&1);
    my $rc=$?;
    LogDebug "host=[$host] rc=[$rc] results=[@results]";

    # remap return code
    if ( $rc != 0 ) {
        LogVerbose "Ran pscmd=[$pscmd]";
        $db_down=1; $db_up=0;
        LogVerbose "host=[$host] dbsid=[$dbsid] rc=[$rc] results=[@results]";
    }
    LogDebug "host=[$host] db_up=[$db_up]";

    LogReturnSub($db_up);
    return $db_up;
}


#===  FUNCTION  =============================================================
#      PURPOSE: query sysbase db for configured "files"
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#
# sybase isql command to query the database
# Note: -b flag suppresses column headers.
#       -w 1024 sets output width
#       -s separator between field
#   su - s7sadm -c "isql -l 10 -Usapsa -SS7S -DS7S -s'|' -w 1024" <<EOF
#
# a sysbase sp_helpdevice entry looks somewhat like this:
#
# |device_name|physical_name|description|status|cntrltype|vdevno|vpn_low|vpn_high|
# --------------------------------------------------------------------------------
# |datasaptools_1|/sybase/S7S/sapdiag/datasaptools_1.dat\
#       |file system device, special, dsync off, directio on, physical disk, 2048.00 MB, Free: 0.00 MB\
#       |2|0|7|0|1048575

###############################################################################
sub sql_get_sybase_db_storage {
    LogEnterSub(@_);

    my $sp_helpdevice_ref = \%sp_helpdevice;

    # NOTE: no indentation
    my $sql_command = <<"EOF";
$SGeSAP_ASEPASSWD
sp_helpdevice
go
exit
EOF

#    # Run the sql query
    LogDebug "sybase sp_helpdevice ...";
    @sql_output = sub_execute_sql_command($sql_command);
    LogDebug "SQL command returned sql_output=[@sql_output]";

    foreach (@sql_output) {

        ## skip uninteresting lines
                                        # only interested in sybase physical disks here
        next unless ($_ =~ / physical disk,/);
        chomp;                          # remove newline
        s/^\|//;                        # remove the leading "|" separator so
                                        # following split doesn't hiccup
        my @line = map {s/^\s+|\s+$//g; $_} split(/\|/, $_);

        LogDebug "Sybase physical disk=[@line]";

        $sp_helpdevice_ref->{"$line[1]"}={
                                'device'   => $line[0],
                                'physical' => $line[1],
                                'desc'     => $line[2],
                                'status'   => $line[3],
                                'cntrl'    => $line[4],
                                'vpn_low'  => $line[5],
                                'vpn_high' => $line[6],
                                }
    }

    foreach my $file ( keys %sp_helpdevice) {
        LogVerbose "sql select returned device file=[$file]";
        push (@dbdatafiles, $file);
    }

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%sp_helpdevice ], ["sp_helpdevice"] ));

    LogReturnSub();
    return;
}


#############################################################################
#
# Get HANA DB info by scanning /sapmnt/<SID>/profile for *HDB* profiles
#
# Arguments:
#   $SID        SID of HDB database
#
# Returns:
#   Array containing $dbsid (usually $SID), $dbtype ("HDB" here) and
#   replication mode (primary, sync or empty)
#
#############################################################################

sub gethdbinstance {

    LogEnterSub( @_ );
    my $SID    = shift;

    #######################################
    # set up global variables and hashes
    #######################################

    # Initialize %addrs      maps IP          -> subnet,
    # Initialize %subnets    maps subnet      -> IP
    %addrs   = ();
    %subnets = ();

    # Initialize %mpsfs      maps mountpoints -> filesystems
    # Initialize %mpsfst     maps mountpoints -> filesystem types
    # Initialize %vxdgshared maps ...
    %mpsfs      = ();
    %mpsfst     = ();
    %vxdgshared = ();

    # directory containing the SAP profiles
    my $profiledir = "/sapmnt/$SID/profile";

    # the default SAP profile
    my $DEFAULTPFL = $profiledir . "/DEFAULT.PFL";

    # convert to lower case and set the sidadm user
    $sidadm = lc($SID) . "adm";

    # Some configs might have a different SID for the DB
    my $dbsid = $SID;

    #######################################
    # Start collecting data
    #######################################

    # get API version to figure out which DBs are supported
    # call only if it is still 0
#TODO
#    $apiversion = getapiversion() if $apiversion==0;

    my $dbtype="";
    @profiles=();

    CollectListOfSIDProfiles( $SID, "/hana/shared/$SID/profile", "HDB" ); # need to check profiles for filtering, writes @profiles

    # Fallback to /usr/sap/$SID, if nothing found
    if (scalar @profiles == 0 ) {
      CollectListOfSIDProfiles( $SID, "/usr/sap/$SID/profile", "HDB" ); # need to check profiles for filtering, writes @profiles
    }

    # Fallback to /sapmnt/$SID, if nothing found
    if (scalar @profiles == 0 ) {
      CollectListOfSIDProfiles( $SID, "/sapmnt/$SID/profile", "HDB" ); # need to check profiles for filtering, writes @profiles
    }

    # if array @profiles > 0 then it must contain a HANA instance
    if (scalar @profiles > 0 ) {
          $dbtype="HDB";
    }

    # in case of unsupported dbtypes dbsid is also invalidated,
    # return immediately to avoid unnecessary processing & output
    return unless $dbtype;

    my $hdbrepl = gethdbreplmode ($SID);

    # introducing global var ($hdbukenv and) $hdbuk
    $hdbuk = "${dbsid}SAPDBCTRL";
    my $hdbukenv = gethdbuserkey($dbsid, "$hdbuk");
    if ($hdbukenv eq "") {
      # Fallback to DEFAULT
      $hdbuk = "DEFAULT";
      $hdbukenv = gethdbuserkey($dbsid, "$hdbuk");
    }

    if ($hdbukenv ne "") {
      my $vhost = (split(':',$hdbukenv))[0];
      #LogVerbose ("hdb uk vhost: $vhost");
      LogVerbose ("HANA user key vhost=[$vhost]");
      # TODO: check if vhost is local hostname, then ignore
    }
    else {
      $hdbuk = "";
    }

    LogReturnSub ("$dbsid,$dbtype,$hdbrepl");

    return ($dbsid,$dbtype,$hdbrepl);
}

#############################################################################
#
# Get HANA DB replication info
#
# Reads global.ini from /usr/sap/<SID>/global/custom/config
#
# Scans for
#  [system_replication]
#  mode = primary | sync
#
# Arguments:
#   $SID        SID of HDB database
#
# Returns: "primary" or "sync" (or whatever comes after "mode =", else ""
#
#############################################################################

sub gethdbreplmode {

    LogEnterSub( @_ );
    my $SID     = shift;
    my $globini = "/usr/sap/${SID}/SYS/global/hdb/custom/config/global.ini";
    my $replmode = "";

    unless (-r $globini) {
      warn "$globini not found";
      LogReturnSub ($replmode);
      return ($replmode);
    }

    open FILE,'<',$globini || warn "Couldn't open $globini";
    my $sysrepsection=0;
    while (<FILE>) {
      #print $_;
      if ($_ =~ m#\[system_replication\]#) {
        #LogVerbose("sysrepsection=1");
        $sysrepsection = 1;
        next;
      }
      if ($sysrepsection) {
        if ($_ =~ m#^mode\s*=\s*(\w+)\s*#) {
          $replmode = $1;
          LogVerbose("Found in global.ini: replmode=[$1]");
          last; # found
        }
        if ($_ =~ m#\[\w+\]#) {
          $sysrepsection = 0;
          last; # next section -> not found
        }
      }
    }
    close FILE;

    LogReturnSub ($replmode);
    return ($replmode);
}
#############################################################################
#
# Get HANA DB userkey environment
#
# Get the hdbuserkey ENV
#
# Arguments:
#   $SID        SID of HDB database
#   $hdbuk      Userkey to check
#
# Calls external programs hdbuserstore (and hdbsql - later ).
#
# Returns:ENV contents or emtpy string
#
#############################################################################
sub gethdbuserkey {

  LogEnterSub( @_ );
  my $SID    = shift;
  my $hdbuk  = shift;
  # my @uklist = ("${SID}SAPDBCTRL","DEFAULT");
  my $sidadm = lc($SID)."adm";
  my $hdbukenv = "";

#  foreach my $uk (@uklist) {
    # print "$uk\n";
    my $cmd=qq[su - $sidadm -c "hdbuserstore LIST $hdbuk " 2>/dev/null |];
    LogVerbose("Execute \$cmd=[$cmd]");
    open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

    while (<$fh>) {
       last if ( $_ =~ "KEY $hdbuk NOT FOUND" );
       if ($_ =~ /^KEY $hdbuk$/) {
         #LogVerbose "Key $hdbuk found";
         LogDebug "Key $hdbuk found";
         # $hdbuk = $uk;
         # assume that ENV: comes next
         if (<$fh> =~ /ENV\s*:\s*(\S+)/) {
            $hdbukenv = $1;
            #LogVerbose "\$hdbukenv = $hdbukenv";
            LogDebug "\$hdbukenv = $hdbukenv";
            last;
         }
       }
    }
    # close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";
    close($fh);
#  }
  LogReturnSub("\$hdbukenv=[$hdbukenv]");
  return ($hdbukenv);
}


#############################################################################
# Collect all SAP profiles found in the $profiledir="/sapmnt/$SID/profile" .
# A SAP profile has the following format <SID>_<INSTNAME><INSTNR>_<HOST>
#############################################################################

sub CollectListOfSIDProfiles {
    LogEnterSub( @_ );

    my $SID        = shift;
    my $profiledir = shift;
    my $filtertype = shift; # maybe undefined, key for InstanceFilters (already checked if allowed)

    LogVerbose("Collecting profiles in profiledir=[$profiledir]");

    # Do (maybe another) NFS check
    # Trick: Build up some (probably) non-exisiting filename to get around NFS caching issues when the cache
    #   still has information on $profiledir, but NFS isn't available anymore. In such a case, a check on $profiledir
    #   will succeed, but the subsequent opendir will hang.
    # (assumes that a check to $profiledir has been done before to ensure that it exists)
    return if checkNFSdiraccess($profiledir."/$$") == 130; # doesn't exist, but may timeout

    # scan for SAP profiles found in a SAP/SID  profile directory

    if ( ! opendir( DIR_HANDLE, $profiledir ) ){
        LogVerbose "Info: No profiledir at [$profiledir]";
        return;
    }

    my @filelist = readdir(DIR_HANDLE);
    closedir(DIR_HANDLE);

    foreach my $file (@filelist) {
        # skip any profiles containing a "." in the filename
        # this eliminates hidden files as well as profiles
        # containing a version number

        my $filtered_out=0;

        if ($file =~ /[.]/) {
            LogDebug("skip profile_file=[$file]");
            next;
        }

        # if a profile name was found that belongs to the SID, save it
        # in array @profiles. NOTE: saving it with absolute pathname because
        # required later
        if ($file =~ /^${SID}_/) {
            $filtered_out=0;
            if (defined $filtertype)
            {
              #LogVerbose ("Using filtertype $filtertype");
              $filtered_out=1;
              if ($file =~ /^${SID}_${InstanceFilters{$filtertype}}[0-9][0-9]_/) {
                LogVerbose ("filtered in ($filtertype): $file");
                $filtered_out=0;
                #  last;
              }
            }
            next if $filtered_out;
            LogVerbose("push onto array \@profiles profile_file=[$profiledir/$file]");
            push( @profiles, "$profiledir/$file" );

            # if this profile also is a ERS profile type , save it in array @ersprofiles
            # NOTE: saving it with absolute pathname because required later
            if ($file =~ /^${SID}_ERS[0-9][0-9]_/) {
                LogVerbose("push onto array \@ersprofiles profile_file=[$profiledir/$file]");
                push( @ersprofiles, "$profiledir/$file" );
            }
        }
    }
}



#############################################################################
#
# checkNFSdiraccess - check if a filespec is a directory and accessible on NFS
#
# Arguments:
#  - filespec
#
# Creates a subprocess for the check avoid the mainprocess to hang on NFS.
# Sets up a signal handler for ALRM signal and uses the function alarm() for timeout purposes.
# On timeout the subprocess is killed with signal INT.
#
# NFS caching effects may lead to wrong results if executed shortly after NFS has gone
# away or is available again
#
# Returns:
# 0 - is a directory (and accessible)
# 2 - not a directory (but accessible)
# 130 - check timed out (subprocess INTerrupted)
#
#############################################################################
sub checkNFSdiraccess {
  LogEnterSub( @_ );
  my $dirtocheck = shift;
  my $childpid = undef;
  my $rc=0;

  $childpid = fork();

  if ($childpid) {
    # in parent
    local $SIG{ALRM} = sub {
          LogError("Signal handler in checkNFSaccessdir triggered - terminate child with: kill INT $childpid");
          kill 'INT', $childpid;
    };
    alarm 5;
    waitpid ($childpid,0);
    my $childrc = $?;
    alarm 0;
    if ($childrc>>8) {
      $rc=2
    }
    elsif ((($childrc>>8) == 0) && ($childrc&127)) {
      LogError("Access failed - timed out on $dirtocheck");
      $rc=130;  # should be 128+$rc&127
    }
    LogReturnSub ("\$rc=$rc");
    return $rc;
  }
  else {
    # child
    exit 0 if (-d $dirtocheck);  # is dir (and accessible)?
    exit 2;  # not a dir
  }
}

#############################################################################
# This function scans a SAP profile file if the High Availablity library
# has been configured in the SAP profile
#     service/halib=/opt/cmcluster/lib/saphpsghalib.so
# and returns the values
#
# Arguments:
#   $file   - SAP profile file to read
#
# Returns if previously not defined:
#   $halib - Found/Not found
#############################################################################
sub GetHalibFromProfile {
    LogEnterSub( @_ );

    my $profile    = shift;
    my $halib   = "";

    LogDebug("parsing file [$profile] ...");
    my $fh = FileHandle->new($profile);
    if ( !defined($fh) ) {
      LogVerbose "ATTENTION: Opening $profile for reading failed - returning";
      return;
    }

    while (<$fh>) {
        if ( m#^service/halib\s*=\s*(\S+)$# ) {
            $halib = $1;
            last;
        }
    }
    $fh->close;
    LogDebug("Subroutine GetHalibFromProfile() returning halib=[$halib] for profile=[$profile]");
    return ($halib);
}


#############################################################################
# HDB_SYSTEM
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: Verifies halib has been configured in an SAP profile
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################

sub verify_for_halib() {
    LogEnterSub(@_);

#   TODO
#   in package conf file: SGeSAP_HDB_SYSTEM[0]="HA1" is an array
#    my $SID = uc( $pkgenv{sgesap_hdb_system} );

    my $SID="";
    if (defined $pkgenv{sgesap_hdb_system}[0] ) {
        $SID = uc( $pkgenv{sgesap_hdb_system}[0] );
        LogVerbose "Scanning for High Availability library [halib] configuration in SAP profiles of SID=[$SID]";
    } else {
        return;
    }
    @hdbvalues=gethdbinstance ($SID);

    my $profile;
    my $hafound="";

    LogDebug("profiles=[@profiles]");

    foreach my $p (@profiles) {
        $profile=$p;
        $hafound=GetHalibFromProfile($p);
        LogDebug("hafound=[$hafound] profile=[$profile]");
        last if ( ! $hafound );
    }

    # if not yet found search the default profiles for halib
    # try new hana profile dir
    if ( ! $hafound) {
        $profile = "/hana/shared/$SID/profile/DEFAULT.PFL";
        $hafound=GetHalibFromProfile($profile) if ( -e $profile ) ;
    }

    if ( $hafound eq "") {
        $profile = "/usr/sap/$SID/profile/DEFAULT.PFL";
        $hafound=GetHalibFromProfile($profile) if ( -e $profile ) ;
    }

    # try old hana profile dir
    if ( ! $hafound) {
        $profile = "/sapmnt/$SID/profile/DEFAULT.PFL";
        $hafound=GetHalibFromProfile($profile) if ( -e $profile ) ;
    }

    if ( ! $hafound) {
        LogInfoF "No High Availability library [halib] configured in SAP profiles for SID=[$SID]";
    } else {
        LogVerbose "Found halib [$hafound] in SAP profile [$profile] for SID=[$SID]";
    }

}

#############################################################################
# SGeSAP SYSTEM / GENERIC
#############################################################################

#===  FUNCTION  =============================================================
#      PURPOSE: Verify lvm configuration for HP-UX
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
###############################################################################
sub verify_lvm_hpux {
    LogEnterSub(@_);

    #---------------------------------------------------
    # if HPUX   scan / create / verify  /etc/lvmtab
    #---------------------------------------------------
    create_lvmtab();

    LogVerbose "Running volume verification for switchnodes=[@switchnodes]";
    foreach my $remotenode (@switchnodes) {

        # if remotenode doesn't match this node
        if ( "$host" ne "$remotenode" ) {

            #---------------------------------------------------
            # create a remote lvmtab to be able to
            # compare the volumegroups by calling cmexec
            #---------------------------------------------------
            create_remotelvmtab($remotenode);

            #---------------------------------------------------
            # Compare volume group for local/remote lvmtab
            #---------------------------------------------------
            foreach my $lvname ( keys %{verify_vxfs} ) {

                my $vgname = dirname($lvname);
                LogVerbose "Comparing volumegroup name=[$vgname]";
                twonode_vgid_compare($host, $remotenode, $vgname);
            }
        }
    }

    LogReturnSub();
    return;
}

#===  FUNCTION  =============================================================
#      PURPOSE: This function is responsible for dispatching sapverify
#               requests based on the arguments of the script.
#   PARAMETERS:
#      RETURNS: None
#  DESCRIPTION: None
#     COMMENTS: None
#============================================================================

sub maindispatcher {
    LogEnterSub(@_);

    if ( "$host" eq "$node" ) {

#        $sg_op_code=""                           unless defined $sg_op_code;
#        LogVerbose("sg_op_code=[$sg_op_code]");

        # set user specified log to file
        setlogtofile($logtofile)                    if defined $logtofile;

        # get/save the script/module that called sapverify
        called_by_script($calledby)                   if defined $calledby;

        # set user specified loglevel
        setverboselevel($verboselevel)              if defined $verboselevel;

        # set user specified timeout
        settimeout($timeo)                          if defined $timeo;

        # check file permissions... for files marked "files_sgesap"....
        verify_files_for_component($filelist)     if defined $filelist;

        # generate checksums for files in filelist marked "clucomp"
        ClucompFileListChksum($clucomp)           if defined $clucomp;

        # list all SGeSAP packages in the cluster
        ShowSGeSAPPackages($info_sgesap)          if defined $info_sgesap;

        # Verify a SGeSAP package e.g. "mdsMO7"
        run_verify_sgesap_package($sgpkg)         if defined $sgpkg;

        # Verify a external instance based on a SAP profile "SID_D23_host"
        run_pf_based_verify($pf_for_verify)       if defined $pf_for_verify;

        # get /etc/lvmtab und output to stdout the lvmtabs can be compared
        create_lvmtab($getremotelvm)              if defined $getremotelvm;

        # get user startup files to stdout so they can be compared
        get_startup_files($host, $userfiles)             if defined $userfiles;

        # list stats of all files that are distributed on each node
#        get_distributed_files ("ALL",$dstrbfiles)   if defined $dstrbfiles;
        list_distributed_files ($dstrbfiles)   if defined $dstrbfiles;

    }
    else {
        #not used
        LogVerbose "Current host=[$host]; Execution on remote cluster node=[$node]";
        my @CMD = ("$SGSBIN/cmexec -n $node -t $timeout $0");
        push( @CMD, @SAVED_ARGV );
        push( @CMD, " 2>&1 |" );
        my $cmd = join( " ", @CMD );

        LogVerbose("Execute \$cmd=[$cmd]");
        open( my $fh, $cmd ) or LogRuntimeError "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

        while (<$fh>) {
            print;
        }

    }

    LogReturnSub();
    return ;

}	# ----------  end of subroutine maindispatcher  ----------

#############################################################################
#
# END OF FUNCTIONS SECTION
#
#############################################################################

#############################################################################
# Potentially fork child and continue with maindispatcher
#
# MAIN SECTION
#
#############################################################################
#
#---------------------------------------------------
# Setup a global timeout handler for this script.
# This is necessary as NFS directories will be
# accessed and a NFS hang will also cause the
# execution of this script to also hang.
#
# Two code paths provided: The first is based on a
# threads implementation of timeout, the second is
# based on a fork (and kill of child if timeout
# exceeded).
#---------------------------------------------------
    # pid for this script
    my $parentpid = $$;
    my $save_child_rc=0;

if ( !$fork ) {

    #---------------------------------------------------
    # Don't fork a child
    #---------------------------------------------------
    LogVerbose("No Child will be forked");

    sub CatchSignal {
        my $signame = shift;
        croak "CatchSignal() was triggered - received SIG $signame";
    }

    $SIG{ALRM} = \&CatchSignal;

    # enable the alarm
    alarm($timeout);

    #---------------------------------------------------
    # Execute the sapverify maindispatcher() function
    #---------------------------------------------------

     maindispatcher();

    # disable the alarm
    alarm 0;

    # if defined the verify part was run
    if (defined $sgpkg) {
        LogVerbose "(sapverify): Statistics for sgpkg=[$sgpkg]:";
        LogVerbose "(sapverify): Number of files verified by sapverify: filecount=[$filecount]";


        # only log warnings instead of errors, so cmapplyconf  always succeeds
        if ($warningmode) {
            LogVerbose "(sapverify): Warnings counted by sapverify: warncount=[$errorcount]";
            timestamp_and_exit(0);
        } else {
            LogVerbose "(sapverify): Errors counted by sapverify: errorcount=[$errorcount]";
            $errorcount > 0 ?  timestamp_and_exit(2) : timestamp_and_exit(0);
        }
    }

}
else {


    #---------------------------------------------------
    # Fork a child
    #---------------------------------------------------
    LogVerbose("Child will be forked");
    eval {

        # setup signal handler. If the alarm has expired
        # kill the child
        local $SIG{ALRM} = sub {
            CountAndLogErrorF "Signal handler triggered - terminate child with: kill INT $childpid";
            kill 'INT', $childpid;
            croak 'Alarm expired';
        };

        # set the alarm
        alarm($timeout);

        if ( $childpid = fork ) {
            LogDebug("In Parent - child has \$childpid=[$childpid])");

            # wait for child to terminate
            waitpid( $childpid, 0 );
            $save_child_rc = $? >> 8;
            LogDebug("In Parent - child returned: \$save_child_rc=[$save_child_rc]");

        }
        else {
            LogDebug("In Child - parent has \$parentpid=[$parentpid]");

            #---------------------------------------------------
            # Execute the maindispatcher() function
            #---------------------------------------------------
            maindispatcher();

            # if defined the verify part was run
            if ( defined $sgpkg ) {

                $verbose = getparam('verbose');
                if (($verbose > 0)  or ($errorcount > 0)) {
                    LogInfo "";
                    LogInfo "(sapverify): Statistics for sgpkg=[$sgpkg]:";
                    LogInfo "(sapverify): Number of files verified by sapverify: filecount=[$filecount]";

                    if ($warningmode) {
                        LogInfoF "(sapverify): Warnings counted by sapverify: warncount=[$errorcount]";
                    } else {
                        LogInfo "(sapverify): Errors counted by sapverify: errorcount=[$errorcount]";
                    }

                }
#               my $datestring = strftime "%b %e %H:%M:%S ", localtime;
# 				print "$datestring $host ********** LEAVE sapverify **********\n";

                if ($warningmode) {
                    timestamp_and_exit(0);
                } else {
                    $errorcount > 0 ? timestamp_and_exit(2) : timestamp_and_exit(0);
                }
            }

        }

        # reset the alarm
        alarm(0);

    };

#    LogDebug "dollar@=[$@]";

    if ( $@ ) {
        if ($@ =~ /Alarm expired/ ) {
                if ($warningmode) {
                    LogInfoF("Timeout! Alarm went off after \$timeout=[$timeout] seconds! - Exiting");
                    timestamp_and_exit(0);
                } else {
                    LogRuntimeError("Timeout! Alarm went off after \$timeout=[$timeout] seconds! - Exiting");
                    timestamp_and_exit(2);
                }
        }

        LogInfo "sapverify eval returned @=[$@]";
        timestamp_and_exit(2);

    }
    else {
        LogDebug("Child returned with: \$save_child_rc=[$save_child_rc]");
    }


    #---------------------------------------------------
    # Done!
    #---------------------------------------------------

    #   logfile not used any  more
    # cleanup_n_close();
    exit $save_child_rc;
}

# Exit with success
exit 0;
